.First.lib <-
function(section, library = paste(getenv("SHOME"), "library", sep = "\\"))
{
	if(missing(library))
		library <- paste(getenv("SHOME"), "library", sep = "\\")


	path <- paste(section, library, "geoS.dll", sep = "\\")

	funcs <- c("diag_quadraticform_XAX", "bilinearform_XAY", "loccoords", "binit", "distdiag", "cor_diag", "kb_sim", "woodandchan", "Xwhittlematern")
	dll.load(path, funcs)
}

# 1 "geoSall.S"
# 21 "geoSall.S"
"try" <-
  function(expr, first = T)
{
  restart(first)
  if(first) {
    first <- F
    expr
  }
  else {
    val <- get.message(even.if.used = T)
    class(val) <- "Error"
    val
  }
}

"get.message" <-
function(new = "", trim = T, even.if.used = F)
{
  if(even.if.used) .Call("s_get_last_message2", new, trim, even.if.used)
  else .Call("s_get_last_message", new, trim)
}

"is.R" <-
  function ()
{
  exists("version") && !is.null(vl <- version$language) && vl == "R"
}



"dist0.krige" <-
function(x0, coords)
{
        if(length(x0) != 2)
                stop(paste("length of x0 is", length(x0), "(it must be 2)"))
        coords[, 1] <- coords[, 1] - x0[1]
        coords[, 2] <- coords[, 2] - x0[2]
        return(sqrt(coords[, 1]^2 + coords[, 2]^2))
}

"dist2full" <-
function(dis)
{
        n <- attr(dis, "Size")
        full <- matrix(0, n, n)
        full[lower.tri(full)] <- dis
        full + t(full)
}

"grf" <-
function(n, grid = "irreg", nx = round(sqrt(n)), ny = round(sqrt(n)), nsim = 1,
        xlims = c(0, 1), ylims = c(0, 1), cov.model = c("exponential", "matern",
        "gaussian", "spherical", "cubic", "wave", "powered.exponential",
        "cauchy", "gneiting", "gneiting.matern", "pure.nugget"), cov.pars =
        stop("covariance parameters (sigmasq and phi) needed"), kappa = 0.5,
        nugget = 0, lambda = 1, aniso.pars = NULL, method = c("cholesky", "svd", "eigen",
        "circular.embedding"), messages.screen = T, ...)
{
        call.fc <- match.call()
        method <- match.arg(method)
        cov.model <- match.arg(cov.model)
        rseed <- .Random.seed
        if(!is.null(kappa))
                if(cov.model == "matern" & kappa == 0.5)
                        cov.model <- "exponential"
        ## Reading the parameters
        tausq <- nugget
        if(is.vector(cov.pars)) {
                sigmasq <- cov.pars[1]
                phi <- cov.pars[2]
                nst <- 1
        }
        else {
                sigmasq <- cov.pars[, 1]
                phi <- cov.pars[, 2]
                nst <- nrow(cov.pars)
        }
        sill.total <- tausq + sum(sigmasq)
        ## Printing some messages
        messa <- grf.aux1(nst, nugget, sigmasq, phi, kappa, cov.model)
        if(messages.screen == T) {
                cat(messa$nst)
                cat(messa$nugget)
                cat(messa$cov.structures)
                cat(paste("grf: decomposition algorithm used is: ", method,
                        "\n"))
        }
        results <- list()
        ##
        ## defining the locations for the simulated data
        ##
        if(is.matrix(grid) | is.data.frame(grid)) {
                results$coords <- as.matrix(grid)
                if(messages.screen)
                        cat("grf: simulation(s) on a grid provided by the user\n"
                                )
        }
        else {
                if(grid == "irreg") {
                        results$coords <- cbind(x = runif(n, xlims[1], xlims[
                                2]), y = runif(n, ylims[1], ylims[2]))
                        if(messages.screen)
                                cat(paste(
                                        "grf: simulation(s) on random locations with ",
                                        n, " points\n"))
                }
                else {
                        results$coords <- as.matrix(expand.grid(x = seq(xlims[
                                1], xlims[2], l = nx), y = seq(ylims[1], ylims[
                                2], l = ny)))
                        if(messages.screen)
                                cat(paste("grf: simulation(s) on a  grid ",
                                        nx, " * ", ny, " with ", (nx * ny),
                                        " points\n"))
                }
        }
        n <- nrow(results$coords)
  ##
  ## transforming to the isotropic space
  ##
  if(!is.null(aniso.pars)) {
    if(length(aniso.pars) != 2 | !is.numeric(aniso.pars))
      stop("anisotropy parameters must be provided as a numeric vector with two elements: the rotation angle (in radians) and the anisotropy ratio (a number greater than 1)")
    if(messages.screen)
      cat("grf: transforming to the isotropic space \n")
    results$coords <- coords.aniso(coords = results$coords,
                                   aniso.pars = aniso.pars)
  }
        ##
        ## simulating data at locations defined by the matrix results$coords
        ##
        if(all(phi) == 0) {
                results$data <- matrix(rnorm((n * nsim), mean = 0, sd = sqrt(
                        sill.total)), nrow = n, ncol = nsim)
        }
        else {
                if(method == "circular.embedding") {
                        if(grid == "irreg")
                                stop("Option for \"circular.embedding\" algorithm only allowed for regular grids. You might have to include the argument grid=\"reg\""
                                        )
                        stepx <- (xlims[2] - xlims[1])/(nx - 1)
                        stepy <- (ylims[2] - ylims[1])/(ny - 1)
                        if(round(100000000. * stepx) != round(100000000. *
                                stepy))
                                stop("grf: distance between points must be the same in X and Y directions"
                                        )
                        temp <- list(n = n, nst = nst, sigmasq = sigmasq, xlims
                                 = xlims, ylims = ylims, stepx = stepx,
                                cov.model = cov.model, phi = phi, kappa = kappa
                                )
                        if(messages.screen)
                                cat("\ngrf: WARNING:\nmessages of the type mtot=XXXXX will appear on your screen. \nIf there are many (3 or more, say) or they run indefinitely, you should stop the simulation and try again with a different grid (e.g. try to add 1 point in each direction)\n"
                                        )
                        grf.aux3 <- function(nsim, temp)
                        {
                                realiz <- rep(0, temp$n)
                                for(i in 1:temp$nst) {
                                        realiz <- realiz + sqrt(temp$sigmasq[
                                                i]) * grf.aux2(xlim = temp$
                                                xlims, ylim = temp$ylims, step
                                                 = temp$stepx, cov.model = temp$
                                                cov.model, phi = temp$phi[
                                                i], kappa = temp$kappa)
                                        NULL
                                }
                                return(realiz)
                        }
                        results$data <- apply(as.matrix(1:nsim), 1, grf.aux3,
                                temp = temp)
                        if(nugget != 0) {
                                results$data <- results$data + matrix(rnorm(
                                        (n * nsim), sd = sqrt(nugget)), ncol =
                                        nsim)
                        }
                }
                else {
                        results$data <- matrix(rnorm((n * nsim)), nrow = n,
                                ncol = nsim)
                        cov.decomp <- t(varcov.spatial(coords = results$coords,
                                cov.model = cov.model, kappa = kappa, nugget =
                                nugget, cov.pars = cov.pars, only.decomposition
                                 = TRUE, func.inv = method)$sqrt.varcov)
                        results$data <- cov.decomp %*% results$data
                }
                if(nsim == 1)
                        results$data <- as.vector(results$data)
        }
        ##
        ## transforming data (Box - Cox)
        ##
        if(lambda != 1) {
                if(lambda != 0)
                        results$data <- (results$data * lambda + 1)^(1/lambda)
                else results$data <- exp(results$data)
                messa$transformation <- paste(
                        "grf: Data transformed (Box-Cox), for lambda =", lambda
                        )
                if(messages.screen)
                        cat(messa$transformation)
                cat("\n")
        }
  ##
  ## back-transformation to the anisotropic space
  ##
  if(!is.null(aniso.pars)) {
    if(messages.screen)
      cat("grf: back-transforming to the anisotropic space \n")
    results$coords <- coords.aniso(coords = results$coords,
                                   aniso.pars = aniso.pars, reverse=TRUE)
  }
  else{aniso.pars <- "no anisotropy parameters provided/used"}
        ##
        ## preparing output
        ##
        if(messages.screen) cat(paste(
                        "grf: End of simulation procedure. Number of realizations:",
                        nsim, "\n"))
        results <- c(results, list(cov.model = cov.model, nugget = nugget,
                cov.pars = cov.pars, kappa = kappa, lambda = lambda, messages
                 = messa, method = method, .Random.seed = rseed, call = call.fc))
        class(results) <- "grf"
        return(results)
}

"grf.aux1" <-
function(nst, nugget, sigmasq, phi, kappa, cov.model)
{
        cov.nst <- paste("grf: Process with ", nst,
                " covariance structure(s)\n")
        cov.nugget <- paste("grf: The nugget effect is: ", nugget, "\n")
        cov.message <- NULL
        for(i in 1:nst) {
                if(phi[i] == 0)
                        cov.message[i] <- paste("grf: covariance model", i,
                                "is a pure nugget effect\n")
                else {
                        if(cov.model == "matern" | cov.model ==
                                "powered.exponential" | cov.model == "cauchy" |
                                cov.model == "gneiting-matern")
                                cov.message[i] <- paste(
                                        "grf: covariance model ", i, " is: ",
                                        cov.model, "(sigmasq=", sigmasq[i],
                                        ", phi=", phi[i], ", kappa = ", kappa,
                                        ")\n", sep = "")
                        else cov.message[i] <- paste("grf: covariance model ",
                                        i, " is: ", cov.model, "(sigmasq=",
                                        sigmasq[i], ", phi=", phi[i], ")\n",
                                        sep = "")
                }
        }
        return(list(nst = cov.nst, nugget = cov.nugget, cov.structures =
                cov.message))
}

"grf.aux2" <-
function(xlim, ylim, step, cov.model, phi, kappa = NULL)
{
        if(!is.null(kappa))
                if(cov.model == "matern" & kappa == 0.5) {
                        cov.model <- "exponential"
                }
        cs <- switch(cov.model,
                spherical = c(1, 1),
                gneiting.matern = c(2, kappa),
                powered.exponential = c(3, kappa),
                exponential = c(3, 1),
                gaussian = c(3, 2),
                matern = c(4, kappa),
                gneiting = c(5, 1),
                cauchy = c(6, kappa),
                wave = c(7, 1))
        covfct <- cs[1]
        shape <- cs[2]
        parameter <- c(phi, shape, 1)
        nx <- c(diff(xlim)/step + 1, diff(ylim)/step + 1)
        storage.mode(nx) <- "integer"
        ln <- 2
        storage.mode(ln) <- "integer"
        covnr <- covfct
        storage.mode(covnr) <- "integer"
        simustep <- step
        storage.mode(simustep) <- "double"
        param <- parameter
        storage.mode(param) <- "double"
        lparam <- length(parameter)
        storage.mode(lparam) <- "integer"
        res <- double(nx[1] * nx[2])
        mm <- integer(ln)
        x <- .C("woodandchan",
                covnr,
                nx,
                ln,
                simustep,
                param,
                lparam,
                res = res,
                m = mm)$res
        cat("\n")
        return(x)
}

"image.ksline" <-
function(obj.ksline, grid = obj.ksline$grid, values = c("est", "kvar",
        "oktrend", "ktrend", "ktetrend"), ...)
{
        values <- match.arg(values)
        grid <- obj.ksline$grid[order(grid[, 2], grid[, 1]), ]
        x <- as.numeric(levels(as.factor(grid[, 1])))
        nx <- length(x)
        y <- as.numeric(levels(as.factor(grid[, 2])))
        ny <- length(y)
        if(nx == ny)
                par(pty = "s")
        if(values == "est")
                m <- matrix(obj.ksline$pred, ncol = ny)
        if(values == "kvar")
                m <- matrix(obj.ksline$krige.var, ncol = ny)
        image(x, y, m, ...)
        return(invisible())
}

"krige.bayes" <-
  function(geodata, coords = geodata$coords, data = geodata$data, locations =
           "no", model = model.control(trend.d = "cte", trend.l = "cte", cov.model
                   = "exponential", kappa = 0.5, aniso.pars = NULL, lambda = 1), prior =
           prior.control(beta.prior = c("flat", "normal", "fixed"), beta = NULL,
                         beta.var = NULL, sill.prior = c("reciprocal", "fixed"), sill = NULL,
                         range.prior = c("uniform", "exponential", "fixed", "squared.reciprocal",
                           "reciprocal"), exponential.prior.par = 1, range = NULL, range.discrete
                         = NULL, nugget.prior = c("fixed", "uniform"), nugget = 0,
                         nugget.discrete = NULL), output = output.control(n.posterior = 1000,
                                                    n.predictive = NULL, moments = TRUE, simulations.predictive = TRUE,
                                                    keep.simulations = TRUE, mean.estimator = TRUE, quantile.estimator =
                                                    NULL, probability.estimator = NULL, signal = FALSE, messages.screen =
                                                    TRUE))
{
  call.fc <- match.call()
  seed <- .Random.seed
  kb.results <- list(posterior = list(), predictive = list())
  ##
  ## reading input
  ##
  cov.model <- model$cov.model
  cov.model.number <- cor.number(cov.model)
  kappa <- model$kappa
  if(cov.model == "powered.exponential" & (kappa <=0 | kappa > 2))
    stop("for power exponential correlation model the parameter kappa must be in the intervel (0,2]\n")
  lambda <- model$lambda
  ##
  beta <- prior$beta
  beta.var <- prior$beta.var
  sill <- prior$sill
  ##range <- prior$range
  nugget <- prior$nugget
  exponential.prior.par <- prior$exponential.prior.par
  ##
  n.posterior <- output$n.posterior
  n.predictive <- output$n.predictive
  moments <- output$moments
  messages.screen <- output$messages.screen
  ##
  ## checking data configuration
  ##
  if(is.vector(coords)) {
    coords <- cbind(coords, 0)
    warning("vector of coordinates: one spatial dimention assumed")
  }
  coords <- as.matrix(coords)
  data.dist<- as.vector(dist(coords))
  data.dist.range <- range(data.dist)
  data.dist.min <- data.dist.range[1]
  data.dist.max <- data.dist.range[2]
  ## check == here
  if(round(1000000000000. * data.dist.min) == 0) stop(
            "This function does not allow two data at same location\n"
            )
  if(all(locations == "no")) {
    if(prior$beta.prior != "fixed" & prior$sill.prior != "fixed" &
       prior$range.prior != "fixed")
      if(messages.screen){
        cat("krige.bayes: no prediction locations provided.")
        cat("             Only Bayesian estimates of model parameters will be returned. \n ")
      }
    else stop("Locations to be predicted not provided. Bayesian parameter estimation allowed only if priors are provided for beta, sill and range\n"
              )
  }
  ##
  ## Cheking priors input
  ##
  if(prior$beta.prior == "fixed" & prior$sill.prior != "fixed")
    stop("option for fixed beta and random sill not implemented\n")
  if(prior$beta.prior == "fixed" & prior$range.prior != "fixed")
    stop("option for fixed beta and random range not implemented\n")
  if(prior$sill.prior == "fixed" & prior$range.prior != "fixed")
    stop("option for fixed sill and random range not implemented\n")
  if(prior$beta.prior != "flat" & prior$sill.prior != "fixed")
    stop("selected prior for beta not allowed for selected prior for sill\n"
         )
  if(prior$beta.prior == "fixed" & is.null(beta)) stop(
       "if beta is fixed, its value must be provided in the argument beta\n"
       )
  if(prior$beta.prior == "normal" & (is.null(beta) | is.null(beta.var)))
    stop("if the prior for beta is normal, the prior mean(s) and prior (co)variance must be provided using the argument beta and priorr$beta.var\n"
         )
  if(prior$sill.prior == "fixed" & is.null(sill))
    stop("if sill is fixed, its value must be provided in the argument sill\n"
         )
  if(prior$range.prior == "fixed" & is.null(prior$range))
    stop("if range is fixed, its value must be provided in the argument range\n"
         )
  if(prior$range.prior != "fixed"){
    if (is.null(prior$range.discrete))
      stop("to include the range as random in the Bayesian analysis the argument range.discrete must be provided\n")
    discrete.diff <- diff(prior$range.discrete)
    if(round(max(1e08 * discrete.diff)) != round(min(1e08 * discrete.diff)))
      stop("the current implementation requires equally spaced values in the argument \"range.discrete\"\n")
    discrete.diff <- NULL
  }
  if(prior$nugget.prior != "fixed"){
    if(is.null(prior$nugget.discrete))
      stop("to include the nugget as random in the Bayesian analysis the argument nugget.discrete must be provided\n")
    if(prior$nugget.prior != "fixed")
      if(any(prior$nugget.discrete > 1))
        warning("when nugget is considered as random in the Bayesian analysis the values in the argument nugget.discrete must be relative nugget: tausq/sigmasq\n")
    discrete.diff <- diff(prior$nugget.discrete)
    if(round(max(1e08 * discrete.diff)) != round(min(1e08 * discrete.diff)))
      stop("the current implementation requires equally spaced values in the argument \"nugget.discrete\"\n")
    discrete.diff <- NULL
  }
  ##
  ## checking output options
  ##
  if(!is.null(output$quantile.estimator)) {
    if(is.numeric(output$quantile.estimator))
      if(any(output$quantile.estimator) < 0 | any(output$
                                                  quantile.estimator) > 1)
        stop("quantiles indicators must be numbers in the interval [0,1]\n"
             )
    if(output$quantile.estimator == TRUE)
      output$quantile.estimator <- c(0.025000000000000001,
                                     0.5, 0.97499999999999998)
  }
  if(!is.null(output$probability.estimator)){
    if(!is.numeric(output$probability.estimator))
      stop("probability.estimator must be a numeric value (or vector) of cut-off value(s)\n")
    if(length(output$probability.estimator)>1 & length(output$probability.estimator) != nrow(locations))
      stop("probability.estimator must either have length 1, or have length = nrow(locations)\n")
  }
  if(lambda != 1 & lambda != 0 & moments) {
    moments <- FALSE
    cat(paste("WARNING: moments cannot be computed for lambda =",
              lambda, ". Argument moments was set to FALSE\n"))
  }
  if(lambda < 0 & output$mean.estimator) {
    output$mean.estimator <- FALSE
    cat("krige.bayes: mean.predictor set to F. The resulting distribution does not has expectation for $lambda < 0$\n"
        )
  }
  ##
  ## Box-Cox transformation of the data
  ##
  if(lambda != 1) {
    if(prior$range.prior == "fixed")
      stop("transformation option available only for the full Bayesian approach\n"
           )
    if(messages.screen)
      cat(paste(
                "krige.bayes: performing Box-Cox's transformation for $lambda=",
                lambda, "\n"))
    if(lambda == 0)
      data <- log(data)
    else data <- ((data^lambda) - 1)/lambda
  }
  ##
  ## Checking the dimension of points to be predict (1 or 2)
  ##
  if(all(locations != "no")) {
    if(is.vector(locations)) {
      if(length(locations) == 2) {
        locations <- t(as.matrix(locations))
        warning("THE FUNCTION IS CONSIDERING THAT YOU HAVE ENTERED WITH 1 POINT TO BE PREDICTED IN A TWO DIMENSION REGION\n"
                )
      }
      else locations <- as.matrix(cbind(locations, 0))
    }
    else locations <- as.matrix(locations)
  }
  ##
  ## Building trend matrices:
  ##
  if(inherits(model$trend.d, "formula") | inherits(model$trend.l,
                                                   "formula")) {
    if(any(locations != "no"))
      if((inherits(model$trend.d, "formula") == FALSE) | (
                    inherits(model$trend.l, "formula") == FALSE))
        stop("model$trend.d and model$trend.l must have similar specification\n"
             )
    if(messages.screen)
      cat("krige.bayes: Kriging with external trend to be performed using covariates provided by the user\n"
          )
  }
  else {
    if(any(locations != "no") & (model$trend.d != model$trend.l))
      {
        stop("model$trend.l is different from model$trend.d\n")
      }
    if(messages.screen) {
      if(model$trend.d == "cte")
        cat("krige.bayes: analysis assuming constant mean\n"
            )
      if(model$trend.d == "1st")
        cat("krige.bayes: analysis assuming a 1st degree polinomial trend\n"
            )
      if(model$trend.d == "2nd")
        cat("krige.bayes: analysis assuming a 2nd degree polinomial trend\n"
            )
    }
  }
  trend.d <- trend.spatial(trend = model$trend.d, coords = coords)
  dimnames(coords) <- list(NULL, NULL)
  dimnames(trend.d) <- list(NULL, NULL)
  if(all(locations != "no")) {
    trend.l <- trend.spatial(trend = model$trend.l, coords =
                             locations)
    dimnames(locations) <- list(NULL, NULL)
    dimnames(trend.l) <- list(NULL, NULL)
    if(is.matrix(trend.l))
      ni <- dim(trend.l)[1]
    else ni <- length(trend.l)
  }
  beta.size <- ncol(trend.d)
  ##
  ## Checking dimensions
  ##
  if(nrow(coords) != length(data)) stop(
           "number of data is different of number of data locations (coordinates)\n"
           )
  if(all(locations != "no")) {
    if(nrow(locations) != nrow(trend.l))
      stop("number of points to be estimated is different of the number of trend points\n"
           )
    dimnames(locations) <- list(NULL, NULL)
  }
  dimnames(coords) <- list(NULL, NULL)
  ##
  ## Anisotropy correction (should be placed after trend matrices be defined)
  ##
  if(!is.null(model$aniso.pars)) {
    if(length(model$aniso.pars) != 2 | !is.numeric(model$aniso.pars
               ))
      stop("anisotropy parameters must be provided as a numeric vector with two elements: the rotation angle and the anisotropy ratio\n"
           )
    if(messages.screen)
      cat("krige.bayes: anisotropy parameters provided and assumed to be constants\n"
          )
    coords <- coords.aniso(coords = coords, aniso.pars = model$
                           aniso.pars)
    if(all(locations != "no"))
      locations <- coords.aniso(coords = locations,
                                aniso.pars = model$aniso.pars)
    data.dist <- as.vector(dist(coords))
  }
  ##
  n <- length(data)
  tausq <- nugget
  if(prior$sill.prior != "fixed")
    sigmasq <- 1
  else sigmasq <- sill
  ##
######################### PART 2 ##############################
  ## Prediction with fixed phi
  ##
  if(prior$range.prior == "fixed") {
    phi <- prior$range
    ## covariance matrix for data points
    invcov <- varcov.spatial(coords=coords, cov.model
                             = cov.model, kappa = kappa, nugget = nugget, cov.pars
                             = c(sigmasq, phi), inv = TRUE, scaled = TRUE,
                             only.inv.lower.diag = TRUE)
    remove("data.dist", frame = sys.nframe())
    ## Some matrices computations
### ttiv <- crossprod(trend.d, invcov)
### ttivtt <- ttiv %*% trend.d
### ittivtt <- solve(ttivtt)
    ttivtt <- as.double(rep(0, beta.size * beta.size))
    ttivtt <- .C ("bilinearform_XAY",



                         as.double(invcov$lower.inverse),
                         as.double(invcov$diag.inverse),
                         as.double(as.vector(trend.d)),
                         as.double(as.vector(trend.d)),
                         as.integer(beta.size),
                         as.integer(beta.size),
                         as.integer(n),
                         res = ttivtt)$res
    attr(ttivtt, "dim") <- c(beta.size, beta.size)
    ittivtt <- solve(ttivtt)
### ttivz <- ttiv %*% data
    ttivz <- as.double(rep(0, beta.size))
    ttivz <- .C("bilinearform_XAY",



                as.double(invcov$lower.inverse),
                as.double(invcov$diag.inverse),
                as.double(as.vector(trend.d)),
                as.double(as.vector(data)),
                as.integer(beta.size),
                as.integer(1),
                as.integer(n),
                res = ttivz)$res
    ## covariance vector between data points and prediction locations
    v0mat <- as.double(rep(0, n * ni))
    v0mat <- .C("loccoords",



                as.double(as.vector(locations[, 1])),
                as.double(as.vector(locations[, 2])),
                as.double(as.vector(coords[, 1])),
                as.double(as.vector(coords[, 2])),
                as.integer(ni),
                as.integer(n),
                res = v0mat)$res
    attr(v0mat, "dim") <- c(n, ni)
    v0mat <- cov.spatial(obj = v0mat, cov.model = cov.model, kappa
                         = kappa, cov.pars = c(1, phi))
### tv0iv <- t(apply(v0mat, 2, crossprod, y = invcov))
### tv0ivz <- tv0iv %*% data
    tv0ivz <- as.double(rep(0, ni))
    tv0ivz <- .C("bilinearform_XAY",



                 as.double(invcov$lower.inverse),
                 as.double(invcov$diag.inverse),
                 as.double(as.vector(v0mat)),
                 as.double(data),
                 as.integer(ni),
                 as.integer(1),
                 as.integer(n),
                 res = tv0ivz)$res
### tv0ivx <- apply(tv0iv, 1, crossprod, y = trend.d)
    tv0ivx <- as.double(rep(0, ni * beta.size))
    tv0ivx <- .C("bilinearform_XAY",



                 as.double(invcov$lower.inverse),
                 as.double(invcov$diag.inverse),
                 as.double(as.vector(v0mat)),
                 as.double(as.vector(trend.d)),
                 as.integer(ni),
                 as.integer(beta.size),
                 as.integer(n),
                 res = tv0ivx)$res
    attr(tv0ivx, "dim") <- c(ni, beta.size)
    tb <- trend.l - tv0ivx
    ## b <- t(tb)
### tv0ivv0 <- diag(tv0iv %*% v0mat)
    tv0ivv0 <- as.double(rep(0, ni))
    tv0ivv0 <- .C("diag_quadraticform_XAX",



                  as.double(invcov$lower.inverse),
                  as.double(invcov$diag.inverse),
                  as.double(as.vector(v0mat)),
                  as.integer(ni),
                  as.integer(n),
                  res = tv0ivv0)$res
    v0mat <- NULL
    ##
########################## PART 2.1 ##############################
    ## Simple kriging (fixed beta and sigmasq)
    ##
    if(prior$beta.prior == "fixed") {
      beta.mean <- paste("SK : beta provided by user: ",
                         beta)
      kb.results$predictive$mean <- as.vector(tv0ivz +
                                              as.vector(tb %*% beta))
      if(output$signal)
        kb.results$predictive$variance <- as.vector(
                                                    sigmasq * (1 - tv0ivv0))
      else kb.results$predictive$variance <- as.vector(nugget +
                                                       sigmasq * (1 - tv0ivv0))
      priors.messa <- c("no priors for parameters",
                        "results corresponds to simple kriging")
      sill.mean <- paste("sill provided by user: ", sill)
    }
    ##
######################### PART 2.2 ##################################
    ## uncertainty in beta and sigmasq
    ##
    if(prior$beta.prior == "flat") {
      beta.mean <- ittivtt %*% ttivz
### predict <- tv0ivz + b %*% beta.mean
      kb.results$predictive$mean <- as.vector(tv0ivz +
                                              as.vector(tb %*% beta.mean))
      ## bi <- apply(tb, 2, crossprod, y = ittivtt)
      ## if(ncol(trend.d) > 1)
      ## bi <- t(bi)
      ## bitb <- diag(bi %*% tb)
      if(beta.size == 1) bitb <- as.vector(tb^2) * as.vector(
           ittivtt) else {
             bitb <- as.double(rep(0, ni))
             bitb <- .C("diag_quadraticform_XAX",



                        as.double(ittivtt[lower.tri(ittivtt)]),

                        as.double(diag(ittivtt)),
                        as.double(as.vector(t(tb))),
                        as.integer(ni),
                        as.integer(beta.size),
                        res = bitb)$res
           }
      if(prior$sill.prior == "fixed") {
        ##
        ## 2.2.1 Ordinary kriging: uncertainty only in beta, flat prior
        ##
        if(output$signal) kb.results$predictive$
        variance <- as.vector(sigmasq *
                              (1 - tv0ivv0 + bitb)) else
        kb.results$predictive$variance <-
          as.vector(nugget + sigmasq *
                    (1 - tv0ivv0 + bitb))
        priors.messa <- c(
                          "uninformative prior for beta",
                          "no prior for sill",
                          "no prior for range",
                          "kriging estimates and variances equivalent to ordinary kriging"
                          )
        sill.mean <- paste("sill provided by user: ",
                           sill)
      }
      else {
        ##
        ## 2.2.2 using a conjugate normal prior
        ##
        residu <- as.vector(data - trend.d %*% beta.mean
                           )
### sill.mean <- (residu %*% invcov %*% residu)/
### (n - length(beta.mean) - 2)
        resivres <- as.double(0)
        resivres <- .C("diag_quadraticform_XAX",



                       as.double(invcov$lower.inverse),
                       as.double(invcov$diag.inverse),
                       as.double(residu),
                       as.integer(1),
                       as.integer(n),
                       res = resivres)$res
        sill.mean <- resivres/(n - length(beta.mean) -
                               2)
        if(output$signal)
          kb.results$predictive$variance <-
            as.vector(sill.mean * (1 -
                                   tv0ivv0 + bitb))
        else kb.results$predictive$variance <-
          as.vector(nugget + sill.mean *
                    (1 - tv0ivv0 + bitb))
        priors.messa <- c(
                          "uninformative prior for beta",
                          "uninformative prior for sill",
                          "kriging estimates (but not variances) equivalent to ordinary kriging"
                          )
      }
    }
    ##
    ## 2.2.3 Uncertainty only in the mean parameter (with normal prior)
    ##
    if(prior$beta.prior == "normal") {
      beta.var <- beta.var/sigmasq
      ibeta.var <- solve(beta.var)
      varbgz <- solve(ibeta.var + ttivtt)
      beta.mean <- varbgz %*% (ibeta.var %*% beta + ttivz)
      kb.results$predictive$mean <- as.vector(tv0ivz + as.vector(tb %*% beta.mean))
### bi <- apply(tb, 2, crossprod, y = varbgz)
### bitb <- diag(bi %*% tb)
      if(beta.size == 1) bitb <- as.vector(tb^2) * as.vector(
           ittivtt) else {
             bitb <- as.double(rep(0, ni))
             bitb <- .C("diag_quadraticform_XAX",



                        as.double(ittivtt[lower.tri(ittivtt)]),

                        as.double(diag(ittivtt)),
                        as.double(as.vector(t(tb))),
                        as.integer(ni),
                        as.integer(beta.size),
                        res = bitb)$res
           }
      if(output$signal)
        kb.results$predictive$variance <- as.vector(
                                                    sigmasq * (1 - tv0ivv0) + bitb)
      else kb.results$predictive$variance <- as.vector(nugget +
                                                       sigmasq * (1 - tv0ivv0) + bitb)
      priors.messa <- c("Normal prior for beta",
                        "no prior for sill", "no prior for range")
      sill.mean <- paste("provided by user: ", sill)
    }
    ##
######################### PART 2.3 ##################################
    ## Preparing output
    ##
### kb.results$predictive$mean <- as.vector(predict)
### kb.results$predictive$variance <- as.vector(krige.var)
    kb.results$posterior$beta.mean <- beta.mean
    kb.results$posterior$sigmasq.mean <- sill.mean
    kb.results$posterior$phi.mean <- paste("provided by user: ",
                                           prior$range)
    kb.results$type.prediction <- priors.messa
  }
  else {
    ##
######################### PART 3 ##############################
    ## considering the uncertainty in the parameter phi
    ##
    if(messages.screen) cat(
                            "krige.bayes: computing the discrete approximation of the posterior of phi/tausq.rel\n"
                            )
    ##
######################### PART 3.1 ##############################
    ## Computing the discrete posterior distribution for phi
    ##
    gl <- as.vector(n - beta.size)
    phidist <- list()
    if(is.null(prior$range.discrete)) {
      phidist$phi <- seq(0, max(data.dist), l = 51)
      warning("argument range.discrete not provided. Default values assumed"
              )
    }
    else phidist$phi <- prior$range.discrete
    if(is.null(prior$nugget.discrete)) {
      nugget.discrete <- nugget
      if(messages.screen)
        cat(paste(
                  "krige.bayes: nugget assumed to be fixed and equals to",
                  nugget, "\n"))
    }
    else nugget.discrete <- prior$nugget.discrete
    n.range <- length(phidist$phi)
    n.nugget <- length(nugget.discrete)
    phi.val <- phidist$phi
    phidist$phi <- as.matrix(expand.grid(phidist$phi,
                                         nugget.discrete))
    dimnames(phidist$phi) <- list(NULL, NULL)
    temp.res <- apply(phidist$phi, 1, krige.bayes.aux1, coords =
                      coords, trend.d = trend.d, data = data, cov.model =
                      cov.model, kappa = kappa, range.prior = prior$
                      range.prior, exponential.prior.par =
                      exponential.prior.par, n = n)
    extract.f <- function(obj)
      {
        return(obj$s2phi)
      }
    phidist$s2 <- as.vector(sapply(temp.res, extract.f))
    attr(phidist$s2, "dim") <- c(n.range, n.nugget)
    extract.f <- function(obj)
      {
        return(obj$logprobphi)
      }
    phidist$logprobphi <- as.vector(sapply(temp.res, extract.f))
    attr(phidist$logprobphi, "dim") <- c(n.range, n.nugget)
    extract.f <- function(obj)
      {
        return(obj$inv.lower)
      }
    inv.lower <- array(t(sapply(temp.res, extract.f)), dim = c(
                                                         n.range, n.nugget, (n * (n - 1))/2), dimnames = list(
                                                                                                NULL, NULL, NULL))
    extract.f <- function(obj)
      {
        return(obj$inv.diag)
      }
    inv.diag <- array(t(sapply(temp.res, extract.f)), dim = c(
                                                        n.range, n.nugget, n), dimnames = list(NULL, NULL,
                                                                                 NULL))
    extract.f <- function(obj)
      {
        return(obj$vbetaphi)
      }
    if(beta.size == 1) {
      phidist$vbeta.vector <- as.vector(sapply(temp.res, extract.f))
      attr(phidist$vbeta.vector, "dim") <- c(n.range,
                                             n.nugget, 1)
    }
    else {
      phidist$vbeta.vector <- as.vector(t(sapply(temp.res, extract.f)))
      attr(phidist$vbeta.vector, "dim") <- c(n.range,
                                             n.nugget, (beta.size^2))
    }
    extract.f <- function(obj)
      {
        return(obj$betaphi)
      }
    if(beta.size == 1) {
      phidist$beta <- as.vector(sapply(temp.res, extract.f))
      attr(phidist$beta, "dim") <- c(n.range, n.nugget, 1)
    }
    else {
      phidist$beta <- as.vector(t(sapply(temp.res, extract.f)))
      attr(phidist$beta, "dim") <- c(n.range, n.nugget,
                                     beta.size)
    }
    temp.res <- NULL
    phidist$logprobphi <- phidist$logprobphi + abs(min(phidist$
                                                       logprobphi))
    phidist$probphi <- exp(phidist$logprobphi)
    sumprob <- sum(phidist$probphi)
    phidist$probphi <- phidist$probphi/sumprob
    ##
######################### PART 3.2 ##############################
    ## Sampling from the (parameters) posterior distribution
    ##
    if(messages.screen) cat(
                            "krige.bayes: sampling from multivariate distribution of the parameters\n"
                            )
    ind <- sample((1:(n.range * n.nugget)), n.posterior, replace =
                  T, prob = as.vector(phidist$probphi))
    ind.unique <- sort(unique(ind))
    ind.length <- length(ind.unique)
    ind.table <- table(ind)
    phi.unique <- phidist$phi[ind.unique, ]
    if(messages.screen) {
      cat("krige.bayes: samples and their frequencies from the distribution of  phi and tau.rel\n"
          )
      print(rbind(phi = phi.unique[, 1], nugget.rel =
                  phi.unique[, 2], frequency = ind.table))
      cat("\n")
    }
    phi.sam <- phidist$phi[ind, ]
    vecpars.back.order <- order(ind)
    vec.s2 <- rep(as.vector(phidist$s2)[ind.unique], ind.table)
    samples.chisq <- rchisq(n.posterior, df = gl)
    sigmasq <- (gl * vec.s2)/samples.chisq
    if(beta.size == 1) {
      samples.beta <- rnorm(n.posterior, mean = 0, sd = 1)
      vec.beta <- rep(as.vector(phidist$beta)[ind.unique],
                      ind.table)
      vec.vbeta <- rep(as.vector(phidist$vbeta.vector)[
                                                       ind.unique], ind.table)
      beta <- vec.beta + sqrt(sigmasq * vec.vbeta) *
        samples.beta
    }
    else {
      ind.beta <- matrix(phidist$beta, ncol = beta.size)[
                                         ind.unique, ]
      ind.beta <- ind.beta[rep(1:ind.length, ind.table),
                           ]
      ind.vbeta <- matrix(phidist$vbeta.vector, ncol =
                          beta.size^2)[ind.unique, ]
      ind.vbeta <- ind.vbeta[rep(1:ind.length, ind.table),
                             ] * sigmasq
      ## print("2.4: try to speed up this bit!")
      temp.res <- apply(ind.vbeta, 1, krige.bayes.aux3,
                        beta.size = beta.size)
      beta <- ind.beta + t(temp.res)
      temp.res <- NULL
    }
    if(beta.size == 1) {
      trend.mean <- mean(beta)
      trend.median <- median(beta)
    }
    else {
      trend.mean <- apply(beta, 2, mean)
      trend.median <- apply(beta, 2, median)
    }
    sill.mean <- mean(sigmasq)
    sill.median <- median(sigmasq)
    range.marg <- apply(phidist$probphi, 1, sum)
    range.marg <- range.marg/(sum(range.marg))
    range.mean <- phi.val %*% range.marg
    range.median <- median(phi.sam[, 1])
    ## check == here
    range.mode <- phi.val[range.marg == max(range.marg)]
    nugget.marg <- apply(phidist$probphi, 2, sum)
    nugget.marg <- nugget.marg/(sum(nugget.marg))
    nugget.mean <- nugget.discrete %*% nugget.marg
    nugget.median <- median(phi.sam[, 2])
    ## check == here
    nugget.mode <- nugget.discrete[nugget.marg == max(nugget.marg)]
    ##
    ## Computing the conditional (on phi and tausq.rel,
    ## the later if the case) modes for beta and sigmasq
    ##
    invcov.mode <- varcov.spatial(dists.lowertri = data.dist,
                                  cov.model = cov.model, kappa = kappa, nugget =
                                  nugget.mode, cov.pars = c(1, range.mode), inv = TRUE,
                                  only.inv.lower.diag = TRUE)
    remove("data.dist", frame = sys.nframe())
    ttivtt.mode <- as.double(rep(0, beta.size * beta.size))
    ttivtt.mode <- .C("bilinearform_XAY",



                      as.double(invcov.mode$lower.inverse),
                      as.double(invcov.mode$diag.inverse),
                      as.double(as.vector(trend.d)),
                      as.double(as.vector(trend.d)),
                      as.integer(beta.size),
                      as.integer(beta.size),
                      as.integer(n),
                      res = ttivtt.mode)$res
    attr(ttivtt.mode, "dim") <- c(beta.size, beta.size)
    ittivtt.mode <- solve(ttivtt.mode)
    ttivz.mode <- as.double(rep(0, beta.size))
    ttivz.mode <- .C("bilinearform_XAY",



                     as.double(invcov.mode$lower.inverse),
                     as.double(invcov.mode$diag.inverse),
                     as.double(as.vector(trend.d)),
                     as.double(as.vector(data)),
                     as.integer(beta.size),
                     as.integer(1),
                     as.integer(n),
                     res = ttivz.mode)$res
    beta.mode.cond <- ittivtt.mode %*% ttivz.mode
    resid.mode <- as.vector(data - trend.d %*% beta.mode.cond)
    sill.mode.cond <- (.C("diag_quadraticform_XAX",



                          as.double(invcov.mode$lower.inverse),
                          as.double(invcov.mode$diag.inverse),
                          as.double(as.vector(resid.mode)),
                          as.integer(1),
                          as.integer(n),
                          res = as.double(0.))$res)/(n - length(beta.mode.cond) +
                            2)
    remove(list = c("invcov.mode", "ttivtt.mode", "ittivtt.mode",
             "ttivz.mode"), frame = sys.nframe())
    if(beta.size == 1)
      kb.results$posterior$beta.summary <- c(mean =
                                             trend.mean, median = trend.median, mode.cond =
                                             beta.mode.cond)
    else kb.results$posterior$beta.summary <- cbind(mean =
                                                    trend.mean, median = trend.median, mode.cond =
                                                    beta.mode.cond)
    kb.results$posterior$sigmasq.summary <- c(mean = sill.mean,
                                              median = sill.median, mode.cond = sill.mode.cond)
    kb.results$posterior$phi.summary <- c(mean = range.mean, median
                                          = range.median, mode = range.mode)
    if(prior$nugget.prior != "fixed")
      kb.results$posterior$tausq.summary <- c(mean =
                                              nugget.mean, median = nugget.median, mode =
                                              nugget.mode)
    else kb.results$posterior$tausq.summary <- paste(
                                                     "fixed tausq with value =", nugget)
    kb.results$posterior$beta.samples <- as.matrix(beta)[
                                                         vecpars.back.order, ]
    beta <- NULL
    if(beta.size == 1)
      kb.results$posterior$beta.samples <- as.vector(
                                                     kb.results$posterior$beta.samples)
    kb.results$posterior$sigmasq.samples <- sigmasq[
                                                    vecpars.back.order]
    sigmasq <- NULL
    kb.results$posterior$phi.samples <- phi.sam[vecpars.back.order,
                                                1]
    if(prior$nugget.prior != "fixed")
      kb.results$posterior$tausq.samples <- phi.sam[
                                                    vecpars.back.order, 2]
    else kb.results$posterior$tausq.samples <- paste(
                                                     "fixed tausq with value =", nugget)
    phi.lev <- unique(phidist$phi[, 1])
    kb.results$posterior$phi.marginal <- data.frame(phi = phi.lev,
                                                    expected = apply(phidist$probphi, 1, sum), sampled =
                                                    as.vector(table(factor(phi.sam[, 1], levels = phi.lev))
                                                              )/n.posterior)
    kb.results$posterior$phi.marginal <- as.matrix(kb.results$
                                                   posterior$phi.marginal)
    nug.lev <- unique(phidist$phi[, 2])
    if(prior$nugget.prior != "fixed")
      kb.results$posterior$nugget.marginal <- data.frame(
                                                         nugget = nug.lev, expected = apply(phidist$
                                                                             probphi, 2, sum), sampled = as.vector(table(
                                                                                                 factor(phi.sam[, 2], levels = nug.lev)))/
                                                         n.posterior)
    else kb.results$posterior$nugget.marginal <- paste(
                                                       "fixed tausq with value =", nugget)
    ##
######################### PART 3.3 ##############################
    ## Predictive distribution: sampling and computation
    ##
    if(all(locations != "no")) {
      if(messages.screen)
        cat("krige.bayes: prediction at locations provided\n"
            )
      if(output$mean.estimator != FALSE | !is.null(output$
           probability.estimator) | !is.null(output$
                                             quantile.estimator))
        prediction.samples <- TRUE
      message.prediction <- character()
      ni <- dim(trend.l)[1]
      if(is.null(n.predictive)) {
        include.it <- FALSE
        n.predictive <- n.posterior
        phi.sam <- phidist$phi[ind, ]
        message.prediction <- c(message.prediction,
                                "phi/tausq.rel samples for the predictive are same as for the posterior"
                                )
        if(messages.screen)
          cat("krige.bayes:", message.prediction,
              "\n")
      }
      else {
        include.it <- TRUE
        ind <- sample((1:(dim(phidist$phi)[1])),
                      n.predictive, replace = TRUE, prob =
                      as.vector(phidist$probphi))
        ind.unique <- sort(unique(ind))
        ind.length <- length(ind.unique)
        ind.table <- table(ind)
        phi.unique <- phidist$phi[ind.unique, ]
        message.prediction <- c(message.prediction,
                                "phi/tausq.rel samples for the predictive are NOT the same as for the posterior "
                                )
        if(messages.screen) {
          cat("krige.bayes:", message.prediction,
              "\n")
          cat("krige.bayes: samples and their frequencies from the distribution of  phi and tau.rel when drawing from the predictive distribution\n"
              )
          print(rbind(phi = phi.unique[, 1],
                      nugget.rel = phi.unique[, 2],
                      frequency = ind.table))
        }
        phi.sam <- phidist$phi[ind, ]
        vecpars.back.order <- order(ind)
      }
      ## d0mat <- apply(locations, 1, d0.krige)
      d0mat <- as.double(rep(0, ni * n))
      d0mat <- .C("loccoords",



                  as.double(as.vector(locations[, 1])),
                  as.double(as.vector(locations[, 2])),
                  as.double(as.vector(coords[, 1])),
                  as.double(as.vector(coords[, 2])),
                  as.integer(ni),
                  as.integer(n),
                  res = d0mat)$res
      attr(d0mat, "dim") <- c(n, ni)
      loc.coincide <- apply(d0mat, 2, function(x)
                            {
                              return(any(x < 1e-10))
                            }
                            )
      if(any(loc.coincide))
        loc.coincide <- (1:ni)[loc.coincide]
      else loc.coincide <- NULL
      if(!is.null(loc.coincide)){
        temp.f <- function(x, data){return(data[x < 1e-10])}
        data.coincide <- apply(d0mat[,loc.coincide, drop=FALSE],2,temp.f, data=data)
      }
      else data.coincide <- NULL
      names(ind.table) <- NULL
      assign(".kb.pl", list(ind.length = ind.length, phi.val
                            = phi.val, nugget.discrete = nugget.discrete,
                            cov.model = cov.model, kappa = kappa, n = n,
                            ni = ni, beta.size = beta.size, inv.lower =
                            inv.lower, inv.diag = inv.diag, trend.d =
                            trend.d, data = data, gl = gl, d0mat = d0mat,
                            loc.coincide = loc.coincide, data.coincide =
                            data.coincide, trend.l = trend.l, signal =
                            output$signal, ind.table = ind.table,
                            messages.screen = messages.screen, locations =
                            locations, cov.model.number = cov.model.number),
             where = 1)
      ##
      ## 3.3.1 Estimating the moments
      ##
      if(moments) {
        if(messages.screen)
          cat("krige.bayes: computing moments of the predictive distributions\n"
              )
        assign(".temp.ap10", 1, where = 1)
        assign(".kb.n.disc", dim(phidist$phi)[1], where
               = 1)
        temp.res <- apply(phidist$phi, 1,
                          krige.bayes.aux10, phidist = phidist)
        remove(list = c(".temp.ap10", ".kb.n.disc"),
               where = 1)
        temp.res <- temp.res %*% as.vector(phidist$
                                           probphi)
        kb.results$predictive$moments <- data.frame(expect.y0 = as.vector(temp.res[1:ni]),
                                                    expect.y0.2 = as.vector(temp.res[(
                                                      ni + 1):(2 * ni)]))
        temp.res <- NULL
        kb.results$predictive$moments$var.y0 <-
          kb.results$predictive$moments$
        expect.y0.2 - ((kb.results$predictive$
                        moments$expect.y0)^2)
        if(lambda == 0) {
          temp <- kb.results$predictive$moments$
          expect.y0
          kb.results$predictive$moments$expect.y0 <-
            exp(temp + 0.5 * (kb.results$
                              predictive$moments$var.y0))
          kb.results$predictive$moments$var.y0 <-
            (exp(2 * temp + kb.results$
                 predictive$moments$var.y0)) *
                   (exp(kb.results$predictive$
                        moments$var.y0) - 1)
          temp <- NULL
          kb.results$predictive$moments$
          expect.y0.2 <- kb.results$
          predictive$moments$var.y0 +
            (kb.results$predictive$moments$
             expect.y0^2)
        }
      }
      ##
      ## 3.3.2 Sampling from the predictive
      ##
      if(output$simulations.predictive) {
        assign(".temp.ap20", 1, where = 1)
        kb.results$predictive$simulations <- matrix(
                                                    unlist(apply(phi.unique, 1,
                                                                 krige.bayes.aux20, phidist = phidist)),
                                                    ncol = n.predictive)
        inv.lower <- inv.diag <- NULL
        remove(".temp.ap20", where = 1)
        if(messages.screen)
          cat("krige.bayes: preparing output of predictive distribution\n"
              )
        ##
        ## Back transforming
        ##
        if(lambda != 1) {
          cat("krige.bayes: Data transformation (Box-Cox) performed. Results returned in the original scale\n"
              )
          if(lambda == 0)
            kb.results$predictive$
          simulations <- exp(
                             kb.results$predictive$
                             simulations)
          else {
            if(lambda > 0)
              kb.results$predictive$
            simulations[
                        kb.results$
                        predictive$
                        simulations <
                        (-1/lambda)] <-
                          (-1/lambda)
            if(lambda < 0)
              kb.results$predictive$
            simulations[
                        kb.results$
                        predictive$
                        simulations >
                        (-1/lambda)] <-
                          (-1/lambda)
            kb.results$predictive$
            simulations <- ((
                             kb.results$predictive$
                             simulations * lambda) +
                            1)^(1/lambda)
          }
        }
        remove(".kb.pl", where = 1)
        ##
        ## 3.3.3 mean estimators
        ##
        if(output$mean.estimator) {
          kb.results$predictive$mean.simulations <-
            as.vector(apply(kb.results$
                            predictive$simulations, 1,
                            mean))
          kb.results$predictive$
          variance.simulations <-
            as.vector(apply(kb.results$
                            predictive$simulations, 1,
                            var))
          message.prediction <- c(
                                  message.prediction,
                                  "mean at each location in $predictive$mean and predicted variances in $predictive$variances"
                                  )
        }
        ##
        ## 3.3.4 quantile estimators
        ##
        if(!is.null(output$quantile.estimator)) {
          kb.results$predictive$quantiles <-
            apply(kb.results$predictive$
                  simulations, 1, quantile, probs
                  = output$quantile.estimator)
          if(length(output$quantile.estimator) >
             1) {
            kb.results$predictive$quantiles <-
              as.data.frame(t(
                              kb.results$predictive$
                              quantiles))
            qname <- rep(0, length(output$
                                   quantile.estimator))
            for(i in 1:length(output$
                              quantile.estimator))
              qname[i] <- paste(
                                "q", 100 *
                                output$
                                quantile.estimator[
                                                   i], sep = "")
            names(kb.results$predictive$
                  quantiles) <- qname
          }
          else {
            kb.results$predictive$quantiles <-
              as.vector(kb.results$
                        predictive$quantiles)
          }
          message.prediction <- c(
                                  message.prediction,
                                  "Predicted quantile(s) at each location returned in $predictive$quantiles"
                                  )
        }
        ##
        ## 3.3.5 probability estimators
        ##
        if(!is.null(output$probability.estimator)) {
          kb.results$predictive$probability <-
            apply(kb.results$predictive$
                  simulations, 1,
                  krige.bayes.aux2, cutoff =
                  output$probability.estimator)
          if(length(output$probability.estimator) >
             1) {
            kb.results$predictive$
            probability <-
              as.data.frame(t(
                              kb.results$predictive$
                              probability))
            pname <- rep(0, length(output$
                                   probability.estimator))
            for(i in 1:length(output$
                              probability.estimator)
                )
              pname[i] <- paste(
                                "cutoff",
                                output$
                                probability.estimator[
                                                      i], sep = "")
            names(kb.results$predictive$
                  probability) <- pname
          }
          else kb.results$predictive$probability <-
            as.vector(kb.results$
                      predictive$probability)
          message.prediction <- c(
                                  message.prediction,
                                  "Estimated probabilities of being less than the provided cutoff(s), at each location, returned in $predictive$probability"
                                  )
        }
        ##
        ## samples from predictive
        ##
        if(output$keep.simulations) kb.results$
        predictive$simulations <-
          kb.results$predictive$
        simulations[,
                    vecpars.back.order] else
        kb.results$predictive$
        simulations <- NULL
        if(include.it) {
          phi.lev <- unique(phidist$phi[, 1])
          kb.results$predictive$phi.marginal <-
            data.frame(phi = phi.lev,
                       expected = apply(phidist$
                         probphi, 1, sum), sampled =
                       as.vector(table(factor(phi.sam[
                                                      , 1], levels = phi.lev)))/
                       n.predictive)
          kb.results$predictive$phi.marginal <-
            as.matrix(kb.results$predictive$
                      phi.marginal)
          nug.lev <- unique(phidist$phi[, 2])
          if(prior$nugget.prior != "fixed")
            data.frame(nugget = nug.lev,
                       expected = apply(
                         phidist$probphi, 2,
                         sum), sampled =
                       as.vector(table(factor(
                                              phi.sam[, 2], levels =
                                              nug.lev)))/n.predictive
                       )
          else kb.results$predictive$
          nugget.marginal <-
            paste(
                  "fixed tausq with value =",
                  nugget)
          kb.results$predictive$nugget.marginal <-
            data.frame(nugget = nug.lev,
                       expected = apply(phidist$
                         probphi, 2, sum), sampled =
                       as.vector(table(factor(phi.sam[
                                                      , 2], levels = nug.lev)))/
                       n.predictive)
        }
      }
      kb.results$predictive$type.prediction <-
        "Spatial prediction performed taking into account uncertainty in trend (mean), sill and range parameters"
      kb.results$message.prediction <- message.prediction
    }
    else {
      kb.results$predictive <-
        "no locations to perform prediction were provided"
      kb.results$message.prediction <-
        "Only Bayesian estimation of model parameters were computed"
      if(messages.screen)
        cat(kb.results$message.prediction)
    }
  }
  if(all(nugget != 0) & prior$nugget.prior == "fixed")
    kb.results$nugget.fixed <- nugget
  kb.results$.Random.seed <- seed
  kb.results$call <- call.fc
  if(version$major <= 4)
    class(kb.results) <- c("krige.bayes", "kriging")
  else oldClass(kb.results) <- c("krige.bayes", "kriging")
  if(messages.screen)
    cat("krige.bayes: done!\n")
  return(kb.results)
}

"krige.bayes.aux1" <-
  function(phinug, coords, trend.d, data, cov.model, kappa, range.prior,
           exponential.prior.par, n)
{
  ## auxiliary function to perform calculation for the function krige.bayes
  phi <- phinug[1]
  nugget <- phinug[2]
  ## check == here
  if(phi == 0) covphi <- list(inverse = diag((1/(1 + nugget)), n),
       log.det.to.half = ((n/2) * log((1 + nugget)))) else
  covphi <- varcov.spatial(coords = coords, cov.model =
                           cov.model, kappa = kappa, nugget = nugget,
                           cov.pars = c(1, phi), inv = T, det = T)
  ttiv <- crossprod(trend.d, covphi$inverse)
  ttivtt <- ttiv %*% trend.d
  vbetaphi <- solve(ttivtt)
  eival.x <- eigen(ttivtt, symmetric = T, only.values = T)
  ttivttdet <- prod(eival.x$values)
  betaphi <- as.vector(vbetaphi %*% ttiv %*% data)
  res <- data - trend.d %*% betaphi
  sqres <- as.vector(crossprod(res, covphi$inverse) %*% res)
  gl <- length(data) - length(betaphi)
  s2phi <- (1/gl) * sqres
  logprobphi <- (-0.5) * log(ttivttdet) - (covphi$log.det.to.half) - (
                                                                      gl/2) * log(s2phi)
  if(range.prior == "reciprocal" & round(100000000. * phi) != 0)
    logprobphi <- logprobphi - log(phi)
  if(range.prior == "squared.reciprocal" & round(100000000. * phi) != 0)
    logprobphi <- logprobphi - log(phi^2)
  if(range.prior == "exponential")
    logprobphi <- logprobphi - exponential.prior.par * phi
  ## probphi <- ttivttdet^(-0.5) * (1/exp(covphi$log.det.to.half)) * s2phi^
  ## ( - gl/2)
  ## if(range.prior == "reciprocal")
  ## probphi <- probphi/phi
  ## if(range.prior == "squared.reciprocal")
  ## probphi <- probphi/(phi^2)
  ## if(range.prior == "exponential")
  ## probphi <- probphi * exp( - exponential.prior.par * phi)
  inv.lower <- covphi$inverse[lower.tri(covphi$inverse)]
  inv.diag <- diag(covphi$inverse)
  results <- list(vbetaphi = vbetaphi, betaphi = betaphi, s2phi = s2phi,
                  logprobphi = logprobphi, inv.lower = inv.lower, inv.diag =
                  inv.diag)
  return(results)
}

"krige.bayes.aux10" <-
  function(phinug, phidist)
{
  ni <- .kb.pl$ni
  n <- .kb.pl$n
  beta.size <- .kb.pl$beta.size
  if(.kb.pl$messages.screen)
    krige.bayes.messages(moments = TRUE, n.disc = .kb.n.disc,
                         .temp.ap = .temp.ap10, ind.length = .kb.pl$ind.length)
  phinug <- as.vector(phinug)
  phi.ind <- order(.kb.pl$phi.val)[round(100000000. * .kb.pl$phi.val) ==
                                   round(100000000. * phinug[1])]
  nug.ind <- order(.kb.pl$nugget.discrete)[round(100000000. * .kb.pl$
                                                 nugget.discrete) == round(100000000. * phinug[2])]
  v0 <- cov.spatial(obj = .kb.pl$d0mat, cov.model = .kb.pl$cov.model,
                    kappa = .kb.pl$kappa, cov.pars = c(1, phinug[1]))
  tb <- as.double(rep(0, ni * beta.size))
  tb <- .C("bilinearform_XAY",



           as.double(as.vector(.kb.pl$inv.lower[phi.ind, nug.ind, ])),

           as.double(as.vector(.kb.pl$inv.diag[phi.ind, nug.ind,
                                               ])),
           as.double(as.vector(v0)),
           as.double(as.vector(.kb.pl$trend.d)),
           as.integer(ni),
           as.integer(beta.size),
           as.integer(n),
           res = tb)$res
  attr(tb, "dim") <- c(ni, beta.size)
  tb <- .kb.pl$trend.l - tb
  tv0ivdata <- as.double(rep(0, ni))
  tv0ivdata <- .C("bilinearform_XAY",



                  as.double(as.vector(.kb.pl$inv.lower[phi.ind, nug.ind, ])),

                  as.double(as.vector(.kb.pl$inv.diag[phi.ind, nug.ind,
                                                      ])),
                  as.double(as.vector(v0)),
                  as.double(.kb.pl$data),
                  as.integer(ni),
                  as.integer(1),
                  as.integer(n),
                  res = tv0ivdata)$res
  tmean <- tv0ivdata + tb %*% as.vector(phidist$beta[phi.ind, nug.ind,
                                                     ])
  tv0ivdata <- NULL
  if(((round(1000000000000. * phinug[2]) == 0)) & (!is.null(.kb.pl$
              loc.coincide)))
    tmean[.kb.pl$loc.coincide] <- .kb.pl$data.coincide
  tv0ivv0 <- as.double(rep(0, ni))
  tv0ivv0 <- .C("diag_quadraticform_XAX",



                as.double(as.vector(.kb.pl$inv.lower[phi.ind, nug.ind, ])),

                as.double(as.vector(.kb.pl$inv.diag[phi.ind, nug.ind,
                                                    ])),
                as.double(as.vector(v0)),
                as.integer(ni),
                as.integer(n),
                res = tv0ivv0)$res
  v0 <- NULL
  s2i <- phidist$s2[phi.ind, nug.ind]
  vbetai <- matrix(as.vector(phidist$vbeta.vector[phi.ind, nug.ind, ]),
                   ncol = beta.size, nrow = beta.size)
  if(beta.size == 1)
    tbivbb <- ((as.vector(tb))^2) * vbetai
  else {
    tbivbb <- as.double(rep(0, ni))
    tbivbb <- .C("diag_quadraticform_XAX",



                 as.double(vbetai[lower.tri(vbetai)]),
                 as.double(diag(vbetai)),
                 as.double(as.vector(t(tb))),
                 as.integer(ni),
                 as.integer(beta.size),
                 res = tbivbb)$res
  }
  if(.kb.pl$signal)
    tvar <- s2i * (1 - tv0ivv0 + tbivbb)
  else tvar <- s2i * ((1 + phinug[2]) - tv0ivv0 + tbivbb)
  tb <- tbivbb <- tv0ivv0 <- NULL
  if(((round(1000000000000. * phinug[2]) == 0) | .kb.pl$signal) & (!
              is.null(.kb.pl$loc.coincide)))
    tvar[.kb.pl$loc.coincide] <- 0
  tvar[tvar < 9.9999999999999998e-17] <- 0
  ## take care here, re-using object!
  tvar <- (.kb.pl$gl/(.kb.pl$gl - 2)) * tvar
  tvar <- tvar + (tmean)^2
  ## here tvar is expc.y0.2 !!!
  assign(".temp.ap10", .temp.ap10 + 1, where=1)
  return(cbind(tmean, tvar))
}

"krige.bayes.aux2" <-
function(x, cutoff)
{
        ## auxiliary function to perform calculation for the function krige.bayes
        ncut <- length(cutoff)
        lx <- length(x)
        if(ncut > 1) {
                result <- rep(0, ncut)
                for(i in 1:ncut)
                        result[i] <- sum(x <= cutoff[i])/lx
        }
        else result <- sum(x <= cutoff)/lx
        return(result)
}

"krige.bayes.aux20" <-
function(phinug, phidist)
{
        ni <- .kb.pl$ni
        n <- .kb.pl$n
        beta.size <- .kb.pl$beta.size
        if(.kb.pl$messages.screen)
                krige.bayes.messages(moments = FALSE, n.disc = .kb.n.disc,
                        .temp.ap = .temp.ap20, ind.length = .kb.pl$ind.length)
        phinug <- as.vector(phinug)
        phi.ind <- order(.kb.pl$phi.val)[round(100000000. * .kb.pl$phi.val) ==
                round(100000000. * phinug[1])]
        nug.ind <- order(.kb.pl$nugget.discrete)[round(100000000. * .kb.pl$
                nugget.discrete) == round(100000000. * phinug[2])]
        v0 <- cov.spatial(obj = .kb.pl$d0mat, cov.model = .kb.pl$cov.model,
                kappa = .kb.pl$kappa, cov.pars = c(1, phinug[1]))
        tb <- as.double(rep(0, ni * beta.size))
        tb <- .C("bilinearform_XAY",



as.double(as.vector(.kb.pl$inv.lower[phi.ind, nug.ind, ])),

                        as.double(as.vector(.kb.pl$inv.diag[phi.ind, nug.ind,
                        ])),
                as.double(as.vector(v0)),
                as.double(as.vector(.kb.pl$trend.d)),
                as.integer(ni),
                as.integer(beta.size),
                as.integer(n),
                res = tb)$res
        attr(tb, "dim") <- c(ni, beta.size)
        tb <- .kb.pl$trend.l - tb
        tv0ivdata <- as.double(rep(0, ni))
        tv0ivdata <- .C("bilinearform_XAY",



as.double(as.vector(.kb.pl$inv.lower[phi.ind, nug.ind, ])),

                        as.double(as.vector(.kb.pl$inv.diag[phi.ind, nug.ind,
                        ])),
                as.double(as.vector(v0)),
                as.double(.kb.pl$data),
                as.integer(ni),
                as.integer(1),
                as.integer(n),
                res = tv0ivdata)$res
        tmean <- tv0ivdata + tb %*% as.vector(phidist$beta[phi.ind, nug.ind,
                ])
        tv0ivdata <- NULL
        if(((round(1000000000000. * phinug[2]) == 0)) & (!is.null(.kb.pl$
                loc.coincide)))
                tmean[.kb.pl$loc.coincide] <- .kb.pl$data.coincide
        s2i <- phidist$s2[phi.ind, nug.ind]
        vbetai <- matrix(as.vector(phidist$vbeta.vector[phi.ind, nug.ind, ]),
                ncol = beta.size, nrow = beta.size)
        if(((round(1000000000000. * phinug[2]) == 0) | .kb.pl$signal) & (!
                is.null(.kb.pl$loc.coincide))) {
                v0 <- v0[, - (.kb.pl$loc.coincide)]
                nloc <- ni - length(.kb.pl$loc.coincide)
                tmean.coincide <- tmean[.kb.pl$loc.coincide]
                tmean <- tmean[ - (.kb.pl$loc.coincide)]
                tb <- tb[ - (.kb.pl$loc.coincide), ]
        }
        else nloc <- ni
        Nsims <- .kb.pl$ind.table[.temp.ap20]
        normalsc <- rnorm((nloc * Nsims))
        chisc <- rchisq(Nsims, df = .kb.pl$gl)
        sqglchi <- sqrt(.kb.pl$gl/chisc)
        if(.kb.pl$signal)
                Dval <- 1.
        else Dval <- 1. + phinug[2]
        if(beta.size == 1) {
                Blower <- 0
                Bdiag <- vbetai
        }
        else {
                Blower <- vbetai[lower.tri(vbetai)]
                Bdiag <- diag(vbetai)
        }
        R0 <- as.double(rep(0., (nloc * (nloc + 1))/2))
        if(.kb.pl$cov.model.number > 10) {
          if(((round(1000000000000. * phinug[2]) == 0) | .kb.pl$signal) &
             (!is.null(.kb.pl$loc.coincide))) {
            R0 <- .C("distdiag",



                     as.double(.kb.pl$locations[ - (.kb.pl$
                                                    loc.coincide), 1]),
                     as.double(.kb.pl$locations[ - (.kb.pl$
                                                    loc.coincide), 2]),
                     as.integer(nloc),
                     out = R0)$out
          }
          else R0 <- .C("distdiag",



                        as.double(.kb.pl$locations[, 1]),
                        as.double(.kb.pl$locations[, 2]),
                        as.integer(ni),
                        out = R0)$out
          R0 <- cov.spatial(R0, cov.pars = c(1, phinug[1]), cov.model =
                            .kb.pl$cov.model, kappa = .kb.pl$kappa)
        }
        else {
          if(((round(1000000000000. * phinug[2]) == 0) | .kb.pl$signal) &
             (!is.null(.kb.pl$loc.coincide))) {
            R0 <- .C("cor_diag",



                     as.double(.kb.pl$locations[ - (.kb.pl$
                                                    loc.coincide), 1]),
                     as.double(.kb.pl$locations[ - (.kb.pl$
                                                    loc.coincide), 2]),
                     as.integer(nloc),
                     as.integer(.kb.pl$cov.model.number),
                     as.double(phinug[1]),
                     as.double(.kb.pl$kappa),
                     res = R0)$res
          }
          else R0 <- .C("cor_diag",



                        as.double(.kb.pl$locations[, 1]),
                        as.double(.kb.pl$locations[, 2]),
                        as.integer(ni),
                        as.integer(.kb.pl$cov.model.number),
                        as.double(phinug[1]),
                        as.double(.kb.pl$kappa),
                        res = R0)$res
        }
        normalsc <- .C("kb_sim",



as.double(tmean),
                out = as.double(normalsc),
                as.double(as.vector(.kb.pl$inv.lower[phi.ind, nug.ind, ])),

                        as.double(as.vector(.kb.pl$inv.diag[phi.ind, nug.ind,
                        ])),
                as.double(as.vector(v0)),
                as.integer(nloc),
                as.integer(n),
                as.double(Dval),
                as.integer(Nsims),
                as.double(sqglchi),
                as.double(s2i),
                as.double(Blower),
                as.double(Bdiag),
                as.double(as.vector(t(tb))),
                as.integer(beta.size),
                as.double(R0))$out
        attr(normalsc, "dim") <- c(nloc, Nsims)
        v0 <- R0 <- tb <- NULL
        assign(".temp.ap20", .temp.ap20 + 1, where=1)
        if(((round(1000000000000. * phinug[2]) == 0) | .kb.pl$signal) & (!
                is.null(.kb.pl$loc.coincide))) {
                result <- matrix(0, nrow = ni, ncol = Nsims)
                result[ - (.kb.pl$loc.coincide), ] <- normalsc
                result[.kb.pl$loc.coincide, ] <- rep(tmean.coincide, Nsims)
                return(result)
        }
        else return(normalsc)
}

"krige.bayes.aux3" <-
##
## This function produces a sample from a multivariate normal distribution
## mean is 0 and cov is a vector of length beta.size^2
function(cov, beta.size)
{
        ## dm <- length(mean)
        ## dc <- dim(cov)[1]
        ## if(dm != dc)
        ## stop("mean and cov must have compatible dimensions\n")
        cov <- matrix(cov, ncol = beta.size)
        cov.svd <- svd(cov)
        cov.decomp <- cov.svd$u %*% (t(cov.svd$u) * sqrt(cov.svd$d))
        zsim <- as.vector(cov.decomp %*% rnorm(beta.size))
        return(zsim)
}

"krige.bayes.messages" <-
  function(moments, n.disc, .temp.ap, ind.length)
{
  if(moments) {
    if(n.disc <= 20)
      cat(paste("computing moments: point", .temp.ap,
                "out of", n.disc, "\n"))
    if(n.disc > 20 & n.disc <= 200){
      if(.temp.ap %% 10 == 1)
        cat(paste("computing moments: point", .temp.ap,
                  "out of", n.disc, "\n"))
      if(.temp.ap == 1) cat("                   (counting for each 10)\n")
    }
    if(n.disc > 200){
      if(.temp.ap %% 100 == 1)
        cat(paste("computing moments: point", .temp.ap,
                  "out of", n.disc, "\n"))
      if(.temp.ap == 1) cat("                   (counting for each 100)\n")
    }
  }
  else {
    if(ind.length <= 20)
      cat(paste("prediction for the parameter set", .temp.ap,
                "out of", ind.length, "\n"))
    if(ind.length > 20 & ind.length <= 200){
      if(.temp.ap %% 10 == 1)
        cat(paste("prediction for the parameter set",
                  .temp.ap, "out of", ind.length, "\n"))
      if(.temp.ap == 1) cat("                   (counting for each 10)\n")
    }
    if(ind.length > 200)
      if(.temp.ap %% 100 == 1){
        cat(paste("prediction for the parameter set",
                  .temp.ap, "out of", ind.length, "\n"))
        if(.temp.ap == 1) cat("                   (counting for each 100)\n")
      }
  }
}

"krige.control" <-
function(type.krige = "ok", beta = NULL, trend.d = "cte", trend.l = "cte",
        cov.model = c("exponential", "matern", "gaussian", "spherical",
        "circular", "cubic", "wave", "powered.exponential", "cauchy",
        "gneiting", "gneiting.matern", "pure.nugget"), cov.pars = stop(
        "covariance parameters (sigmasq and phi) should be provided"), kappa =
        0.5, nugget = 0, micro.scale = 0, dist.epsilon = 1e-10, aniso.pars =
        NULL, lambda = 1, signal = FALSE, n.samples.backtransform = 500, n.sim
         = 0)
{
        cov.model <- match.arg(cov.model)
        return(list(type.krige = type.krige, beta = beta, trend.d = trend.d,
                trend.l = trend.l, cov.model = cov.model, cov.pars = cov.pars,
                kappa = kappa, nugget = nugget, micro.scale = micro.scale,
                dist.epsilon = dist.epsilon, aniso.pars = aniso.pars, lambda =
                lambda, signal = signal, n.samples.backtransform =
                n.samples.backtransform, n.sim = n.sim))
}

"krige.conv" <-
  function(geodata, coords = geodata$coords, data = geodata$data, locations,
           krige = krige.control(type.krige, beta = NULL, trend.d, trend.l,
             cov.model, cov.pars, kappa = 0.5, nugget = 0, micro.scale = 0,
             dist.epsilon = 1e-10, aniso.pars = NULL, lambda = 1, signal = FALSE,
             n.samples.backtransform = 500, n.sim = 0), messages.screen = TRUE,
           ...)
{
  call.fc <- match.call()
  ##
  ## reading input
  ##
  cov.model <- krige$cov.model
  kappa <- krige$kappa
  lambda <- krige$lambda
  ##
  beta <- krige$beta
  cov.pars <- krige$cov.pars
  nugget <- krige$nugget
  ##
  signal <- krige$signal
  n.sim <- krige$n.sim
  n.samples.backtransform <- krige$n.samples.backtransform
  micro.scale <- krige$micro.scale
  ##
  ## checking input
  ##
  if(micro.scale > nugget) stop(
                                "micro.scale must be in the interval [0, nugget]"
                                )
  if(krige$type.krige != "ok" & krige$type.krige != "OK" & krige$
     type.krige != "o.k." & krige$type.krige != "O.K." & krige$
     type.krige != "sk" & krige$type.krige != "SK" & krige$
     type.krige != "s.k." & krige$type.krige != "S.K.")
    stop("wrong option in the argument type.krige. It should be \"OK\" or \"SK\"(if ordinary or simple kriging is to be performed)"
         )
  if(krige$type.krige == "ok" | krige$type.krige == "OK" | krige$
     type.krige == "o.k." | krige$type.krige == "O.K.")
    beta.prior <- "flat"
  if(krige$type.krige == "sk" | krige$type.krige == "SK" | krige$
     type.krige == "s.k." | krige$type.krige == "S.K.") {
    if(is.null(beta) | !is.numeric(beta))
      stop("argument beta must be provided in order to perform simple kriging"
           )
    beta.prior <- "deg"
  }
  ##
  if(is.vector(coords)) {
    coords <- cbind(coords, 0)
    warning("vector of coordinates: one spatial dimention assumed")
  }
  coords <- as.matrix(coords)
  if(is.vector(locations)) {
    if(length(locations) == 2) {
      locations <- t(as.matrix(locations))
      if(messages.screen)
        warning("assuming that there is 1 prediction point"
                )
    }
    else {
      warning("vector of locations: one spatial dimention assumed"
              )
      locations <- as.matrix(cbind(locations, 0))
    }
  }
  else locations <- as.matrix(locations)
  dimnames(coords) <- list(NULL, NULL)
  dimnames(locations) <- list(NULL, NULL)
  ##
  if(inherits(krige$trend.d, "formula") | inherits(krige$trend.l,
                                                   "formula")) {
    if((inherits(krige$trend.d, "formula") == FALSE) | (inherits(
                  krige$trend.l, "formula") == FALSE))
      stop("krige$trend.d and krige$trend.l must have similar specification"
           )
    if(messages.screen)
      cat("krige.conv: Kriging with external trend to be performed using covariates provided by the user\n"
          )
  }
  else {
    if(krige$trend.d != krige$trend.l) {
      stop("krige$trend.l is different from krige$trend.d")
    }
    if(messages.screen) {
      if(krige$trend.d == "cte" & krige$type.krige == "sk")
        cat("krige.conv: Simple kriging to be performed with constant mean provided by the user\n"
            )
      if(krige$trend.d == "cte" & krige$type.krige == "ok")
        cat("krige.conv: Ordinary kriging to be performed filtering a constant mean\n"
            )
      if(krige$trend.d == "1st")
        cat("krige.conv: Trend (or universal) kriging to be performed filtering a 1st degree polinomial trend\n"
            )
      if(krige$trend.d == "2nd")
        cat("krige.conv: Trend (or universal) kriging to be performed filtering a 2nd degree polinomial trend\n"
            )
    }
  }
  trend.d <- trend.spatial(trend = krige$trend.d, coords = coords)
  beta.size <- ncol(trend.d)
  trend.l <- trend.spatial(trend = krige$trend.l, coords = locations)
  ##
  ## Anisotropy correction (should be placed AFTER trend.d/trend.l
  ##
  if(!is.null(krige$aniso.pars)) {
    if(length(krige$aniso.pars) != 2 | !is.numeric(krige$aniso.pars
               ))
      stop("anisotropy parameters must be provided as a numeric vector with two elements: the rotation angle (in radians) and the anisotropy ratio (a number greater than 1)"
           )
    if(messages.screen)
      cat("krige.conv: anisotropy correction performed\n")
    coords <- coords.aniso(coords = coords, aniso.pars = krige$
                           aniso.pars)
    locations <- coords.aniso(coords = locations, aniso.pars =
                              krige$aniso.pars)
  }
  ##
  ## Box-Cox transformation
  ##
  if(lambda != 1) {
    if(messages.screen)
      cat("krige.conv: Box-Cox's transformation (Box-Cox) performed on the data.\n"
          )
    if(lambda == 0)
      data <- log(data)
    else data <- ((data^lambda) - 1)/lambda
  }
  ##
  ## setting covariance parameters
  ##
  tausq <- nugget
  if(is.vector(cov.pars)) {
    sigmasq <- cov.pars[1]
    phi <- cov.pars[2]
  }
  else {
    sigmasq <- cov.pars[, 1]
    phi <- cov.pars[, 2]
  }
  sill.partial <- micro.scale + sum(sigmasq)
  sill.total <- tausq + sum(sigmasq)
  n <- length(data)
  ni <- nrow(trend.l)
  ##
  ## starting kriging calculations
  ##
  kc.result <- list()
  invcov <- varcov.spatial(coords = coords, cov.model = cov.model, kappa
                           = kappa, nugget = nugget, cov.pars = cov.pars, inv = TRUE,
                           only.inv.lower.diag = TRUE)
  temp <- as.double(rep(0, beta.size * beta.size))
  temp <- .C("bilinearform_XAY",



             as.double(invcov$lower.inverse),
             as.double(invcov$diag.inverse),
             as.double(as.vector(trend.d)),
             as.double(as.vector(trend.d)),
             as.integer(beta.size),
             as.integer(beta.size),
             as.integer(n),
             res = temp)$res
  attr(temp, "dim") <- c(beta.size, beta.size)
  ittivtt <- solve(temp)
  temp <- NULL
  if(beta.prior == "flat") {
    temp <- as.double(rep(0, beta.size))
    temp <- .C("bilinearform_XAY",



               as.double(invcov$lower.inverse),
               as.double(invcov$diag.inverse),
               as.double(as.vector(trend.d)),
               as.double(as.vector(data)),
               as.integer(beta.size),
               as.integer(1),
               as.integer(n),
               res = temp)$res
    beta.flat <- ittivtt %*% temp
    temp <- NULL
  }
  v0mat <- as.double(rep(0, n * ni))
  v0mat <- .C("loccoords",



              as.double(as.vector(locations[, 1])),
              as.double(as.vector(locations[, 2])),
              as.double(as.vector(coords[, 1])),
              as.double(as.vector(coords[, 2])),
              as.integer(ni),
              as.integer(n),
              res = v0mat)$res
  attr(v0mat, "dim") <- c(n, ni)
  if(n.sim > 0) {
    ## checking coincident data points and prediction locations
    loc.coincide <- apply(v0mat, 2,
                          function(x, min.dist){any(x < min.dist)},
                          min.dist = krige$dist.epsilon)
    if(any(loc.coincide))
      loc.coincide <- (1:ni)[loc.coincide]
    else loc.coincide <- NULL
    if(!is.null(loc.coincide)) {
      temp.f <- function(x, data, dist.eps)
        {
          return(data[x < dist.eps])
        }
      data.coincide <- apply(v0mat[, loc.coincide, drop=FALSE], 2, temp.f,
                             data = data, dist.eps = krige$dist.epsilon)
    }
    else data.coincide <- NULL
  }
  else locations <- NULL
  if(signal){
    v0mat <- ifelse(v0mat < krige$dist.epsilon, sill.partial,
                    cov.spatial(obj = v0mat, cov.model = cov.model,
                                kappa = kappa, cov.pars = cov.pars))
  }
  else{
    v0mat <- ifelse(v0mat < krige$dist.epsilon, sill.total,
                    cov.spatial(obj = v0mat, cov.model = cov.model,
                                kappa = kappa, cov.pars = cov.pars))
  }
  tv0ivv0 <- as.double(rep(0, ni))
  tv0ivv0 <- .C("diag_quadraticform_XAX",



                as.double(invcov$lower.inverse),
                as.double(invcov$diag.inverse),
                as.double(as.vector(v0mat)),
                as.integer(ni),
                as.integer(n),
                res = tv0ivv0)$res
  tb <- as.double(rep(0, ni * beta.size))
  tb <- .C("bilinearform_XAY",



           as.double(invcov$lower.inverse),
           as.double(invcov$diag.inverse),
           as.double(as.vector(v0mat)),
           as.double(as.vector(trend.d)),
           as.integer(ni),
           as.integer(beta.size),
           as.integer(n),
           res = tb)$res
  attr(tb, "dim") <- c(ni, beta.size)
  tb <- trend.l - tb
  if(beta.prior == "deg") {
    tv0ivdata <- as.double(rep(0, ni))
    tv0ivdata <- .C("bilinearform_XAY",



                    as.double(invcov$lower.inverse),
                    as.double(invcov$diag.inverse),
                    as.double(as.vector(v0mat)),
                    as.double(data),
                    as.integer(ni),
                    as.integer(1),
                    as.integer(n),
                    res = tv0ivdata)$res
    if(n.sim == 0) {
      remove(list = c("v0mat", "invcov"), frame = sys.nframe())
    }
    kc.result$predict <- tv0ivdata + as.vector(tb %*% beta)
    if(n.sim == 0) {
      remove(list = c("tb", "tv0ivdata"), frame = sys.nframe())
    }
    else {
      remove(list = c("tv0ivdata"), frame = sys.nframe())
    }
    if(krige$signal)
      kc.result$krige.var <- as.vector(sill.partial - tv0ivv0
                                       )
    else kc.result$krige.var <- as.vector(sill.total - tv0ivv0)
    beta.est <- "Simple kriging performed (beta provided by user)"
  }
  if(beta.prior == "flat") {
    tv0ivdata <- as.double(rep(0, ni))
    tv0ivdata <- .C("bilinearform_XAY",



                    as.double(invcov$lower.inverse),
                    as.double(invcov$diag.inverse),
                    as.double(as.vector(v0mat)),
                    as.double(data),
                    as.integer(ni),
                    as.integer(1),
                    as.integer(n),
                    res = tv0ivdata)$res
    if(n.sim == 0) {
      remove(list = c("v0mat", "invcov"), frame = sys.nframe())
    }
    kc.result$predict <- tv0ivdata + as.vector(tb %*% beta.flat)
    tv0ivdata <- NULL
    if(beta.size == 1)
      bitb <- as.vector(tb^2) * as.vector(ittivtt)
    else {
      bitb <- as.double(rep(0, ni))
      bitb <- .C("diag_quadraticform_XAX",



                 as.double(ittivtt[lower.tri(ittivtt)]),
                 as.double(diag(ittivtt)),
                 as.double(as.vector(t(tb))),
                 as.integer(ni),
                 as.integer(beta.size),
                 res = bitb)$res
    }
    if(krige$signal)
      kc.result$krige.var <- as.vector(sill.partial - tv0ivv0 +
                                       bitb)
    else kc.result$krige.var <- as.vector(sill.total - tv0ivv0 + bitb
                                          )
    kc.result$beta.est <- beta.flat
    remove("bitb", frame=sys.nframe())
  }
  remove("tv0ivv0", frame=sys.nframe())
  if(any(round(kc.result$krige.var, dig=12) < 0))
    cat("krige.conv: negative kriging variance found! Investigate why this is happening.\n"
        )
  message <- "krige.conv: Kriging performed using global neighbourhood"
  if(messages.screen)
    cat(paste(message, "\n"))
############## Sampling from the resulting distribution #####################
  if(n.sim > 0) {
    if(messages.screen)
      cat("krige.conv: sampling from the predictive distribution (conditional simulations)\n"
        )
    if(length(cov.pars) > 2) {
      reduce.var <- as.double(rep(0, ni * ni))
      reduce.var <- .C("bilinearform_XAY",



                       as.double(invcov$lower.inverse),
                       as.double(invcov$diag.inverse),
                       as.double(as.vector(v0mat)),
                       as.double(as.vector(v0mat)),
                       as.integer(ni),
                       as.integer(ni),
                       as.integer(n),
                       res = reduce.var)$res
      attr(reduce.var, "dim") <- c(ni, ni)
      remove(list = c("v0mat", "invcov"), frame = sys.nframe())
      if(beta.prior == "flat") {
        if(beta.size == 1)
          ok.add.var <- outer(as.vector(tb),
                              as.vector(tb)) * as.vector(
                                                         ittivtt)
        else {
          b <- t(tb)
          tb <- NULL
          ok.add.var <- as.double(rep(0, ni *
                                      ni))
          ok.add.var <- .C("bilinearform_XAY",




                           as.double(ittivtt[lower.tri(
                                                       ittivtt)]),
                           as.double(diag(ittivtt)),
                           as.double(as.vector(b)),
                           as.double(as.vector(b)),
                           as.integer(ni),
                           as.integer(ni),
                           as.integer(beta.size),
                           res = ok.add.var)$res
          attr(ok.add.var, "dim") <- c(ni, ni)
          b <- NULL
        }
        reduce.var <- reduce.var + ok.add.var
        remove("ok.add.var", frame = sys.nframe())
      }
      varcov <- varcov.spatial(coords = locations, cov.model
                               = cov.model, cov.pars = cov.pars, kappa =
                               kappa, nugget = nugget)$varcov
      varcov <- round(varcov, dig = 12) - reduce.var
      remove("reduce.var", frame = sys.nframe())
      kc.result$simulations <- kc.result$predict + (t(chol(
                                                           varcov)) %*% matrix(rnorm(ni * n.sim), ncol =
                                                                               n.sim))
    }
    else {
      if(((round(1000000000000. * nugget) == 0) | signal) &
         (!is.null(loc.coincide))) {
        v0mat <- v0mat[, - (loc.coincide)]
        nloc <- ni - length(loc.coincide)
        tmean.coincide <- kc.result$predict[
                                            loc.coincide]
        tmean <- kc.result$predict[ - (loc.coincide)]
        tb <- tb[ - (loc.coincide), ]
      }
      else {
        nloc <- ni
        tmean <- kc.result$predict
      }
      normalsc <- rnorm((nloc * n.sim))
      if(signal)
        Dval <- 1. + micro.scale
      else Dval <- 1. + (nugget/cov.pars[1])
      if(beta.size == 1) {
        Blower <- 0
        if(beta.prior == "flat")
          Bdiag <- ittivtt
        else Bdiag <- 0.
      }
      else {
        Blower <- ittivtt[lower.tri(ittivtt)]
        if(beta.prior == "flat")
          Bdiag <- diag(ittivtt)
        else Bdiag <- rep(0, beta.size)
      }
      R0 <- as.double(rep(0., (nloc * (nloc + 1))/2))
      if(((round(1000000000000. * nugget) == 0) | signal) &
         (!is.null(loc.coincide))) {
        R0 <- .C("distdiag",



                 as.double(locations[ - (loc.coincide),
                                     1]),
                 as.double(locations[ - (loc.coincide),
                                     2]),
                 as.integer(nloc),
                 out = R0)$out
      }
      else R0 <- .C("distdiag",



                    as.double(locations[, 1]),
                    as.double(locations[, 2]),
                    as.integer(ni),
                    out = R0)$out
      locations <- NULL
      R0 <- cov.spatial(R0, cov.pars = cov.pars, cov.model =
                        cov.model, kappa = kappa)
      normalsc <- .C("kb_sim",




                     as.double(tmean),
                     out = as.double(as.vector(normalsc)),
                     as.double(invcov$lower.inverse),
                     as.double(invcov$diag.inverse),
                     as.double(as.vector(v0mat)),
                     as.integer(nloc),
                     as.integer(n),
                     as.double(Dval),
                     as.integer(n.sim),
                     as.double(rep(1, n.sim)),
                     as.double(cov.pars[1]),
                     as.double(Blower),
                     as.double(Bdiag),
                     as.double(as.vector(t(tb))),
                     as.integer(beta.size),
                     as.double(R0))$out
      attr(normalsc, "dim") <- c(nloc, n.sim)
      remove(list = c("v0mat", "invcov", "R0", "tb"), frame
             = sys.nframe())
      if(((round(1000000000000. * nugget) == 0) | signal) &
         (!is.null(loc.coincide))) {
        kc.result$simulations <- matrix(0, nrow = ni,
                                        ncol = n.sim)
        kc.result$simulations[ - (loc.coincide), ] <-
          normalsc
        kc.result$simulations[loc.coincide, ] <- rep(
                                                      tmean.coincide, n.sim)
      }
      else kc.result$simulations <- normalsc
      normalsc <- NULL
    }
    if(lambda != 1) {
      cat("krige.conv: back-transforming the simulated values\n"
          )
      if(any(kc.result$simulations < -1/lambda))
        warning("Truncation in the back-transformation: there are simulated values less than (- 1/lambda) in the normal scale."
                )
      if(lambda == 0)
        kc.result$simulations <- ifelse(kc.result$
                                        simulations > -1/lambda, exp(kc.result$
                                                                     simulations), -1/lambda)
      if(lambda > 0)
        kc.result$simulations <- ifelse(kc.result$
                                        simulations > -1/lambda, ((kc.result$
                                                                   simulations * lambda) + 1)^(1/lambda),
                                        -1/lambda)
      if(lambda < 0)
        warning("back transformation not performed (negative value of lambda)"
                )
    }
  }
######## Back - transforming predictions###############################
  if(lambda != 1) {
    if(messages.screen)
      cat("krige.conv: back-transforming the predictions according to the (Box-Cox) parameter lambda\n"
          )
    kc.result$transf.predict <- kc.result$predict
    kc.result$transf.krige.var <- kc.result$krige.var
    if(lambda == 0 & beta.prior == "deg") {
      ## don't change the order of the next two commands!!!'
      kc.result$predict <- exp(kc.result$transf.predict +
                               0.5 * kc.result$krige.var)
      kc.result$krige.var <- (kc.result$predict^2) * (exp(
                                                          kc.result$krige.var) - 1)
    }
    if(lambda > 0 | (lambda == 0 & beta.prior == "flat")) {
      if(messages.screen)
        cat("krige.conv: back-transformation done by sampling from the resulting (normal) predictive distribution (inspect results carefully, run the function more than once and check for stability of the results\n"
            )
      ap.warn <- options()$warn
      options(warn = -1)
      temp.data <- matrix(rnorm(ni * n.samples.backtransform,
                                mean = kc.result$predict, sd = sqrt(kc.result$
                                                            krige.var)), nrow = ni)
      options(warn = ap.warn)
      ind.zero <- (round(1000000000000. * kc.result$
                         transf.krige.var) == 0)
      temp.data[ind.zero, ] <- kc.result$transf.predict[
                                                         ind.zero]
      ind.zero <- NULL
      if(lambda == 0)
        temp.data <- exp(temp.data)
      else {
        temp.data[temp.data < -1/lambda] <- -1/lambda
        temp.data <- ((temp.data * lambda) + 1)^(1/
                                                 lambda)
      }
      kc.result$predict <- as.vector(apply(temp.data, 1,
                                           mean))
      kc.result$krige.var <- as.vector(apply(temp.data, 1,
                                             var))
    }
    if(lambda < 0) {
      cat("krige.conv: resulting distribution has no mean for lambda < 0 - back transformation not performed\n"
          )
    }
  }
  kc.result <- c(kc.result, list(message = message, call = call.fc))
############################
  class(kc.result) <- "kriging"
  return(kc.result)
}

"ksline" <-
function(geodata, coords = geodata$coords, data = geodata$data, locations,
        cov.pars = stop("covariance parameters (sigmasq and phi) needed"),
        nugget = 0, micro.scale = 0, cov.model = c("exponential", "matern",
        "gaussian", "spherical", "cubic", , "wave", "powered.exponential",
        "cauchy", "gneiting", "gneiting.matern", "pure.nugget"), kappa = 0.5,
        lambda = 1, n.samples.backtransform = 500, m0 = "ok", nwin = "full",
        trend = 1, d = 2, ktedata = NULL, ktelocations = NULL, aniso.pars = F,
        signal = F, messages.screen = T, dist.epsilon = 1e-10)
{
        ##
        ###### 1. Reading input information #####
        ##
        call.fc <- match.call()
        if(lambda != 1) {
                if(messages.screen == T)
                        cat("ksline: Data transformation (Box-Cox) performed.\n"
                                )
                if(lambda == 0)
                        data <- log(data)
                else data <- ((data^lambda) - 1)/lambda
        }
        cov.model <- match.arg(cov.model)
        coords <- as.matrix(coords)
        locations <- as.matrix(locations)
        dimnames(coords) <- list(NULL, NULL)
        dimnames(locations) <- list(NULL, NULL)
        if(!is.null(ktedata) & !is.null(ktelocations) & m0 != "kte") {
                cat("ksline: external variable (covariate) provided. Kriging ste to KTE\n"
                        )
                m0 <- "kte"
        }
        ##
        ##
        ## Correcting anisotropy
        ##
        if(any(aniso.pars) == T) {
                coords.c <- coords.aniso(coord = coords, aniso.pars =
                        aniso.pars)
                locations.c <- coords.aniso(coord = locations, aniso.pars =
                        aniso.pars)
        }
        else {
                coords.c <- coords
                locations.c <- locations
        }
        ##### 2. Preparing KTE matrices #####
        ##
        if(m0 == "kte") {
                ktedata <- as.matrix(ktedata)
                ktelocations <- as.matrix(ktelocations)
                dimnames(ktedata) <- list(NULL, NULL)
                dimnames(ktelocations) <- list(NULL, NULL)
        }
        ##
        ##### 3. Preparing matrices to store results ######
        ##
        n <- length(data)
        ni <- length(locations[, 1])
        tausq <- nugget
        sigmasq <- cov.pars[1]
        phi <- cov.pars[2]
        ##
        ##### 4. Kriging with full neighbourhood #####
        ##
        if(nwin == "full") {
                est <- rep(0, ni)
                dif <- rep(0, ni)
                kvar <- rep(0, ni)
                sumw <- rep(0, ni)
                wofmean <- rep(0, ni)
                iv <- varcov.spatial(coords = coords.c, cov.model = cov.model,
                        kappa = kappa, nugget = nugget, cov.pars = cov.pars,
                        inv = T, det = F, func.inv = "chol",
                        only.decomposition = F)$inverse
                av <- mean(data)
                sd <- sqrt(var(data))
                one <- rep(1, n)
                tone <- t(one)
                toneiv <- crossprod(one, iv)
                den <- solve(toneiv %*% one)
                ml <- den %*% toneiv %*% data
                kmsd <- sqrt(den)
                means <- c(average = av, stdev = sd, kmean = ml, kmsd = kmsd)
                if(m0 != "kt") {
                        mktlocations <- "Constant trend"
                        beta <- ml
                }
                else {
                        mktlocations <- rep(0, ni)
                        if(m0 == "kt" & trend == 1) {
                                if(d == 1) {
                                        xmat <- cbind(rep(1, n), coords[, 2])
                                        xmati <- cbind(rep(1, ni), locations[
                                                , 2])
                                }
                                else {
                                        xmat <- cbind(rep(1, n), coords[, 1],
                                                coords[, 2])
                                        xmati <- cbind(rep(1, ni), locations[
                                                , 1], locations[, 2])
                                }
                                txiv <- crossprod(xmat, iv)
                                iviv <- solve(txiv %*% xmat)
                                beta <- iviv %*% txiv %*% data
                                mkt <- xmat %*% beta
                        }
                        if(m0 == "kt" & trend == 2) {
                                if(d == 1) {
                                        xmat <- cbind(rep(1, n), coords[, 2],
                                                (coords[, 2])^2)
                                        xmati <- cbind(rep(1, ni), locations[
                                                , 2], (locations[, 2])^2)
                                }
                                else {
                                        xmat <- cbind(rep(1, n), coords[, 1],
                                                coords[, 2], (coords[, 1])^
                                                2, (coords[, 2])^2, coords[
                                                , 1] * coords[, 2])
                                        xmati <- cbind(rep(1, ni), locations[
                                                , 1], locations[, 2], (
                                                locations[, 1])^2, (locations[
                                                , 2])^2, locations[, 1] *
                                                locations[, 2])
                                }
                                txiv <- crossprod(xmat, iv)
                                iviv <- solve(txiv %*% xmat)
                                beta <- iviv %*% txiv %*% data
                                mkt <- xmat %*% beta
                        }
                }
                ##
                #### 4.1 Preparing matrices for KT (UK) #####
                ##
                if(m0 != "kte") mktelocations <- "No external trend" else {
                        if(m0 == "kte") {
                                ##sxy <- paste(coords[,1],coords[,2],sep=',') : the data coordinates
                                ##sxyi <- paste(coordsi[,1],coordsi[,2],sep=',') : the locations+est.drift
                                ##row.names(xyi) <- scoordsi
                                ##coordsi[scoords,]
                                mktelocations <- rep(0, ni)
                                xmat <- cbind(rep(1, n), ktedata)
                                xmati <- cbind(rep(1, ni), ktelocations)
                                txiv <- crossprod(xmat, iv)
                                iviv <- solve(txiv %*% xmat)
                                beta <- iviv %*% txiv %*% data
                                mkte <- xmat %*% beta
                        }
                }
                for(i in 1:ni) {
                        if(messages.screen == T) {
                                if(ni < 11)
                                        cat(paste("ksline: kriging location: ",
                                                i, "out of", ni, "\n"))
                                else {
                                        if(ni < 101 & (i %% 10 == 1))
                                                cat(paste(
                                                        "ksline: kriging location: ",
                                                        i, "out of", ni, "\n"))
                                        if(ni > 100 & i %% 100 == 1)
                                                cat(paste(
                                                        "ksline: kriging location: ",
                                                        i, "out of", ni, "\n"))
                                        if(i == ni)
                                                cat(paste(
                                                        "ksline: kriging location: ",
                                                        i, "out of", ni, "\n"))
                                }
                        }
                        coords0 <- cbind((coords.c[, 1] - locations.c[i, 1]),
                                (coords.c[, 2] - locations.c[i, 2]))
                        dm0 <- sqrt(coords0[, 1]^2 + coords0[, 2]^2)
                        ## coords0 <- rbind(locations[i, ], coords)
                        ## dm0 <- dist(coords0)[1:n]
                        v0 <- cov.spatial(obj = dm0, cov.model = cov.model,
                                kappa = kappa, cov.pars = cov.pars)
                        if(micro.scale == 0)
                                v0[dm0 < dist.epsilon] <- sigmasq
                        else v0[dm0 < dist.epsilon] <- micro.scale + sigmasq
                        tv0 <- t(v0)
                        v0iv <- crossprod(v0, iv)
                        v0ivv0 <- v0iv %*% v0
                        skw <- v0iv
                        wofmean[i] <- 1 - sum(skw)
                        ##
                        ###### 4.2.1 Simple kriging with known mean
                        ##
                        if(is.numeric(m0) == T) {
                                dif[i] <- skw %*% (data - m0)
                                est[i] <- m0 + dif[i]
                                if(signal == T)
                                        kvar[i] <- sigmasq - v0ivv0
                                else kvar[i] <- tausq + sigmasq - v0ivv0
                                sumw[i] <- sum(skw)
                        }
                        ##
                        ###### 4.2.2 Simple kriging with data average mean
                        ##
                        if(m0 == "av") {
                                dif[i] <- skw %*% (data - av)
                                est[i] <- av + dif[i]
                                if(signal == T)
                                        kvar[i] <- sigmasq - v0ivv0
                                else kvar[i] <- tausq + sigmasq - v0ivv0
                                sumw[i] <- sum(((tone/n) + skw - ((skw %*% one %*%
                                        tone)/n)))
                        }
                        ##
                        ###### 4.2.3 Ordinary kriging (or SK with G.L.S. mean)
                        ##
                        if(m0 == "ok") {
                                dif[i] <- skw %*% (data - ml)
                                est[i] <- ml + dif[i]
                                redu <- as.vector(1 - toneiv %*% v0)
                                if(signal == T)
                                        kvar[i] <- sigmasq - v0ivv0 + (redu %*%
                                                den %*% redu)
                                else kvar[i] <- tausq + sigmasq - v0ivv0 + (
                                                redu %*% den %*% redu)
                                sumw[i] <- sum((den %*% one + tv0 - v0iv %*%
                                        one %*% den %*% tone) %*% iv)
                        }
                        ##
                        ###### 4.2.4 Universal Kriging (or Kriging with trend model)
                        ##
                        if(m0 == "kt") {
                                dif[i] <- skw %*% (data - mkt)
                                est[i] <- xmati[i, ] %*% beta + dif[i]
                                redu <- as.vector(xmati[i, ]) - as.vector(
                                        txiv %*% v0)
                                if(signal == T)
                                        kvar[i] <- sigmasq - v0ivv0 + (redu %*%
                                                iviv %*% redu)
                                else kvar[i] <- tausq + sigmasq - v0ivv0 + (
                                                redu %*% iviv %*% redu)
                                sumw[i] <- sum(skw + xmati[i, ] %*% iviv %*%
                                        txiv - skw %*% xmat %*% iviv %*% txiv)
                                mktlocations[i] <- xmati[i, ] %*% beta
                        }
                        ##
                        ###### 4.2.5 Kriging with external trend
                        ##
                        if(m0 == "kte") {
                                dif[i] <- skw %*% (data - mkte)
                                est[i] <- xmati[i, ] %*% beta + dif[i]
                                redu <- as.vector(xmati[i, ]) - as.vector(
                                        txiv %*% v0)
                                if(signal == T)
                                        kvar[i] <- sigmasq - v0ivv0 - (redu %*%
                                                iviv %*% redu)
                                else kvar[i] <- tausq + sigmasq - v0ivv0 + (
                                                redu %*% iviv %*% redu)
                                sumw[i] <- sum(skw + xmati[i, ] %*% iviv %*%
                                        txiv - skw %*% xmat %*% iviv %*% txiv)
                                mktelocations[i] <- xmati[i, ] %*% beta
                        }
                }
                message <- "Kriging performed in global neighbourhood"
                if(messages.screen == T)
                        cat(paste(message, "\n"))
                results <- list(predict = est, krige.var = kvar, dif = dif,
                        summary = means, ktrend = mktlocations, ktetrend =
                        mktelocations, beta = beta, wofmean = wofmean)
        }
        else {
                ##
                #### 5.1 Preparing vectors to store results and matrices #####
                ##
                nwin <- min(n, nwin)
                avwin <- rep(0, ni)
                sdwin <- rep(0, ni)
                mlwin <- rep(0, ni)
                kmsdwin <- rep(0, ni)
                estwin <- rep(0, ni)
                difwin <- rep(0, ni)
                kvarwin <- rep(0, ni)
                sumwwin <- rep(0, ni)
                wofmean <- rep(0, ni)
                if(m0 != "kt")
                        mkt <- "Constant position trend"
                else mkt <- rep(0, ni)
                if(m0 != "kte")
                        mkte <- "No external trend"
                else mkte <- rep(0, ni)
                if(m0 != "kt" & m0 != "kte")
                        betawin <- "No polynomial or external trend"
                if(m0 == "kt") {
                        if(trend == 1) {
                                if(d == 1)
                                        xmati <- cbind(rep(1, ni), locations[
                                                , 2])
                                else xmati <- cbind(rep(1, ni), locations[
                                                , 1], locations[, 2])
                        }
                        if(trend == 2) {
                                if(d == 1)
                                        xmati <- cbind(rep(1, ni), locations[
                                                , 2], locations[, 2]^2)
                                else xmati <- cbind(rep(1, ni), locations[
                                                , 1], locations[, 2], (
                                                locations[, 1])^2, (locations[
                                                , 2])^2, locations[, 1] *
                                                locations[, 2])
                        }
                        betawin <- matrix(0, nrow = (ncol(xmati) * ni), ncol =
                                ncol(xmati))
                }
                if(m0 == "kte") {
                        xmati <- cbind(rep(1, ni), ktelocations)
                        if(is.vector(ktedata) == T)
                                betawin <- matrix(0, nrow = (2 * ni), ncol = 2)
                        else betawin <- matrix(0, nrow = ((ncol(ktedata) + 1) *
                                        ni), ncol = (ncol(ktedata) + 1))
                }
                ##
                #### 5.2 Kriging calculations: #####
                ##
                for(i in 1:ni) {
                        temp.win <- ksline.aux1(coords = coords, coords.c =
                                coords.c, data = data, n = n, locations =
                                locations[i, ], locations.c = locations.c[
                                i, ], cov.pars = cov.pars, nugget = nugget,
                                cov.model = cov.model, kappa = kappa, m0 = m0,
                                nwin = nwin, trend = trend, d = d, ktedata =
                                ktedata, ktelocations = ktelocations,
                                micro.scale = micro.scale, location.number = i,
                                xmati = xmati[i, ], mkte = NULL, mkt = NULL,
                                betawin = NULL, signal = signal, dist.epsilon
                                 = dist.epsilon)
                        avwin[i] <- temp.win$avwin
                        sdwin[i] <- temp.win$sdwin
                        mlwin[i] <- temp.win$mlwin
                        kmsdwin[i] <- temp.win$kmsdwin
                        estwin[i] <- temp.win$estwin
                        difwin[i] <- temp.win$difwin
                        kvarwin[i] <- temp.win$kvarwin
                        sumwwin[i] <- temp.win$sumwwin
                        wofmean[i] <- temp.win$wofmean
                        if(m0 == "kt")
                                mkt[i] <- temp.win$mkt
                        if(m0 == "kte")
                                mkte[i] <- temp.win$mkte
                        if(m0 == "kt" | m0 == "kte")
                                betawin[i, ] <- temp.win$betawin
                        if(messages.screen == T) {
                                if(ni < 11)
                                        cat(paste("ksline: kriging location: ",
                                                i, "out of", ni, "\n"))
                                else {
                                        if(ni < 101 & (i %% 10 == 1))
                                                cat(paste(
                                                        "ksline: kriging location: ",
                                                        i, "out of", ni, "\n"))
                                        if(ni > 100 & i %% 100 == 1)
                                                catt(paste(
                                                        "ksline: kriging location: ",
                                                        i, "out of", ni, "\n"))
                                        if(i == ni)
                                                cat(paste(
                                                        "ksline: kriging location: ",
                                                        i, "out of", ni, "\n"))
                                }
                        }
                }
                message <- "kriging performed in moving neighbourhood"
                if(messages.screen == T)
                        print(paste(message, "\n"))
                results <- list(predict = estwin, krige.var = kvarwin, dif =
                        difwin, avtrend = avwin, sd = sdwin, oktrend = mlwin,
                        oksd = kmsdwin, ktrend = mkt, ktetrend = mkte, beta =
                        betawin, sumw = sumwwin, wofmean = wofmean)
        }
        results$krige.var <- round(results$krige.var, dig = 8)
        if(lambda != 1) {
                if(messages.screen == T)
                        cat("ksline: Back-transforming the predictions according to the (Box-Cox) parameter lambda\n"
                                )
                if(lambda == 0) {
                        predict.transf <- results$predict
                        results$predict <- exp(predict.transf) - 0.5 * results$
                                krige.var
                        results$krige.var <- (exp(2 * predict.transf - results$
                                krige.var)) * (exp(results$krige.var) - 1)
                }
                if(lambda > 0) {
                        if(messages.screen == T)
                                cat("ksline: Back-transformation done by sampling from the resulting (normal) predictive distribution\n"
                                        )
                        ap.warn <- options()$warn
                        options(warn = -1)
                        temp.data <- matrix(rnorm(ni * n.samples.backtransform,
                                mean = results$predict, sd = sqrt(results$
                                krige.var)), nrow = ni)
                        options(warn = ap.warn)
                        temp.data[(results$krige.var == 0), ] <- results$
                                predict[(results$krige.var == 0)]
                        temp.data[temp.data < -1/lambda] <- -1/lambda
                        temp.data <- ((temp.data * lambda) + 1)^(1/lambda)
                        # temp.data[is.na(temp.data)] <- Inf
                        results$predict <- as.vector(apply(temp.data, 1, mean))
                        results$krige.var <- as.vector(apply(temp.data, 1,
                                var))
                }
                if(lambda < 0) {
                        cat("ksline: Resulting distribution has no mean for lambda < 0 - back transformation not performed\n"
                                )
                }
        }
        results$locations <- locations
        results$message <- message
        results$call <- call.fc
        if(version$major <= 4)
          class(results) <- c("kriging", "ksline")
        else oldClass(results) <- c("kriging", "ksline")
        return(results)
}

"ksline.aux1" <-
function(coords, coords.c, data, locations, locations.c, n, cov.pars, nugget,
        cov.model, kappa, m0, nwin, trend, d, ktedata, ktelocations,
        micro.scale, messages, location.number, xmati, mkte, mkt, betawin,
        signal, dist.epsilon = dist.epsilon)
{
        i <- location.number
        sigmasq <- cov.pars[1]
        phi <- cov.pars[2]
        tausq <- nugget
        coords0 <- cbind((coords.c[, 1] - locations.c[1]), (coords.c[, 2] -
                locations.c[2]))
        dm0 <- sqrt(coords0[, 1]^2 + coords0[, 2]^2)
        coordswin <- coords[order(dm0)[1:nwin], ]
        coordswin.c <- coords.c[order(dm0)[1:nwin], ]
        datawin <- data[order(dm0)[1:nwin]]
        ivwin <- varcov.spatial(coords = coordswin.c, cov.model = cov.model,
                kappa = kappa, nugget = nugget, cov.pars = cov.pars, inv = T,
                det = F, func.inv = "chol", only.decomposition = F)$inverse
        avwin <- mean(datawin)
        sdwin <- sqrt(var(datawin))
        onewin <- rep(1, nwin)
        toneivwin <- crossprod(onewin, ivwin)
        denwin <- solve(toneivwin %*% onewin)
        mlwin <- denwin %*% toneivwin %*% datawin
        kmsdwin <- sqrt(denwin)
        coords0win <- cbind((coordswin[, 1] - locations[1]), (coordswin[, 2] -
                locations[2]))
        coords0win.c <- cbind((coordswin.c[, 1] - locations.c[1]), (coordswin.c[
                , 2] - locations.c[2]))
        dm0win <- sqrt(coords0win.c[, 1]^2 + coords0win.c[, 2]^2)
        v0win <- cov.spatial(obj = dm0win, cov.model = cov.model, kappa = kappa,
                cov.pars = cov.pars)
        if(micro.scale == T)
                v0win[dm0win < dist.epsilon] <- sigmasq
        else v0win[dm0win < dist.epsilon] <- micro.scale + sigmasq
        skwwin <- crossprod(v0win, ivwin)
        wofmean <- 1 - sum(skwwin)
        if(m0 == "kt" & trend == 1) {
                if(d == 1)
                        xmatwin <- cbind(rep(1, nwin), coordswin[, 2])
                else xmatwin <- cbind(rep(1, nwin), coordswin[, 1], coordswin[
                                , 2])
                txivwin <- crossprod(xmatwin, ivwin)
                ivivwin <- solve(txivwin %*% xmatwin)
                betawin <- ivivwin %*% txivwin %*% datawin
                mktwin <- xmatwin %*% betawin
                mkt <- xmati %*% betawin
        }
        if(m0 == "kt" & trend == 2) {
                if(d == 1)
                        xmatwin <- cbind(rep(1, nwin), coordswin[, 2], (
                                coordswin[, 2])^2)
                else xmatwin <- cbind(rep(1, nwin), coordswin[, 1], (coordswin[
                                , 1])^2, coordswin[, 2], (coordswin[, 2])^
                                2, coordswin[, 1] * coordswin[, 2])
                xmatwin.cent <- xmatwin
                xmatwin.cent[, 2] <- xmatwin.cent[, 2] - mean(xmatwin[, 2])
                xmatwin.cent[, 3] <- xmatwin.cent[, 3] - mean(xmatwin[, 3])
                ivivwin <- solve(crossprod(xmatwin.cent, ivwin) %*%
                        xmatwin.cent)
                txivwin <- crossprod(xmatwin.cent, ivwin)
                betawin <- ivivwin %*% txivwin %*% datawin
                betawin <- mean(datawin) - crossprod(betawin, c(0, mean(xmatwin[
                        , 2]), mean(xmatwin[, 3])))
                mktwin <- xmatwin %*% betawin
                mkt <- xmati %*% betawin
        }
        if(m0 == "kte") {
                if(is.vector(ktedata) == T)
                        ktedatawin <- ktedata[order(dm0)[1:nwin]]
                else ktedatawin <- ktedata[order(dm0)[1:nwin], ]
                xmatwin <- cbind(rep(1, nwin), ktedatawin)
                ivivwin <- solve(crossprod(xmatwin, ivwin) %*% xmatwin)
                txivwin <- crossprod(xmatwin, ivwin)
                betawin <- ivivwin %*% txivwin %*% datawin
                mktewin <- xmatwin %*% betawin
                mkte <- xmati %*% betawin
        }
        ##
        ###### Simple kriging with known mean
        ##
        if(is.numeric(m0) == T) {
                difwin <- skwwin %*% (data - m0)
                estwin <- m0win + difwin
                if(signal == T)
                        kvarwin <- sigmasq - crossprod(v0win, ivwin) %*% v0win
                else kvarwin <- tausq + sigmasq - crossprod(v0win, ivwin) %*%
                                v0win
                sumwwin <- sum(skwwin)
        }
        ##
        ###### 4.2.2 Simple kriging with data average mean
        ##
        if(m0 == "av") {
                difwin <- skwwin %*% (datawin - avwin)
                estwin <- avwin + difwin
                if(signal == T)
                        kvarwin <- sigmasq - crossprod(v0win, ivwin) %*% v0win
                else kvarwin <- tausq + sigmasq - crossprod(v0win, ivwin) %*%
                                v0win
                sumwwin <- sum(((t(onewin)/nwin) + skwwin - ((skwwin %*% onewin %*%
                        t(onewin))/n)))
        }
        ##
        ###### Ordinary kriging (or SK with G.L.S. mean)
        ##
        if(m0 == "ok") {
                difwin <- skwwin %*% (datawin - mlwin)
                estwin <- mlwin + difwin
                redu <- as.vector(1 - toneivwin %*% v0win)
                if(signal == T)
                        kvarwin <- sigmasq - v0win %*% ivwin %*% v0win + (
                                redu %*% denwin %*% redu)
                else kvarwin <- tausq + sigmasq - v0win %*% ivwin %*% v0win +
                                (redu %*% denwin %*% redu)
                sumwwin <- sum((denwin %*% onewin + t(v0win) - crossprod(v0win,
                        ivwin) %*% onewin %*% denwin %*% t(onewin)) %*% ivwin)
        }
        ##
        ###### Universal Kriging (or Kriging with trend model)
        ##
        if(m0 == "kt") {
                difwin <- skwwin %*% (datawin - mktwin)
                estwin <- mkt + difwin
                xmati <- as.vector(xmati)
                redu <- as.vector(xmati) - as.vector(txivwin %*% v0win)
                if(signal == T)
                        kvarwin <- sigmasq - v0win %*% ivwin %*% v0win + (
                                redu %*% ivivwin %*% redu)
                else kvarwin <- tausq + sigmasq - v0win %*% ivwin %*% v0win +
                                (redu %*% ivivwin %*% redu)
                sumwwin <- sum(skwwin + xmati %*% ivivwin %*% txivwin - skwwin %*%
                        xmatwin %*% ivivwin %*% txivwin)
        }
        ##
        ###### Kriging with external trend
        ##
        if(m0 == "kte") {
                difwin <- skwwin %*% (datawin - mktewin)
                estwin <- mkte + difwin
                xmati <- as.vector(xmati)
                redu <- as.vector(xmati) - as.vector(txivwin %*% v0win)
                if(signal == T)
                        kvarwin <- sigmasq - v0win %*% ivwin %*% v0win + (
                                redu %*% ivivwin %*% redu)
                else kvarwin <- tausq + sigmasq - v0win %*% ivwin %*% v0win +
                                (redu %*% ivivwin %*% redu)
                sumwwin <- sum(skwwin + xmati %*% ivivwin %*% txivwin - skwwin %*%
                        xmatwin %*% ivivwin %*% txivwin)
        }
        ##
        ##
        ##
        results <- list(avwin = avwin, sdwin = sdwin, mlwin = mlwin, kmsdwin =
                kmsdwin, wofmean = wofmean, betawin = betawin, mkt = mkt, mkte
                 = mkte, difwin = difwin, estwin = estwin, kvarwin = kvarwin,
                sumwwin = sumwwin)
        return(results)
}

"levelplot.ksline" <-
function(obj.ksline, grid = obj.ksline$grid, values = c("est", "kvar",
        "oktrend", "ktrend", "ktetrend"), ...)
{
        ##
        ## obj.ksline: tipically an object generated by the function ksline
        ## grid : regular grid where the interpolation was performed
        ## values : valeus to be plotted. The options are:
        ### - est: the kriging estimates (default option)
        ### - kvar : the kriging variances
        values <- match.arg(values)
        ## grid <- obj.ksline$grid[order(grid[, 2], grid[, 1]), ]
        ## if(nx == ny)
        ## par(pty = "s")
        if(values == "est") m <- obj.ksline$pred
        if(values == "kvar")
                m <- obj.ksline$krige.var
        levelplot(m ~ obj.ksline$grid[, 1] * obj.ksline$grid[, 2], , xlab =
                "Coord X", ylab = "Coord Y", ...)
}

"likfit.old" <-
  function (geodata, coords=geodata$coords, data=geodata$data, trend = "cte",
            ini, fix.nugget = FALSE, nugget = 0,
            cov.model = c("matern", "exponential", "gaussian", "spherical",
              "circular", "cubic", "wave", "powered.exponential",
              "cauchy", "gneiting", "gneiting.matern", "pure.nugget"),
            kappa = 0.5, fix.lambda = TRUE, lambda = 1, method = "ML",
            predicted = FALSE, residuals = FALSE,
            automatic.refit = FALSE, range.limits,
            messages.screen = TRUE, ...)
{
        ## stop("trend must be expecified as `cte`, `1st`, `2nd` or as formulae of the type: ~ X where X is a matrix with the trend variables (covariates. A formula of the type ~X-1 will define a vector of means (which can be different for each location)"
        call.fc <- match.call()
        ftau <- nugget
        fixtau <- fix.nugget
        cov.model <- match.arg(cov.model)
        if(cov.model == "pure.nugget")
                ini <- 0
        if(!is.null(kappa))
                if(cov.model == "matern" & kappa == 0.5)
                        cov.model <- "exponential"
        coords <- as.matrix(coords)
        dists.vec <- as.vector(dist(coords))
        range.dist <- range(dists.vec)
        max.dist <- max(range.dist)
        min.dist <- min(range.dist)
        if(missing(range.limits)){
          lower.phi <- 0
          upper.phi <- + Inf
        }
        else {
                lower.phi <- range.limits[1]
                upper.phi <- range.limits[2]
        }
        z <- as.vector(data)
        if(fix.lambda == T) {
                if(lambda == 1) {
                        log.jacobian <- 0
                }
                else {
                        if(any(data <= 0))
                                stop("Transformation option not allowed when there are zeros or negative data"
                                        )
                        if(any(data^(lambda - 1) <= 0))
                                log.jacobian <- log(prod(data^(lambda - 1)))
                        else log.jacobian <- sum(log(data^(lambda - 1)))
                        if(lambda == 0)
                                z <- log(z)
                        else z <- ((z^lambda) - 1)/lambda
                }
        }
        n <- length(z)
        if((2 * n) != length(coords))
                stop("Number of locations does not match with number of data")
        reduce.pars <- 0
        if(method == "REML" | method == "reml" | method == "rml")
                method <- "RML"
        if(method == "ML" | method == "ml")
                method <- "ML"
        # Checking initial parameters vector size:
        if(all(ini == 0)) {
                cov.model <- "pure.nugget"
                cat("likfit: all initial values equal to zero. Model without spatial correlation will be fitted\n"
                        )
        }
        else {
                if(is.matrix(ini)) {
                        inilength <- dim(ini)[2]
                        if(fixtau == F & inilength != 3)
                                stop("wrong number of columns for ini (must be 3)"
                                        )
                        if(fixtau == T & inilength != 2)
                                stop("wrong number of columns for ini (must be 2)"
                                        )
                }
                else {
                        inilength <- length(ini)
                        if(fixtau == F & inilength != 3)
                                stop("wrong length for ini (must be 3)")
                        if(fixtau == T & inilength != 2)
                                stop("wrong length for ini (must be 2)")
                }
        }
        xmat <- trend.spatial(trend = trend, coords = coords)
        fit.ols <- lm(z ~ xmat - 1)
        trend.ols <- list(coefficients = fit.ols$coefficients)
        var.z <- sum((fit.ols$residual)^2)/(n - length(fit.ols$coefficients))
        dimnames(xmat) <- list(NULL, NULL)
        txmat <- t(xmat)
        beta.size <- dim(xmat)[2]
        ## List with objects to be passed to the minimisation function
        temp.list <- list(z = z, xmat = xmat, txmat = txmat, fixtau = fixtau,
                ftau = ftau, method = method, kappa = kappa, cov.model =
                cov.model, beta.size = beta.size, lower.phi = lower.phi,
                dists.lowertri = dists.vec, var.z = var.z, fix.lambda =
                fix.lambda, n = n)
        if(fix.lambda == T) {
                temp.list$lambda <- lambda
                temp.list$log.jacobian <- log.jacobian
        }
        ## Searching for best initial value
        if((cov.model == "pure.nugget") | all(ini == 0)) {
                lik.results <- likfit.nospatial(temp.list, ...)
                if(fix.nugget == F)
                        temp.pars <- c(lik.results$tausqhat, 0, 0)
                else temp.pars <- c(ftau, (lik.results$tausqhat - ftau), 0)
                lambda <- lik.results$lambda
        }
        else {
                if(is.matrix(ini) | is.data.frame(ini)) {
                        ini <- as.matrix(ini)
                        if(messages.screen == T)
                                cat("likfit: searching for the best initial value\n"
                                        )
                        ini.search <- ini
                        if(fix.nugget == T)
                                ini.search <- cbind(nugget, ini.search)
                        if(length(lambda) == 1)
                                ini.search <- cbind(ini.search, lambda)
                        else {
                                temp <- ini.search
                                for(i in 1:(length(lambda) - 1)) {
                                        ini.search <- rbind(ini.search, temp)
                                }
                                ini.search <- cbind(ini.search, rep(lambda,
                                        each = dim(temp)[1]))
                        }
                        dimnames(ini.search) <- list(NULL, NULL)
                        loglik.ini <- round(100000000. * apply(ini.search,
                                1, loglik.spatial, temp.list))
                        ini.max <- as.vector(ini.search[loglik.ini == max(
                                loglik.ini), ])
                        if(fixtau == T) {
                                ini <- as.vector(ini.max[2:3])
                        }
                        else {
                                ini <- as.vector(ini.max[1:3])
                        }
                        if(messages.screen == T) {
                                cat("likfit: best initial value:\n")
                                names(ini.max) <- c("nugget", "sill", "range",
                                        "lambda")
                                print(ini.max)
                        }
                        if(fix.lambda == F)
                                lambda <- as.vector(ini.max[4])
                }
                ## Finding the ML/REML estimates using numerical methods
                if(messages.screen == T) {
                        cat("likfit: Initialising likelihood maximisation using nlminb.\n"
                                )
                        cat("likfit: WARNING: This step can be time demanding!\n"
                                )
                        cat("------------------------------------------------------------\n"
                                )
                        cat("Consider providing additional (non-default) arguments for minimisation function nlminb using the argument ...\n"
                                )
                        cat("For more details see documentation for the function nlminb.control.\n"
                                )
                        cat("------------------------------------------------------------\n")
                        "nice" <-
                          function (x, decimal = 2, fixed = FALSE)
                            {
                              ergb <- x
                              index <- (x != 0) & is.finite(x)
                              if (fixed)
                                n <- 0
                              else n <- floor(log(abs(x[index]))/log(10))
                              ergb[index] <- trunc(x[index]/10^(n - decimal)) * 10^(n - decimal)
                              return(ergb)
                            }

                        cat(paste(
                                "likfit: range of values for the parameter phi:",
                                nice(lower.phi), "and", nice(upper.phi), "\n"))
                }
                if(fixtau == F | ftau != 0) {
                        ## case 1: parameters = (\sigma^2, \phi) ; nugget fixed and different from 0
                        if(fixtau == T & ftau != 0) {
                                ## if(messages.screen == T)
                                ## print("covariance parameters used in the minimization algorithm  are $sigma^2$ and $phi$")
                                lower.sigmasq <- 0.01 * var.z
                                if(fix.lambda == T) {
                                        lik.results <- nlminb(ini, proflik.ftau,
                                                lower = c(0, lower.phi), upper
                                                 = c( + Inf, upper.phi),
                                                temp.list = temp.list, ...)
                                }
                                else {
                                        lik.results <- nlminb(c(ini, lambda),
                                                proflik.ftau, lower = c(0,
                                                lower.phi, -3), upper = c(
                                                 + Inf, upper.phi, 3),
                                                temp.list = temp.list, ...)
                                        lambda <- lik.results$parameters[3]
                                        lik.results$parameters <- as.vector(
                                                lik.results$parameters[1:2])
                                        if(lambda == 0)
                                                z <- log(as.vector(data))
                                        else z <- (((as.vector(data))^lambda) -
                                                        1)/lambda
                                }
                                lik.results$parameters <- temp.pars <-
                                        as.vector(c(ftau, lik.results$
                                        parameters))
                                if(messages.screen == T) {
                                        cat(paste(lik.results$message, "\n"))
                                }
                        }
                        ## case 2: parameters = (\tau^2, \sigma^2, \phi)
                        if(fixtau == F) {
                                ini.m <- c(ini[1]/ini[2], ini[3])
                                ## if(messages.screen == T)
                                ## print("parameters used in the minimization function are the ratio (\tausq/sigmasq) and phi")
                                if(fix.lambda == T) {
                                        lik.results <- nlminb(ini.m,
                                                proflik.nug, lower = c(0,
                                                lower.phi), upper = c( + Inf,
                                                upper.phi), temp.list =
                                                temp.list,
                                                ...)
                                }
                                else {
                                        lik.results <- nlminb(c(ini.m, lambda),
                                                proflik.nug, lower = c(0,
                                                lower.phi, -3), upper = c(
                                                 + Inf, upper.phi, 3),
                                                temp.list = temp.list, ...)
                                        lambda <- lik.results$parameters[3]
                                        lik.results$parameters <- as.vector(
                                                lik.results$parameters[1:2])
                                        if(lambda == 0)
                                                z <- log(as.vector(data))
                                        else z <- (((as.vector(data))^lambda) -
                                                        1)/lambda
                                }
                                if(messages.screen == T) {
                                        cat(paste(lik.results$message, "\n"))
                                }
                                if(automatic.refit == T & (lik.results$
                                        parameters[1] < 0.01)) {
                                        if(messages.screen == T)
                                                cat(paste(
                                                        "likfit: WARNING: ratio of estimates $\tau^2/$sigma^2$ < 0.01 (",
                                                        round(lik.results$
                                                        parameters[1], dig = 4),
                                                        "\n"))
                                        reduce.pars <- 1
                                        temp.list$ftau <- 0
                                        temp.list$fixtau <- T
                                        ini.m <- ini[3]
                                        if(fix.lambda == T) {
                                                lik.results <- nlminb(
                                                        lik.results$parameters[
                                                        2], proflik.phi, lower
                                                         = lower.phi, upper =
                                                        upper.phi, temp.list =
                                                        temp.list, ...)
                                        }
                                        else {
                                                lik.results <- nlminb(c(
                                                        lik.results$parameters[
                                                        2], lambda),
                                                        proflik.phi, lower = c(
                                                        lower.phi, -3), upper
                                                         = c(upper.phi, 3),
                                                        temp.list = temp.list,
                                                        ...)
                                                lambda <- as.vector(lik.results$
                                                        parameters[2])
                                                lik.results$parameters <-
                                                        as.vector(lik.results$
                                                        parameters[1])
                                                if(lambda == 0)
                                                        z <- log(as.vector(
                                                                data))
                                                else z <- (((as.vector(data))^
                                                                lambda) - 1)/
                                                                lambda
                                        }
                                        if(messages.screen == T)
                                                cat("likfit: model re-fitted without nugget effect (tausq = 0)\n"
                                                        )
                                        lik.results$parameters <- as.vector(
                                                c(0, lik.results$parameters))
                                        if(messages.screen == T) {
                                                cat(paste(lik.results$message,
                                                        "\n"))
                                        }
                                }
                                nugget.rel <- lik.results$parameters[1]
                                if(lik.results$parameters[2] < 1e-08)
                                        icovhat <- diag(n)
                                else icovhat <- varcov.spatial(coords = coords,
                                                cov.model = cov.model, kappa =
                                                kappa, nugget = nugget.rel,
                                                cov.pars = c(1, lik.results$
                                                parameters[2]), inv = T, det =
                                                F)$inverse
                                txiv <- crossprod(xmat, icovhat)
                                sigmasqhat <- (z %*% (icovhat - crossprod(
                                        txiv, solve(txiv %*% xmat)) %*% txiv) %*%
                                        z)/n
                                if(method == "RML")
                                        sigmasqhat <- (sigmasqhat * n)/(n -
                                                beta.size)
                                nuggethat <- lik.results$parameters[1] *
                                        sigmasqhat
                                lik.results$parameters <- temp.pars <-
                                        as.vector(c(nuggethat, sigmasqhat,
                                        lik.results$parameters[2]))
                        }
                        if((automatic.refit == T & (lik.results$parameters[
                                3] <= lower.phi)) | lik.results$parameters[
                                3] < 1e-08) {
                                if(messages.screen == TRUE) {
                                        cat("likfit: WARNING: phi estimate < minimum value allowed\n"
                                                )
                                        cat("likfit: model re-fitted without spatial correlation (phi=0)\n"
                                                )
                                }
                                reduce.pars <- 2
                                lik.results <- likfit.nospatial(temp.list,
                                        ...)
                                lambda <- lik.results$lambda
                                if(fix.nugget == T) {
                                        lik.results$parameters <- temp.pars <-
                                                as.vector(c(ftau, (lik.results$
                                                tausqhat - ftau), 0))
                                }
                                else {
                                        lik.results$parameters <- temp.pars <-
                                                as.vector(c(lik.results$
                                                tausqhat, 0, 0))
                                }
                        }
                }
                else {
                        ## case 3: parameters = $(\sigma^2, \phi)$ ; fixed nugget: $\tau^2= 0$
                        ini.m <- ini[2]
                        if(fix.lambda == T) {
                                lik.results <- nlminb(ini.m, proflik.phi, lower
                                         = lower.phi, upper = upper.phi,
                                        temp.list = temp.list, ...)
                        }
                        else {
                                lik.results <- nlminb(c(ini.m, lambda),
                                        proflik.phi, lower = c(lower.phi, -3
                                        ), upper = c(upper.phi, 3), temp.list
                                         = temp.list, ...)
                                lambda <- as.vector(lik.results$parameters[
                                        2])
                                lik.results$parameters <- as.vector(lik.results$
                                        parameters[1])
                                if(lambda == 0)
                                        z <- log(as.vector(data))
                                else z <- (((as.vector(data))^lambda) - 1)/
                                                lambda
                        }
                        if(messages.screen == T) {
                                cat(paste(lik.results$message, "\n"))
                        }
                        if(automatic.refit == T & (lik.results$parameters <=
                                lower.phi)) {
                                cat("likfit: WARNING: phi estimate < minimum value allowed\n"
                                        )
                                cat("likfit: model re-fitted without spatial correlation (phi=0)\n"
                                        )
                                reduce.pars <- 1
                                lik.results <- likfit.nospatial(temp.list,
                                        ...)
                                lambda <- lik.results$lambda
                                lik.results$parameters <- temp.pars <-
                                        as.vector(c(0, lik.results$tausqhat,
                                        0))
                        }
                        else {
                                if(lik.results$parameters < 1e-08)
                                        icovhat <- diag(n)
                                else icovhat <- varcov.spatial(coords = coords,
                                                cov.model = cov.model, kappa =
                                                kappa, nugget = 0, cov.pars = c(
                                                1, lik.results$parameters),
                                                inv = T, det = F)$inverse
                                txiv <- crossprod(xmat, icovhat)
                                sigmasqhat <- (z %*% (icovhat - crossprod(
                                        txiv, solve(txiv %*% xmat)) %*% txiv) %*%
                                        z)/n
                                if(method == "RML")
                                        sigmasqhat <- (sigmasqhat * n)/(n -
                                                beta.size)
                                temp.pars <- as.vector(c(0, sigmasqhat,
                                        lik.results$parameters))
                                lik.results$parameters <- as.vector(c(
                                        sigmasqhat, lik.results$parameters))
                        }
                }
        }
        if(messages.screen == T) {
                cat("likfit: end of likelihood maximisation\n")
        }
        if(any(temp.pars < 0)){
          temp.pars <- round(temp.pars, dig=14)
          lik.results$estimate <- round(lik.results$estimate, dig=14)
        }
        loglik <- - lik.results$objective
        npars <- length(trend.ols$coefficients) + length(ini) - reduce.pars
        if(fix.lambda == F)
                npars <- npars + 1
        AIC <- loglik - npars
        BIC <- loglik - 0.5 * log(n) * npars
        if(any(temp.pars[2:3]) != 0)
                invcov <- varcov.spatial(coords = coords, cov.model = cov.model,
                        kappa = kappa, nugget = temp.pars[1], cov.pars =
                        temp.pars[2:3], inv = T, det = F)$inverse
        else invcov <- diag((1/temp.pars[1]), n)
        txmatinvcov <- crossprod(xmat, invcov)
        beta <- solve(txmatinvcov %*% xmat) %*% txmatinvcov %*% z
        beta.var <- solve(txmatinvcov %*% xmat)
        if(residuals == T | predicted == T) {
                ## print("computing predicted values and residuals")
                trend.est <- as.vector(xmat %*% beta)
                residuals.trend <- as.vector(z - trend.est)
                if(any(temp.pars[2:3]) != 0)
                        covmat.signal <- varcov.spatial(coords = coords,
                                cov.model = cov.model, kappa = kappa, nugget =
                                0, cov.pars = temp.pars[2:3])$varcov
                else covmat.signal <- diag(n)
                signal.est <- as.vector(covmat.signal %*% invcov %*%
                        residuals.trend)
                predict.est <- trend.est + signal.est
                residuals.est <- as.vector(z - predict.est)
                residuals.std <- as.vector(invcov %*% residuals.est)
                residuals.trend.std <- as.vector(invcov %*% residuals.trend)
                s2.trend <- (crossprod(residuals.trend, invcov) %*%
                        residuals.trend)/(n - length(beta))
                s2 <- (crossprod(residuals.est, invcov) %*% residuals.est)/
                        (n - length(beta))
        }
        if(messages.screen == T)
                cat("likfit: preparing output\n")
        results <- list()
        results$cov.model <- cov.model
        results$nugget <- temp.pars[1]
        results$cov.pars <- as.vector(c(sigmasq = temp.pars[2], phi = temp.pars[
                3]))
        if(is.null(kappa))
                results$kappa <- "not used"
        else results$kappa <- kappa
        results$beta <- as.vector(beta)
        results$beta.var <- beta.var
        if(length(results$beta.var) == 1)
                results$beta.var <- as.vector(results$beta.var)
        if(length(results$beta) > 1) {
                if(inherits(trend, "formula"))
                        beta.names <- c("intercept", paste("covar", 1:(ncol(
                                xmat) - 1), sep = ""))
                else if(trend == "1st")
                        beta.names <- c("1", "x", "y")
                else if(trend == "2nd")
                        beta.names <- c("1", "x", "y", "x2", "xy", "y2")
                names(results$beta) <- beta.names
        }
        results$lambda <- lambda
        results$loglik <- loglik
        results$npars <- npars
        results$AIC <- AIC
        results$BIC <- BIC
        results$trend.ols <- trend.ols$coefficients
        names(results$trend.ols) <- names(results$beta)
        if(residuals == T) {
                results$s2 <- s2
                results$s2.trend <- s2.trend
        }
        if(predicted == T)
                results$predicted <- cbind(predicted = predict.est, trend.est
                         = trend.est, signal.est = signal.est)
        if(residuals == T)
                results$residuals <- round(cbind(residuals = residuals.est,
                        resid.trend = residuals.trend, resid.std =
                        residuals.std, resid.trend.std = residuals.trend.std),
                        dig = 12)
        if(fix.lambda == F) {
                if(lambda == 1) {
                        log.jacobian <- 0
                }
                else {
                        if(any(data^(lambda - 1) <= 0))
                                log.jacobian <- log(prod(data^(lambda - 1)))
                        else log.jacobian <- sum(log(data^(lambda - 1)))
                }
        }
        results$info.lambda <- list(fix.lambda = fix.lambda, log.jacobian =
                log.jacobian)
        lik.results$parameters <- NULL
        lik.results$aux <- NULL
        lik.results$objective <- NULL
        results$info <- lik.results
        results$max.dist <- max.dist
        results$trend.matrix <- xmat
        results$method <- method
        results$call <- call.fc
        class(results) <- "variomodel"
        if(messages.screen == T) {
                cat("likfit: model parameters are:\n")
                cat(paste("covariance model:", cov.model))
                if(cov.model == "matern" | cov.model == "powered.exponential" |
                        cov.model == "cauchy" | cov.model == "gneiting.matern"
                        )
                        cat(paste(" with kappa =", kappa))
                if(!is.null(kappa))
                        if(cov.model == "matern" & kappa == 0.5)
                                cat(" (exponential)")
                cat("\n")
                print(c(nugget = results$nugget, sill = results$cov.pars[1],
                        range = results$cov.pars[2]))
                if(fix.lambda == F)
                        cat(paste("Box-Cox transformation parameter:", round(
                                results$lambda, dig=3), "\n"))
                if((results$cov.pars[1] < (0.01 * (results$nugget + results$
                        cov.pars[1]))) & (results$cov.pars[2] > 0))
                        cat("\nWARNING: estimated sill is less than 1 hundredth of the total variance (nugget + sill). Consider re-examine the model excall.fcuding spatial dependence\n"
                                )
                if((results$cov.pars[2] > (10 * max.dist)) & (results$cov.pars[
                        1] > 0))
                        cat("\nWARNING: estimated range is more than 10 times bigger than the biggest distance between two points. Consider re-examine the model:\n 1) excluding spatial dependence if estimated sill is too low and/or \n 2) taking trends (covariates) into account\n"
                                )
                if((results$cov.pars[2] < (0.10000000000000001 * min.dist)) &
                        (results$cov.pars[1] > 0))
                        cat("\nWARNING: estimated range is less than 1 tenth of the minimum distance between two points. Consider re-examine the model excluding spatial dependence\n"
                                )
        }
        return(results)
}

"likfit.nospatial" <-
function(temp.list, ...)
{
        results <- list()
        z <- temp.list$z
        n <- temp.list$n
        beta.size <- temp.list$beta.size
        xmat <- temp.list$xmat
        txmat <- temp.list$txmat
        ixx <- solve(crossprod(xmat))
        if(temp.list$fix.lambda == F) {
                results <- nlminb(1, proflik.lambda, lower = -2, upper = 2,
                        temp.list = temp.list,
                        ...)
                results$lambda <- results$parameters
                if(results$lambda == 1) {
                        temp.list$log.jacobian <- 0
                }
                else {
                        if(any(z <= 0))
                                stop("Transformation option not allowed when there are zeros or negative data"
                                        )
                        if(any(z^(results$lambda - 1) <= 0))
                                temp.list$log.jacobian <- log(prod(z^(results$
                                        lambda - 1)))
                        else temp.list$log.jacobian <- sum(log(z^(results$
                                        lambda - 1)))
                        if(results$lambda == 0)
                                z <- log(z)
                        else z <- ((z^results$lambda) - 1)/results$lambda
                }
        }
        else {
                results$lambda <- temp.list$lambda
                results$message <- "RELATIVE FUNCTION CONVERGENCE"
        }
        ssres <- (z %*% (diag(n) - xmat %*% solve(crossprod(xmat)) %*% txmat) %*%
                z)
        if(temp.list$method == "ML") {
                results$tausqhat <- ssres/n
                if(temp.list$fix.lambda == T)
                        results$objective <- ((n/2) * log(2 * pi) + (n/2) *
                                log(results$tausqhat) + (n/2) - temp.list$
                                log.jacobian)
        }
        if(temp.list$method == "RML") {
                results$tausqhat <- (ssres/(n - beta.size))
                if(temp.list$fix.lambda == T) {
                        results$objective <- (((n - beta.size)/2) * log(2 *
                                pi) + ((n - beta.size)/2) * log(results$
                                tausqhat) + (n/2) - temp.list$log.jacobian)
                }
        }
        return(results)
}

"lines.grid.bayes" <-
  function(obj, max.dist, length = 100, summary.posterior = c("mode", "median",
                                          "mean"), ...)
{
  spost <- match.arg(summary.posterior)
  if(is.null(obj$call$cov.model))
    cov.model <- "exponential"
  else {
    cov.model <- obj$call$cov.model
    if(obj$call$cov.model == "matern" | obj$call$cov.model ==
       "powered.exponential" | obj$call$cov.model == "cauchy" |
       obj$call$cov.model == "gneiting-matern")
      kappa <- obj$call$kappa
    else kappa <- NULL
  }
  distance <- seq(0, max.dist, length = length)
  cov.pars <- c(obj$sigmasq.summary[spost], obj$phi.summary[spost])
  names(cov.pars) <- NULL
  if(any(obj$tausq.summary != 0))
    nugget <- obj$tausq.summary[spost] * cov.pars[1]
  else nugget <- 0
  names(nugget) <- NULL
  sill.total <- nugget + cov.pars[1]
  gamma <- (nugget + cov.pars[1]) - cov.spatial(distance, cov.model =
                                                cov.model, kappa = kappa, cov.pars = cov.pars)
  lines(distance, gamma, ...)
}

"lines.krige.bayes" <-
  function(obj, max.dist, length = 100, summary.posterior = c("mode", "median", "mean"), ...)
{
  spost <- match.arg(summary.posterior)
  if(is.null(obj$call$cov.model))
    cov.model <- "exponential"
  else {
    cov.model <- obj$call$cov.model
    if(obj$call$cov.model == "matern" | obj$call$cov.model == "powered.exponential" | obj$
       call$cov.model == "cauchy" | obj$call$cov.model == "gneiting-matern")
      kappa <- obj$call$kappa
    else kappa <- NULL
  }
  distance <- seq(0, max.dist, length = length)
  if(spost == "mode")
    spost1 <- "mode.cond"
  else spost1 <- spost
  cov.pars <- c(obj$posterior$sigmasq.summary[spost1], obj$posterior$phi.summary[spost])
  names(cov.pars) <- NULL
  if(is.numeric(obj$posterior$tausq.summary))
    nugget <- obj$posterior$tausq.summary[spost] * cov.pars[1]
  else nugget <- 0
  names(nugget) <- NULL
  sill.total <- nugget + cov.pars[1]
  gamma <- (nugget + cov.pars[1]) - cov.spatial(distance, cov.model = cov.model, kappa = kappa,
                                                cov.pars = cov.pars)
  lines(distance, gamma, ...)
}

"loglik.spatial" <-
## This function computes the log-likelihood for a (transformed) Gaussian random field for known parameters
function(pars, temp.list)
{
        pars <- as.vector(pars)
        tausq <- pars[1]
        sigmasq <- pars[2]
        phi <- pars[3]
        lambda <- pars[4]
        z <- temp.list$z
        n <- temp.list$n
        if(temp.list$fix.lambda == F) {
                if(lambda == 1) {
                        temp.list$log.jacobian <- 0
                }
                else {
                        if(any(z < 0))
                                stop("Transformation option not allowed when there are zeros or negative data"
                                        )
                        if(any(z^(lambda - 1) <= 0))
                                temp.list$log.jacobian <- log(prod(z^(lambda -
                                        1)))
                        else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
                        if(lambda == 0)
                                z <- log(z)
                        else z <- ((z^lambda) - 1)/lambda
                }
        }
        beta.size <- temp.list$beta.size
        kappa <- temp.list$kappa
        xmat <- temp.list$xmat
        txmat <- temp.list$txmat
        covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
                cov.model = temp.list$cov.model, kappa = kappa, nugget = tausq,
                cov.pars = c(sigmasq, phi), scaled = F, inv = T, det = T,
                sqrt.inv = T)
        six <- crossprod(covinf$sqrt.inverse, xmat)
        xix <- crossprod(six)
        if(length(as.vector(xix)) == 1) {
                ixix <- 1/xix
                choldet <- 0.5 * log(xix)
        }
        else {
                chol.xix <- chol(xix)
                i.chol.xix <- solve.upper(chol.xix)
                ixix <- i.chol.xix %*% t(i.chol.xix)
                choldet <- sum(log(diag(chol.xix)))
        }
        xi <- crossprod(xmat, covinf$inverse)
        ssresmat <- z %*% (covinf$inverse - crossprod(xi, ixix) %*% xi) %*% z
        if(temp.list$method == "ML") {
                loglik <- ( - (n/2) * log(2 * pi) - covinf$log.det.to.half -
                        0.5 * ssresmat + temp.list$log.jacobian)
        }
        if(temp.list$method == "RML") {
                xx.eigen <- eigen(crossprod(temp.list$xmat), symmetric = T, only.values
                         = T)
                loglik <- ( - ((n - beta.size)/2) * log(2 * pi) + 0.5 * sum(
                        log(xx.eigen$values)) - covinf$log.det.to.half - (
                        0.5) * ssresmat - choldet + temp.list$log.jacobian)
        }
        # }
        return(as.vector(loglik))
}

"loss.olsvario" <-
function(theta, .global.list)
{
  u <- .global.list$u
  v <- .global.list$v
  kappa <- .global.list$kappa
  if(.global.list$fix.nugget == FALSE) {
    tausq <- theta[1]
    sigmasq <- theta[2]
    phi <- theta[3]
    tau1 <- tausq
  }
  else {
    tausq <- .global.list$nugget
    sigmasq <- theta[1]
    phi <- theta[2]
    tau1 <- tausq + 1
  }
        sill.total <- sigmasq + tausq
  if(.global.list$fix.nugget == FALSE) {
    gamma <- sill.total - cov.spatial(u, cov.model = .global.list$cov.model, kappa
                                      = kappa, cov.pars = c(theta[2], theta[3]))
  }
  else {
    gamma <- sill.total - cov.spatial(u, cov.model = .global.list$cov.model, kappa
                                      = kappa, cov.pars = c(theta[1], theta[2]))
  }
  loss <- sum((v - gamma)^2)
  return(loss)
}

"loss.wlsvario" <-
function(theta, .global.list)
{
  u <- .global.list$u
  v <- .global.list$v
  n <- .global.list$n
  kappa <- .global.list$kappa
  if (.global.list$fix.nugget == FALSE) {
    tausq <- theta[1]
    sigmasq <- theta[2]
    phi <- theta[3]
    tau1 <- tausq
  }
  else {
    tausq <- .global.list$nugget
    sigmasq <- theta[1]
    phi <- theta[2]
    tau1 <- tausq + 1
  }
  sill.total <- sigmasq + tausq
  if (.global.list$fix.nugget == FALSE) {
    gamma <- sill.total - cov.spatial(u, cov.model = .global.list$cov.model, kappa
                                      = kappa, cov.pars = c(theta[2], theta[3]))
  }
  else {
    gamma <- sill.total - cov.spatial(u, cov.model = .global.list$cov.model,
                                      kappa = .global.list$kappa, cov.pars = c(theta[1], theta[2]))
  }
  if(.global.list$weight == "npairs")
    loss <- sum(n * (v - gamma)^2)
  if(.global.list$weight == "cressie")
    loss <- sum((n/(gamma^2)) * (v - gamma)^2)
  return(loss)
}

"matern" <-
  function(u, phi, kappa)
{
  if(kappa == 0.5) {
    covs <- exp( - u/phi)
  }
  else {
    if(kappa != floor(kappa)) {
      kappa <- floor(kappa)
      cat(paste("matern: kappa must be an integer value, changing value to kappa =", kappa, "\n"))
    }
    if(!is.loaded("Xwhittlematern"))
      dyn.load("/home/ribeiro/geoS/woodandchan.o")
    covs <- .C("Xwhittlematern",
               u = as.double(u),
               as.integer(length(u)),
               as.double(phi),
               as.double(kappa))$u
    ## check == here
    covs[u == 0] <- 1
  }
  return(covs)
}

"model.control" <-
function(trend.d = "cte", trend.l = "cte", cov.model = c("exponential",
        "matern", "gaussian", "spherical", "circular", "cubic", "wave",
        "powered.exponential", "cauchy", "gneiting", "gneiting.matern",
        "pure.nugget"), kappa = 0.5, aniso.pars = NULL, lambda = 1)
{
        cov.model <- match.arg(cov.model)
        return(list(trend.d = trend.d, trend.l = trend.l, cov.model = cov.model,
                kappa = kappa, aniso.pars = aniso.pars, lambda = lambda))
}


"olsfit" <-
function(vario, ini.cov.pars, cov.model = c("exponential", "matern", "gaussian",
        "spherical", "cubic", "wave", "powered.exponential", "cauchy",
        "gneiting", "gneiting.matern", "pure.nugget"), fix.nugget = F, nugget = 0,
         kappa = NULL,
        simul.number = NULL, max.dist = "all", minimisation.function = c("nlminb", "nls"),
        lower = 0, messages.screen = TRUE)
{
  call.fc <- match.call()
  ini <- ini.cov.pars
  minimisation.function <- match.arg(minimisation.function)
  cov.model <- match.arg(cov.model)
  if (is.matrix(vario$v) & is.null(simul.number))
    stop("object in vario$v is a matrix. This function works for only 1 empirical variogram at once")
  if (!is.null(simul.number))
    vario$v <- vario$v[, simul.number]
  ##
  ## Checking and/or preparing initial values
  ##
  if(is.matrix(ini)) {
    inilength <- dim(ini)[2]
    if(inilength != 2)
      stop("wrong number of columns for ini (must be 2)")
    if(fix.nugget == FALSE)
      ini <- as.matrix(expand.grid(nugget, as.numeric(names(table(ini[,1]))),
                                   as.numeric(names(table(ini[,1])))))
  }
  else {
    inilength <- length(ini)
    if(inilength != 2)
      stop("wrong length for ini (must be 2)")
    if(fix.nugget == FALSE){
      if(length(nugget) == 1)
        ini <- c(ini, nugget)
      else
        ini <- as.matrix(nugget, expand.grid(ini[1], ini[2]))
    }
  }
  ##
  ## Setting maximum distance
  ##
  if(max.dist == "all")
    data <- list(u = vario$u, v = vario$v)
  else data <- list(u = vario$u[vario$u <= max.dist], v = vario$v[vario$
                                                        u <= max.dist])
  ##
  ## Preparing lists for the minimiser
  ##
  .global.list <- list(u = data$u, v = data$v, fix.nugget = fix.nugget,
                       nugget = nugget, kappa = kappa,
                       cov.model = cov.model, m.f = minimisation.function)
  if(minimisation.function == "nlm")
    assign(".global.lower", lower, where=1)
  ##
  ## Choosing the best initial value
  ##
  if(is.matrix(ini) | is.data.frame(ini)) {
    ini <- as.matrix(ini)
    if(messages.screen) cat("olsfit: searching for the best initial value\n")
    ini.search <- ini
    dimnames(ini.search) <- list(NULL, NULL)
    ols.ini <- round(100000000. * apply(ini.search, 1, loss.olsvario, .global.list = .global.list))
    ini <- as.vector(ini.search[ols.ini == min(ols.ini),])
    if(messages.screen){
      cat("olsfit: best initial value:\n")
      if(fix.nugget == TRUE){
        names(ini) <- c("sill", "range")
        print(ini)
      }
      else{
        names(ini) <- c("nugget", "sill", "range")
        print(ini)
      }
    }
    remove(".global.lower", where=1)
  }
  ##
  ## minimisation using "nls"
  ##
  if(minimisation.function == "nls") {
    data$kappa <- kappa
    data$cov.model <- cov.model
    if(fix.nugget == F) {
      result <- nls(v ~ ((sigmasq + tausq) - cov.spatial(
                                                         u, cov.model = cov.model, kappa = kappa,
                                                         cov.pars = c(sigmasq, phi))), start = list(
                                                                                         tausq = ini[1], sigmasq = ini[2], phi = ini[
                                                                                                                             3]), data = data)
      nugget <- result$parameter[1]
      cov.pars <- as.vector(result$parameter[2:3])
      value <- vario$v - ((nugget + cov.pars[1]) -
                          cov.spatial(vario$u, cov.model = cov.model,
                                      kappa = kappa, cov.pars = cov.pars))
    }
    else {
      data$ftau <- nugget
      result <- nls(v ~ ((sigmasq + ftau) - cov.spatial(
                                                        u, cov.model = cov.model, kappa = kappa,
                                                        cov.pars = c(sigmasq, phi))), start = list(
                                                                                        sigmasq = ini[1], phi = ini[2]), data = data)
      cov.pars <- as.vector(result$parameter[1:2])
      value <- vario$v - ((nugget + cov.pars[1]) -
                          cov.spatial(vario$u, cov.model = cov.model,
                                      kappa = kappa, cov.pars = cov.pars))
    }
    value <- sum(value * value)
    message <- "nls does not provides convergence message"
  }
  ##
  ## minimisation using "nlminb"
  ##
  if(minimisation.function == "nlminb") {
    result <- nlminb(ini, loss.olsvario, lower = lower, .global.list = .global.list)
    value <- result$objective
    message <- result$message
    if(fix.nugget == F) {
      nugget <- result$parameter[1]
      cov.pars <- as.vector(result$parameter[2:3])
    }
    else {
      cov.pars <- as.vector(result$parameter[1:2])
    }
  }
  dmax <- max(vario$u)
  estimation <- list(nugget = nugget, cov.pars = cov.pars, cov.model =
                     cov.model, kappa = kappa, value = value, max.dist = dmax,
                     minimisation.function = minimisation.function, message =
                     message)
  estimation$method <- "OLS"
  estimation$call <- call.fc
  class(estimation) <- "variomodel"
  if(messages.screen) {
    cat("olsfit: estimated model parameters are:\n")
    cat(paste("covariance model:", cov.model))
    if(cov.model == "matern" | cov.model == "powered.exponential" |
       cov.model == "cauchy" | cov.model == "gneiting.matern"
       )
      cat(paste(" with kappa =", kappa))
    if(!is.null(kappa))
      if(cov.model == "matern" & kappa == 0.5)
        cat(" (exponential)")
    cat("\n")
    print(c(nugget = estimation$nugget, sill = estimation$cov.pars[
                                          1], range = estimation$cov.pars[2]))
  }
  return(estimation)
}

"output.control" <-
function(n.posterior = 1000, n.predictive = NULL, simulations.predictive = TRUE,
        keep.simulations = TRUE, mean.estimator = TRUE, quantile.estimator =
        NULL, probability.estimator = NULL, messages.screen = TRUE, signal =
        FALSE, moments = TRUE)
{
        return(list(n.posterior = n.posterior, n.predictive = n.predictive,
                moments = moments, simulations.predictive =
                simulations.predictive, keep.simulations = keep.simulations,
                mean.estimator = mean.estimator, quantile.estimator =
                quantile.estimator, probability.estimator =
                probability.estimator, signal = signal, messages.screen =
                messages.screen))
}

"plot.geodata" <-
function(geodata, coords = geodata$coords, data = geodata$data, trend = "cte",
        lambda = 1, col.data = 1, window = F, ...)
{
        par.ori <- par()
        coords <- as.matrix(coords)
        data <- as.matrix(data)
        data <- data[, col.data]
        if(lambda != 1) {
                if(lambda == 0)
                        data <- log(data)
                else data <- ((data^lambda) - 1)/lambda
        }
        if(window == T)
                trellis.device()
        ## motifA4()
        par(mfrow = c(2, 2))
        par(pty = "s")
        ## plot(coords, xlab = "Coord X", ylab = "Coord Y", pch = 3, cex = 1.2)
        ##,main='\nData locations'
        ## plot(coords, xlab = "Coord X", ylab = "Coord Y",main='\nData locations',pch=3)
        data.quantile <- quantile(data)
        plot(coords, xlab = "Coord X", ylab = "Coord Y", type = "n")
        if(length(unique(data.quantile)) == length(data.quantile)) {
                points(coords[(data <= data.quantile[2]), ], pch = 1, cex =
                        0.59999999999999998, col = 2)
                points(coords[((data > data.quantile[2]) & (data <=
                        data.quantile[3])), ], pch = 2, cex =
                        1.3999999999999999, col = 4)
                points(coords[((data > data.quantile[3]) & (data <=
                        data.quantile[4])), ], pch = 3, cex = 1.7, col = 7)
                points(coords[(data > data.quantile[4]), ], pch = 4, cex = 2,
                        col = 8)
        }
        else {
                points(coords, pch = 16)
        }
        par(pty = "m")
        par(mar = c(0, 1, 0, 0.5))
        xyzplot(coords = coords, data = data, ...)
        ## plot(variog(coords = coords, data = data, trend = trend, ...))
        par(mar = c(5, 5, 1, 0.5))
        plot(coords[, 1], data, xlab = "Coord X", cex = 1)
        plot(coords[, 2], data, xlab = "Coord Y", cex = 1)
        par(par.ori)
        return(invisible())
}

"plot.proflik" <-
function(obj.proflik, pages = c("user", "one", "two"), uni.only = F, bi.only =
        F, type.bi = c("contour", "persp"), conf.int = c(0.90000000000000002,
        0.94999999999999996), yaxis.lims = c("conf.int", "as.computed"), by.col
         = T, log.scale = F, par.mar.persp = c(9.5, 1, 1, 0), ...)
{
        ## par.mfrow.ori <- par()$mfrow
        ## par.mfcol.ori <- par()$mfcol
        pages <- match.arg(pages)
        if(all(is.character(yaxis.lims)))
                yaxis.lims <- match.arg(yaxis.lims)
        type.bi <- match.arg(type.bi)
        n.uni <- obj.proflik$n.uni
        n.bi <- obj.proflik$n.bi
        if(n.bi == 0)
                uni.only <- T
        if((uni.only == F) & (bi.only == F))
                np <- n.uni + n.bi
        if((uni.only == T) & (bi.only == F))
                np <- n.uni
        if((uni.only == F) & (bi.only == T))
                np <- n.bi
        if(n.uni == 0 & np > 0)
                bi.only <- T
        if(n.bi == 0 & np > 0)
                uni.only <- T
        if(pages == "one") {
                if(np >= 1 & np < 4)
                        par(mfrow = c(np, 1))
                if(np >= 4) {
                        if(by.col == T)
                                par(mfcol = c(ceiling(np/2), 2))
                        else par(mfrow = c(ceiling(np/2), 2))
                }
        }
        if(pages == "two") {
                if(n.uni > 1 & n.uni < 4)
                        par(mfrow = c(n.uni, 1))
                if(n.uni >= 4)
                        par(mfrow = c(ceiling(n.uni/2), 2))
        }
        if(bi.only == F) {
                for(i in 1:n.uni) {
                        if(obj.proflik$method == "ML")
                                ylabm <- "profile log-likelihood"
                        else ylabm <- "profile log-(restricted) likelihood"
                        if(all(conf.int) != F) {
                                if(!is.numeric(conf.int) | any(conf.int > 1))
                                        stop("argument conf.int must be numerical (scalar or vector) with values between 0 and 1"
                                                )
                                conf.int.drop <- obj.proflik[[i]][[3]][2] -
                                        0.5 * qchisq(conf.int, 1)
                        }
                        if(all(is.character(yaxis.lims))) {
                                if(yaxis.lims == "conf.int")
                                        lik.lims <- c(min(conf.int.drop),
                                                obj.proflik[[i]][[3]][2])
                                else lik.lims <- c(min(obj.proflik[[i]][[2]]),
                                                obj.proflik[[i]][[3]][2])
                        }
                        else lik.lims <- yaxis.lims
                        if(log.scale == T) {
                                plot(log(obj.proflik[[i]][[1]]), obj.proflik[[
                                        i]][[2]], type = "b", xlab = paste(
                                        "log-", names(obj.proflik[[i]])[1]),
                                        ylab = ylabm, ylim = lik.lims, ...)
                                lines(log(c(obj.proflik[[i]][[3]][1],
                                        obj.proflik[[i]][[3]][1])), c(min(
                                        lik.lims), obj.proflik[[i]][[3]][2]),
                                        lty = 2)
                        }
                        else {
                                plot(obj.proflik[[i]][[1]], obj.proflik[[i]][[
                                        2]], type = "b", xlab = names(
                                        obj.proflik[[i]])[1], ylab = ylabm,
                                        ylim = lik.lims, ...)
                                lines(c(obj.proflik[[i]][[3]][1], obj.proflik[[
                                        i]][[3]][1]), c(min(lik.lims),
                                        obj.proflik[[i]][[3]][2]), lty = 2)
                        }
                        abline(h = conf.int.drop, lty = 3)
                }
        }
        if(uni.only == F) {
                if(pages == "two") {
                        if(n.bi >= 1 & n.bi < 4)
                                par(mfrow = c(n.bi, 1))
                        if(n.bi >= 4)
                                par(mfrow = c(ceiling(n.bi/2), 2))
                }
                for(i in 1:n.bi) {
                        if(type.bi == "contour") {
                                if(log.scale == T) {
                                        contour(log(obj.proflik[[(n.uni + i)]][[
                                                1]]), log(obj.proflik[[(n.uni +
                                                i)]][[2]]), matrix(obj.proflik[[
                                                (n.uni + i)]][[3]], ncol =
                                                length(obj.proflik[[(n.uni +
                                                i)]][[2]])), xlab = paste(
                                                "log-", names(obj.proflik[[
                                                (n.uni + i)]][1])), ylab =
                                                paste("log-", names(obj.proflik[[
                                                (n.uni + i)]][2]), ...))
                                        points(log(t(obj.proflik[[(n.uni + i)]][[
                                                4]][1:2])))
                                }
                                else {
                                        contour(obj.proflik[[(n.uni + i)]][[
                                                1]], obj.proflik[[(n.uni + i)]][[
                                                2]], matrix(obj.proflik[[(
                                                n.uni + i)]][[3]], ncol =
                                                length(obj.proflik[[(n.uni +
                                                i)]][[2]])), xlab = names(
                                                obj.proflik[[(n.uni + i)]][
                                                1]), ylab = names(obj.proflik[[
                                                (n.uni + i)]][2]), ...)
                                        points(t(obj.proflik[[(n.uni + i)]][[
                                                4]][1:2]))
                                }
                        }
                        if(type.bi == "persp") {
                                par.mar.ori <- par()$mar
                                par(mar = par.mar.persp)
                                if(obj.proflik$method == "ML")
                                        zlabm <- "profile log-likelihood"
                                else zlabm <- "profile (restricted) log-likelihood"
                                zlimm <- range(obj.proflik[[(n.uni + i)]][[
                                        3]])
                                zlimm[1] <- 1.01 * zlimm[1]
                                minlik <- min(obj.proflik[[(n.uni + i)]][[3]])
                                if(log.scale == T) {
                                        pp <- persp(log(obj.proflik[[(n.uni +
                                                i)]][[1]]), log(obj.proflik[[
                                                (n.uni + i)]][[2]]), matrix(
                                                obj.proflik[[(n.uni + i)]][[
                                                3]], ncol = length(obj.proflik[[
                                                (n.uni + i)]][[2]])), xlab =
                                                paste("log-", names(obj.proflik[[
                                                (n.uni + i)]][1])), ylab =
                                                paste("log-", names(obj.proflik[[
                                                (n.uni + i)]][2])), zlab =
                                                zlabm, box = F, ...)
                                        pp1 <- perspp(x = log(c(obj.proflik[[
                                                (n.uni + i)]][[4]][1], min(
                                                obj.proflik[[(n.uni + i)]][[
                                                1]]))[c(1, 1, 1, 2)]), y = log(
                                                c(obj.proflik[[(n.uni + i)]][[
                                                4]][2], min(obj.proflik[[(
                                                n.uni + i)]][[2]]))[c(1, 1,
                                                2, 1)]), z = c(minlik,
                                                obj.proflik[[(n.uni + i)]][[
                                                4]][3])[c(1, 2, 1, 1)], pp)
                                        segments(log(pp1$x[1]), log(pp1$y[
                                                1]), log(pp1$x[2]), log(pp1$
                                                y[2]), lwd = 2)
                                }
                                else {
                                        pp <- persp(obj.proflik[[(n.uni + i)]][[
                                                1]], obj.proflik[[(n.uni + i)]][[
                                                2]], matrix(obj.proflik[[(
                                                n.uni + i)]][[3]], ncol =
                                                length(obj.proflik[[(n.uni +
                                                i)]][[2]])), xlab = names(
                                                obj.proflik[[(n.uni + i)]][
                                                1]), ylab = names(obj.proflik[[
                                                (n.uni + i)]][2]), zlab = zlabm,
                                                box = F, ...)
                                        pp1 <- perspp(x = c(obj.proflik[[(
                                                n.uni + i)]][[4]][1], min(
                                                obj.proflik[[(n.uni + i)]][[
                                                1]]))[c(1, 1, 1, 2)], y = c(
                                                obj.proflik[[(n.uni + i)]][[
                                                4]][2], min(obj.proflik[[(
                                                n.uni + i)]][[2]]))[c(1, 1,
                                                2, 1)], z = c(minlik,
                                                obj.proflik[[(n.uni + i)]][[
                                                4]][3])[c(1, 2, 1, 1)], pp)
                                        segments(pp1$x[1], pp1$y[1], pp1$x[
                                                2], pp1$y[2], lwd = 2)
                                }
                                par(mar = par.mar.ori)
                        }
                }
        }
        ## par(mfrow=par.mfrow.ori)
        ## par(mfcol=par.mfcol.ori)
        return(invisible())
}

"plot.symbols" <-
function(obj, data = obj$data, col.data = 1, polygon = NULL, window = F, ...)
{
        if(class(obj) == "geodata" | class(obj) == "grf") {
                coords <- as.matrix(obj$coords)
                data <- as.matrix(data)
                data <- data[, col.data]
        }
        else {
                if(is.list(obj)) {
                        coords <- as.matrix(obj$coords)
                        data <- as.matrix(obj$data)
                        data <- data[, col.data]
                        print("Object assumed to be a list with point coordinates in the first component and data in the second."
                                )
                }
                else {
                        if(is.data.frame(obj) | is.matrix(obj)) {
                                coords <- as.matrix(obj[, 1:2])
                                data <- obj[, 2 + col.data]
                                cat("plot.symbols: Object assumed to be a matrix with point coordinates in the first and second columns.\n"
                                        )
                        }
                }
        }
        if(window == T)
                trellis.device()
        ## par(mfrow = c(2, 2))
        ## par.ori <- par()
        par(pty = "s")
        par(mar = c(5, 5, 0, 0))
        ## plot(coords, xlab = "Coord X", ylab = "Coord Y", pch = 3, cex = 1.2)
        ##,main='\nData locations'
        ## plot(coords, xlab = "Coord X", ylab = "Coord Y",main='\nData locations',pch=3)
        data.quantile <- quantile(data)
        if(!is.null(polygon)) {
                x.lim <- range(polygon[, 1])
                y.lim <- range(polygon[, 2])
                if(diff(x.lim) > diff(y.lim))
                        y.lim <- y.lim + c( - ((diff(x.lim) - diff(y.lim))/
                                2), (diff(x.lim) - diff(y.lim))/2)
                if(diff(y.lim) > diff(x.lim))
                        x.lim <- x.lim + c( - ((diff(y.lim) - diff(x.lim))/
                                2), (diff(y.lim) - diff(x.lim))/2)
                plot(x.lim, y.lim, xlab = "Coord X", ylab = "Coord Y", type =
                        "n")
                lines(polygon)
        }
        else plot(coords, xlab = "Coord X", ylab = "Coord Y", type = "n")
        points(coords[(data <= data.quantile[2]), ], pch = 1, cex =
                0.59999999999999998, col = 2)
        points(coords[((data > data.quantile[2]) & (data <= data.quantile[
                3])), ], pch = 2, cex = 1.3999999999999999, col = 4)
        points(coords[((data > data.quantile[3]) & (data <= data.quantile[
                4])), ], pch = 3, cex = 1.7, col = 7)
        points(coords[(data > data.quantile[4]), ], pch = 4, cex = 2, col = 8)
        ## par(par.ori)
        return(invisible())
}

"plot.variogram" <-
  function(obj, max.dist = max(obj$u), ylim = "default", type = "b", scaled = F,
           var.lines = F, envelope.obj = NULL, bin.cloud = F, ...)
{
  if(obj$output.type == "bin")
    type <- "b"
  if(obj$output.type == "smooth")
    type <- "l"
  if(obj$output.type == "cloud")
    type <- "p"
  if(bin.cloud == T && type != "b")
    stop("object must be a binned variogram with option bin.cloud=T"
         )
  if(bin.cloud == T && all(is.na(obj$bin.cloud)))
    stop("object must be a binned variogram with option bin.cloud=T"
         )
  if(bin.cloud == T && any(!is.na(obj$bin.cloud)))
    boxplot(obj$bin.cloud, varwidth = T, names = as.character(
                                           obj$u), xlab = "midpoints of distance class", ylab =
            paste("variogram values / ", obj$estimator.type,
                  "estimator"), ...)
  else {
    u <- obj$u[obj$u <= max.dist]
    if(!is.matrix(obj$v)) {
      v <- obj$v[obj$u <= max.dist]
      ymax <- max(obj$v[obj$u <= max.dist])
    }
    else {
      v <- obj$v[obj$u <= max.dist, 1]
      ymax <- max(obj$v[obj$u <= max.dist, ])
    }
    if(!is.null(envelope.obj))
      ymax <- max(envelope.obj$v.upper, ymax)
    if(scaled)
      v <- v/obj$var.mark
    if(is.numeric(ylim))
      plot(u, v, xlim = c(0, max.dist), ylim = ylim, xlab =
           "Distance", ylab = "Semi-variance", type = type,
           ...)
    else plot(u, v, xlim = c(0, max.dist), ylim = c(0, ymax), xlab
              = "Distance", ylab = "Semi-variance", type =
              type, ...)
    if(var.lines == T) {
      if(scaled)
        abline(h = 1, lty = 3)
      else abline(h = obj$var.mark, lty = 3)
    }
    if(is.matrix(obj$v)) {
      for(k in 2:ncol(obj$v)) {
        v <- obj$v[obj$u <= max.dist, k]
        lines(u, v, type = "b")
      }
    }
    if(!is.null(envelope.obj)) {
      lines(u, envelope.obj$v.lower, lty = 4)
      lines(u, envelope.obj$v.upper, lty = 4)
    }
  }
  return(invisible())
}

"points.geodata" <-
  function(geodata, coords = geodata$coords, data = geodata$data, col.data = 1,
           pt.sizes = c("data.proportional", "rank.proportional", "quintiles",
             "quartiles", "deciles", "equal"), cex.min, cex.max, pch.seq, col.seq,
           add.to.plot = F, round.quantiles = F, graph.pars = FALSE, ...)
{
  if(is.R()){
    par.now <- par()
    on.exit(par(par.now))
  }
  pt.sizes <- match.arg(pt.sizes)
  if(add.to.plot == F) {
    plot(apply(coords, 2, range), type = "n", ...)
  }
  if(missing(cex.min))
    cex.min <- 0.5
  if(missing(cex.max))
    cex.max <- 2
  if(is.matrix(data))
    data <- as.vector(data[, col.data])
  graph.list <- list()
  if(pt.sizes == "quintiles" | pt.sizes == "quartiles" | pt.sizes ==
     "deciles") {
    if(pt.sizes == "quintiles") {
      n.quant <- 5
      if(missing(col.seq))
        col.seq <- c(6, 4, 7, 5, 8)
    }
    if(pt.sizes == "quartiles") {
      n.quant <- 4
      if(missing(col.seq))
        col.seq <- c(6, 4, 7, 8)
    }
    if(pt.sizes == "deciles") {
      n.quant <- 10
      if(missing(col.seq))
        col.seq <- c(6, 4, 2, 9, 11, 14, 12, 7, 5,
                     8)
    }
    if(missing(pch.seq))
      pch.seq <- rep(16, n.quant)
    cex.pt <- seq(cex.min, cex.max, l = n.quant)
    data.quantile <- quantile(data, probs = seq(0, 1, by = (1/
                                      n.quant)))
    if(round.quantiles == T) {
      data.quantile[1] <- floor(data.quantile[1])
      data.quantile[n.quant + 1] <- ceiling(data.quantile[
                                                          n.quant + 1])
      data.quantile <- round(data.quantile)
    }
    graph.list$quantiles <- data.quantile
    graph.list$cex <- cex.pt
    graph.list$col <- col.seq
    graph.list$pch <- pch.seq
    data.quantile[1] <- 0.90000000000000002 * data.quantile[1]
    for(i in 1:n.quant) {
      if(add.to.plot){
        points(coords[((data > data.quantile[i]) & (data <=
                                                    data.quantile[i + 1])), ], pch = pch.seq[
                                                                                  i], cex = cex.pt[i], col = col.seq[i], ...)
        points(coords[((data > data.quantile[i]) & (data <=
                                                    data.quantile[i + 1])), ], pch = 1, cex = 1.2 *
               cex.pt[i], col = 1, ...)
      }
      else{
        points(coords[((data > data.quantile[i]) & (data <=
                                                    data.quantile[i + 1])), ], pch = pch.seq[
                                                                                  i], cex = cex.pt[i], col = col.seq[i])
        points(coords[((data > data.quantile[i]) & (data <=
                                                    data.quantile[i + 1])), ], pch = 1, cex = 1.2 *
               cex.pt[i], col = 1)
      }
    }
  }
  if(pt.sizes == "rank.proportional") {
    if(missing(pch.seq))
      pch.seq <- 16
    data.quantile <- range(data)
    n <- length(data)
    coords.order <- coords[order(data), ]
    data.order <- data[order(data)]
    pt.size <- seq(cex.min, cex.max, l = n)
    graph.list$cex <- c(cex.min, cex.max)
    graph.list$pch <- unique(range(pch.seq))
    for(i in 1:n){
      if(add.to.plot)
        points(coords.order[i, , drop = F], cex = pt.size[
                                               i], pch = pch.seq, ...)
      else
        points(coords.order[i, , drop = F], cex = pt.size[
                                               i], pch = pch.seq)
    }
  }
  if(pt.sizes == "data.proportional") {
    if(missing(pch.seq))
      pch.seq <- 16
    r.y <- range(data)
    n <- length(data)
    size <- cex.min + ((data - r.y[1]) * (cex.max - cex.min))/
      (r.y[2] - r.y[1])
    graph.list$cex <- c(cex.min, cex.max)
    graph.list$pch <- unique(range(pch.seq))
    for(i in 1:n){
      if(add.to.plot)
        points(coords[i, , drop = F], cex = size[i], pch =
               pch.seq, ...)
      else
        points(coords[i, , drop = F], cex = size[i], pch =
               pch.seq)
    }
  }
  if(pt.sizes == "equal") {
    if(missing(pch.seq))
      pch.seq <- 16
    if(add.to.plot)
      points(coords, pch = pch.seq, ...)
    else
      points(coords, pch = pch.seq)
  }
  if(graph.pars == T)
    return(graph.list)
  else return(invisible())
}

"prior.control" <-
function(beta.prior = c("flat", "normal", "fixed"), beta = NULL, beta.var =
        NULL, sill.prior = c("reciprocal", "fixed"), sill = NULL, range.prior
         = c("uniform", "exponential", "fixed", "squared.reciprocal",
        "reciprocal"), exponential.prior.par = 1, range = NULL, range.discrete
         = NULL, nugget.prior = c("fixed", "uniform"), nugget = 0,
        nugget.discrete = NULL)
{
        beta.prior <- match.arg(beta.prior)
        sill.prior <- match.arg(sill.prior)
        range.prior <- match.arg(range.prior)
        nugget.prior <- match.arg(nugget.prior)
        return(list(beta.prior = beta.prior, beta = beta, beta.var = beta.var,
                sill.prior = sill.prior, sill = sill, range.prior = range.prior,
                exponential.prior.par = exponential.prior.par, range = range,
                range.discrete = range.discrete, nugget.prior = nugget.prior,
                nugget = nugget, nugget.discrete = nugget.discrete))
}

"proflik" <-
function(obj.likfit, geodata, coords = geodata$coords, data = geodata$data,
        nugget.values, nugget.rel.values, sill.values, range.values,
        lambda.values, sillrange.values = T, sillnugget.values = T,
        rangenugget.values = T, sillnugget.rel.values = F,
        rangenugget.rel.values = F, silllambda.values = F, rangelambda.values
         = T, nuggetlambda.values = F, nugget.rellambda.values = F, uni.only =
        T, bi.only = F, ...)
{
        ##
        ## 1. setting arguments
        ##
        call.fc <- match.call()
        n.cov.pars <- obj.likfit$npars - length(obj.likfit$beta)
        if(obj.likfit$transform.info$fix.lambda == F)
                n.cov.pars <- n.cov.pars - 1
        if(missing(sill.values))
                sill.values <- F
        if(missing(range.values))
                range.values <- F
        if(!is.null(obj.likfit$call$fix.nugget))
                if(obj.likfit$call$fix.nugget == T)
                        nugget.values <- nugget.rel.values <- F
        if(missing(nugget.values))
                nugget.values <- F
        if(missing(nugget.rel.values))
                nugget.rel.values <- F
        if(missing(lambda.values) | obj.likfit$transform.info$fix.lambda == T)
                lambda.values <- F
        if(uni.only == T) {
                sillrange.values <- sillnugget.values <- rangenugget.values <-
                        sillnugget.rel.values <- rangenugget.rel.values <-
                        silllambda.values <- rangelambda.values <-
                        nugget.rellambda.values <- nuggetlambda.values <- F
        }
        else {
                if(all(sillrange.values == T)) {
                        if(all(sill.values == F) | all(range.values == F)) {
                                sillrange.values <- F
                                stop("if argument sillrange.values = T sill.values and range.values must be provided. Alternatively a matrix can be provided in sillrange.values"
                                        )
                        }
                        else sillrange.values <- as.matrix(expand.grid(
                                        sill.values, range.values))
                }
                if(n.cov.pars == 2) {
                        sillnugget.values <- rangenugget.values <-
                                sillnugget.rel.values <- rangenugget.rel.values <-
                                nugget.rellambda.values <- nuggetlambda.values <-
                                F
                }
                else {
                        if(all(sillnugget.values == T)) {
                                if(all(sill.values == F) | all(nugget.values ==
                                        F)) {
                                        sillnugget.values <- F
                                        stop("if argument sillnugget.values = T sill.values and nugget.values must be provided. Alternatively a matrix can be provided in sillnugget.values"
                                                )
                                }
                                else sillnugget.values <- as.matrix(expand.grid(
                                                sill.values, nugget.values))
                        }
                        if(all(rangenugget.values == T)) {
                                if(all(range.values == F) | all(nugget.values ==
                                        F)) {
                                        rangenugget.values <- F
                                        stop("if argument rangenugget.values = T range.values and nugget.values must be provided. Alternatively a matrix can be provided in rangenugget.values"
                                                )
                                }
                                else rangenugget.values <- as.matrix(
                                                expand.grid(range.values,
                                                nugget.values))
                        }
                        if(all(sillnugget.rel.values == T)) {
                                if(all(sill.values == F) | all(
                                        nugget.rel.values == F)) {
                                        sillnugget.rel.values <- F
                                        stop("if argument sillnugget.rel.values = T sill.values and nugget.rel.values must be provided. Alternatively a matrix can be provided in sillnugget.rel.values"
                                                )
                                }
                                else sillnugget.rel.values <- as.matrix(
                                                expand.grid(sill.values,
                                                nugget.rel.values))
                        }
                        if(all(rangenugget.rel.values == T)) {
                                if(all(range.values == F) | all(
                                        nugget.rel.values == F)) {
                                        rangenugget.rel.values <- F
                                        stop("if argument rangenugget.rel.values = T range.values and nugget.rel.values must be provided. Alternatively a matrix can be provided in rangenugget.rel.values"
                                                )
                                }
                                else rangenugget.rel.values <- as.matrix(
                                                expand.grid(range.values,
                                                nugget.rel.values))
                        }
                        if(obj.likfit$transform.info$fix.lambda == T) {
                                if(all(nuggetlambda.values == T)) {
                                        if(all(lambda.values == F) | all(
                                                nugget.values == F)) {
                                                nuggetlambda.values <- F
                                                stop("if argument nuggetlambda.values = T lambda.values and nugget.values must be provided. Alternatively a matrix can be provided in nuggetlambda.values"
                                                        )
                                        }
                                        else nuggetlambda.values <- as.matrix(
                                                        expand.grid(
                                                        lambda.values,
                                                        nugget.values))
                                }
                                if(all(nugget.rellambda.values == T)) {
                                        if(all(lambda.values == F) | all(
                                                nugget.rel.values == F)) {
                                                nugget.rellambda.values <- F
                                                stop("if argument nugget.rellambda.values = T lambda.values and nugget.rel.values must be provided. Alternatively a matrix can be provided in nugget.rellambda.values"
                                                        )
                                        }
                                        else nugget.rellambda.values <-
                                                        as.matrix(expand.grid(
                                                        lambda.values,
                                                        nugget.rel.values))
                                }
                        }
                }
                if(obj.likfit$transform.info$fix.lambda == T)
                        silllambda.values <- rangelambda.values <- F
                else {
                        if(all(silllambda.values == T)) {
                                if(all(sill.values == F) | all(lambda.values ==
                                        F)) {
                                        silllambda.values <- F
                                        stop("if argument silllambda.values = T sill.values and lambda.values must be provided. Alternatively a matrix can be provided in silllambda.values"
                                                )
                                }
                                else silllambda.values <- as.matrix(expand.grid(
                                                sill.values, lambda.values))
                        }
                        if(all(rangelambda.values == T)) {
                                if(all(range.values == F) | all(lambda.values ==
                                        F)) {
                                        rangelambda.values <- F
                                        stop("if argument rangelambda.values = T range.values and lambda.values must be provided. Alternatively a matrix can be provided in rangelambda.values"
                                                )
                                }
                                else rangelambda.values <- as.matrix(
                                                expand.grid(range.values,
                                                lambda.values))
                        }
                }
        }
        ##
        ## 2. data preparation
        ##
        trend <- as.matrix(obj.likfit$trend.matrix)
        data <- as.vector(data)
        dimnames(trend) <- list(NULL, NULL)
        if(obj.likfit$transform.info$fix.lambda == T) {
                if(obj.likfit$lambda != 1) {
                        if(any(data <= 0))
                                stop("Data transformation not allowed when there are zeros or negative data"
                                        )
                        if(obj.likfit$lambda == 0)
                                data <- log(data)
                        else data <- ((data^obj.likfit$lambda) - 1)/obj.likfit$
                                        lambda
                }
        }
        n <- length(data)
        dists.vec <- dist(coords)
        d <- range(dists.vec)
        min.dist <- d[1]
        max.dist <- d[2]
        tausq <- obj.likfit$nugget
        sigmasq <- obj.likfit$cov.pars[1]
        tausq.rel <- tausq/sigmasq
        phi <- obj.likfit$cov.pars[2]
        lambda <- obj.likfit$lambda
        loglik <- obj.likfit$loglik
        sill.total <- sigmasq + tausq
        n.uni <- 0
        n.bi <- 0
        lower.phi <- 0
        upper.phi <- 1000 * max.dist
        lower.sigmasq <- 0.01 * sill.total
        result <- list()
        temp.list <- list(n = n, z = data, beta.size = dim(trend)[2], kappa =
                obj.likfit$kappa, xmat = trend, txmat = t(trend), method =
                obj.likfit$method, dists.lowertri = dists.vec, cov.model =
                obj.likfit$cov.model, fix.lambda = obj.likfit$transform.info$
                fix.lambda, lambda = obj.likfit$lambda, lower.phi = lower.phi,
                upper.phi = upper.phi, lower.sigmasq = lower.sigmasq, phi.est
                 = phi, tausq.rel.est = tausq.rel, tausq.est = tausq,
                sigmasq.est = sigmasq)
        if(obj.likfit$transform.info$fix.lambda == T) {
                temp.list$log.jacobian <- obj.likfit$transform.info$log.jacobian
        }
        ##
        ## 3. One-dimentional profile likelihoods
        ##
        ##
        ## 3.1 Profile for sigmasq
        ##
        if(bi.only == F) {
                if(any(sill.values != F)) {
                        n.uni <- n.uni + 1
                        cat("proflik: computing profile likelihood for the sill\n"
                                )
                        if(n.cov.pars == 2) {
                                if(tausq == 0) {
                                        if(any(lambda.values != F)) {
                                                temp.list$ini.grid <- as.matrix(
                                                        expand.grid(seq(min(
                                                        range.values), max(
                                                        range.values), l = 10),
                                                        seq(min(lambda.values),
                                                        max(lambda.values),
                                                        l = 5)))
                                        }
                                        else {
                                                temp.list$ini.grid <- as.matrix(
                                                        seq(min(range.values),
                                                        max(range.values),
                                                        l = 10))
                                        }
                                        dimnames(temp.list$ini.grid) <- list(
                                                NULL, NULL)
                                        pl.sigmasq <- apply(matrix(sill.values,
                                                ncol = 1), 1, proflik.aux2,
                                                temp.list)
                                        temp.list$ini.grid <- NULL
                                }
                                else {
                                        stop("not yet implemented for fixed nugget != 0"
                                                )
                                }
                        }
                        if(n.cov.pars == 3) {
                                if(any(lambda.values != F)) {
                                        ini.grid <- as.matrix(expand.grid(
                                                seq(min(range.values), max(
                                                range.values), l = 6), seq(
                                                0, 2 * tausq.rel, l = 4), seq(
                                                -1, 1, l = 5)))
                                }
                                else {
                                        ini.grid <- as.matrix(expand.grid(
                                                seq(min(range.values), max(
                                                range.values), l = 10), seq(
                                                0, 2 * tausq.rel, l = 4)))
                                }
                                dimnames(ini.grid) <- list(NULL, NULL)
                                temp.list$ini.grid <- ini.grid
                                pl.sigmasq <- apply(matrix(sill.values, ncol =
                                        1), 1, proflik.aux9, temp.list)
                                temp.list$ini.grid <- NULL
                        }
                        v.ord <- order(c(sigmasq, sill.values))
                        result$sill <- list(sill = c(sigmasq, sill.values)[
                                v.ord], proflik.sill = c(loglik, pl.sigmasq)[
                                v.ord], est.sill = c(sigmasq, loglik))
                }
                ##
                ## 3.2 Profile for phi
                ##
                if(any(range.values != F)) {
                        n.uni <- n.uni + 1
                        cat("proflik: computing profile likelihood for the range\n"
                                )
                        if(n.cov.pars == 2) {
                                if(tausq == 0) {
                                        temp.list$nugget <- 0
                                        pl.phi <- apply(matrix(range.values,
                                                ncol = 1), 1, proflik.aux0,
                                                temp.list)
                                        temp.list$nugget <- NULL
                                }
                                else {
                                        stop("not yet implemented for fixed nugget != 0"
                                                )
                                }
                        }
                        if(n.cov.pars == 3) {
                                pl.phi <- apply(matrix(range.values, ncol = 1),
                                        1, proflik.aux7, temp.list)
                        }
                        v.ord <- order(c(phi, range.values))
                        result$range <- list(range = c(phi, range.values)[
                                v.ord], proflik.range = c(loglik, pl.phi)[
                                v.ord], est.range = c(phi, loglik))
                }
                ##
                ## 3.3 Profile for \tau^2
                ##
                if(n.cov.pars == 3) {
                        if(any(nugget.values != F)) {
                                n.uni <- n.uni + 1
                                cat("proflik: computing profile likelihood for the nugget\n"
                                        )
                                pl.tausq <- apply(matrix(nugget.values, ncol =
                                        1), 1, proflik.aux11, temp.list)
                                v.ord <- order(c(tausq, nugget.values))
                                result$nugget <- list(nugget = c(tausq,
                                        nugget.values)[v.ord], proflik.nugget
                                         = c(loglik, pl.tausq)[v.ord],
                                        est.nugget = c(tausq, loglik))
                        }
                        ##
                        ## 3.4 Profile for relative \tau^2
                        ##
                        if(any(nugget.rel.values != F)) {
                                cat("proflik: computing profile likelihood for the relative nugget\n"
                                        )
                                n.uni <- n.uni + 1
                                pl.tausq.rel <- apply(matrix(nugget.rel.values,
                                        ncol = 1), 1, proflik.aux5, temp.list)
                                v.ord <- order(c(tausq.rel, nugget.rel.values))
                                result$nugget.rel <- list(nugget.rel = c(
                                        tausq.rel, nugget.rel.values)[v.ord],
                                        proflik.nugget = c(loglik, pl.tausq.rel
                                        )[v.ord], est.nugget.rel = c(tausq.rel,
                                        loglik))
                        }
                }
                ##
                ## 3.5 Profile for \lambda
                ##
                if(any(lambda.values != F)) {
                        temp.list$coords <- coords
                        if(dim(temp.list$xmat)[2] == 1 & all(temp.list$xmat ==
                                1))
                                temp.list$trend.mat <- "cte"
                        else {
                                assign(".temp.xmat",(temp.list$xmat[, -1]), where=1)
                                temp.list$trend.mat <- ~ .temp.xmat
                        }
                        n.uni <- n.uni + 1
                        cat("proflik: computing profile likelihood for the transformation parameter\n"
                                )
                        if(n.cov.pars == 2) {
                                if(tausq == 0) {
                                        temp.list$fixtau <- T
                                        temp.list$ini <- c(sigmasq, phi)
                                        pl.lambda <- apply(as.matrix(
                                                lambda.values), 1,
                                                proflik.aux23, temp.list)
                                        temp.list$fixtau <- temp.list$ini <-
                                                NULL
                                }
                                else {
                                        stop("not yet implemented for fixed nugget != 0"
                                                )
                                }
                        }
                        if(n.cov.pars == 3) {
                                temp.list$fixtau <- F
                                temp.list$ini <- phi
                                pl.lambda <- apply(matrix(lambda.values, ncol
                                         = 1), 1, proflik.aux23, temp.list)
                                temp.list$fixtau <- temp.list$ini <- NULL
                        }
                        v.ord <- order(c(lambda, lambda.values))
                        result$lambda <- list(lambda = c(lambda, lambda.values)[
                                v.ord], proflik.lambda = c(loglik, pl.lambda)[
                                v.ord], est.lambda = c(lambda, loglik))
                        if(exists(".temp.xmat", w = 1))
                                remove(".temp.xmat", where = 1)
                }
        }
        ##
        ## 4. Two-dimentional profile likelihoods
        ##
        ##
        ## 4.1 Profile for \sigma^2 and \phi
        ##
        ##
        ##
        if(uni.only == F) {
                if(any(sillrange.values != F)) {
                        n.bi <- n.bi + 1
                        cat("proflik: computing 2-D profile likelihood for the sill and range parameters\n"
                                )
                        if(n.cov.pars == 2) {
                                if(tausq == 0) {
                                        temp.list$nugget <- 0
                                        if(temp.list$fix.lambda == T) {
                                                pl.sigmasqphi <- apply(cbind(
                                                        0, sillrange.values,
                                                        1), 1, loglik.spatial,
                                                        temp.list = temp.list)
                                        }
                                        else {
                                                pl.sigmasqphi <- apply(
                                                        sillrange.values, 1,
                                                        proflik.aux28,
                                                        temp.list)
                                        }
                                        temp.list$nugget <- NULL
                                }
                                else {
                                        stop("not yet implemented for fixed nugget != 0"
                                                )
                                }
                        }
                        if(n.cov.pars == 3) {
                                pl.sigmasqphi <- apply(sillrange.values, 1,
                                        proflik.aux13, temp.list)
                        }
                        names(pl.sigmasqphi) <- NULL
                        result$sillrange <- list(sill = as.numeric(levels(
                                as.factor(sillrange.values[, 1]))), range =
                                as.numeric(levels(as.factor(sillrange.values[
                                , 2]))), proflik.sillrange = pl.sigmasqphi,
                                est.sillrange = c(sigmasq, phi, loglik))
                }
                ##
                ## 4.2 Profile for \sigma^2 and \tau^2
                ##
                if(any(sillnugget.values != F)) {
                        n.bi <- n.bi + 1
                        cat("proflik: computing 2-D profile likelihood for the sill and nugget\n"
                                )
                        if(obj.likfit$transform.info$fix.lambda == F)
                                ini.grid <- as.matrix(expand.grid(seq(min(
                                        range.values), max(range.values), l =
                                        10), seq(-1, 1, l = 5)))
                        else ini.grid <- as.matrix(seq(min(range.values), max(
                                        range.values), l = 10))
                        dimnames(ini.grid) <- list(NULL, NULL)
                        temp.list$ini.grid <- ini.grid
                        pl.sigmasqtausq <- apply(sillnugget.values, 1,
                                proflik.aux15, temp.list)
                        temp.list$ini.grid <- NULL
                        names(pl.sigmasqtausq) <- NULL
                        result$sillnugget <- list(sill = as.numeric(levels(
                                as.factor(sillnugget.values[, 1]))), nugget =
                                as.numeric(levels(as.factor(sillnugget.values[
                                , 2]))), proflik.sillnugget = pl.sigmasqtausq,
                                est.sillrange = c(sigmasq, tausq, loglik))
                }
                ##
                ## 4.3 Profile for \phi and \tau^2
                ##
                if(any(rangenugget.values != F)) {
                        n.bi <- n.bi + 1
                        cat("proflik: computing 2-D profile likelihood for the range and nugget\n"
                                )
                        temp.list$ini.grid <- as.matrix(seq(sigmasq/4, 5 *
                                sigmasq, l = 15))
                        pl.phitausq <- apply(rangenugget.values, 1,
                                proflik.aux17, temp.list)
                        temp.list$ini.grid <- NULL
                        names(pl.phitausq) <- NULL
                        result$rangenugget <- list(range = as.numeric(levels(
                                as.factor(rangenugget.values[, 1]))), nugget =
                                as.numeric(levels(as.factor(rangenugget.values[
                                , 1]))), proflik.rangenugget = pl.phitausq,
                                est.rangenugget = c(phi, tausq, loglik))
                }
                ##
                ## 4.4 Profile for \sigma^2 and \tau^2_{rel}
                ##
                if(any(sillnugget.rel.values != F)) {
                        n.bi <- n.bi + 1
                        cat("proflik: computing 2-D profile likelihood for the sill and relative nugget parameters\n"
                                )
                        if(temp.list$fix.lambda == F)
                                ini.grid <- as.matrix(expand.grid(seq(min(
                                        range.values), max(range.values), l =
                                        10), seq(-1, 1, l = 5)))
                        else ini.grid <- as.matrix(seq(min(range.values), max(
                                        range.values), l = 10))
                        dimnames(ini.grid) <- list(NULL, NULL)
                        temp.list$ini.grid <- ini.grid
                        pl.sigmasqtausq.rel <- apply(sillnugget.rel.values,
                                1, proflik.aux19, temp.list)
                        names(pl.sigmasqtausq.rel) <- NULL
                        temp.list$ini.grid <- NULL
                        result$sillnugget.rel <- list(sill = as.numeric(levels(
                                as.factor(sillnugget.rel.values[, 1]))),
                                nugget.rel = as.numeric(levels(as.factor(
                                sillnugget.rel.values[, 2]))),
                                proflik.sillnugget.rel = pl.sigmasqtausq.rel,
                                est.sillrange.rel = c(sigmasq, tausq.rel,
                                loglik))
                }
                ##
                ## 4.5 Profile for \phi and \tau^2_{rel}
                ##
                if(any(rangenugget.rel.values != F)) {
                        n.bi <- n.bi + 1
                        cat("proflik: computing 2-D profile likelihood for the range and relative nugget parameters\n"
                                )
                        pl.phitausq.rel <- apply(rangenugget.rel.values, 1,
                                proflik.aux30, temp.list)
                        names(pl.phitausq.rel) <- NULL
                        result$rangenugget.rel <- list(range = as.numeric(
                                levels(as.factor(rangenugget.rel.values[, 1]))),
                                nugget.rel = as.numeric(levels(as.factor(
                                rangenugget.rel.values[, 2]))),
                                proflik.rangenugget.rel = pl.phitausq.rel,
                                "est.rangenugget .rel" = c(phi, tausq.rel,
                                loglik))
                }
        }
        ##
        ## 4.6 Profile for \sigma^2 and \lambda
        ##
        if(any(silllambda.values != F)) {
                n.bi <- n.bi + 1
                cat("proflik: computing 2-D profile likelihood for the sill and transformation parameters\n"
                        )
                if(n.cov.pars == 2) {
                        ini.grid <- as.matrix(seq(min(range.values), max(
                                range.values), l = 10))
                        dimnames(ini.grid) <- list(NULL, NULL)
                        temp.list$ini.grid <- ini.grid
                        if(tausq == 0) {
                                temp.list$nugget <- 0
                                pl.sigmasqlambda <- apply(silllambda.values,
                                        1, proflik.aux24, temp.list)
                                temp.list$ini.grid <- temp.list$nugget <- NULL
                        }
                        else {
                                stop("not yet implemented for fixed nugget != 0"
                                        )
                        }
                }
                if(n.cov.pars == 3) {
                        ini.grid <- as.matrix(expand.grid(seq(min(range.values),
                                max(range.values), l = 10), seq(0, 1, l = 5)))
                        dimnames(ini.grid) <- list(NULL, NULL)
                        temp.list$ini.grid <- ini.grid
                        pl.sigmasqlambda <- apply(sigmasqlambda.values, 1,
                                proflik.aux27, temp.list)
                        temp.list$ini.grid <- NULL
                }
                names(pl.sigmasqlambda) <- NULL
                result$silllambda <- list(sill = as.numeric(levels(as.factor(
                        silllambda.values[, 1]))), lambda = as.numeric(levels(
                        as.factor(silllambda.values[, 2]))), proflik.silllambda
                         = pl.sigmasqlambda, est.silllambda = c(sigmasq, lambda,
                        loglik))
        }
        ##
        ## 4.7 Profile for \phi and \lambda
        ##
        if(any(rangelambda.values != F)) {
                n.bi <- n.bi + 1
                cat("proflik: computing 2-D profile likelihood for the range and transformation parameters\n"
                        )
                if(n.cov.pars == 2) {
                        if(tausq == 0) {
                                temp.list$nugget <- 0
                                pl.philambda <- apply(rangelambda.values, 1,
                                        proflik.aux29, temp.list)
                                temp.list$nugget <- NULL
                        }
                        else {
                                stop("not yet implemented for fixed nugget != 0"
                                        )
                        }
                }
                if(n.cov.pars == 3) {
                        pl.philambda <- apply(rangelambda.values, 1,
                                proflik.aux31, temp.list)
                }
                names(pl.philambda) <- NULL
                result$rangelambda <- list(range = as.numeric(levels(as.factor(
                        rangelambda.values[, 1]))), lambda = as.numeric(levels(
                        as.factor(rangelambda.values[, 2]))),
                        proflik.rangelambda = pl.philambda, est.rangelambda = c(
                        phi, lambda, loglik))
        }
        ##
        ## ## 4.8 Profile for \tau^2 and \lambda
        ## #
        if(any(nuggetlambda.values != F)) {
                n.bi <- n.bi + 1
                cat("proflik: computing 2-D profile likelihood for the nugget and transformation parameters\n"
                        )
                pl.nuggetlambda <- apply(nuggetlambda.values, 1, proflik.aux32,
                        temp.list)
                names(pl.nuggetlambda) <- NULL
                result$nuggetlambda <- list(nugget = as.numeric(levels(
                        as.factor(nuggetlambda.values[, 1]))), lambda =
                        as.numeric(levels(as.factor(nuggetlambda.values[, 2]))),
                        proflik.nuggetlambda = pl.nuggetlambda,
                        est.nuggetlambda = c(tausq, lambda, loglik))
        }
        ##
        ## 4.9 2-D Profile for \tau^2_{rel} and \lambda
        ##
        if(any(nugget.rellambda.values != F)) {
                n.bi <- n.bi + 1
                pl.nugget.rellambda <- apply(nugget.rellambda.values, 1,
                        proflik.aux33, temp.list)
                names(pl.nugget.rellambda) <- NULL
                result$nugget.rellambda <- list(nugget.rel = as.numeric(levels(
                        as.factor(nugget.rellambda.values[, 1]))), lambda =
                        as.numeric(levels(as.factor(nugget.rellambda.values[
                        , 2]))), proflik.nugget.rellambda = pl.nugget.rellambda,
                        est.nugget.rellambda = c(tausq.rel, lambda, loglik))
        }
        result$n.uni <- n.uni
        result$n.bi <- n.bi
        result$method <- obj.likfit$method
        result$call <- call.fc
        class(result) <- "proflik"
        return(result)
}

"proflik.aux0" <-
function(phi, temp.list)
{
        ## This function computes the value of the profile likelihood for the correlation parameter \phi when nugget effect is not included in the model.
        ## It requires the minimisation of the function wrt \phi for each value of \lambda, if this transformation parameter is included in the model
        ## This is an auxiliary function called by likfit.proflik
        ##
        if(temp.list$fix.lambda == T) proflik <- proflik.aux1(phi = phi,
                        temp.list = temp.list) else {
                temp.list$phi <- phi
                proflik <- - (nlminb(temp.list$lambda, proflik.aux1.1, lower
                         = -3, upper = 3, temp.list = temp.list)$objective)
        }
        return(proflik)
}

"proflik.aux1" <-
function(phi, temp.list)
{
        ## This function computes the value of the profile likelihood for the correlation function scale parameter \phi when nugget effect = 0
        z <- temp.list$z
        n <- temp.list$n
        beta.size <- temp.list$beta.size
        kappa <- temp.list$kappa
        xmat <- temp.list$xmat
        covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
                cov.model = temp.list$cov.model, kappa = kappa, nugget =
                temp.list$nugget, cov.pars = c(1, phi), inv = T, det = T)
        txiv <- crossprod(xmat, covinf$inverse)
        ixix <- solve(txiv %*% xmat)
        ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
                z
        if(temp.list$method == "ML") {
                proflik <- - (n/2) * log(2 * pi) - covinf$log.det.to.half -
                        (n/2) * log(ssresmat/n) - (n/2) + temp.list$
                        log.jacobian
        }
        if(temp.list$method == "RML") {
                eigentrem <- eigen(ixix, symmetric = T, only.values = T)
                proflik <- - ((n - beta.size)/2) * log(2 * pi) - covinf$
                        log.det.to.half - ((n - beta.size)/2) * log(ssresmat/
                        n) - (n/2) + 0.5 * sum(log(eigentrem$values)) +
                        temp.list$log.jacobian
        }
        return(proflik)
}

"proflik.aux1.1" <-
function(lambda, temp.list)
{
        ## This function computes the value of the profile likelihood for the correlation function scale parameter \phi when nugget effect = 0
        z <- temp.list$z
        n <- temp.list$n
        phi <- temp.list$phi
        if(lambda == 1) {
                temp.list$log.jacobian <- 0
        }
        else {
                if(any(z <= 0))
                        stop("Transformation option not allowed when there are zeros or negative data"
                                )
                if(any(z^(lambda - 1) <= 0))
                        temp.list$log.jacobian <- log(prod(z^(lambda - 1)))
                else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
                if(lambda == 0)
                        z <- log(z)
                else z <- ((z^lambda) - 1)/lambda
        }
        beta.size <- temp.list$beta.size
        kappa <- temp.list$kappa
        xmat <- temp.list$xmat
        covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
                cov.model = temp.list$cov.model, kappa = kappa, nugget =
                temp.list$nugget, cov.pars = c(1, phi), inv = T, det = T)
        txiv <- crossprod(xmat, covinf$inverse)
        ixix <- solve(txiv %*% xmat)
        ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
                z
        if(temp.list$method == "ML") {
                neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + (
                        n/2) * log(ssresmat/n) + (n/2) - temp.list$log.jacobian
        }
        if(temp.list$method == "RML") {
                eigentrem <- eigen(ixix, symmetric = T, only.values = T)
                neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
                        log.det.to.half + ((n - beta.size)/2) * log(ssresmat/
                        n) + (n/2) - 0.5 * sum(log(eigentrem$values)) -
                        temp.list$log.jacobian
        }
        return(as.vector(neglik))
}

"proflik.aux10" <-
function(phitausq.rel.lambda, temp.list)
{
        phitausq.rel.lambda <- as.vector(phitausq.rel.lambda)
        if(length(phitausq.rel.lambda) == 3)
                include.lambda <- TRUE
        else include.lambda <- FALSE
        z <- temp.list$z
        n <- length(z)
        phi <- phitausq.rel.lambda[1]
        tausq <- phitausq.rel.lambda[2]
        if(include.lambda) {
                lambda <- phitausq.rel.lambda[3]
                if(lambda == 1) {
                        temp.list$log.jacobian <- 0
                }
                else {
                        if(any(z <= 0))
                                stop("Transformation option not allowed when there are zeros or negative data"
                                        )
                        if(any(z^(lambda - 1) <= 0))
                                temp.list$log.jacobian <- log(prod(z^(lambda -
                                        1)))
                        else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
                        if(lambda == 0)
                                z <- log(z)
                        else z <- ((z^lambda) - 1)/lambda
                }
        }
        beta.size <- temp.list$beta.size
        kappa <- temp.list$kappa
        xmat <- temp.list$xmat
        sigmasq <- temp.list$sigmasq
        covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
                cov.model = temp.list$cov.model, kappa = kappa, nugget = tausq,
                cov.pars = c(1, phi), inv = T, det = T)
        txiv <- crossprod(xmat, covinf$inverse)
        ixix <- solve(txiv %*% xmat)
        ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
                z
        if(temp.list$method == "ML") {
                neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + (
                        n/2) * log(sigmasq) + (0.5/sigmasq) * ssresmat -
                        temp.list$log.jacobian
        }
        if(temp.list$method == "RML") {
                eigentrem <- eigen(ixix, symmetric = T, only.values = T)
                neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
                        log.det.to.half + ((n - beta.size)/2) * log(sigmasq) +
                        (0.5/sigmasq) * ssresmat - 0.5 * sum(log(eigentrem$
                        values)) - temp.list$log.jacobian
        }
        return(as.vector(neglik))
}

"proflik.aux11" <-
function(tausq, temp.list)
{
        ## This function computes the value of the profile likelihood for the parameter \tau^2.
        ## It requires the minimisation of the function wrt \sigma^2, \phi and \lambda (if the case) for each value of \tau^2.
        ## This is an auxiliary function called by proflik.
        temp.list$nugget <- as.vector(tausq)
        if(temp.list$fix.lambda == T) {
                sigmasqphi.res <- nlminb(c(temp.list$sigmasq.est, temp.list$
                        phi.est), proflik.aux12, lower = c(temp.list$
                        lower.sigmasq, temp.list$lower.phi), temp.list =
                        temp.list)
        }
        else {
                sigmasqphi.res <- nlminb(c(temp.list$sigmasq.est, temp.list$
                        phi.est, temp.list$lambda), proflik.aux12, lower = c(
                        temp.list$lower.sigmasq, temp.list$lower.phi, -3),
                        upper = c( + Inf, + Inf, 3), temp.list = temp.list)
        }
        return( - sigmasqphi.res$objective)
}

"proflik.aux12" <-
function(sigmasqphi.lambda, temp.list)
{
        ## This function computes the value of the profile likelihood for the nugget parameter \tau^2, minimizing the likelihood wrt correlation function scale parameter \phi (range), the random field scale parameter \sigma^2 (sill) and the transformation parameter \lambda.
        sigmasqphi.lambda <- as.vector(sigmasqphi.lambda)
        if(length(sigmasqphi.lambda) == 3)
                include.lambda <- TRUE
        else include.lambda <- FALSE
        z <- temp.list$z
        n <- temp.list$n
        sigmasq <- sigmasqphi.lambda[1]
        phi <- sigmasqphi.lambda[2]
        if(include.lambda) {
                lambda <- sigmasqphi.lambda[3]
                if(lambda == 1) {
                        temp.list$log.jacobian <- 0
                }
                else {
                        if(any(z <= 0))
                                stop("Transformation option not allowed when there are zeros or negative data"
                                        )
                        if(any(z^(lambda - 1) <= 0))
                                temp.list$log.jacobian <- log(prod(z^(lambda -
                                        1)))
                        else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
                        if(lambda == 0)
                                z <- log(z)
                        else z <- ((z^lambda) - 1)/lambda
                }
        }
        beta.size <- temp.list$beta.size
        kappa <- temp.list$kappa
        xmat <- temp.list$xmat
        covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
                cov.model = temp.list$cov.model, kappa = kappa, nugget =
                temp.list$nugget, cov.pars = c(sigmasq, phi), inv = T, det = T,
                sca = F)
        txiv <- crossprod(xmat, covinf$inverse)
        ixix <- solve(txiv %*% xmat)
        ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
                z
        if(temp.list$method == "ML") {
                neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + 0.5 *
                        (ssresmat) - temp.list$log.jacobian
        }
        if(temp.list$method == "RML") {
                eigentrem <- eigen(ixix, symmetric = T, only.values = T)
                neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
                        log.det.to.half + 0.5 * (ssresmat) - 0.5 * sum(log(
                        eigentrem$values)) - temp.list$log.jacobian
        }
        return(as.vector(neglik))
}

"proflik.aux13" <-
function(sigmasqphi, temp.list)
{
        ## This function computes the value of the 2-D profile likelihood for the parameters \sigma^2 and \phi when the nugget is included.
        ## It requires the minimisation of the function wrt \tau^2 and \lambda (if the case) for each value of (\sigma^2, \phi)
        ## This is an auxiliary function called by likfit.proflik
        temp.list$sigmasqphi <- as.vector(sigmasqphi)
        if(temp.list$fix.lambda == T) {
                tausq.res <- nlminb(temp.list$tausq.est, proflik.aux14, lower
                         = 0, temp.list = temp.list)
        }
        else {
                tausq.res <- nlminb(c(temp.list$tausq.est, temp.list$lambda),
                        proflik.aux14, lower = c(0, -3), upper = c( + Inf,
                        3), temp.list = temp.list)
        }
        return( - tausq.res$objective)
}

"proflik.aux14" <-
function(tausq.lambda, temp.list)
{
        ## This function computes the value of the 2-D profile likelihood for the parameters (\sigma^2, \phi), minimizing the likelihood wrt the nugget parameter \tau^2.
        ## This functions is called by the auxiliary function proflik.aux13
        if(length(tausq.lambda) == 2) include.lambda <- TRUE else
                        include.lambda <- FALSE
        z <- temp.list$z
        n <- temp.list$n
        tausq <- tausq.lambda[1]
        if(include.lambda) {
                lambda <- tausq.lambda[2]
                if(lambda == 1) {
                        temp.list$log.jacobian <- 0
                }
                else {
                        if(any(z <= 0))
                                stop("Transformation option not allowed when there are zeros or negative data"
                                        )
                        if(any(z^(lambda - 1) <= 0))
                                temp.list$log.jacobian <- log(prod(z^(lambda -
                                        1)))
                        else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
                        if(lambda == 0)
                                z <- log(z)
                        else z <- ((z^lambda) - 1)/lambda
                }
        }
        beta.size <- temp.list$beta.size
        kappa <- temp.list$kappa
        xmat <- temp.list$xmat
        covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
                cov.model = temp.list$cov.model, kappa = kappa, nugget = tausq,
                cov.pars = temp.list$sigmasqphi, inv = T, det = T)
        txiv <- crossprod(xmat, covinf$inverse)
        ixix <- solve(txiv %*% xmat)
        ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
                z
        if(temp.list$method == "ML") {
                neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + 0.5 *
                        ssresmat - temp.list$log.jacobian
        }
        if(temp.list$method == "RML") {
                eigentrem <- eigen(ixix, symmetric = T, only.values = T)
                neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
                        log.det.to.half + 0.5 * ssresmat - 0.5 * sum(log(
                        eigentrem$values)) - temp.list$log.jacobian
        }
        return(as.vector(neglik))
}

"proflik.aux15" <-
function(sigmasqtausq, temp.list)
{
        ## This function computes the value of the 2-D profile likelihood for the parameters \sigma^2 and \tau^2
        ## It requires the minimisation of the function wrt \phi and also \lambda (if the case) for each value of (\sigma^2, \tau^2)
        ## This is an auxiliary function called by likfit.proflik
        temp.list$sigmasqtausq <- as.vector(sigmasqtausq)
        if(temp.list$fix.lambda == T) {
                phi.res <- nlminb(temp.list$phi.est, proflik.aux16, lower =
                        temp.list$lower.phi, temp.list = temp.list)
        }
        else {
                phi.res <- nlminb(c(temp.list$phi.est, temp.list$lambda),
                        proflik.aux16, lower = c(temp.list$lower.phi, -3),
                        upper = c( + Inf, 3), temp.list = temp.list)
        }
        return( - phi.res$objective)
}

"proflik.aux16" <-
function(phi.lambda, temp.list)
{
        ## This function computes the value of the 2-D profile likelihood for the sill and nugget parameters (\sigma^2,\tau^2), minimising the profile likelihood wrt correlation function scale parameter \phi
        ## This is an auxiliary function called by likfit.aux15
        if(length(phi.lambda) == 2) include.lambda <- TRUE else include.lambda <-
                        FALSE
        z <- temp.list$z
        n <- temp.list$n
        phi <- phi.lambda[1]
        if(include.lambda) {
                lambda <- phi.lambda[2]
                if(lambda == 1) {
                        temp.list$log.jacobian <- 0
                }
                else {
                        if(any(z <= 0))
                                stop("Transformation option not allowed when there are zeros or negative data"
                                        )
                        if(any(z^(lambda - 1) <= 0))
                                temp.list$log.jacobian <- log(prod(z^(lambda -
                                        1)))
                        else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
                        if(lambda == 0)
                                z <- log(z)
                        else z <- ((z^lambda) - 1)/lambda
                }
        }
        beta.size <- temp.list$beta.size
        kappa <- temp.list$kappa
        xmat <- temp.list$xmat
        covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
                cov.model = temp.list$cov.model, kappa = kappa, nugget =
                temp.list$sigmasqtausq[2], cov.pars = c(temp.list$sigmasqtausq[
                1], phi), inv = T, det = T)
        txiv <- crossprod(xmat, covinf$inverse)
        ixix <- solve(txiv %*% xmat)
        ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
                z
        if(temp.list$method == "ML") {
                neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + 0.5 *
                        ssresmat - temp.list$log.jacobian
        }
        if(temp.list$method == "RML") {
                eigentrem <- eigen(ixix, symmetric = T, only.values = T)
                neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
                        log.det.to.half + 0.5 * ssresmat - 0.5 * sum(log(
                        eigentrem$values)) - temp.list$log.jacobian
        }
        return(as.vector(neglik))
}

"proflik.aux17" <-
function(phitausq, temp.list)
{
        ## This function computes the value of the 2-D profile likelihood for the parameters (\phi, \tau^2)
        ## It requires the minimisation of the function wrt \sigma^2 and \lambda (if the case) for each value of (\phi, \tau^2)
        ## This is an auxiliary function called by likfit.proflik
        temp.list$phitausq <- as.vector(phitausq)
        if(temp.list$fix.lambda == T) {
                sigmasq.res <- nlminb(temp.list$sigmasq.est, proflik.aux18,
                        lower = temp.list$lower.sigmasq, temp.list = temp.list)
        }
        else {
                sigmasq.res <- nlminb(c(temp.list$sigmasq.est, temp.list$lambda
                        ), proflik.aux18, lower = c(temp.list$lower.sigmasq,
                        -3), upper = c( + Inf, 3), temp.list = temp.list)
        }
        return( - sigmasq.res$objective)
}

"proflik.aux18" <-
function(sigmasq.lambda, temp.list)
{
        ## This function computes the value of the 2-D profile likelihood for the range and nugget parameters (\phi, \tau^2), minimising the likelihood wrt the random field scale parameter \sigma^2 (sill) ant the transformation parameter \lambda.
        ## This is an auxiliary function called by likfit.aux17.
        if(length(sigmasq.lambda) == 2) include.lambda <- TRUE else
                        include.lambda <- FALSE
        z <- temp.list$z
        n <- temp.list$n
        sigmasq <- sigmasq.lambda[1]
        if(include.lambda) {
                lambda <- sigmasq.lambda[2]
                if(lambda == 1) {
                        temp.list$log.jacobian <- 0
                }
                else {
                        if(any(z <= 0))
                                stop("Transformation option not allowed when there are zeros or negative data"
                                        )
                        if(any(z^(lambda - 1) <= 0))
                                temp.list$log.jacobian <- log(prod(z^(lambda -
                                        1)))
                        else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
                        if(lambda == 0)
                                z <- log(z)
                        else z <- ((z^lambda) - 1)/lambda
                }
        }
        beta.size <- temp.list$beta.size
        kappa <- temp.list$kappa
        xmat <- temp.list$xmat
        covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
                cov.model = temp.list$cov.model, kappa = kappa, nugget =
                temp.list$phitausq[2], cov.pars = c(sigmasq, temp.list$phitausq[
                1]), inv = T, det = T)
        txiv <- crossprod(xmat, covinf$inverse)
        ixix <- solve(txiv %*% xmat)
        ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
                z
        if(temp.list$method == "ML") {
                neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + 0.5 *
                        ssresmat - temp.list$log.jacobian
        }
        if(temp.list$method == "RML") {
                eigentrem <- eigen(ixix, symmetric = T, only.values = T)
                neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
                        log.det.to.half + 0.5 * ssresmat - 0.5 * sum(log(
                        eigentrem$values)) - temp.list$log.jacobian
        }
        return(as.vector(neglik))
}

"proflik.aux19" <-
function(sigmasqtausq.rel, temp.list)
{
        ## This function computes the value of the 2-D profile likelihood for the parameters (\sigma^2, \tau^2_{rel})
        ## It requires the minimisation of the function wrt \phi and \lambda (if the case) for each value of (\sigma^2, \tau^2_{rel})
        ## This is an auxiliary function called by likfit.proflik
        temp.list$sigmasqtausq.rel <- as.vector(sigmasqtausq.rel)
        if(temp.list$fix.lambda == T) {
                phi.res <- nlminb(temp.list$phi.est, proflik.aux20, lower =
                        temp.list$lower.phi, temp.list = temp.list)
        }
        else {
                phi.res <- nlminb(c(temp.list$phi.est, temp.list$lambda),
                        proflik.aux20, lower = c(temp.list$lower.phi, -3),
                        upper = c( + Inf, 3), temp.list = temp.list)
        }
        return( - phi.res$objective)
}

"proflik.aux2" <-
function(sigmasq, temp.list)
{
        ## This function computes the value of the profile likelihood for the random field scale (variance) parameter \sigma^2 when nugget effect is not included in the model.
        ## It requires the minimisation of the function wrt \phi for each value of \sigma^2
        ## This is an auxiliary function called by likfit.proflik
        ##
        temp.list$sigmasq <- as.vector(sigmasq)
        ## phi.res <- nlminb(temp.list$phi.est, proflik.aux3, lower =
        ## temp.list$lower.phi, temp.list = temp.list)
        if(temp.list$fix.lambda == T) {
                ini.lik <- round(100000000. * apply(temp.list$ini.grid, 1,
                        proflik.aux3, temp.list))
                ini <- as.vector(temp.list$ini.grid[ini.lik == min(ini.lik),
                        ])
                phi.res <- nlminb(ini, proflik.aux3, lower = temp.list$
                        lower.phi, temp.list = temp.list)
        }
        else {
                ini.lik <- round(100000000. * apply(temp.list$ini.grid, 1,
                        proflik.aux3, temp.list))
                ini <- as.vector(temp.list$ini.grid[ini.lik == min(ini.lik),
                        ])
                phi.res <- nlminb(ini, proflik.aux3, lower = c(temp.list$
                        lower.phi, -3), upper = c( + Inf, 3), temp.list =
                        temp.list)
        }
        return( - phi.res$objective)
}

"proflik.aux20" <-
function(phi.lambda, temp.list)
{
        ## This function computes the value of the 2-D profile likelihood for the sill and relative nugget parameters (\sigma^2, \tau^2_{rel}), minimising the likelihood wrt the correlation function scale parameter \phi and the transformation parameter \lambda.
        ## This is an auxiliary function called by likfit.aux19.
        if(length(phi.lambda) == 2) include.lambda <- TRUE else include.lambda <-
                        FALSE
        phi.lambda <- as.vector(phi.lambda)
        sigmasqtausq.rel <- as.vector(temp.list$sigmasqtausq.rel)
        sigmasq <- sigmasqtausq.rel[1]
        tausq.rel <- sigmasqtausq.rel[2]
        phi <- phi.lambda[1]
        lambda <- phi.lambda[2]
        z <- temp.list$z
        n <- temp.list$n
        if(include.lambda) {
                lambda <- phi.lambda[2]
                if(lambda == 1) {
                        temp.list$log.jacobian <- 0
                }
                else {
                        if(any(z <= 0))
                                stop("Transformation option not allowed when there are zeros or negative data"
                                        )
                        if(any(z^(lambda - 1) <= 0))
                                temp.list$log.jacobian <- log(prod(z^(lambda -
                                        1)))
                        else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
                        if(lambda == 0)
                                z <- log(z)
                        else z <- ((z^lambda) - 1)/lambda
                }
        }
        beta.size <- temp.list$beta.size
        kappa <- temp.list$kappa
        xmat <- temp.list$xmat
        covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
                cov.model = temp.list$cov.model, kappa = kappa, nugget =
                tausq.rel, cov.pars = c(1, phi), inv = T, det = T)
        txiv <- crossprod(xmat, covinf$inverse)
        ixix <- solve(txiv %*% xmat)
        ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
                z
        if(temp.list$method == "ML") {
                neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + (
                        n/2) * log(sigmasq) + (0.5/sigmasq) * ssresmat
                ssresmat - temp.list$log.jacobian
        }
        if(temp.list$method == "RML") {
                eigentrem <- eigen(ixix, symmetric = T, only.values = T)
                neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
                        log.det.to.half + (n/2) * log(sigmasq) + (0.5/sigmasq) *
                        ssresmat - 0.5 * sum(log(eigentrem$values)) - temp.list$
                        log.jacobian
        }
        return(as.vector(neglik))
}

"proflik.aux21" <-
function(phitausq.rel, temp.list)
{
        ## This function computes the value of the 2-D profile likelihood for the parameters (\phi, \tau^2_{rel})
        ## This is an auxiliary function called by likfit.proflik
        phitausq.rel <- as.vector(phitausq.rel)
        phi <- phitausq.rel[1]
        tausq.rel <- phitausq.rel[2]
        z <- temp.list$z
        n <- temp.list$n
        beta.size <- temp.list$beta.size
        kappa <- temp.list$kappa
        xmat <- temp.list$xmat
        txmat <- temp.list$txmat
        covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
                cov.model = temp.list$cov.model, kappa = kappa, nugget =
                tausq.rel, cov.pars = c(1, phi), inv = T, det = T)
        ixix <- solve(txmat %*% covinf$inverse %*% xmat)
        ssresmat <- z %*% (covinf$inverse - covinf$inverse %*% xmat %*% ixix %*%
                txmat %*% covinf$inverse) %*% z
        sigmasq.hat <- ssresmat/n
        if(temp.list$method == "ML") {
                proflik <- - (n/2) * log(2 * pi) - covinf$log.det.to.half -
                        (n/2) * log(sigmasq.hat) - (n/2) - temp.list$
                        log.jacobian
        }
        if(temp.list$method == "RML") {
                eigentrem <- eigen(ixix, symmetric = T, only.values = T)
                proflik <- - ((n - beta.size)/2) * log(2 * pi) - covinf$
                        log.det.to.half - (n/2) * log(sigmasq.hat) - (n/2) +
                        0.5 * sum(log(eigentrem$values)) - temp.list$
                        log.jacobian
        }
        return(proflik)
}

"proflik.aux21.1" <-
function(lambda, temp.list)
{
        ## This function computes the value of the 2-D profile likelihood for the parameters (\phi, \tau^2_{rel})
        ## This requires minimasation wrt to the transformation parameter \lambda
        ## This is an auxiliary function called by likfit.proflik
        phitausq.rel <- as.vector(temp.list$phitausq.rel)
        phi <- phitausq.rel[1]
        tausq.rel <- phitausq.rel[2]
        z <- temp.list$z
        n <- temp.list$n
        if(lambda == 1) {
                temp.list$log.jacobian <- 0
        }
        else {
                if(any(z <= 0))
                        stop("Transformation option not allowed when there are zeros or negative data"
                                )
                if(any(z^(lambda - 1) <= 0))
                        temp.list$log.jacobian <- log(prod(z^(lambda - 1)))
                else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
                if(lambda == 0)
                        z <- log(z)
                else z <- ((z^lambda) - 1)/lambda
        }
        beta.size <- temp.list$beta.size
        kappa <- temp.list$kappa
        xmat <- temp.list$xmat
        txmat <- temp.list$txmat
        covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
                cov.model = temp.list$cov.model, kappa = kappa, nugget =
                tausq.rel, cov.pars = c(1, phi), inv = T, det = T)
        ixix <- solve(txmat %*% covinf$inverse %*% xmat)
        ssresmat <- z %*% (covinf$inverse - covinf$inverse %*% xmat %*% ixix %*%
                txmat %*% covinf$inverse) %*% z
        sigmasq.hat <- ssresmat/n
        if(temp.list$method == "ML") {
                neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + (
                        n/2) * log(sigmasq.hat) + (n/2) + temp.list$
                        log.jacobian
        }
        if(temp.list$method == "RML") {
                eigentrem <- eigen(ixix, symmetric = T, only.values = T)
                neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
                        log.det.to.half + (n/2) * log(sigmasq.hat) + (n/2) -
                        0.5 * sum(log(eigentrem$values)) + temp.list$
                        log.jacobian
        }
        return(as.vector(neglik))
}

"proflik.aux22" <-
function(sigmasq, temp.list)
{
        ## This function computes the value of the 2-D profile likelihood for the range and nugget parameters (\phi, \tau^2), minimising the likelihood wrt the random field scale parameter \sigma^2 (sill)
        ## This is an auxiliary function called by likfit.aux17
        z <- temp.list$z
        n <- temp.list$n
        beta.size <- temp.list$beta.size
        kappa <- temp.list$kappa
        xmat <- temp.list$xmat
        covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
                cov.model = temp.list$cov.model, kappa = kappa, nugget =
                temp.list$phitausq.rel[2], cov.pars = c(sigmasq, temp.list$
                phitausq[1]), inv = T, det = T)
        txiv <- crossprod(xmat, covinf$inverse)
        ixix <- solve(txiv %*% xmat)
        ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
                z
        if(temp.list$method == "ML") {
                neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + 0.5 *
                        ssresmat - temp.list$log.jacobian
        }
        if(temp.list$method == "RML") {
                eigentrem <- eigen(ixix, symmetric = T, only.values = T)
                neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
                        log.det.to.half + 0.5 * ssresmat - 0.5 * sum(log(
                        eigentrem$values)) - temp.list$log.jacobian
        }
        return(as.vector(neglik))
}

"proflik.aux23" <-
function(lambda, temp.list)
{
        ## This function computes the value of the profile likelihood for the transformation parameter \lambda
        ## It requires the minimisation of the function wrt \phi and \tau^2 and sigma^2 for each value of \lambda
        ## This is an auxiliary function called by proflik
        lambda <- as.vector(lambda)
        if(temp.list$fixtau == F) {
                if(lambda == 0)
                        data.l <- log(temp.list$z)
                else data.l <- ((temp.list$z^lambda) - 1)/lambda
                var.l <- var(data.l)
                ini.l <- c(0.10000000000000001 * var.l, var.l, temp.list$ini)
        }
        else ini.l <- temp.list$ini
        ## if(dim(temp.list$xmat)[2] == 1 & all(temp.list$xmat == 1))
        ## trend.mat <- "cte"
        ## else{
        ## .temp.xmat <<- (temp.list$xmat[, -1])
        ## trend.mat <- ~.temp.xmat
        ## }
        lambda.res <- likfit(coords = temp.list$coords, data = temp.list$z,
                ini = ini.l, trend = temp.list$trend.mat, fix.nugget =
                temp.list$fixtau, method = temp.list$method, cov.model =
                temp.list$cov.model, kappa = temp.list$kappa, fix.lambda = T,
                lambda = lambda, messages.screen = F)$loglik
        return(lambda.res)
}

"proflik.aux24" <-
function(sigmasqlambda, temp.list)
{
        ## This function computes the value of the 2-D profile likelihood for the parameters (\sigma^2, \lambda) when there is no nugget effect (\tau^2 = 0, fixed)
        ## It requires the minimisation of the function wrt \phi for each value of (\sigma^2, \lambda)
        ## This is an auxiliary function called by proflik
        sigmasqlambda <- as.vector(sigmasqlambda)
        temp.list$sigmasq <- sigmasqlambda[1]
        lambda <- sigmasqlambda[2]
        if(lambda == 1) {
                temp.list$log.jacobian <- 0
        }
        else {
                if(any(temp.list$z <= 0))
                        stop("Transformation option not allowed when there are zeros or negative data"
                                )
                temp.list$log.jacobian <- sum(log(temp.list$z^(lambda - 1)))
                if(lambda == 0)
                        temp.list$z <- log(temp.list$z)
                else temp.list$z <- ((temp.list$z^lambda) - 1)/lambda
        }
        phi.res <- nlminb(temp.list$phi.est, proflik.aux3, lower = temp.list$
                lower.phi, temp.list = temp.list)
        return( - phi.res$objective)
}

"proflik.aux25" <-
function(sigmasqlambda, temp.list)
{
        ## This function computes the value of the 2-D profile likelihood for the parameters (\sigma^2, \lambda), when the nugget (\tau^2) is included in the model.
        ## It requires the minimisation of the function wrt \phi and \tau^2 for each value of (\sigma^2, \lambda)
        ## This is an auxiliary function called by proflik
        sigmasqlambda <- as.vector(sigmasqlambda)
        temp.list$sigmasq <- sigmasqlambda[1]
        lambda <- sigmasqlambda[2]
        if(lambda == 1) {
                temp.list$log.jacobian <- 0
        }
        else {
                stop("Transformation option not allowed when there are zeros or negative data"
                        )
                temp.list$log.jacobian <- sum(log(temp.list$z^(lambda - 1)))
                if(lambda == 0)
                        temp.list$z <- log(temp.list$z)
                else temp.list$z <- ((temp.list$z^lambda) - 1)/lambda
        }
        phitausq.rel.res <- nlminb(c(temp.list$phi.est, temp.list$tausq.rel.est
                ), proflik.aux10, lower = c(temp.list$lower.phi, 0), temp.list
                 = temp.list)
        return( - phitausq.rel.res$objective)
}

"proflik.aux27" <-
function(sigmasqlambda, temp.list)
{
        ## This function computes the value of the 2-D profile likelihood for sill \sigma^2 and the transformation parameter \lambda
        ## It requires the minimisation of the function wrt \phi and \tau^2 and for each value of (\sigma^2,\lambda)
        ## This is an auxiliary function called by proflik.
        sigmasqlambda <- as.vector(sigmasqlambda)
        temp.list$sigmasq <- sigmasqlambda[1]
        lambda <- sigmasqlambda[2]
        if(lambda == 1) {
                temp.list$log.jacobian <- 0
        }
        else {
                temp.list$fix.lambda <- T
                if(any(temp.list$z^(lambda - 1) <= 0))
                        temp.list$log.jacobian <- log(prod(temp.list$z^(lambda -
                                1)))
                else temp.list$log.jacobian <- sum(log(temp.list$z^(lambda -
                                1)))
                if(lambda == 0)
                        temp.list$z <- log(temp.list$z)
                else temp.list$z <- ((temp.list$z^lambda) - 1)/lambda
        }
        phitausq.rel.res <- nlminb(c(temp.list$phi.est, temp.list$tausq.rel.est
                ), proflik.aux10, lower = c(temp.list$lower.phi, 0), temp.list
                 = temp.list)
        return( - phitausq.rel.res$objective)
}

"proflik.aux28" <-
function(sigmasqphi, temp.list)
{
        ## This function computes the value of the 2-D profile likelihood for the random field scale (variance) parameter \sigma^2 and the correlation function parameter \phi when nugget effect is not included in the model.
        ## It requires the minimisation of the function wrt \lambda for each value of (\sigma^2, \phi)
        ## This is an auxiliary function called by likfit.proflik
        ##
        ini.seq <- seq(-2, 2, l = 10)
        temp.list$sigmasqphi <- sigmasqphi
        lambda.lik <- apply(as.matrix(ini.seq), 1, proflik.aux4, temp.list =
                temp.list)
        ini <- ini.seq[lambda.lik == max(lambda.lik)]
        lambda.res <- nlminb(ini, proflik.aux4, lower = -3, upper = 3,
                temp.list = temp.list)
        temp.list$sigmasqphi <- NULL
        return( - lambda.res$objective)
}

"proflik.aux29" <-
function(philambda, temp.list)
{
        ## This function computes the value of the profile likelihood for the parameters (\phi, \lambda) when nugget effect is not included in the model.
        ## This is an auxiliary function called by proflik.
        ##
        philambda <- as.vector(philambda)
        phi <- philambda[1]
        lambda <- philambda[2]
        if(lambda == 1) {
                temp.list$log.jacobian <- 0
        }
        else {
                if(any(temp.list$z^(lambda - 1) <= 0))
                        temp.list$log.jacobian <- log(prod(temp.list$z^(lambda -
                                1)))
                else temp.list$log.jacobian <- sum(log(temp.list$z^(lambda -
                                1)))
                if(lambda == 0)
                        temp.list$z <- log(temp.list$z)
                else temp.list$z <- ((temp.list$z^lambda) - 1)/lambda
        }
        proflik <- proflik.aux1(phi = phi, temp.list = temp.list)
        temp.list$z <- temp.list$data
        return(proflik)
}

"proflik.aux3" <-
function(phi.lambda, temp.list)
{
        ## This function computer the negative of the likelihood function for the correlation function scale parameter \phi and tenasformation parameter \lambda only for models with fixed nugget effect (i.e., when it is not a parameter to be estimated)
        ## This function is used when computing the profile likelihood for \sigma^2
        ## This is an auxiliary function called by proflik.aux2
        ## phi <- pmax(phi, temp.list$lower.phi)
        if(length(phi.lambda) == 2) include.lambda <- TRUE else include.lambda <-
                        FALSE
        sigmasq <- temp.list$sigmasq
        z <- temp.list$z
        n <- length(z)
        phi <- phi.lambda[1]
        if(include.lambda) {
                lambda <- phi.lambda[2]
                if(lambda == 1) {
                        temp.list$log.jacobian <- 0
                }
                else {
                        if(any(z <= 0))
                                stop("Transformation option not allowed when there are zeros or negative data"
                                        )
                        if(any(z^(lambda - 1) <= 0))
                                temp.list$log.jacobian <- log(prod(z^(lambda -
                                        1)))
                        else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
                        if(lambda == 0)
                                z <- log(z)
                        else z <- ((z^lambda) - 1)/lambda
                }
        }
        beta.size <- temp.list$beta.size
        kappa <- temp.list$kappa
        xmat <- temp.list$xmat
        covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
                cov.model = temp.list$cov.model, kappa = kappa, nugget = 0,
                cov.pars = c(1, phi), inv = T, det = T)
        txiv <- crossprod(xmat, covinf$inverse)
        ixix <- solve(txiv %*% xmat)
        ssresmat <- as.vector(z %*% (covinf$inverse - crossprod(txiv, ixix) %*%
                txiv) %*% z)
        if(temp.list$method == "ML") {
                neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + (
                        n/2) * log(sigmasq) + (0.5/sigmasq) * ssresmat -
                        temp.list$log.jacobian
        }
        if(temp.list$method == "RML") {
                eigentrem <- eigen(ixix, symmetric = T, only.values = T)
                neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
                        log.det.to.half + ((n - beta.size)/2) * log(sigmasq) +
                        (0.5/sigmasq) * ssresmat - 0.5 * sum(log(eigentrem$
                        values)) - temp.list$log.jacobian
        }
        return(as.vector(neglik))
}

"proflik.aux30" <-
function(phitausq.rel, temp.list)
{
        ## This function computes the value of the profile likelihood for the correlation parameter \phi when nugget effect is not included in the model.
        ## It requires the minimisation of the function wrt \phi for each value of \lambda, if this transformation parameter is included in the model
        ## This is an auxiliary function called by likfit.proflik
        ##
        phitausq.rel <- as.vector(phitausq.rel)
        if(temp.list$fix.lambda == T)
                proflik <- proflik.aux21(phitausq.rel = phitausq.rel, temp.list
                         = temp.list)
        else {
                temp.list$phitausq.rel <- phitausq.rel
                proflik <- - (nlminb(temp.list$lambda, proflik.aux21.1, lower
                         = -3, upper = 3, temp.list = temp.list)$objective)
                temp.list$phitausq.rel <- NULL
        }
        return(proflik)
}

"proflik.aux31" <-
function(philambda, temp.list)
{
        ## This function computes the value of the 2-D profile likelihood for range \phi and the transformation parameter \lambda.
        ## It requires the minimisation of the function wrt \tau^2_{rel} and for each value of (\phi,\lambda).
        ## This is an auxiliary function called by proflik.
        philambda <- as.vector(philambda)
        temp.list$phi <- philambda[1]
        lambda <- philambda[2]
        if(lambda == 1) {
                temp.list$log.jacobian <- 0
        }
        else {
                if(any(temp.list$z^(lambda - 1) <= 0))
                        temp.list$log.jacobian <- log(prod(temp.list$z^(lambda -
                                1)))
                else temp.list$log.jacobian <- sum(log(temp.list$z^(lambda -
                                1)))
                if(lambda == 0)
                        temp.list$z <- log(temp.list$z)
                else temp.list$z <- ((temp.list$z^lambda) - 1)/lambda
        }
        tausq.rel.res <- nlminb(temp.list$tausq.rel.est, proflik.aux8, lower =
                0, temp.list = temp.list)
        return( - tausq.rel.res$objective)
}

"proflik.aux32" <-
function(tausqlambda, temp.list)
{
        ## This function computes the value of the 2-D profile likelihood for nugget \tau^2 and the transformation parameter \lambda.
        ## It requires the minimisation of the function wrt \phi and \sigma^2 and for each value of (\tau^2,\lambda).
        ## This is an auxiliary function called by proflik.
        tausqlambda <- as.vector(tausqlambda)
        temp.list$nugget <- tausqlambda[1]
        lambda <- tausqlambda[2]
        if(lambda == 1) {
                temp.list$log.jacobian <- 0
        }
        else {
                if(any(temp.list$z^(lambda - 1) <= 0))
                        temp.list$log.jacobian <- log(prod(temp.list$z^(lambda -
                                1)))
                else temp.list$log.jacobian <- sum(log(temp.list$z^(lambda -
                                1)))
                if(lambda == 0)
                        temp.list$z <- log(temp.list$z)
                else temp.list$z <- ((temp.list$z^lambda) - 1)/lambda
        }
        sigmasqphi.res <- nlminb(c(temp.list$sigmasq.est, temp.list$phi.est),
                proflik.aux12, lower = c(temp.list$lower.sigmasq, temp.list$
                lower.phi), temp.list = temp.list)
        return( - sigmasqphi.res$objective)
}

"proflik.aux33" <-
function(tausq.rellambda, temp.list)
{
        ## This function computes the value of the 2-D profile likelihood for nugget \tau^2 and the transformation parameter \lambda.
        ## It requires the minimisation of the function wrt \phi for each value of (\tau^2,\lambda).
        ## This is an auxiliary function called by proflik.
        tausq.rellambda <- as.vector(tausq.rellambda)
        temp.list$nugget.rel <- tausq.rellambda[1]
        lambda <- tausq.rellambda[2]
        if(lambda == 1) {
                temp.list$log.jacobian <- 0
        }
        else {
                if(any(temp.list$z^(lambda - 1) <= 0))
                        temp.list$log.jacobian <- log(prod(temp.list$z^(lambda -
                                1)))
                else temp.list$log.jacobian <- sum(log(temp.list$z^(lambda -
                                1)))
                if(lambda == 0)
                        temp.list$z <- log(temp.list$z)
                else temp.list$z <- ((temp.list$z^lambda) - 1)/lambda
        }
        phi.res <- nlminb(temp.list$phi.est, proflik.aux6, lower = temp.list$
                lower.phi, temp.list = temp.list)
        return( - phi.res$objective)
}

"proflik.aux4" <-
function(lambda, temp.list)
{
        ## This function computer the values of the profile likelihood function for the parameters \phi and \sigma^2 for models with nugget effect = 0, including the tranformation parameter \lambda
        ## This is an auxiliary function called by proflik.aux28
        ##
        sigmasqphi <- as.vector(temp.list$sigmasqphi)
        sigmasq <- sigmasqphi[1]
        phi <- sigmasqphi[2]
        z <- temp.list$z
        n <- temp.list$n
        if(lambda > 0.94999999999999996 & lambda < 1.05)
                lambda <- 1
        if(lambda == 1) {
                log.jacobian <- 0
        }
        else {
                if(any(z <= 0))
                        stop("Transformation option not allowed when there are zeros or negative data"
                                )
                if(any(z^(lambda - 1) <= 0))
                        log.jacobian <- log(prod(z^(lambda - 1)))
                else log.jacobian <- sum(log(z^(lambda - 1)))
                if(lambda == 0)
                        z <- log(z)
                else z <- ((z^lambda) - 1)/lambda
        }
        beta.size <- temp.list$beta.size
        kappa <- temp.list$kappa
        xmat <- temp.list$xmat
        covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
                cov.model = temp.list$cov.model, kappa = kappa, nugget =
                temp.list$nugget, cov.pars = c(sigmasq, phi), scaled = F, inv =
                T, det = T, sqrt.inv = T)
        six <- crossprod(covinf$sqrt.inverse, xmat)
        xix <- crossprod(six)
        if(length(as.vector(xix)) == 1) {
                ixix <- 1/xix
                choldet <- 0.5 * log(xix)
        }
        else {
                chol.xix <- chol(xix)
                ixix <- chol2inv(chol.xix)
                choldet <- sum(log(diag(chol.xix)))
        }
        xi <- crossprod(xmat, covinf$inverse)
        ssresmat <- z %*% (covinf$inverse - crossprod(xi, ixix) %*% xi) %*% z
        if(temp.list$method == "ML") {
                neglik <- ((n/2) * log(2 * pi) + covinf$log.det.to.half + 0.5 *
                        ssresmat - log.jacobian)
        }
        if(temp.list$method == "RML") {
                xx.eigen <- eigen(crossprod(xmat), symmetric = T, only.values
                         = T)
                neglik <- (((n - beta.size)/2) * log(2 * pi) - 0.5 * sum(log(
                        xx.eigen$values)) + covinf$log.det.to.half + (0.5) *
                        ssresmat + choldet + log.jacobian)
        }
        return(as.vector(neglik))
}

"proflik.aux5" <-
function(tausq.rel, temp.list)
{
        ## This function computes the value of the profile likelihood for the parameter \tau^2_{rel}.
        ## It requires the minimisation of the function wrt \phi and \lambda (if the case) for each value of \tau^2_{rel}.
        ## This is an auxiliary function called by proflik.
        temp.list$nugget.rel <- as.vector(tausq.rel)
        if(temp.list$fix.lambda == T) {
                phi.res <- nlminb(temp.list$phi.est, proflik.aux6, lower =
                        temp.list$lower.phi, temp.list = temp.list)
        }
        else {
                phi.res <- nlminb(c(temp.list$phi.est, temp.list$lambda),
                        proflik.aux6, lower = c(temp.list$lower.phi, -3), upper
                         = c( + Inf, 3), temp.list = temp.list)
        }
        return( - phi.res$objective)
}

"proflik.aux6" <-
function(phi.lambda, temp.list)
{
        ## This function computes the value of the profile likelihood for the relative nugget parameter \tau^2_{rel}, minimizing the likelihood wrt correlation function scale parameter \phi (range) and the transformation parameter \lambda.
        if(length(phi.lambda) == 2) include.lambda <- TRUE else include.lambda <-
                        FALSE
        phi.lambda <- as.vector(phi.lambda)
        phi <- phi.lambda[1]
        z <- temp.list$z
        n <- temp.list$n
        if(include.lambda) {
                lambda <- phi.lambda[2]
                if(lambda == 1) {
                        temp.list$log.jacobian <- 0
                }
                else {
                        if(any(z <= 0))
                                stop("Transformation option not allowed when there are zeros or negative data"
                                        )
                        if(any(z^(lambda - 1) <= 0))
                                temp.list$log.jacobian <- log(prod(z^(lambda -
                                        1)))
                        else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
                        if(lambda == 0)
                                z <- log(z)
                        else z <- ((z^lambda) - 1)/lambda
                }
        }
        beta.size <- temp.list$beta.size
        kappa <- temp.list$kappa
        xmat <- temp.list$xmat
        covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
                cov.model = temp.list$cov.model, kappa = kappa, nugget =
                temp.list$nugget.rel, cov.pars = c(1, phi), inv = T, det = T)
        txiv <- crossprod(xmat, covinf$inverse)
        ixix <- solve(txiv %*% xmat)
        ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
                z
        if(temp.list$method == "ML") {
                neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + (
                        n/2) * log(ssresmat/n) + (n/2) - temp.list$log.jacobian
        }
        if(temp.list$method == "RML") {
                eigentrem <- eigen(ixix, symmetric = T, only.values = T)
                neglik <- ((n - beta.size)/2) * log(2 * pi) + covinf$
                        log.det.to.half + ((n - beta.size)/2) * log(ssresmat/
                        n) + (n/2) - 0.5 * sum(log(eigentrem$values)) -
                        temp.list$log.jacobian
        }
        return(as.vector(neglik))
}

"proflik.aux7" <-
function(phi, temp.list)
{
        ## This function computes the value of the profile likelihood for the parameter \phi when the nugget \tau^2 is included in the model
        ## It requires the minimisation of the function wrt relative \tau^2_{rel} for each value of \phi
        ## This is an auxiliary function called by proflik.
        temp.list$phi <- as.vector(phi)
        if(temp.list$fix.lambda == T) {
                tausq.rel.res <- nlminb(temp.list$tausq.rel.est, proflik.aux8,
                        lower = 0, temp.list = temp.list)
        }
        else {
                tausq.rel.res <- nlminb(c(temp.list$tausq.rel.est, temp.list$
                        lambda), proflik.aux8, lower = c(0, -3), upper = c(
                         + Inf, 3), temp.list = temp.list)
        }
        return( - tausq.rel.res$objective)
}

"proflik.aux8" <-
function(tausq.rel.lambda, temp.list)
{
        ## This function computes the value of the profile likelihood for the correlation function scale parameter \phi, minimizing the likelihood wrt relative nugget parameter \tau^2_{rel}
        if(length(tausq.rel.lambda) == 2) include.lambda <- TRUE else
                        include.lambda <- FALSE
        z <- temp.list$z
        n <- temp.list$n
        phi <- temp.list$phi
        tausq.rel <- tausq.rel.lambda[1]
        if(include.lambda) {
                lambda <- tausq.rel.lambda[2]
                if(lambda == 1) {
                        temp.list$log.jacobian <- 0
                }
                else {
                        if(any(z <= 0))
                                stop("Transformation option not allowed when there are zeros or negative data"
                                        )
                        if(any(z^(lambda - 1) <= 0))
                                temp.list$log.jacobian <- log(prod(z^(lambda -
                                        1)))
                        else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
                        if(lambda == 0)
                                z <- log(z)
                        else z <- ((z^lambda) - 1)/lambda
                }
        }
        beta.size <- temp.list$beta.size
        kappa <- temp.list$kappa
        xmat <- temp.list$xmat
        covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
                cov.model = temp.list$cov.model, kappa = kappa, nugget =
                tausq.rel, cov.pars = c(1, phi), inv = T, det = T)
        txiv <- crossprod(xmat, covinf$inverse)
        ixix <- solve(txiv %*% xmat)
        ssresmat <- z %*% (covinf$inverse - crossprod(txiv, ixix) %*% txiv) %*%
                z
        if(temp.list$method == "ML") {
                neglik <- (n/2) * log(2 * pi) + covinf$log.det.to.half + (
                        n/2) * log(ssresmat/n) + (n/2) - temp.list$log.jacobian
        }
        if(temp.list$method == "RML") {
                eigentrem <- eigen(ixix, symmetric = T, only.values = T)
                neglik <- (((n - beta.size)/2) * log(2 * pi) + covinf$
                        log.det.to.half + ((n - beta.size)/2) * log(ssresmat/
                        n) + (n/2) - 0.5 * sum(log(eigentrem$values))) -
                        temp.list$log.jacobian
        }
        return(as.vector(neglik))
}

"proflik.aux9" <-
function(sigmasq, temp.list)
{
        ## This function computes the value of the profile likelihood for the parameter \sigma^2 when \tau^2 is included in the model
        ## It requires the minimisation of the function wrt \phi and \tau^2 for each value of \sigma^2
        ## This is an auxiliary function called by likfit.proflik
        temp.list$sigmasq <- as.vector(sigmasq)
        if(temp.list$fix.lambda == T) {
                phitausq.rel.res <- nlminb(c(temp.list$phi.est, temp.list$
                        tausq.rel.est), proflik.aux10, lower = c(temp.list$
                        lower.phi, 0), temp.list = temp.list)
        }
        else {
                phitausq.rel.res <- nlminb(c(temp.list$phi.est, temp.list$
                        tausq.rel.est, temp.list$lambda), proflik.aux10, lower
                         = c(temp.list$lower.phi, 0, -3), upper = c( + Inf,
                         + Inf, 3), temp.list = temp.list)
        }
        return( - phitausq.rel.res$objective)
}

"proflik.cov" <-
function(theta, temp.list)
{
        z <- temp.list$z
        n <- temp.list$n
        beta.size <- temp.list$beta.size
        kappa <- temp.list$kappa
        xmat <- temp.list$xmat
        if(temp.list$fixtau == T) {
                tausq <- temp.list$ftau
                sigmasq <- theta[1]
                phi <- theta[2]
        }
        else {
                tausq <- theta[1]
                sigmasq <- theta[2]
                phi <- theta[3]
        }
        sill.total <- sigmasq + tausq
        covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
                cov.model = temp.list$cov.model, kappa = kappa, nugget = (
                tausq/sigmasq), cov.pars = c(1, phi), sqrt.inv = T, inv = T,
                det = T)
        six <- crossprod(covinf$sqrt.inverse, xmat)
        xix <- crossprod(six)
        if(length(as.vector(xix)) == 1) {
                ixix <- 1/xix
                choldet <- 0.5 * log(xix)
        }
        else {
                chol.xix <- chol(xix)
                i.chol.xix <- solve.upper(chol.xix)
                ixix <- i.chol.xix %*% t(i.chol.xix)
                choldet <- sum(log(diag(chol.xix)))
        }
        xi <- crossprod(xmat, covinf$inverse)
        ssresmat <- z %*% (covinf$inverse - crossprod(xi, ixix) %*% xi) %*% z
        if(temp.list$method == "ML") {
                neglik <- ((n/2) * log(2 * pi) + covinf$log.det.to.half + (
                        n/2) * log(ssresmat/n) + (n/2)) - temp.list$
                        log.jacobian
        }
        if(temp.list$method == "RML") {
                neglik <- (((n - beta.size)/2) * log(2 * pi) + covinf$
                        log.det.to.half + ((n - beta.size)/2) * log(ssresmat/
                        n) + (n/2) + choldet) - temp.list$log.jacobian
        }
        return(as.vector(neglik))
}

"proflik.ftau" <-
function(theta, temp.list)
{
        if(length(theta) == 3)
                include.lambda <- TRUE
        else include.lambda <- FALSE
        z <- temp.list$z
        n <- temp.list$n
        lambda <- theta[3]
        if(include.lambda) {
                lambda <- theta[3]
                if(lambda == 1) {
                        temp.list$log.jacobian <- 0
                }
                else {
                        if(any(z < 0))
                                stop("Transformation option not allowed when there are zeros or negative data"
                                        )
                        if(any(z^(lambda - 1) <= 0))
                                temp.list$log.jacobian <- log(prod(z^(lambda -
                                        1)))
                        else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
                        if(lambda == 0)
                                z <- log(z)
                        else z <- ((z^lambda) - 1)/lambda
                }
        }
        beta.size <- temp.list$beta.size
        kappa <- temp.list$kappa
        xmat <- temp.list$xmat
        txmat <- temp.list$txmat
        ftau <- temp.list$ftau
        sigmasq <- theta[1]
        sill.total <- ftau + sigmasq
        phi <- theta[2]
        covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
                cov.model = temp.list$cov.model, kappa = kappa, nugget = ftau,
                cov.pars = c(sigmasq, phi), inv = T, det = T, sqrt.inv = T)
        six <- crossprod(covinf$sqrt.inverse, xmat)
        xix <- crossprod(six)
        if(length(as.vector(xix)) == 1) {
                ixix <- 1/xix
                choldet <- 0.5 * log(xix)
        }
        else {
                chol.xix <- chol(xix)
                i.chol.xix <- solve.upper(chol.xix)
                ixix <- i.chol.xix %*% t(i.chol.xix)
                choldet <- sum(log(diag(chol.xix)))
        }
        xi <- crossprod(xmat, covinf$inverse)
        ssresmat <- z %*% (covinf$inverse - crossprod(xi, ixix) %*% xi) %*% z
        if(temp.list$method == "ML") {
                neglik <- ((n/2) * log(2 * pi) + covinf$log.det.to.half + 0.5 *
                        ssresmat - temp.list$log.jacobian)
        }
        if(temp.list$method == "RML") {
                xx.eigen <- eigen(crossprod(xmat), symmetric = T, only.values
                         = T)
                neglik <- (((n - beta.size)/2) * log(2 * pi) + covinf$
                        log.det.to.half + 0.5 * ssresmat + choldet - 0.5 * sum(
                        log(xx.eigen$values)) - temp.list$log.jacobian)
        }
        return(as.vector(neglik))
}

"proflik.lambda" <-
function(lambda, temp.list)
{
        z <- temp.list$z
        n <- temp.list$n
        if(lambda == 1) {
                temp.list$log.jacobian <- 0
        }
        else {
                if(any(z < 0))
                        stop("Transformation option not allowed when there are zeros or negative data"
                                )
                if(any(z^(lambda - 1) <= 0))
                        temp.list$log.jacobian <- log(prod(z^(lambda - 1)))
                else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
                if(lambda == 0)
                        z <- log(z)
                else z <- ((z^lambda) - 1)/lambda
        }
        beta.size <- temp.list$beta.size
        kappa <- temp.list$kappa
        xmat <- temp.list$xmat
        txmat <- temp.list$txmat
        ixx <- solve(crossprod(xmat))
        tausqhat <- (z %*% (diag(n) - xmat %*% ixx %*% txmat) %*% z)/n
        if(temp.list$method == "ML")
                neglik <- ((n/2) * log(2 * pi) + (n/2) * log(tausqhat) + (
                        n/2) - temp.list$log.jacobian)
        if(temp.list$method == "RML") {
                eigentrem <- eigen(ixx, symmetric = T, only.values = T)
                neglik <- (((n - beta.size)/2) * log(2 * pi) + ((n - beta.size)/
                        2) * log(tausqhat) + (n/2) - 0.5 * sum(log(eigentrem$
                        values)) - temp.list$log.jacobian)
        }
        return(as.vector(neglik))
}

"proflik.nug" <-
function(theta, temp.list)
{
        if(length(theta) == 3)
                include.lambda <- TRUE
        else include.lambda <- FALSE
        z <- temp.list$z
        n <- temp.list$n
        if(include.lambda) {
                lambda <- theta[3]
                if(lambda == 1) {
                        temp.list$log.jacobian <- 0
                }
                else {
                        if(any(z < 0))
                                stop("Transformation option not allowed when there are zeros or negative data"
                                        )
                        if(any(z^(lambda - 1) <= 0))
                                temp.list$log.jacobian <- log(prod(z^(lambda -
                                        1)))
                        else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
                        if(lambda == 0)
                                z <- log(z)
                        else z <- ((z^lambda) - 1)/lambda
                }
        }
        beta.size <- temp.list$beta.size
        kappa <- temp.list$kappa
        xmat <- temp.list$xmat
        txmat <- temp.list$txmat
        tausq.rel <- theta[1]
        phi <- theta[2]
        covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
                cov.model = temp.list$cov.model, kappa = kappa, nugget =
                tausq.rel, cov.pars = c(1, phi), det = T, inv = T, sqrt.inv = T
                )
        six <- crossprod(covinf$sqrt.inverse, xmat)
        xix <- crossprod(six)
        if(length(as.vector(xix)) == 1) {
                ixix <- 1/xix
                choldet <- 0.5 * log(xix)
        }
        else {
                chol.xix <- chol(xix)
                i.chol.xix <- solve.upper(chol.xix)
                ixix <- i.chol.xix %*% t(i.chol.xix)
                choldet <- sum(log(diag(chol.xix)))
        }
        xi <- crossprod(xmat, covinf$inverse)
        ssresmat <- z %*% (covinf$inverse - crossprod(xi, ixix) %*% xi) %*% z
        if(temp.list$method == "ML") {
                neglik <- ((n/2) * log(2 * pi) + covinf$log.det.to.half + (
                        n/2) * log(ssresmat/n) + (n/2) - temp.list$log.jacobian
                        )
        }
        if(temp.list$method == "RML") {
                xx.eigen <- eigen(crossprod(xmat), symmetric = T, only.values
                         = T)
                neglik <- (((n - beta.size)/2) * log(2 * pi) + covinf$
                        log.det.to.half + ((n - beta.size)/2) * log(ssresmat/
                        (n - beta.size)) + (n/2) + choldet - 0.5 * sum(log(
                        xx.eigen$values)) - temp.list$log.jacobian)
        }
        return(as.vector(neglik))
}

"proflik.phi" <-
function(theta, temp.list)
{
        if(length(theta) == 2)
                include.lambda <- TRUE
        else include.lambda <- FALSE
        z <- as.vector(temp.list$z)
        n <- temp.list$n
        phi <- theta[1]
        if(include.lambda) {
                lambda <- theta[2]
                if(lambda == 1) {
                        temp.list$log.jacobian <- 0
                }
                else {
                        if(any(z <= 0))
                                stop("Transformation option not allowed when there are zeros or negative data"
                                        )
                        if(any(z^(lambda - 1) <= 0))
                                temp.list$log.jacobian <- log(prod(z^(lambda -
                                        1)))
                        else temp.list$log.jacobian <- sum(log(z^(lambda - 1)))
                        if(lambda == 0)
                                z <- log(z)
                        else z <- ((z^lambda) - 1)/lambda
                }
        }
        beta.size <- temp.list$beta.size
        kappa <- temp.list$kappa
        xmat <- temp.list$xmat
        txmat <- temp.list$txmat
        covinf <- varcov.spatial(dists.lowertri = temp.list$dists.lowertri,
                cov.model = temp.list$cov.model, kappa = kappa, nugget = 0,
                cov.pars = c(1, phi), inv = T, det = T, sqrt.inv = T)
        six <- crossprod(covinf$sqrt.inverse, xmat)
        xix <- crossprod(six)
        if(length(as.vector(xix)) == 1) {
                ixix <- 1/xix
                choldet <- 0.5 * log(xix)
        }
        else {
                chol.xix <- chol(xix)
                i.chol.xix <- solve.upper(chol.xix)
                ixix <- i.chol.xix %*% t(i.chol.xix)
                choldet <- sum(log(diag(chol.xix)))
        }
        xi <- crossprod(xmat, covinf$inverse)
        ssresmat <- z %*% (covinf$inverse - crossprod(xi, ixix) %*% xi) %*% z
        if(temp.list$method == "ML") {
                neglik <- ((n/2) * log(2 * pi) + covinf$log.det.to.half + (
                        n/2) * log(ssresmat/n) + (n/2) - temp.list$log.jacobian
                        )
        }
        if(temp.list$method == "RML") {
                xx.eigen <- eigen(crossprod(xmat), symmetric = T, only.values
                         = T)
                neglik <- (((n - beta.size)/2) * log(2 * pi) + covinf$
                        log.det.to.half + ((n - beta.size)/2) * log(ssresmat/
                        (n - beta.size)) + (n/2) + choldet - 0.5 * sum(log(
                        xx.eigen$values)) - temp.list$log.jacobian)
        }
        return(as.vector(neglik))
}

"rfm.bin" <-
function(cloud, l = 15, uvec = "default", nugget.tolerance = 0, estimator.type
         = c("classical", "robust"), bin.cloud = F, max.dist)
{
        ##
        ## binned f*.rfm function
        ##
        ## cloud : output from f*.rfm function with option 'cloud'
        ## uvec : n-element vector of values to define the binning; the values of uvec defines the center of the bins
        ### if uvec[1] > 0 first bin is for 0< u <= uvec[2]-0.5*(uvec[2]-uvec[1])
        ### if uvec[1] = 0 first bin is for 0< u <= 0.5*uvec[1] and uvec[1] is replaced by the midpoint of this interval
        ## nugget.tolerance: pairs separated for a distance less than that va;lue are considered to be at same location (for direct estimation of nugget effect)
        ## result is a list with components:
        ## u : uvec
        ## v : averages of empirical (semi)variogram ordinates within bins
        ## n : numbers of empirical (semi)variogram ordinates contributing
        ## to each bin
        ### if (all(uvec == "default"))
        ### uvec <- seq(0, max(cloud$u), l = l)
        estimator.type <- match.arg(estimator.type)
        ### if(nugget.tolerance > 0) {
        ### dnug <- mean(cloud$u[cloud$u <= nugget.tolerance])
        ### cloud$u[cloud$u <= nugget.tolerance] <- 0
        ### uvec <- uvec[uvec > nugget.tolerance]
        ### }
        ### u <- c(0, uvec)
        ### n <- length(u)
        if(all(uvec == "default")) uvec <- seq(0, max.dist, l = 15)
        ubin <- c(0, uvec)
        nvec <- length(ubin)
        d <- 0.5 * diff(ubin[2:nvec])
        bins.lim <- c(0, (ubin[2:(nvec - 1)] + d), (d[nvec - 2] + ubin[nvec]))
        if(uvec[1] == 0 & nugget.tolerance == 0)
                uvec[1] <- (bins.lim[1] + bins.lim[2])/2
        if(nugget.tolerance > 0) {
                bins.lim <- c(0, nugget.tolerance, bins.lim[bins.lim >
                        nugget.tolerance])
                uvec <- c(0, (bins.lim[ - (1:2)] - 0.5 * diff(bins.lim)[-1]))
        }
        nbins <- nc <- length(bins.lim) - 1
        if(is.null(max.dist))
                max.dist <- max(bins.lim)
        min.dist <- min(cloud$u)
        ## d <- 0.5 * (u[3:n] - u[2:(n - 1)])
        ## low <- c(0, (u[2:(n - 1)] + d))
        ## high <- c((u[3:n] - d), (d[n - 2] + u[n]))
        ## nc <- n - 1
        if(!is.matrix(cloud$v)) {
                vbin <- rep(0, nc)
                nbin <- rep(0, nc)
                sdbin <- rep(0, nc)
                if(bin.cloud == T)
                        bins.clouds <- list()
                for(i in 1:nc) {
                        ind <- (cloud$u > bins.lim[i]) & (cloud$u <= bins.lim[
                                i + 1])
                        vbin[i] <- mean(cloud$v[ind])
                        if(bin.cloud == T)
                                bins.clouds[[i]] <- cloud$v[ind]
                        nbin[i] <- sum(ind)
                        if(estimator.type == "robust")
                                vbin[i] <- ((vbin[i])^4)/(0.91400000000000003 +
                                        (0.98799999999999999/nbin[i]))
                        if(nbin[i] > 0)
                                sdbin[i] <- sqrt(var(cloud$v[ind]))
                        else sdbin[i] <- NA
                        NULL
                }
                ## check == here
                if(uvec[1] == 0) uvec[1] <- (bins.lim[1] + bins.lim[2])/2
                ## check == here
                if(min.dist == 0) {
                        # check == here
                        ind <- (cloud$u == 0)
                        n.zero <- sum(ind)
                        v.zero <- mean(cloud$v[ind])
                        if(bin.cloud == T) {
                                bins.clouds[2:(length(bins.clouds) + 1)] <-
                                        bins.clouds[1:nc]
                                bins.clouds[[1]] <- cloud$v[ind]
                        }
                        if(estimator.type == "robust")
                                v.zero <- ((v.zero)^4)/(0.91400000000000003 +
                                        (0.98799999999999999/n.zero))
                        if(n.zero > 0)
                                sd.zero <- sqrt(var(cloud$v[ind]))
                        else sd.zero <- NA
                        uvec <- c(0, uvec)
                        vbin <- c(v.zero, vbin)
                        nbin <- c(n.zero, nbin)
                        sdbin <- c(sd.zero, sdbin)
                }
                u <- uvec[!is.na(vbin)]
                v <- vbin[!is.na(vbin)]
                n <- nbin[!is.na(vbin)]
                sd <- sdbin[!is.na(vbin)]
                if(bin.cloud == T)
                        bins.clouds <- bins.clouds[!is.na(vbin)]
        }
        else {
                if(bin.cloud == T)
                        stop("option bins.cloud=T allowed only for 1 variable")
                nvcols <- ncol(cloud$v)
                vbin <- matrix(0, nrow = nc, ncol = nvcols)
                nbin <- rep(0, nc)
                sdbin <- matrix(0, nrow = nc, ncol = nvcols)
                for(i in 1:nc) {
                        ind <- (cloud$u >= bins.lim[i]) & (cloud$u < bins.lim[
                                i + 1])
                        nbin[i] <- sum(ind)
                        for(j in 1:nvcols) {
                                vbin[i, j] <- mean(cloud$v[ind, j])
                                if(estimator.type == "robust")
                                        vbin[i, j] <- ((vbin[i, j])^4)/(
                                                0.91400000000000003 + (
                                                0.98799999999999999/nbin[i]))
                                if(nbin[i] > 0)
                                        sdbin[i, j] <- sqrt(var(cloud$v[ind,
                                                j]))
                                else sdbin[i, j] <- NA
                        }
                        NULL
                }
                ## check == here
                if(uvec[1] == 0) uvec[1] <- (bins.lim[1] + bins.lim[2])/2
                ## check == here
                if(min.dist == 0) {
                        v.zero <- rep(0, nvcols)
                        n.zero <- rep(0, nvcols)
                        sd.zero <- rep(0, nvcols)
                        for(j in 1:nvcols) {
                                ## check == here
                                ind <- (cloud$u == 0)
                                n.zero[j] <- sum(ind)
                                v.zero[j] <- mean(cloud$v[ind, j])
                                if(estimator.type == "robust")
                                        v.zero[j] <- ((v.zero[j])^4)/(
                                                0.91400000000000003 + (
                                                0.98799999999999999/n.zero[
                                                j]))
                                if(n.zero[j] > 0)
                                        sd.zero[j] <- sqrt(var(cloud$v[ind,
                                                j]))
                                else sd.zero[j] <- NA
                                uvec <- c(0, uvec)
                                vbin <- rbind(v.zero, vbin)
                                nbin <- c(n.zero, nbin)
                                sdbin <- rbind(sd.zero, sdbin)
                        }
                }
                u <- uvec[!is.na(vbin[, 1])]
                n <- nbin[!is.na(vbin[, 1])]
                v <- matrix(0, nrow = length(u), ncol = nvcols)
                sd <- matrix(0, nrow = length(u), ncol = nvcols)
                for(j in 1:nvcols) {
                        v[, j] <- vbin[!is.na(vbin[, j]), j]
                        ## sd[, j] <- sdbin[!is.na(sdbin[, j]), j]
                        sd[, j] <- sdbin[!is.na(vbin[, j]), j]
                }
        }
        if(nugget.tolerance > 0) {
                u[1] <- nugget.tolerance
        }
        result <- list(u = u, v = v, n = n, sd = sd, output.type = "bin")
        if(!is.matrix(cloud$v) && bin.cloud == T)
                result$bin.cloud <- bins.clouds
        if(!is.null(class(cloud)))
                class(result) <- class(cloud)
        return(result)
}

"rfm.tests" <-
function(f.data, f.data.mean, f.values, f.mean, f.n, test, outliers.p,
        cloud.data, cloud.sim)
{
        results <- list()
        f.total <- sum(f.n)
        results$limits <- apply(f.values, 1, range)
        results$tests <- NULL
        pondera <- function(obj, w = f.n)
        {
                (obj %*% w)/sum(w)
        }
        if(any(test == 1)) {
                ind.out <- (f.data >= results$limits[2, ] | f.data <= results$
                        limits[1, ])
                t1 <- sum(ind.out)
                ind.out <- NULL
                results$tests <- c(results$tests, T1 = t1)
        }
        if(any(test == 2)) {
                ind.out <- apply(f.values, 2, range)
                ind.out <- ind.out[2, ] - ind.out[1, ]
                ind.data <- range(f.data)
                ind.data <- ind.data[2] - ind.data[1]
                t2 <- 100 * (sum(ind.out < ind.data)/length(ind.out))
                results$tests <- c(results$tests, T2 = t2)
                ind.out <- NULL
                ind.data <- NULL
        }
        if(any(test > 2)) {
                f.values <- abs(t(t(f.values) - f.mean))
                t.f.values <- t(f.values)
                f.data <- abs(f.data - f.data.mean)
        }
        if(any(test == 3)) {
                ind.data <- max(f.data)
                ind.out <- apply(f.values, 2, max)
                t3 <- 100 * (sum(ind.out < ind.data)/length(ind.out))
                results$tests <- c(results$tests, T3 = t3)
                ind.out <- NULL
                ind.data <- NULL
        }
        if(any(test == 4 | test == 5 | test == 8 | test == 9)) {
                f.data2 <- f.data^2
                f.values2 <- f.values^2
                if(any(test == 4)) {
                        ind.data <- sum(f.data2)
                        ind.out <- apply(f.values2, 2, sum)
                        t4 <- 100 * (sum(ind.out < ind.data)/length(ind.out))
                        results$tests <- c(results$tests, T4 = t4)
                        ind.out <- NULL
                        ind.data <- NULL
                }
                if(any(test == 5)) {
                        ind.data <- pondera(f.data2, w = f.n)
                        ind.out <- apply(f.values2, 2, pondera, w = f.n)
                        t5 <- 100 * (sum(ind.out < ind.data)/length(ind.out))
                        results$tests <- c(results$tests, T5 = t5)
                }
                ind.out <- NULL
                ind.data <- NULL
        }
        if(any(test == 6 | test == 7)) {
                if(any(test == 6)) {
                        ind.data <- sum(f.data)
                        ind.out <- apply(f.values, 2, sum)
                        t6 <- 100 * (sum(ind.out < ind.data)/length(ind.out))
                        results$tests <- c(results$tests, T6 = t6)
                        ind.out <- NULL
                        ind.data <- NULL
                }
                if(any(test == 7)) {
                        ind.data <- pondera(f.data, w = f.n)
                        ind.out <- apply(f.values, 2, pondera, w = f.n)
                        t7 <- 100 * (sum(ind.out < ind.data)/length(ind.out))
                        results$tests <- c(results$tests, T7 = t7)
                }
                ind.out <- NULL
                ind.data <- NULL
        }
        if(any(test == 8 | test == 9)) {
                if(outliers.p == 0)
                        ref.data <- max(f.data)
                else ref.data <- quantile(f.data, prob = (1 - outliers.p))
                f.data.r <- ifelse(f.data <= ref.data, (0.5 * f.data2), (
                        ref.data * f.data - (ref.data^2/2)))
                if(outliers.p == 0)
                        ref.values <- apply(f.values, 2, max)
                else ref.values <- apply(f.values, 2, quantile, prob = (1 -
                                outliers.p))
                f.values.r <- ifelse((t(t.f.values <= ref.values)), (0.5 *
                        f.values2), (t(t.f.values * ref.values - (ref.values^
                        2/2))))
                if(any(test == 8)) {
                        ind.data <- sum(f.data.r)
                        ind.out <- apply(f.values.r, 2, sum)
                        t8 <- 100 * (sum(ind.out < ind.data)/length(ind.out))
                        results$tests <- c(results$tests, T8 = t8)
                        ind.out <- NULL
                        ind.data <- NULL
                }
                if(any(test == 9)) {
                        ind.data <- pondera(f.data.r, w = f.n)
                        ind.out <- apply(f.values.r, 2, pondera, w = f.n)
                        t9 <- 100 * (sum(ind.out < ind.data)/length(ind.out))
                        results$tests <- c(results$tests, T9 = t9)
                }
                ind.out <- NULL
                ind.data <- NULL
        }
        if(any(test > 9)) {
                cloud.sim <- abs(t(t(cloud.sim) - f.mean))
                t.cloud.sim <- t(cloud.sim)
                cloud.data <- abs(cloud.data - f.data.mean)
                if(any(test == 10 | test == 12)) {
                        cloud.data2 <- cloud.data^2
                        cloud.sim2 <- cloud.sim^2
                }
                if(any(test == 10)) {
                        ind.data <- sum(cloud.data2)
                        ind.out <- apply(cloud.sim2, 2, sum)
                        t10 <- 100 * (sum(ind.out < ind.data)/length(ind.out))
                        results$tests <- c(results$tests, T10 = t10)
                        ind.out <- NULL
                        ind.data <- NULL
                }
                if(any(test == 11)) {
                        ind.data <- sum(cloud.data)
                        ind.out <- apply(cloud.sim, 2, sum)
                        t11 <- 100 * (sum(ind.out < ind.data)/length(ind.out))
                        results$tests <- c(results$tests, T11 = t11)
                        ind.out <- NULL
                        ind.data <- NULL
                }
                if(any(test == 12)) {
                        if(outliers.p == 0)
                                ref.data <- max(f.data)
                        else ref.data <- quantile(cloud.data, prob = (1 -
                                        outliers.p))
                        cloud.data.r <- ifelse(cloud.data < ref.data, (0.5 *
                                cloud.data2), (ref.data * cloud.data - (
                                ref.data^2/2)))
                        if(outliers.p == 0)
                                ref.data <- apply(cloud.sim, 2, max)
                        else ref.values <- apply(cloud.sim, 2, quantile, prob
                                         = (1 - outliers.p))
                        cloud.sim.r <- ifelse((t(t.cloud.sim <= ref.values)),
                                (0.5 * cloud.sim2), (t(t.cloud.sim * ref.values -
                                (ref.values^2/2))))
                        ind.data <- sum(cloud.data.r)
                        ind.out <- apply(cloud.sim.r, 2, sum)
                        t12 <- 100 * (sum(ind.out < ind.data)/length(ind.out))
                        results$tests <- c(results$tests, T12 = t12)
                        ind.out <- NULL
                        ind.data <- NULL
                }
        }
        results$tests <- round(results$tests)
        return(results)
}

"validation.spatial" <-
function(coords, observed, predicted, variance.predicted = NULL, n.worse = 5)
{
        residu <- observed - predicted
        results <- cbind(coords, observed = observed, predicted = predicted,
                residuals = residu)
        if(!is.null(variance.predicted)) {
                sd.resid <- residu/sqrt(variance.predicted)
                results <- cbind(results, sd.resid = sd.resid)
        }
        return(results)
}

"vario.matern" <-
  function(u, tausq, sigmasq, phi, kappa)
{
  ##
  ## variogram corresponding to matern correlation function
  ##
  tausq + sigmasq * (1 - matern(u, phi, kappa))
}

"variog" <-
  function(geodata, coords = geodata$coords, data = geodata$data, uvec =
           "default", trend = "cte", lambda = 1, option = c("bin", "cloud",
                                                   "smooth"), estimator.type = c("classical", "robust"), nugget.tolerance
           = 0, max.dist = NULL, pairs.min = 2, bin.cloud = F, ...)
{
  ##
  ## reading and checking input
  ##
  class.fc <- match.call()
  option <- match.arg(option)
  estimator.type <- match.arg(estimator.type)
  coords <- as.matrix(coords)
  data <- as.matrix(data)
  data.var <- apply(data, 2, var)
  n.data <- nrow(coords)
  n.datasets <- ncol(data)
  if(ncol(data) == 1)
    data <- as.vector(data)
  ##
  ## Box-Cox transformation
  ##
  if(lambda != 1) {
    if(lambda == 0)
      data <- log(data)
    else data <- ((data^lambda) - 1)/lambda
  }
  ##
  ## Removing OLS trend
  ##
  xmat <- trend.spatial(trend = trend, coords = coords)
        if(trend != "cte") {
                if(is.vector(data)) {
                        data <- lm(data ~ xmat - 1)$residuals
                        names(data) <- NULL
                }
                else {
                        only.res <- function(y, x)
                        {
                                lm(y ~ x - 1)$residuals
                        }
                        data <- apply(data, 2, only.res, x = xmat)
                }
        }
        u <- dist(as.matrix(coords))
        if(option == "bin" & bin.cloud == FALSE) {
                if(!is.null(max.dist))
                        umax <- max(u[u < max.dist])
                else umax <- max(u)
                if(all(uvec == "default"))
                        uvec <- seq(0, umax, l = 15)
                ubin <- c(0, uvec)
                nvec <- length(ubin)
                d <- 0.5 * diff(ubin[2:nvec])
                bins.lim <- c(0, (ubin[2:(nvec - 1)] + d), (d[nvec - 2] + ubin[
                        nvec]))
                if(uvec[1] == 0 & nugget.tolerance == 0)
                        uvec[1] <- (bins.lim[1] + bins.lim[2])/2
                if(nugget.tolerance > 0) {
                        bins.lim <- c(0, nugget.tolerance, bins.lim[bins.lim >
                                nugget.tolerance])
                        uvec <- c(0, (bins.lim[ - (1:2)] - 0.5 * diff(bins.lim)[
                                -1]))
                }
                nbins <- length(bins.lim) - 1
                if(is.null(max.dist))
                        max.dist <- max(bins.lim)
                temp.list <- list(n.data = n.data, coords = coords, data = data,
                        nbins = nbins, bins.lim = bins.lim, estimator.type =
                        estimator.type, max.dist = max.dist)
                bin.f <- function(data, temp.list)
                {
                        cbin <- vbin <- sdbin <- rep(0, temp.list$nbins)
                        result <- .C("binit",
                                as.integer(temp.list$n.data),
                                as.double(as.vector(temp.list$coords[, 1])),

                                        as.double(as.vector(temp.list$coords[
                                        , 2])),
                                as.double(as.vector(data)),
                                as.integer(temp.list$nbins),
                                as.double(as.vector(temp.list$bins.lim)),
                                as.integer(temp.list$estimator.type == "robust"
                                        ),
                                as.double(temp.list$max.dist),
                                cbin = as.integer(cbin),
                                vbin = as.double(vbin),
                                as.integer(TRUE),
                                sdbin = as.double(sdbin))[c("vbin", "cbin",
                                "sdbin")]
                        return(result)
                }
                result <- array(unlist(lapply(as.data.frame(data), bin.f,
                        temp.list = temp.list)), dim = c(nbins, 3, n.datasets))
                indp <- (result[, 2, 1] >= pairs.min)
                result <- list(u = uvec[indp], v = result[indp, 1, ], n =
                        result[indp, 2, 1], sd = result[indp, 3, ], bins.lim
                         = bins.lim, ind.bin = indp)
        }
        else {
                if(is.matrix(data)) {
                        v <- matrix(0, nrow = length(u), ncol = ncol(data))
                        for(i in 1:ncol(data)) {
                                v[, i] <- as.vector(dist(data[, i]))
                                if(estimator.type == "robust")
                                        v[, i] <- v[, i]^(0.5)
                                else v[, i] <- (v[, i]^2)/2
                        }
                        if(!is.null(max.dist)) {
                                v <- v[u <= max.dist, ]
                                u <- u[u <= max.dist]
                        }
                }
                else {
                        v <- as.vector(dist(data))
                        if(estimator.type == "robust")
                                v <- v^(0.5)
                        else v <- (v^2)/2
                        if(is.numeric(max.dist)) {
                                v <- v[u <= max.dist]
                                u <- u[u <= max.dist]
                        }
                }
                if(option == "cloud") {
                        result <- list(u = u, v = v)
                }
                if(option == "bin") {
                        if(!is.null(max.dist))
                                umax <- max(u[u < max.dist])
                        else umax <- max(u)
                        result <- rfm.bin(cloud = list(u = u, v = v),
                                estimator.type = estimator.type, uvec = uvec,
                                nugget.tolerance = nugget.tolerance, bin.cloud
                                 = bin.cloud, max.dist = umax, ...)
                        indp <- rep(TRUE, length(u))
                        if(pairs.min > 0) {
                                indp <- (result$n >= pairs.min)
                                if(is.matrix(result$v)) {
                                        result$v <- result$v[indp, ]
                                        result$sd <- result$sd[indp, ]
                                }
                                else {
                                        result$v <- result$v[indp]
                                        result$sd <- result$sd[indp]
                                }
                                result$u <- result$u[indp]
                                result$n <- result$n[indp]
                        }
                        result$ind.bin <- indp
                }
                if(option == "smooth") {
                        if(is.matrix(v))
                                stop("smooth not yet available for several variables"
                                        )
                        temp <- ksmooth(u, v, ...)
                        result <- list(u = temp[[1]], v = temp[[2]])
                }
        }
        result <- c(result, list(var.mark = data.var, output.type = option,
                estimator.type = estimator.type, n.data = n.data, lambda =
                lambda, trend = trend, call = class.fc))
        class(result) <- "variogram"
        return(result)
}

"wireframe.grf" <-
function(obj, sim.number = 1, ...)
{
        obj$data <- as.matrix(obj$data)
        wireframe(obj$data[, sim.number] ~ obj$coords[, 1] * obj$coords[, 2],
                xlab = "Coord X", ylab = "Coord Y", zlab = "Simulated", ...)
}

"wireframe.ksline" <-
function(obj.ksline, grid = obj.ksline$grid, values = c("est", "kvar",
        "oktrend", "ktrend", "ktetrend"), ...)
{
        values <- match.arg(values)
        if(values == "est") m <- obj.ksline$pred
        if(values == "kvar")
                m <- obj.ksline$krige.var
        wireframe(m ~ obj.ksline$grid[, 1] * obj.ksline$grid[, 2], xlab =
                "Coord X", ylab = "Coord Y", zlab = "Simulated", ...)
}

"wlsfit" <-
function(vario, ini.cov.pars, cov.model = c("exponential", "matern", "gaussian",
        "spherical", "cubic", "wave", "powered.exponential", "cauchy",
        "gneiting", "gneiting.matern", "pure.nugget"), fix.nugget = F, nugget = 0,
         kappa = NULL,
        simul.number = NULL, max.dist = "all", minimisation.function = c("nlminb", "ms"),
        lower = 0, weight = c("npairs", "cressie"),
        messages.screen = TRUE)
{
  call.fc <- match.call()
  ini <- ini.cov.pars
  weight <- match.arg(weight)
  minimisation.function <- match.arg(minimisation.function)
  cov.model <- match.arg(cov.model)
  if (is.matrix(vario$v) & is.null(simul.number))
    stop("object in vario$v is a matrix. This function works for only 1 empirical variogram")
  if (!is.null(simul.number))
    vario$v <- vario$v[, simul.number]
  ##
  ## Checking and/or preparing initial values
  ##
  if(is.matrix(ini)) {
    inilength <- dim(ini)[2]
    if(inilength != 2)
      stop("wrong number of columns for ini (must be always 2)")
    if(fix.nugget == FALSE)
      ini <- as.matrix(expand.grid(nugget, as.numeric(names(table(ini[,1]))),
                                   as.numeric(names(table(ini[,1])))))
  }
  else {
    inilength <- length(ini)
    if(inilength != 2)
      stop("wrong length for ini (must be 2)")
    if(fix.nugget == FALSE){
      if(length(nugget) == 1)
        ini <- c(ini, nugget)
      else
      ini <- as.matrix(expand.grid(nugget, ini[1], ini[2]))
    }
  }
  ##
  ## Setting maximum distance
  ##
  if(max.dist == "all")
    data <- list(u = vario$u, v = vario$v, n = vario$n)
  else data <- list(u = vario$u[vario$u <= max.dist], v = vario$v[vario$
                                                        u <= max.dist], n = vario$n[vario$u <= max.dist])
  holdit <- data$v
  holdit[is.na(holdit)] <- 0
  data$v <- holdit
  ##
  ## Preparing lists for the minimiser
  ##
  .global.list <- list(u = data$u, v = data$v, n = data$n,
                       fix.nugget = fix.nugget,
                       nugget = nugget,
                       kappa = kappa, cov.model = cov.model,
                       weight = weight, m.f = minimisation.function)
  if(minimisation.function == "nlm")
    assign(".global.lower", lower, where=1)
  ##
  ## Searchin for best initial value
  ##
  if(is.matrix(ini) | is.data.frame(ini)) {
    ini <- as.matrix(ini)
    if(messages.screen) cat("wlsfit: searching for the best initial value\n")
    ini.search <- ini
    dimnames(ini.search) <- list(NULL, NULL)
    wls.ini <- round(100000000. * apply(ini.search, 1, loss.wlsvario,
                                        .global.list = .global.list))
    ini <- as.vector(ini.search[wls.ini == min(wls.ini),])
    if(messages.screen){
      cat("wlsfit: best initial value:\n")
      if(fix.nugget == TRUE){
        names(ini) <- c("sill", "range")
        print(ini)
      }
      else{
        names(ini) <- c("nugget", "sill", "range")
        print(ini)
      }
    }
    remove(".global.lower", where=1)
  }
  ##
  ## Minimisation using "ms"
  ##
  if(minimisation.function == "ms") {
    if(fix.nugget == F) {
      if(weight == "npairs") {
        result <- ms( ~ n * ((v - ((sigmasq + tausq) -
                                   cov.spatial(u, cov.model = cov.model,
                                               kappa = kappa, cov.pars = c(sigmasq,
                                                                phi))))^2), start = list(tausq = ini[
                                                                                           1], sigmasq = ini[2], phi = ini[3]),
                     data = data)
      }
      if(weight == "cressie") {
        cov.vec <- cov.spatial(u, cov.model = cov.model,
                               kappa = kappa, cov.pars = c(sigmasq,
                                                phi))
        result <- ms( ~ (n/(cov.vec^2)) * ((v - ((
                                                  sigmasq + tausq) - cov.vec))^2), start
                     = list(tausq = ini[1], sigmasq = ini[
                                              2], phi = ini[3]), data = data)
      }
      nugget <- result$parameter[1]
      cov.pars <- as.vector(result$parameter[2:3])
    }
    else {
      data$nugget <- nugget
      if(weight == "npairs") {
        result <- ms( ~ n * ((v - ((sigmasq + nugget) -
                                   cov.spatial(u, cov.model = cov.model,
                                               kappa = kappa, cov.pars = c(sigmasq,
                                                                phi))))^2), start = list(sigmasq = ini[
                                                                                           1], phi = ini[2]), data = data)
      }
      if(weight == "cressie") {
        cov.vec <- cov.spatial(u, cov.model = cov.model,
                               kappa = kappa, cov.pars = c(sigmasq,
                                                phi))
        result <- ms( ~ (n/(cov.vec^2)) * ((v - ((
                                                  sigmasq + tausq) - cov.vec))^2), start
                     = list(tausq = ini[1], sigmasq = ini[
                                              2], phi = ini[3]), data = data)
      }
      cov.pars <- as.vector(result$parameter[1:2])
    }
    value <- result$value
    value <- sum(value)
    message <- paste("ms does not return convergence message")
  }
  ##
  ## Minimisation isung "nlminb"" "
  ##
  if(minimisation.function == "nlminb") {
    result <- nlminb(ini, loss.wlsvario, lower = lower, .global.list = .global.list)
    value <- result$objective
    message <- result$message
    if(fix.nugget == F) {
      nugget <- result$parameter[1]
      cov.pars <- as.vector(result$parameter[2:3])
    }
    else {
      cov.pars <- as.vector(result$parameter[1:2])
    }
  }
  estimation <- list(nugget = nugget, cov.pars = cov.pars, cov.model =
                     cov.model, kappa = kappa, value = value, max.dist = max(data$
                                                                u), minimisation.function = minimisation.function, message =
                     message)
  estimation$method <- "WLS"
  estimation$call <- call.fc
  class(estimation) <- "variomodel"
  if(messages.screen) {
    cat("wlsfit: estimated model parameters are:\n")
    cat(paste("covariance model:", cov.model))
    if(cov.model == "matern" | cov.model == "powered.exponential" |
       cov.model == "cauchy" | cov.model == "gneiting.matern"
       )
      cat(paste(" with kappa =", kappa))
    if(!is.null(kappa))
      if(cov.model == "matern" & kappa == 0.5)
        cat(" (exponential)")
    cat("\n")
    print(c(nugget = estimation$nugget, sill = estimation$cov.pars[
                                          1], range = estimation$cov.pars[2]))
  }
  return(estimation)
}

"xyzplot" <-
function(geodata, coords = geodata$coords, data = geodata$data, eyes, planepts
         = F, ...)
{
        xr <- range(coords[, 1])
        yr <- range(coords[, 2])
        zr <- range(data)
        if(missing(eyes))
                eyes <- c(8, 8, 3)
        eyex <- min(xr) - eyes[1] * (xr[2] - xr[1])
        eyey <- min(yr) - eyes[2] * (yr[2] - yr[1])
        eyez <- min(zr) + 2 * eyes[3] * (zr[2] - zr[1])
        eye <- c(eyex, eyey, eyez)
        ##eye_c(-113.6,29,20000)
        pp <- persp(xr, yr, matrix(zr[1], ncol = 2, nrow = 2), zlim = zr, eye
                 = eye, box = F, xlab = "Coord X", ylab = "Coord Y", zlab =
                "data")
        ptsin3d <- perspp(coords[, 1], coords[, 2], data, pp)
        ptsonplane <- perspp(coords[, 1], coords[, 2], rep(zr[1], length(data)),
                pp)
        points(ptsin3d, pch = 16, ...)
        if(planepts) {
                points(ptsonplane)
        }
        segments(ptsin3d$x, ptsin3d$y, ptsonplane$x, ptsonplane$y)
}

"image.krige.bayes" <-
  function (obj, locations,
            values.to.plot = c("moments.mean", "moments.variance",
              "mean.simulations", "variance.simulations",
              "quantiles", "probability", "simulation"),
            number.col, ...)
{
  if(is.null(locations))
    stop("prediction locations must be provided")
  locations <- locations[order(locations[, 2], locations[,1]), ]
  x <- as.numeric(levels(as.factor(locations[, 1])))
  nx <- length(x)
  y <- as.numeric(levels(as.factor(locations[, 2])))
  ny <- length(y)
  if (nx == ny)
    par(pty = "s")
  if (is.numeric(values.to.plot)){
    image(x,y,matrix(values.to.plot, ncol=ny), ...)
  }
  else{
    values <- match.arg(values.to.plot)
    switch(values,
           moments.mean =
           {
             values <- matrix(obj$predictive$moments$expect.y0, ncol = ny);
             cat("plotting map the mean of the predictive distribution\n")
           },
           moments.variance = {
             values <- matrix(obj$predictive$moments$var.y0, ncol = ny);
             cat("plotting map the variance of the predictive distribution\n")
           },
           mean.simulations=
           {
             values <- matrix(obj$predictive$mean.simulations, ncol = ny);
             cat("plotting map the mean of the simulations from the predictive distribution\n")
           },
           variance.simulations =
           {
             values <- matrix(obj$predictive$variance.simulations, ncol = ny);
             cat("plotting map the variance of the predictive distribution\n")
           },
           quantile =
           {
             if(!is.vector(obj$predictive$quantiles))
               if(missing(number.col))
                 stop("argument number.col must be provided")
               else
                 values <- matrix(obj$predictive$quantiles[,number.col], ncol = ny)
             else
               values <- matrix(obj$predictive$quantiles, ncol = ny);
             cat("plotting map a quantile of the predictive distribution\n")
           },
           probability =
           {
             if(!is.vector(obj$predictive$probability))
               if(missing(number.col))
                 stop("argument number.col must be provided")
               else
                 values <- matrix(obj$predictive$probability[,number.col], ncol = ny)
             else
               values <- matrix(obj$predictive$probability, ncol = ny);
             cat("plotting map a simulation of the predictive distribution\n")
           },
           simulation =
           {
             values <- matrix(obj$predictive$simulations[,number.col], ncol = ny);
             cat("plotting map the variance of the predictive distribution\n")
           },
           stop("wrong specification for values to plot")
           )
    image(x, y, values, ...)
  }
  return(invisible())
}

"persp.krige.bayes" <-
  function (obj, locations,
            values.to.plot = c("moments.mean", "moments.variance",
              "mean.simulations", "variance.simulations",
              "quantiles", "probability", "simulation"), number.col, ...)
{
  if(is.null(locations))
    stop("prediction locations must be provided")
  locations <- locations[order(locations[, 2], locations[,1]), ]
  x <- as.numeric(levels(as.factor(locations[, 1])))
  nx <- length(x)
  y <- as.numeric(levels(as.factor(locations[, 2])))
  ny <- length(y)
  if (nx == ny)
    par(pty = "s")
  if (is.numeric(values.to.plot)){
    persp(x,y,matrix(values.to.plot, ncol=ny), ...)
  }
  else{
    values <- match.arg(values.to.plot)
    switch(values,
           moments.mean =
           {
             values <- matrix(obj$predictive$moments$expect.y0, ncol = ny);
             cat("plotting map the mean of the predictive distribution\n")
           },
           moments.variance = {
             values <- matrix(obj$predictive$moments$var.y0, ncol = ny);
             cat("plotting map the variance of the predictive distribution\n")
           },
           mean.simulations=
           {
             values <- matrix(obj$predictive$mean.simulations, ncol = ny);
             cat("plotting map the mean of the simulations from the predictive distribution\n")
           },
           variance.simulations =
           {
             values <- matrix(obj$predictive$variance.simulations, ncol = ny);
             cat("plotting map the variance of the predictive distribution\n")
           },
           quantile =
           {
             if(!is.vector(obj$predictive$quantiles))
               if(missing(number.col))
                 stop("argument number.col must be provided")
               else
                 values <- matrix(obj$predictive$quantiles[,number.col], ncol = ny)
             else
               values <- matrix(obj$predictive$quantiles, ncol = ny);
             cat("plotting map a quantile of the predictive distribution\n")
           },
           probability =
           {
             if(!is.vector(obj$predictive$probability))
               if(missing(number.col))
                 stop("argument number.col must be provided")
               else
                 values <- matrix(obj$predictive$probability[,number.col], ncol = ny)
             else
               values <- matrix(obj$predictive$probability, ncol = ny);
             cat("plotting map a simulation of the predictive distribution\n")
           },
           simulation =
           {
             values <- matrix(obj$predictive$simulations[,number.col], ncol = ny);
             cat("plotting map the variance of the predictive distribution\n")
           },
           stop("wrong specification for values to plot")
           )
    persp(x, y, values, ...)
  }
  return(invisible())
}



##
## "wrappers" for pieces of C code used in geoR/geoS
##

"bilinearformXAY" <-
  function(X, lowerA, diagA, Y)
{
  nA <- length(diagA)
  nX <- length(X)/nA
  nY <- length(Y)/nA
  if(length(lowerA) != (nA * (nA -1)/2))
    stop("lowerA and diagA have incompatible dimentions")
  out <- .C("bilinearform_XAY",



            as.double(as.vector(lowerA)),
            as.double(as.vector(diagA)),
            as.double(as.vector(X)),
            as.double(as.vector(Y)),
            as.integer(nX),
            as.integer(nY),
            as.integer(nA),
            res=as.double(rep(0,(nX*nY))))$res
  attr(out, "dim") <- c(nX, nY)
  return(out)
}

"diagquadraticformXAX" <-
  function(X, lowerA, diagA)
{
  nA <- length(diagA)
  nX <- length(X)/nA
  if(length(lowerA) != (nA * (nA -1)/2))
    stop("lowerA and diagA have incompatible dimentions")
  out <- .C("diag_quadraticform_XAX",



            as.double(as.vector(lowerA)),
            as.double(as.vector(diagA)),
            as.double(as.vector(X)),
            as.integer(nX),
            as.integer(nA),
            res = as.double(rep(0,nX)))$res
  return(out)
}

"loccoords" <-
  function(coords, locations)
{
  ## Computes a matrix for which each row has the distances between
  ## each point in 'locations' to all the points in 'coords'
  coords <- as.matrix(coords)
  locations <- as.matrix(locations)
  dimc <- dim(coords)
  diml <- dim(locations)
  if((dimc[2] != 2) | (diml[2] != 2))
    stop("coords and locations must have two columns")
  nc <- dimc[1]
  nl <- diml[1]
  out <- .C("loccoords",



            as.double(as.vector(locations[,1])),
            as.double(as.vector(locations[,2])),
            as.double(as.vector(coords[,1])),
            as.double(as.vector(coords[,2])),
            as.integer(nl),
            as.integer(nc),
            res = as.double(rep(0, nc*nl)))$res
  attr(out, "dim") <- c(nc, nl)
  return(out)
}

"distdiag" <-
  function(coords)
{
  ## returns the lower triangle of the matrix with euclidean distances
  ## between pairs of points, including the diagonal.
  ##
  coords <- as.matrix(coords)
  dimc <- dim(coords)
  if(dimc[2] == 1 & dimc[1] == 2)
    return(0)
  else{
    if(dimc[2] != 2)
      stop("coords must have two columns")
    nc <- dimc[1]
    out <- .C("distdiag",



              as.double(coords[,1]),
              as.double(coords[,2]),
              as.integer(nc),
              res = as.double(rep(0, (nc * (nc+1)/2))))$res
    return(out)
  }
}
"lines.grf" <-
  function (obj, max.dist = max(dist(obj$coords)), length = 100, 
            lwd = 2, ...) 
{
  if(is.R()) require(mva)
  if (obj$cov.model == "matern" | obj$cov.model == "powered.exponential" | 
      obj$cov.model == "cauchy" | obj$cov.model == "gneiting-matern") 
    kappa <- obj$kappa
  else kappa <- NULL
  distance <- seq(0, max.dist, length = length)
  if (is.vector(obj$cov.pars)) 
    sill.total <- obj$nugget + obj$cov.pars[1]
  else sill.total <- obj$nugget + sum(obj$cov.pars[, 1])
  gamma <- sill.total - cov.spatial(distance, cov.model = obj$cov.model, 
                                  kappa = kappa, cov.pars = obj$cov.pars)
  lines(distance, gamma, lwd = lwd, ...)
  return(invisible())
}

"cov.spatial" <-
  function(obj, cov.model = c("exponential", "matern", "gaussian", "spherical",
                  "circular", "cubic", "wave", "powered.exponential",
                  "cauchy", "gneiting", "gneiting.matern", "pure.nugget"),
           kappa, cov.pars = stop("no cov.pars argument provided"))
{
  ##
  ## checking/reading input
  ##
  cov.model <- match.arg(cov.model)
  if(cov.model == "matern" | cov.model == "powered.exponential" | 
     cov.model == "cauchy" | cov.model == "gneiting.matern"){
    if(missing(kappa))
      stop("for matern, powered.exponential, cauchy, gneiting and gneiting.matern covariance functions the parameter kappa must be provided")
    if(cov.model == "gneiting.matern" & length(kappa) != 2)
      stop("gneiting.matern correlation function model requires a vector with 2 parameters in the argument kappa")
    if((cov.model == "matern" | cov.model == "powered.exponential" | 
        cov.model == "cauchy") & length(kappa) != 1)
      stop("for this choice of  correlation function model kappa should be a scalar parameter")
  }
  if(is.vector(cov.pars))
    sigmasq <- cov.pars[1]
  else sigmasq <- cov.pars[, 1]
  if(is.vector(cov.pars))
    phi <- cov.pars[2]
  else phi <- cov.pars[, 2]
  if(is.vector(cov.pars))
    ns <- 1
  else ns <- nrow(cov.pars)
  covs <- array(0, dim = dim(obj))
  ##
  ## computing correlations/covariances
  ##
  for(i in 1:ns) {
    if(phi[i] < 1e-12)
      cov.model <- "pure.nugget"
    cov.values <- switch(cov.model,
                  pure.nugget = rep(0, length(obj)),
                  wave = (1/obj) * (phi[i] * sin(obj/phi[i])),
                  exponential = exp( - (obj/phi[i])),
                  matern = matern(u = obj, phi = phi[i], kappa = kappa),
                  gaussian = exp( - ((obj/phi[i])^2)),
                  spherical = ifelse(obj < phi[i], (1 - 1.5 * (obj/phi[i]) +
                    0.5 * (obj/phi[i])^3), 0),
                  circular = {
                    obj.sc <- obj/phi[i];
                    obj.sc[obj.sc > 1] <- 1;
                    ifelse(obj < phi[i], (1 - (2 * ((obj.sc) *
                                                    sqrt(1 - ((obj.sc)^2)) +
                                                    asin(obj.sc)))/pi), 0)
                  },
                  cubic = {
                    obj.sc <- obj/phi[i];
                    ifelse(obj < phi[i], (1 - (7 * (obj.sc^2) -
                                               8.75 * (obj.sc^3) +
                                               3.5 * (obj.sc^5) -
                                               0.75 * (obj.sc^7))), 0)
                  },
                  powered.exponential = {
                    if(kappa > 2 | kappa <= 0)
                      stop("for power exponential correlation model the parameter kappa must be in the intervel (0,2]"
                           );
                    exp( - ((obj/phi[i])^kappa))
                  },
                  cauchy = (1 + (obj/phi[i])^2)^(-1 * kappa),
                  gneiting = {
                    obj.sc <- obj/phi[i];
                    t2 <- (1 - obj.sc);
                    t2 <- ifelse(t2 > 0, (t2^8), 0);
                    (1 + 8 * obj.sc + 25 * (obj.sc^2) + 32 * (obj.sc^
                                                              3)) * t2
                  },
                  gneiting.matern = { 
                    obj.sc <- obj/(phi[i] * kappa[2]);
                    t2 <- (1 - obj.sc);
                    t2 <- ifelse(t2 > 0, (t2^8), 0);
                    cov.values <- (1 + 8 * obj.sc + 25 * (obj.sc^2) + 32 * (obj.sc^3)) * t2;
                    cov.values * matern(u = obj, phi = phi[i], kappa = kappa[1])
                  },
                  stop("wrong or no specification of cov.model")
                  )
    cov.values <- sigmasq[i] * cov.values
    covs <- covs + cov.values
  }
  covs[obj < 1e-15] <- sum(sigmasq)
  return(covs)
}

"cor.number" <- 
  function(cov.model= c("exponential", "matern", "gaussian",
             "spherical", "circular", "cubic", "wave",
             "powered.exponential", "cauchy", "gneiting",
             "gneiting.matern", "pure.nugget"))
{
###	WARNING: codes above must be the same as in the C code
###              "cor_diag"
  
  cov.model <- match.arg(cov.model)
  cornumber <- switch(cov.model,
                      pure.nugget = as.integer(1),
                      exponential = as.integer(2),
                      spherical = as.integer(3),
                      gaussian = as.integer(4),
                      wave = as.integer(5),
                      cubic = as.integer(6),
                      powered.exponential = as.integer(7),
                      cauchy = as.integer(8),
                      gneiting = as.integer(9),
                      circular = as.integer(10),
                      matern = as.integer(11),
                      gneiting.matern = as.integer(12),
                      stop("wrong or no specification of cov.model")
                      )
  return(cornumber)
}

"read.geodata" <-
  function(file, header = FALSE, coords.col= 1:2, data.col = 3,
           data.names = NULL, covar.col = NULL, covar.names = "header", ...)
{
  obj <- read.table(file = file, header = header, ...)
  if(covar.names == "header"){
    if(!is.null(covar.col)){
      col.names <- names(obj)
      covar.names <- col.names[covar.col]
    }
    else covar.names <- NULL
  }
  res <- as.geodata(obj = obj, coords.col = coords.col, data.col = data.col,
                    covar.col = covar.col, covar.names = covar.names)
  return(res)
}

"as.geodata" <-
  function(obj, coords.col = 1:2, data.col = 3, data.names = NULL, 
           covar.col = NULL, covar.names = "obj.names")
{
  if(!is.matrix(obj) & !is.data.frame(obj))
    stop("object must be a matrix or data.frame")
  if(!is.null(data.names) & length(data.col) < 2)
    stop("data.names allowed only if there is more than 1 column of data")
  res <- list()
  res$coords <- as.matrix(obj[,coords.col])
  res$data <- as.matrix(obj[,data.col])
  if(length(data.col) == 1) res$data <- as.vector(res$data)
  else{
    res$data <- as.data.frame(res$data)
    if(!is.null(data.names)) names(res$data) <- data.names
  }
  if(!is.null(covar.col)){
    res[[3]] <- as.data.frame(as.matrix(obj[,covar.col]))
    if(covar.names == "obj.names"){
      if(is.matrix(obj))      
        col.names <- dimnames(obj)[2]
      if(is.data.frame(obj))      
        col.names <- names(obj)
    }
    if(length(covar.col) == 1){
      if(covar.names == "obj.names")
        names(res)[3] <- col.names[covar.col]
      else
        names(res)[3] <- covar.names
    }
    else{
      names(res)[3] <- "covariates"
      if(covar.names == "obj.names")
        names(res[[3]]) <- col.names[covar.col]
      else
        names(res[[3]]) <- covar.names
    }
  }
  class(res) <- "geodata"
  return(res)
}


"image.grf" <-
  function (obj, sim.number = 1, ...) 
{
  x <- as.numeric(levels(as.factor(obj$coords[, 1])))
  nx <- length(x)
  y <- as.numeric(levels(as.factor(obj$coords[, 2])))
  ny <- length(y)
  obj$data <- as.matrix(obj$data)
  n <- nrow(obj$data)
  if (nx * ny != n) 
    stop("Probably irregular grid")
  m <- matrix(obj$data[, sim.number], ncol = ny)
  image(x, y, m, ...)
  return(invisible())
}

"image.kriging" <-
  function (kriging.obj, locations = kriging.obj$locations, values = kriging.obj$predict, coords.data, ...) 
{
##  par.ori <- par(no.readonly = TRUE)
##  on.exit(par(par.ori))
  if(is.null(locations))
    stop("prediction locations must be provided")
  if(is.null(values))
    stop("data.to be plotted must be provided")
  locations <- locations[order(locations[, 2], locations[,1]), ]
  x <- as.numeric(levels(as.factor(locations[, 1])))
  nx <- length(x)
  y <- as.numeric(levels(as.factor(locations[, 2])))
  ny <- length(y)
  if  (nx == ny) 
    par(pty = "s")
  values <- matrix(values, ncol = ny)
  image(x, y, values, ...)
  if(!missing(coords.data))
    points(coords.data)
  return(invisible())
}

"lines.variogram" <-
function (obj, max.dist = max(obj$u), type = "o", scaled = FALSE, ...) 
{
  if (scaled) 
    obj$v <- obj$v/obj$var.mark
  if (!is.matrix(obj$v)) 
    lines(obj$u[obj$u <= max.dist], obj$v[obj$u <= max.dist], 
          type = type, ...)
  else {
    for (j in 1:ncol(obj$v)) lines(obj$u[obj$u <= max.dist], 
                                   obj$v[obj$u <= max.dist, j], type = type, ...)
  }
}
"lines.variomodel" <-
function (obj, max.dist = obj$max.dist, n.points = 100, scaled = FALSE,...)
{
  if (scaled) 
    obj$v <- obj$v/obj$var.mark
  if (is.null(max.dist)) 
    stop("argument max.dist needed for this object")
  if (obj$cov.model == "matern" | obj$cov.model == "powered.exponential" | 
      obj$cov.model == "cauchy" | obj$cov.model == "gneiting-matern") 
    kappa <- obj$kappa
  else kappa <- NULL
  distance <- seq(0, max.dist, length = n.points)
  if (is.vector(obj$cov.pars)) 
    sill.total <- obj$nugget + obj$cov.pars[1]
  else sill.total <- obj$nugget + sum(obj$cov.pars[, 1])
  if (obj$cov.pars[2] < 1e-12)
    gamma <- rep(sill.total, n.points)
  else
    gamma <- sill.total - cov.spatial(distance, cov.model = obj$cov.model, 
                                    kappa = kappa, cov.pars = obj$cov.pars)
  lines(distance, gamma, ...)
}

"persp.kriging" <- 
  function(kriging.obj, locations = kriging.obj$locations, values = kriging.obj$
           predict, ...)
{
  locations <- locations[order(locations[, 2], locations[, 1]),  ]
  x <- as.numeric(levels(as.factor(locations[, 1])))
  nx <- length(x)
  y <- as.numeric(levels(as.factor(locations[, 2])))
  ny <- length(y)
  if(nx == ny)
    par(pty = "s")
  values <- matrix(values, ncol = ny)
  persp(x, y, values, ...)
  return(invisible())
}

"persp.grf" <- 
function(obj, sim.number = 1, ...)
{
	x <- as.numeric(levels(as.factor(obj$coords[, 1])))
	nx <- length(x)
	y <- as.numeric(levels(as.factor(obj$coords[, 2])))
	ny <- length(y)
	obj$data <- as.matrix(obj$data)
	n <- nrow(obj$data)
	if(nx * ny != n)
		stop("Probably irregular grid")
	m <- matrix(obj$data[, sim.number], ncol = ny)
	persp(x, y, m, ...)
	return(invisible())
}

"coords.aniso" <- 
  function(coords, aniso.pars, reverse=FALSE)
{
  coords <- as.matrix(coords)
  n <- nrow(coords)
  rotation <- aniso.pars[1]
  ratio <- aniso.pars[2]
  if(ratio < 1){
    ratio <- round(ratio, dig=8)
    if(ratio < 1)
      stop("anisotropy ratio must be greater than 1")
  }
  rm <- matrix(c(cos(rotation), -sin(rotation),
                 sin(rotation), cos(rotation)),
               ncol = 2)
  tm <- diag(c(1, 1/ratio))
  if(reverse)
    coords.mod <- coords %*% solve(rm %*% tm)
  else
    coords.mod <- coords %*% rm %*% tm
  return(coords.mod)
}


"plot.grf" <-
  function (obj, model.line = TRUE, plot.grid = FALSE, ylim="default", ...) 
{
  nsim <- ncol(obj$data)
  if (plot.grid) 
    plot(obj$coords[, 1], obj$coords[, 2], xlab = "Coord X", ylab = "Coord Y")
  if (is.vector(obj$cov.pars)) 
    sill.total <- obj$nugget + obj$cov.pars[1]
  else sill.total <- obj$nugget + sum(obj$cov.pars[, 1])
  if (obj$lambda != 1){
    if (obj$lambda == 0) data <- log(obj$data)
    else data <- ((obj$data^obj$lambda)-1)/obj$lambda
  }
  else
    data <- obj$data          
  sim.bin <- variog(obj)
  plot(sim.bin, var.lines = FALSE, ylim = c(0,max(sim.bin$v, sill.total)), ...)
  if (model.line){
    var.model <- list(nugget = obj$nugget, cov.pars = obj$cov.pars, 
                      kappa = obj$kappa, max.dist = max(sim.bin$u),
                      cov.model = obj$cov.model)
    lines.variomodel(var.model, lwd = 3)
  }
  return(invisible())
}

"summary.variomodel" <-
  function(obj)
{
  cat("Summary of the parameter estimation\n")
  cat("-----------------------------------\n")
  if(obj$method == "ML")
    method <- "maximum likelihood"
  if(obj$method == "RML")
    method <- "restricted maximum likelihood"
  if(obj$method == "OLS")
    method <- "ordinary least squares"
  if(obj$method == "WLS")
    method <- "weighted least squares"
  cat(paste("Estimation method:", method, "\n"))
  cat("\n")
  cat(paste("Covariance model:", obj$cov.model))
  if(obj$cov.model == "matern" | obj$cov.model == "powered.exponential" |
     obj$cov.model == "cauchy" | obj$cov.model == "gneiting.matern"
     )
    cat(paste(" with kappa = ", obj$kappa))
  if(!is.null(obj$kappa))
    if(obj$cov.model == "matern" & obj$kappa == 0.5)
      cat(" (exponential)")
  cat("\n")
  cat("Covariance parameters:\n")
  print(c(nugget = obj$nugget, sill = obj$cov.pars[1], range = obj$
          cov.pars[2]))
  cat("\n")
  if(obj$method == "ML" | obj$method == "RML") {
    if(obj$lambda != 1) {
      if(exists("obj$call$fix.lambda"))
        if(obj$call$fix.lambda == FALSE){
          cat(paste(
                    "Estimated Box-Cox transformation parameter:",
                    round(obj$lambda, dig=3)))
        }
        else cat(paste("Fixed Box-Cox transformation parameter:",
                       obj$lambda))
      cat("\n")
      cat("\n")
    }
    cat("mean parameter(s): ")
    if(length(obj$beta) == 1){
      cat(obj$beta)
      cat("\n")
    }
    else{
      cat("\n")
      print(obj$beta)
    }
    cat("\n")
    print(c(log.L = obj$loglik, AIC = obj$AIC, BIC = obj$BIC, 
            n.params = as.integer(obj$npars)))
  }
  return(invisible())
}

"trend.spatial" <-
  function(trend, coords=NULL)
{
  if(inherits(trend, "formula")){
    trend.mat <- model.matrix(trend)
  }
  else{
    if(trend == "cte")
      trend.mat <- as.matrix(rep(1, nrow(coords)))
    else
      if(trend == "1st")
        trend.mat <- cbind(1, coords)
      else
        if(trend == "2nd")
          trend.mat <- cbind(1, coords, coords[, 1]^2, coords[
                                                              , 2]^2, coords[, 1] * coords[, 2])
	else
          stop("external trend must be provided for data locations to be estimated using the argments `trend.d` and `trend.l`. Both (trend.d and trend.l) must be the same `cte`, `1st`, `2nd` or given by a formula of the type ~X where X is a matrix or vector of covariates."
               )
  }
  trend.mat <- as.matrix(trend.mat)
  dimnames(trend.mat) <- list(NULL, NULL)
  return(trend.mat)
}

#"dist0.krige" <-
#function (x0, coords) 
#{
#  if (length(x0) != 2) 
#    stop(paste("length of x0 is", length(x0), "(it must be 2)"))
#  coords[, 1] <- coords[, 1] - x0[1]
#  coords[, 2] <- coords[, 2] - x0[2]
#  return(sqrt(coords[, 1]^2 + coords[, 2]^2))
#}


"polygrid" <- 
  function(xgrid, ygrid, poly, vec.inout = F)
{
  if(is.R())
  	require(splancs)
  if(exists("inout")){
    xygrid <- expand.grid(x = xgrid, y = ygrid)
    ind <- as.vector(inout(xygrid, poly))
    xypoly <- xygrid[ind == T,  ]
    if(vec.inout == F)
      return(xypoly)
    else return(list(xypoly = xypoly, vec.inout = ind))
  }
  else{
    cat("ERROR: cannot run the function\n")
    cat("package \"splancs\" should be installed/loaded")
    return(invisible())
  }
}

#"variog.env" <-
#  function (obj.variog, coords, model.pars, nsim = 99, messages.screen = TRUE)  
#{
#  cat("This function has been made obsolete\n")
#  cat("There are now two functions for variogram envelops:\n")
#  cat(" - variog.env.model:\n")
#  cat("       the same as the previous variog.env, based on the model")
#  cat(" - variog.env.mc:\n")
#  cat("       the new one based on permutations of the data")
#  return(invisible())
#}

"variog.model.env" <-
  function(geodata, coords = geodata$coords, obj.variog,
           model.pars, nsim = 99, save.sim = FALSE,
           messages.screen = TRUE) 
{
  call.fc <- match.call()
  obj.variog$v <- NULL
  ##
  ## reading input
  ##
  if(!is.null(model.pars$beta)) beta <- model.pars$beta
  else beta <- 0
  if(!is.null(model.pars$cov.model))
    cov.model <- model.pars$cov.model
  else cov.model <- "exponential"
  if(!is.null(model.pars$kappa)) kappa <- model.pars$kappa
  else kappa <- 0.5
  if(!is.null(model.pars$nugget)) nugget <- model.pars$nugget
  else nugget <- 0
  cov.pars <- model.pars$cov.pars
  if(!is.null(obj.variog$estimator.type))
    estimator.type <- obj.variog$estimator.type
  else estimator.type <- "classical"
  if (obj.variog$output.type != "bin") 
    stop("envelops can be computed only for binned variogram")
  ##
  ## generating simulations from the model with parameters provided
  ##
  if (messages.screen) 
    cat(paste("variog.env: generating", nsim, "simulations (with ",
              obj.variog$n.data, 
              "points each) using the function grf\n"))
  simula <- grf(obj.variog$n.data, grid = as.matrix(coords),
                cov.model = cov.model, cov.pars = cov.pars,
                nugget = nugget, kappa = kappa, nsim = nsim,
                messages.screen = FALSE, lambda = obj.variog$lambda)
  if(messages.screen)
    cat("variog.env: adding the mean or trend\n")
  x.mat <- trend.spatial(trend=obj.variog$trend, coords=coords)
  simula$data <- as.vector(x.mat %*% beta) + simula$data
  ##
  ## computing empirical variograms for the simulations
  ##
  if (messages.screen) 
    cat(paste("variog.env: computing the empirical variogram for the", 
              nsim, "simulations\n"))
  nbins <- length(obj.variog$bins.lim) - 1
  if(is.R()){
    bin.f <- function(sim){
      cbin <- vbin <- sdbin <- rep(0, nbins)  
      temp <- .C("binit",
                 as.integer(obj.variog$n.data),
                 as.double(as.vector(coords[,1])),
                 as.double(as.vector(coords[,2])),
                 as.double(as.vector(sim)),
                 as.integer(nbins),
                 as.double(as.vector(obj.variog$bins.lim)),
                 as.integer(estimator.type == "robust"),
                 as.double(max(obj.variog$u)),
                 as.double(cbin),
                 vbin = as.double(vbin),
                 as.integer(FALSE),
                 as.double(sdbin)
                 )$vbin
      return(temp)
    }
    simula.bins <- apply(simula$data, 2, bin.f)
    simula.bins <- simula.bins[obj.variog$ind.bin,]
  }
  else{
    bin.f <- function(sim, nbins, n.data, coords, bins.lim, estimator.type, max.u){
      cbin <- vbin <- sdbin <- rep(0, nbins)  
      temp <- .C("binit",
                 as.integer(n.data),
                 as.double(as.vector(coords[,1])),
                 as.double(as.vector(coords[,2])),
                 as.double(as.vector(sim)),
                 as.integer(nbins),
                 as.double(as.vector(bins.lim)),
                 as.integer(estimator.type == "robust"),
                 as.double(max.u),
                 as.double(cbin),
                 vbin = as.double(vbin),
                 as.integer(FALSE),
                 as.double(sdbin)
                 )$vbin
      return(temp)
    }
    simula.bins <- apply(simula$data, 2, bin.f, nbins=nbins, n.data=obj.variog$n.data, coords=coords, bins.lim=obj.variog$bins.lim, estimator.type=estimator.type, max.u=max(obj.variog$u))
    simula.bins <- simula.bins[obj.variog$ind.bin,]
  }
  if(save.sim == FALSE) simula$data <- NULL
  ##
  ## computing envelops
  ##
  if (messages.screen) 
    cat("variog.env: computing the envelops\n")
  limits <- apply(simula.bins, 1, range)
  res.env <- list(u = obj.variog$u, v.lower = limits[1, ],
                  v.upper = limits[2,])
  if(save.sim) res.env$simulated <- simula$data
  res.env$call <- call.fc
  class(res.env) <- "variogram.envelope"
  return(res.env)
}

"variog.mc.env" <-
  function (geodata, coords = geodata$coords, data = geodata$data,
            obj.variog, nsim = 99, save.sim = FALSE,
            messages.screen = TRUE) 
{
  call.fc <- match.call()
  ##
  ## Checking input
  ##
  obj.variog$v <- NULL
  if((is.matrix(data) | is.data.frame(data)))
    if(ncol(data) > 1)
      stop("envelops can be computed for only one data set at once")
  if(!is.null(obj.variog$estimator.type))
    estimator.type <- obj.variog$estimator.type
  else estimator.type <- "classical"
  ##
  ## generating several "data-sets" by permutating data values
  ##
  if (messages.screen) 
    cat(paste("variog.env: generating", nsim,
              "simulations by permutating data values\n"))
  simula <- list(coords = coords)
  n.data <- length(data)
  perm.f <- function(i, data, n.data){return(data[sample(1:n.data)])}
  simula$data <- apply(as.matrix(1:nsim),1, perm.f, data=data, n.data=n.data) 
  ##
  ## computing empirical variograms for the simulations
  ##
  if (messages.screen) 
    cat(paste("variog.env: computing the empirical variogram for the", 
              nsim, "simulations\n"))
  nbins <- length(obj.variog$bins.lim) - 1
  if(is.R()){
    bin.f <- function(sim){
      cbin <- vbin <- sdbin <- rep(0, nbins)  
      temp <- .C("binit",
                 as.integer(obj.variog$n.data),
                 as.double(as.vector(coords[,1])),
                 as.double(as.vector(coords[,2])),
                 as.double(as.vector(sim)),
                 as.integer(nbins),
                 as.double(as.vector(obj.variog$bins.lim)),
                 as.integer(estimator.type == "robust"),
                 as.double(max(obj.variog$u)),
                 as.double(cbin),
                 vbin = as.double(vbin),
                 as.integer(FALSE),
                 as.double(sdbin)
                 )$vbin
      return(temp)
    }
    simula.bins <- apply(simula$data, 2, bin.f)
    simula.bins <- simula.bins[obj.variog$ind.bin,]
  }
  else{
    bin.f <- function(sim, nbins, n.data, coords, bins.lim, estimator.type, max.u){
      cbin <- vbin <- sdbin <- rep(0, nbins)  
      temp <- .C("binit",
                 as.integer(n.data),
                 as.double(as.vector(coords[,1])),
                 as.double(as.vector(coords[,2])),
                 as.double(as.vector(sim)),
                 as.integer(nbins),
                 as.double(as.vector(bins.lim)),
                 as.integer(estimator.type == "robust"),
                 as.double(max.u),
                 as.double(cbin),
                 vbin = as.double(vbin),
                 as.integer(FALSE),
                 as.double(sdbin)
                 )$vbin
      return(temp)
    }
    simula.bins <- apply(simula$data, 2, bin.f, nbins=nbins, n.data=obj.variog$n.data, coords=coords, bins.lim=obj.variog$bins.lim, estimator.type=estimator.type, max.u=max(obj.variog$u))
    simula.bins <- simula.bins[obj.variog$ind.bin,]
  }
  if(save.sim == FALSE) simula$data <- NULL
  ##
  ## computing envelops
  ##
  if (messages.screen) 
    cat("variog.env: computing the envelops\n")
  limits <- apply(simula.bins, 1, range)
  res.env <- list(u = obj.variog$u, v.lower = limits[1, ],
                  v.upper = limits[2,])
  if(save.sim) res.env$simulations <- simula$data
  res.env$call <- call.fc
  class(res.env) <- "variogram.envelope"
  return(res.env)
}

"lines.variogram.envelope" <-
  function(obj, lty=3, ...)
{
  lines(obj$u, obj$v.lower, ...)
  lines(obj$u, obj$v.upper, ...)
  return(invisible())
}

"varcov.spatial" <-
  function(coords = NULL, dists.lowertri = NULL,
           cov.model = c("exponential","matern", "gaussian", "spherical",
             "circular", "cubic", "wave", "powered.exponential",
             "cauchy", "gneiting", "gneiting.matern", "pure.nugget"),
           kappa = NULL, nugget = 0, cov.pars = stop("no cov.pars argument"),
           inv = FALSE, det = FALSE,
           func.inv = c("cholesky", "eigen", "svd", "solve"),
           scaled = FALSE, only.decomposition = FALSE, sqrt.inv = FALSE,
           try.another.decomposition = TRUE, only.inv.lower.diag = FALSE)
{
  if(is.R()) require(mva)
  func.inv <- match.arg(func.inv)
  cov.model <- match.arg(cov.model)
  if(only.inv.lower.diag) inv <- TRUE
  if(is.null(coords) & is.null(dists.lowertri))
    stop("one of the arguments, coords or dists.lowertri must be provided")
  if(!is.null(coords) & !is.null(dists.lowertri))
    stop("only ONE argument, either coords or dists.lowertri must be provided")
  if(!is.null(coords))
    n <- nrow(coords)
  if(!is.null(dists.lowertri)){
    n <- as.integer(round(0.5 * (1 + sqrt(1 + 8 * length(dists.lowertri)))))
  }
  tausq <- nugget
  if(is.vector(cov.pars)){
    sigmasq <- cov.pars[1]
    phi <- cov.pars[2]
  }
  else{
    sigmasq <- cov.pars[, 1]
    phi <- cov.pars[, 2]
  }
  if(!is.null(coords)) {
    dists.lowertri <- as.vector(dist(coords))
  }
  if(round(1e12*min(dists.lowertri)) == 0)
    warning("Two or more pairs of data at coincident (or very close) locations. \nThis can cause matrices operations to crash!\n")
  ##
  ## Building the covariance matrix
  ##
  varcov <- matrix(0, n, n)
  if(scaled) {
    if(all(phi < 1e-12))
      varcov <- diag(x=(1 + (tausq/sum(sigmasq))),n)
    else{
      if(is.vector(cov.pars))
        cov.pars.sc <- c(1, phi)
      else cov.pars.sc <- cbind(1, phi)
      covvec <- cov.spatial(obj = dists.lowertri, cov.model = 
                            cov.model, kappa = kappa, cov.pars = cov.pars.sc)
      varcov[lower.tri(varcov)] <- covvec
      varcov <- t(varcov)
      varcov[lower.tri(varcov)] <- covvec
      if(is.R()){
        remove("covvec")
        ##gc(verbose=FALSE)
      }
      else remove("covvec", frame=sys.nframe())
      diag(varcov) <- 1 + (tausq/sum(sigmasq))
    }
  }
  else {
    if(all(sigmasq < 1e-10) | all(phi < 1e-10)){
      varcov <- diag(x=(tausq + sum(sigmasq)), n)
    }
    else{
      covvec <- cov.spatial(obj = dists.lowertri, cov.model = 
                            cov.model, kappa = kappa, cov.pars = cov.pars)
      varcov[lower.tri(varcov)] <- covvec
      varcov <- t(varcov)
      varcov[lower.tri(varcov)] <- covvec
      if(is.R()){
        remove("covvec")
        ##gc(verbose=FALSE)
      }
      else remove("covvec", frame=sys.nframe())
      diag(varcov) <- tausq + sum(sigmasq)
    }
  }
  ##
  ## Computing decompositions, inverses and determinants
  ## 
  ##  if(test.decomp == TRUE & func.inv == "cholesky"){
  ##    varcov.eig <- eigen(varcov, symmetric = TRUE, only.values = TRUE))
  ##    if(any(varcov.eig$val < 1e-12)){
  ##      ##      func.inv <- "eigen"
  ##      diag(varcov) <- 1.0001 * diag(varcov)
  ##    }
  ##  }
  if(inv | det | only.decomposition | sqrt.inv | only.inv.lower.diag) {
    if(func.inv == "cholesky") {
      varcov.sqrt <- try(chol(varcov))
      if(!is.numeric(varcov.sqrt)){
        if(try.another.decomposition) func.inv <- "eigen"
        else{print(varcov.sqrt[1]);stop()}
      }
      else{
        if(only.decomposition | inv)
          if(is.R()){remove("varcov")}
          else remove("varcov", frame=sys.nframe())          
        if(only.decomposition == FALSE) {
          if(det)
            cov.logdeth <- sum(log(diag(varcov.sqrt)))
          if(sqrt.inv)
            inverse.sqrt <- solve(varcov.sqrt)
          if(inv) {
            if(is.R()){
              invcov <- chol2inv(varcov.sqrt)
              if(!sqrt.inv) remove("varcov.sqrt")
              ##  gc(verbose=FALSE)
            }
            else{
              invcov.sqrt <- solve.upper(varcov.sqrt)
              invcov <- invcov.sqrt %*% t(invcov.sqrt)
              if(!sqrt.inv) remove("varcov.sqrt", frame=sys.nframe())
            }
          }
        }
      }
    }
    if(func.inv == "svd") {
      varcov.svd <- svd(varcov, nv = 0)
      cov.logdeth <- try(sum(log(sqrt(varcov.svd$d))))
      if(!is.numeric(cov.logdeth)){
        if(try.another.decomposition) func.inv <- "eigen"
        else{print(cov.logdeth[1]);stop()}
      }
      else{
        if(only.decomposition | inv)
          if(is.R()){remove("varcov")}
          else remove("varcov", frame=sys.nframe())          
        if(only.decomposition)
          varcov.sqrt <- t(varcov.svd$u %*% (t(varcov.svd$
                                               u) * sqrt(varcov.svd$d)))
        if(inv){
          invcov <- t(varcov.svd$u %*% (t(varcov.svd$
                                          u) * (1/varcov.svd$d)))
        }
        if(sqrt.inv)
          inverse.sqrt <- t(varcov.svd$u %*% (t(varcov.svd$
                                                u) * (1/sqrt(varcov.svd$d))))
      }
    }
    if(func.inv == "solve") {
      if(det)
        stop("the option func.inv == \"solve\" does not allow computation of determinants. \nUse func.inv = \"chol\",\"svd\" or \"eigen\"\n"
             )
      invcov <- try(solve(varcov))
      if(!is.numeric(cov.logdeth)){
        if(try.another.decomposition) func.inv <- "eigen"
        else{print(invcov[1]);stop()}
      }
      if(is.R()){remove("varcov")}
      else remove("varcov", frame=sys.nframe())          
    }
    if(func.inv == "eigen") {
      ##      if(!exists("varcov.eig"))
      varcov.eig <- eigen(varcov, symmetric = TRUE)
      cov.logdeth <- try(sum(log(sqrt(varcov.eig$val))))
      if(!is.numeric(cov.logdeth)){
        diag(varcov) <- 1.0001 * diag(varcov)
        varcov.eig <- eigen(varcov, symmetric = TRUE)
        cov.logdeth <- sum(log(sqrt(varcov.eig$val)))
      }
      else{
        if(only.decomposition | inv)
          if(is.R()){remove("varcov")}
          else remove("varcov", frame=sys.nframe())          
        if(only.decomposition)
          varcov.sqrt <- (varcov.eig$vec %*%
                          diag(sqrt(varcov.eig$val)) %*%
                          t(varcov.eig$vec))
        if(inv)
          invcov <- (varcov.eig$vec %*% diag(1/varcov.eig$val) %*%
                     t(varcov.eig$vec))
        if(sqrt.inv)
          inverse.sqrt <- (varcov.eig$vec %*% diag(1/sqrt(varcov.eig$val)) %*%
                           t(varcov.eig$vec))
      }
    }
  }
  ##
  ## Preparing output
  ##
  if(only.decomposition == FALSE) {
    if(det) {
      if(inv) {
	if(only.inv.lower.diag)
	  result <- list(lower.inverse = invcov[
                           lower.tri(invcov)], 
			 diag.inverse = diag(invcov),
			 log.det.to.half = cov.logdeth)
        else result <- list(inverse = invcov, 
                            log.det.to.half = cov.logdeth)
      }
      else {
	result <- list(varcov = varcov, log.det.to.half
		       = cov.logdeth)
      }
      if(sqrt.inv)
	result$sqrt.inverse <- inverse.sqrt
    }
    else{
      if(inv){
	if(only.inv.lower.diag)
	  result <- list(lower.inverse = invcov[lower.tri(invcov)],
			 diag.inverse = diag(invcov))
        else{
          if(sqrt.inv)
            result <- list(inverse=invcov, sqrt.inverse = inverse.sqrt)
          else
            result <- list(inverse=invcov)
        }
      }
      else
	result <- list(varcov=varcov)
    }
  }
  else result <- list(sqrt.varcov=varcov.sqrt)
  return(result)
}

"plot.geodata" <-
  function (geodata, coords = geodata$coords, data = geodata$data,
            trend = "cte", lambda=1, col.data = 1, window.new = FALSE,
            weights.divide = NULL, ...) 
{
  if(is.R()) par.ori <- par(no.readonly=TRUE)
  else par.ori <- par()
  on.exit(par(par.ori))
  coords <- as.matrix(geodata$coords)
  data <- as.matrix(data)
  data <- data[, col.data]
  if (window.new){
    if(is.R()) X11()
    else trellis.device()
  }
  if (lambda != 1){
    if(lambda == 0)
      data <- log(data)
    else
      data <- ((data^lambda)-1)/lambda
  }
  if(!is.null(weights.divide)){
    if(length(weights.divide) != length(data))
      stop("length of weights.divide must be equals to the length of data")
    data <- data/weights.divide
  }
  par(mfrow = c(2, 2))
  par(pty = "s")
  par(mar = c(4, 4, 0, .5))
  data.quantile <- quantile(data)
  plot(coords, xlab = "Coord X", ylab = "Coord Y", type = "n")
  if(length(unique(data.quantile)) == length(data.quantile)) {
    if(is.R()){
      points(coords[(data <= data.quantile[2]), ], cex = 0.4, col = "blue")
      points(coords[((data > data.quantile[2]) & (data <= data.quantile[3])), 
                    ], cex = 0.6, col = "green")
      points(coords[((data > data.quantile[3]) & (data <= data.quantile[4])), 
                    ], cex = 0.9, col = "yellow")
      points(coords[(data > data.quantile[4]), ], cex = 1.2, col = "red")
    }
    else{
      points(coords[(data <= data.quantile[2]),  ], pch = 1, cex = 
             0.59999999999999998, col = 2)
      points(coords[((data > data.quantile[2]) & (data <= 
                                                  data.quantile[3])),  ], pch = 2, cex = 
             1.3999999999999999, col = 4)
      points(coords[((data > data.quantile[3]) & (data <= 
                                                  data.quantile[4])),  ], pch = 3, cex = 1.7, col = 7)
      points(coords[(data > data.quantile[4]),  ], pch = 4, cex = 2,
             col = 8)
    }    
  }
  else {
    points(coords, pch = 16)
  }
  par(pty = "m")
  if(is.R()) par(mar = c(4, 4, 1, 1))
  else par(mar = c(0, 1, 0, 0.5))
###  plot(variog(coords=coords, data=data, trend=trend,...))
  if(is.R()){
    if(require(scatterplot3d) == FALSE){
      hist(data)
      warning("plot.geodata: a 3d plot will be draw instead of the histogram if the package \"scatterplot3d\" is available")
    }
    else
      scatterplot3d(x=coords[,1], y=coords[,2], z=data, box=F, type="h", pch=16, xlab = "Coord X", ylab = "Coord Y",  ...)
  }
  else xyzplot(coords = coords, data = data, ...)
  if(!is.R()) par(mar = c(5, 5, 1, 0.5))
  plot(coords[, 1], data, xlab = "Coord X", cex = 1)
  plot(coords[, 2], data, xlab = "Coord Y", cex = 1)
  return(invisible())
}

"likfit" <-
  function (geodata, coords=geodata$coords, data=geodata$data,
            trend = "cte", ini.cov.pars,
            fix.nugget = FALSE, nugget = 0, 
            fix.kappa = TRUE, kappa = 0.5, 
            fix.lambda = TRUE, lambda = 1, 
            fix.psiA = TRUE, psiA = 0, 
            fix.psiR = TRUE, psiR = 1, 
            cov.model = c("matern", "exponential", "gaussian",
              "spherical", "circular", "cubic", "wave",
              "powered.exponential", "cauchy", "gneiting",
              "gneiting.matern", "pure.nugget"),
            method = "ML",
            components = FALSE, nospatial = TRUE,
            limits = likfit.limits(), messages.screen = TRUE, ...) 
{
  ##
  ## Checking input
  ##
  if(is.R()) require(mva)
  call.fc <- match.call()
  temp.list <- list()
  ##
  cov.model <- match.arg(cov.model)
  temp.list$cov.model <- cov.model
  ##
  if (method == "REML" | method == "reml" | method == "rml") 
    method <- "RML"
  if(method == "ML" | method == "ml")
    method <- "ML"
  temp.list$method <- method
  if(is.matrix(ini.cov.pars) | is.data.frame(ini.cov.pars)){
    ini.cov.pars <- as.matrix(ini.cov.pars)
    if(nrow(ini.cov.pars) == 1)
      ini.cov.pars <- as.vector(ini.cov.pars)
    else{
      if((cov.model != "pure.nugget") & (ncol(ini.cov.pars) != 2))
        stop("\nini.cov.pars must be a matrix or data.frame with 2 components: \ninitial values for sigmasq and phi")
    }
  }
  if(is.vector(ini.cov.pars)){
    if((cov.model != "pure.nugget") & (length(ini.cov.pars) != 2))
      stop("\nini.cov.pars must be a vector with 2 components: \ninitial values for sigmasq and phi")
  }
  if(fix.kappa & !is.null(kappa))
    if(cov.model == "matern" & kappa == 0.5)
      cov.model <- "exponential"
  coords <- temp.list$coords <- as.matrix(coords)
  n <- temp.list$n <- length(data)
  if ((2*n) != length(coords))
    stop("\nnumber of locations does not match with number of data")
  temp.list$xmat <- trend.spatial(trend=trend, coords=coords)
  beta.size <- temp.list$beta.size <- dim(temp.list$xmat)[2]
  ##
  ## Checking for multiple initial values for preliminar search of   
  ## best initial value
  ##
  if(is.matrix(ini.cov.pars) | (length(nugget) > 1) | (length(kappa) > 1) | (length(lambda) > 1) | (length(psiR) > 1) | (length(psiA) > 1)){
    if(messages.screen)
      cat("likfit: searching for best initial value ...")
    .likGRF.dists.vec <<- as.vector(dist(coords))
    temp.list$z <- as.vector(data)
    temp.list$trend <- trend
    ini.temp <- matrix(ini.cov.pars, ncol=2)
    grid.ini <- as.matrix(expand.grid(sigmasq=unique(ini.temp[,1]), phi=unique(ini.temp[,2]), tausq=unique(nugget), kappa=unique(kappa), lambda=unique(lambda), psiR=unique(psiR), psiA=unique(psiA)))
    temp.f <- function(parms, temp.list){
      return(loglik.GRF(coords=temp.list$coords, data=temp.list$z, cov.model=temp.list$cov.model, cov.pars=parms[1:2], nugget=parms["tausq"], kappa=parms["kappa"], lambda=parms["lambda"], psiR=parms["psiR"], psiA=parms["psiA"], trend= temp.list$trend, method=temp.list$method, compute.dists=F))
    }
    grid.lik <- apply(grid.ini, 1, temp.f, temp.list=temp.list)
    ini.temp <- grid.ini[which(grid.lik == max(grid.lik)),]
    if(messages.screen){
      cat(" selected values:\n")
      print(round(as.vector(ini.temp), dig=3))
    }
    names(ini.temp) <- NULL
    ini.cov.pars <- ini.temp[1:2]
    nugget <- ini.temp[3]
    kappa <- ini.temp[4]
    lambda <- ini.temp[5]
    psiR <- ini.temp[6]
    psiA <- ini.temp[7]
    grid.ini <- NULL
    temp.list$trend <- NULL
    if(is.R()) {remove(".likGRF.dists.vec", pos=1); gc(verbose=FALSE)}    
    else remove(".likGRF.dists.vec", where=1)    
  }
  ##
  tausq <- nugget
  ##
  ## Box-Cox transformation for fixed lambda
  ##
  if(fix.lambda) {
    if(round(lambda, dig=4) == 1) {
      temp.list$log.jacobian <- 0
      temp.list$z <- as.vector(data)
    }
    else {
      if(any(data <= 0))
        stop("Transformation option not allowed when there are zeros or negative data")
      Jdata <- data^(lambda - 1)
      if(any(Jdata <= 0))
        temp.list$log.jacobian <- log(prod(Jdata))
      else temp.list$log.jacobian <- sum(log(Jdata))
      Jdata <- NULL
      if(is.R()) gc(verbose=FALSE)
      if(round(lambda, dig=4) == 0)
        temp.list$z <- log(data)
      else temp.list$z <- ((data^lambda) - 1)/lambda
    }
  }
  else{
    temp.list$z <- as.vector(data)
    temp.list$log.jacobian <- NULL
  }
  ##
  ## Coordinates transformation for fixed anisotropy parameters
  ##
  if(fix.psiR & fix.psiA){
    if(psiR != 1 | psiA != 0)
      coords <- coords.aniso(coords, aniso.pars=c(psiA, psiR))
    if(is.R()) assign(".likGRF.dists.vec", as.vector(dist(coords)), pos=1)
    else assign(".likGRF.dists.vec", as.vector(dist(coords)), where=1)
    range.dist <- range(.likGRF.dists.vec)
    max.dist <- max(range.dist)
    min.dist <- min(range.dist)
  }
  ##
  ##
  ##
  ini <- ini.cov.pars[2]
  ##  fixed.pars <- NULL
  lower.optim <- c(limits$phi["lower"])
  upper.optim <- c(limits$phi["upper"])
  fixed.values <- list()
  if(fix.nugget) {
    ##    fixed.pars <- c(fixed.pars, 0)
    fixed.values$tausq <- nugget
  }
  else {
    ini <- c(ini, nugget/ini.cov.pars[1])
    lower.optim <- c(lower.optim, limits$tausq.rel["lower"])
    upper.optim <- c(upper.optim, limits$tausq.rel["upper"])
  }
  if(fix.kappa){
##    fixed.kappa <- c(fixed.pars, kappa)
    fixed.values$kappa <- kappa
  }
  else {
    ini <- c(ini, kappa)
    lower.optim <- c(lower.optim, limits$kappa["lower"])
    upper.optim <- c(upper.optim, limits$kappa["upper"])
  }
  if(fix.lambda){
##    fixed.pars <- c(fixed.pars, lambda)
    fixed.values$lambda <- lambda
  }
  else {
    ini <- c(ini, lambda)
    lower.optim <- c(lower.optim, limits$lambda["lower"])
    upper.optim <- c(upper.optim, limits$lambda["upper"])
  }
  if(fix.psiR){
##    fixed.pars <- c(fixed.pars, psiR)
    fixed.values$psiR <- psiR
  }
  else {
    ini <- c(ini, psiR)
    lower.optim <- c(lower.optim, limits$psiR["lower"])
    upper.optim <- c(upper.optim, limits$psiR["upper"])
  }
  if(fix.psiA){
##    fixed.pars <- c(fixed.pars, psiA)
    fixed.values$psiA <- psiA
  }
  else {
    ini <- c(ini, psiA)
    lower.optim <- c(lower.optim, limits$psiA["lower"])
    upper.optim <- c(upper.optim, limits$psiA["upper"])
  }
  ## This must be here, after the previous ones:
  if(fix.nugget & nugget > 0){
    ## Warning: Inverting order here, ini will be now: c(phi,sigmasg)
    ini <- c(ini, ini.cov.pars[1])
    lower.optim <- c(lower.optim, limits$sigmasq["lower"])
    upper.optim <- c(upper.optim, limits$sigmasq["upper"])
##    fixed.pars <- c(fixed.pars, ini.cov.pars[1])
##    fixed.values$sigmasq <- 0
  }
  names(ini) <- NULL
  ##
  ip <- list(f.tausq = fix.nugget, f.kappa = fix.kappa, f.lambda = fix.lambda,
                  f.psiR = fix.psiR, f.psiA = fix.psiA)
  ##  
  if (messages.screen == TRUE) {
    cat("-----------------------------------------------------------------------\n")
    cat("likfit: Initialising likelihood maximisation using the function ")
    if(is.R()) cat("optim.\n") else cat("nlminb.\n")
    cat("likfit: Use control() to pass arguments for the maximisation function.")
    cat("\n        For more details see documentation for ")
    if(is.R()) cat("optim.\n") else cat("nlminb.\n")        
    cat("likfit: It is highly advisable to run this function several\n        times with different initial values for the parameters.\n")
    cat("likfit: WARNING: This step can be time demanding!\n")
    cat("-----------------------------------------------------------------------\n")
  }
  npars <- beta.size + 2 + sum(unlist(ip)==FALSE)
  if(is.R()){
    lik.minim <- optim(par = ini, fn = negloglik.GRF, method="L-BFGS-B",
                       lower=lower.optim, upper=upper.optim,
                       fp=fixed.values, ip=ip, temp.list = temp.list, ...)
  }
  else{
    lik.minim <- nlminb(ini, negloglik.GRF,
                        lower=lower.optim, upper=upper.optim,
                        fp=fixed.values, ip=ip, temp.list = temp.list, ...)
  }
  ##
  if(messages.screen == TRUE) 
    cat("likfit: end of numerical maximisation.\n")
  par.est <- lik.minim$par
  if(any(par.est < 0)) par.est <- round(par.est, dig=12)
  phi <- par.est[1]
  ##
  ## Values of the maximised likelihood
  ##
  if(is.R())
    value.min <- lik.minim$value
  else
    value.min <- lik.minim$objective
  if(method == "ML"){
    if(ip$f.tausq & (tausq > 0))
      loglik.max <-  (- value.min) + (n/2)*(-log(2*pi))
    else
      loglik.max <-  (- value.min) + (n/2)*(-log(2*pi) + log(n) -1)
  }
  if(method == "RML"){
    xx.eigen <- eigen(crossprod(temp.list$xmat), symmetric = TRUE, only.values = TRUE)
    if(ip$f.tausq & (tausq > 0))
      loglik.max <- (- value.min) - ((n-beta.size)/2)*(log(2*pi)) +
        0.5 * sum(log(xx.eigen$values))
    else
      loglik.max <- (- value.min) - ((n-beta.size)/2)*(log(2*pi)) +
        ((n-beta.size)/2)*(log(n-beta.size)) - ((n-beta.size)/2) +
          0.5 * sum(log(xx.eigen$values))
  }
  ##
  ## Assigning values for estimated parameters
  ##
  if(ip$f.tausq & ip$f.kappa & ip$f.lambda & ip$f.psiR & !ip$f.psiA){
    psiA <- par.est[2]
  }
  if(ip$f.tausq & ip$f.kappa & ip$f.lambda & !ip$f.psiR & ip$f.psiA){
    psiR <- par.est[2]
  }
  if(ip$f.tausq & ip$f.kappa & ip$f.lambda & !ip$f.psiR & !ip$f.psiA){
    psiR <- par.est[2]
    psiA <- par.est[3]
  }
  if(ip$f.tausq & ip$f.kappa & !ip$f.lambda & ip$f.psiR & ip$f.psiA){
    lambda  <- par.est[2]
  }
  if(ip$f.tausq & ip$f.kappa & !ip$f.lambda & ip$f.psiR & !ip$f.psiA){
    lambda  <- par.est[2]
    psiA <- par.est[3]
  }
  if(ip$f.tausq & ip$f.kappa & !ip$f.lambda & !ip$f.psiR & ip$f.psiA){
    lambda  <- par.est[2]
    psiR <- par.est[3]
  }
  if(ip$f.tausq & ip$f.kappa & !ip$f.lambda & !ip$f.psiR & !ip$f.psiA){
    lambda  <- par.est[2]
    psiR <- par.est[3]
    psiA <- par.est[4]
  }
  if(ip$f.tausq & !ip$f.kappa & ip$f.lambda & ip$f.psiR & ip$f.psiA){
    kappa  <-  par.est[2]
  }
  if(ip$f.tausq & !ip$f.kappa & ip$f.lambda & ip$f.psiR & !ip$f.psiA){
    kappa  <-  par.est[2]
    psiA <- par.est[3]
  }
  if(ip$f.tausq & !ip$f.kappa & ip$f.lambda & !ip$f.psiR & ip$f.psiA){
    kappa  <-  par.est[2]
    psiR <- par.est[3]
  }
  if(ip$f.tausq & !ip$f.kappa & ip$f.lambda & !ip$f.psiR & !ip$f.psiA){
    kappa  <-  par.est[2]
    psiR <- par.est[3]
    psiA <- par.est[4]
  }
  if(ip$f.tausq & !ip$f.kappa & !ip$f.lambda & ip$f.psiR & ip$f.psiA){
    kappa <-  par.est[2]
    lambda <- par.est[3]
  }
  if(ip$f.tausq & !ip$f.kappa & !ip$f.lambda & ip$f.psiR & !ip$f.psiA){
    kappa <-  par.est[2]
    lambda <- par.est[3]
    psiA <- par.est[4]
  }
  if(ip$f.tausq & !ip$f.kappa & !ip$f.lambda & !ip$f.psiR & ip$f.psiA){
    kappa <-  par.est[2]
    lambda <- par.est[3]
    psiR<- par.est[4]
  }
  if(ip$f.tausq & !ip$f.kappa & !ip$f.lambda & !ip$f.psiR & !ip$f.psiA){
    kappa <-  par.est[2]
    lambda <- par.est[3]
    psiR<- par.est[4]
    psiA<- par.est[5]
  }
  if(!ip$f.tausq & ip$f.kappa & ip$f.lambda & ip$f.psiR & ip$f.psiA){
    tausq <- par.est[2]
  }
  if(!ip$f.tausq & ip$f.kappa & ip$f.lambda & ip$f.psiR & !ip$f.psiA){
    tausq <- par.est[2]
    psiA<- par.est[3]
  }
  if(!ip$f.tausq & ip$f.kappa & ip$f.lambda & !ip$f.psiR & ip$f.psiA){
    tausq <- par.est[2]
    psiR<- par.est[3]
  }
  if(!ip$f.tausq & ip$f.kappa & ip$f.lambda & !ip$f.psiR & !ip$f.psiA){
    tausq <- par.est[2]
    psiR<- par.est[3]
    psiA<- par.est[4]
  }
  if(!ip$f.tausq & ip$f.kappa & !ip$f.lambda & ip$f.psiR & ip$f.psiA){
    tausq <- par.est[2]
    lambda <- par.est[3]
  }
  if(!ip$f.tausq & ip$f.kappa & !ip$f.lambda & ip$f.psiR & !ip$f.psiA){
    tausq <- par.est[2]
    lambda <- par.est[3]
    psiA <- par.est[4]
  }
  if(!ip$f.tausq & ip$f.kappa & !ip$f.lambda & !ip$f.psiR & ip$f.psiA){
    tausq <- par.est[2]
    lambda <- par.est[3]
    psiR <- par.est[4]
  }
  if(!ip$f.tausq & ip$f.kappa & !ip$f.lambda & !ip$f.psiR & !ip$f.psiA){
    tausq <- par.est[2]
    lambda <- par.est[3]
    psiR <- par.est[4]
    psiA <- par.est[5]
  }
  if(!ip$f.tausq & !ip$f.kappa & ip$f.lambda & ip$f.psiR & ip$f.psiA){
    tausq <- par.est[2]
    kappa <-  par.est[3]
  }
  if(!ip$f.tausq & !ip$f.kappa & ip$f.lambda & ip$f.psiR & !ip$f.psiA){
    tausq <- par.est[2]
    kappa <-  par.est[3]
    psiA <- par.est[4]
  }
  if(!ip$f.tausq & !ip$f.kappa & ip$f.lambda & !ip$f.psiR & ip$f.psiA){
    tausq <- par.est[2]
    kappa <-  par.est[3]
    psiR <- par.est[4]
  }
  if(!ip$f.tausq & !ip$f.kappa & ip$f.lambda & !ip$f.psiR & !ip$f.psiA){
    tausq <- par.est[2]
    kappa <-  par.est[3]
    psiR <- par.est[4]
    psiA <- par.est[5]
  }
  if(!ip$f.tausq & !ip$f.kappa & !ip$f.lambda & ip$f.psiR & ip$f.psiA){
    tausq <- par.est[2]
    kappa <-  par.est[3]
    lambda <- par.est[4]
  }
  if(!ip$f.tausq & !ip$f.kappa & !ip$f.lambda & ip$f.psiR & !ip$f.psiA){
    tausq <- par.est[2]
    kappa <-  par.est[3]
    lambda <- par.est[4]
    psiA <- par.est[5]
  }
  if(!ip$f.tausq & !ip$f.kappa & !ip$f.lambda & !ip$f.psiR & ip$f.psiA){
    tausq <- par.est[2]
    kappa <-  par.est[3]
    lambda <- par.est[4]
    psiR <- par.est[5]
  }
  if(!ip$f.tausq & !ip$f.kappa & !ip$f.lambda & !ip$f.psiR & !ip$f.psiA){
    tausq <- par.est[2]
    kappa <-  par.est[3]
    lambda <- par.est[4]
    psiR <- par.est[5]
    psiA <- par.est[6]
  }
  ##
  if(fix.nugget & nugget > 0){
    sigmasq <- par.est[length(par.est)]
    if(sigmasq > 1e-12) tausq <- nugget/sigmasq
    check.sigmasq <- TRUE
  }
  else check.sigmasq <- FALSE
  ##
  ##
  ## Transforming data acccording to the estimated lambda (Box-Cox) parameter
  ##
  if(!fix.lambda) {
    if(round(lambda, dig=4) == 1) {
      log.jacobian.max <- 0
      temp.list$z <- data
    }
    else {
      if(any(data^(lambda - 1) <= 0))
        log.jacobian.max <- log(prod(data^(lambda - 1)))
      else log.jacobian.max <- sum(log(data^(lambda - 1)))
      temp.list$z <- ((data^lambda)-1)/lambda
    }
  }
  else{
    log.jacobian.max <- temp.list$log.jacobian
  }
  ##
  ## Transforming coords for estimated anisotropy (if the case)
  ##
  if(fix.psiR & fix.psiA){
    if(is.R()) remove(".likGRF.dists.vec", pos=1)
    else{
      remove(".likGRF.dists.vec", where=1)
    }
  }
  else{
    if(psiR != 1 | psiA != 0)
      coords <- coords.aniso(coords, aniso.pars=c(psiA, psiR))
    .likGRF.dists.vec <- as.vector(dist(coords))
    range.dist <- range(.likGRF.dists.vec)
    max.dist <- max(range.dist)
    min.dist <- min(range.dist)
    remove(".likGRF.dists.vec")
  }      
  if(is.R()) gc(verbose=FALSE)
  ##
  ## Computing estimated beta and tausq/sigmasq (if the case)
  ##
  if((phi < 1e-12))
    siv <- diag(x=1/sqrt((1+tausq)), n)
  else{
    if(check.sigmasq){
      if(sigmasq < 1e-12){
        if(!fix.nugget)
          siv <- diag(x=1/sqrt((1+tausq)), n)
        else
          siv <- diag(x=1/sqrt((tausq)), n)          
      }
      else
        siv <- varcov.spatial(coords = coords, cov.model = cov.model,
                              kappa = kappa,
                              nugget = tausq, cov.pars = c(1, phi),
                              inv=TRUE, sqrt.inv = TRUE,
                              det = FALSE)$sqrt.inverse
    }
    else
      siv <- varcov.spatial(coords = coords, cov.model = cov.model,
                            kappa = kappa,
                            nugget = tausq, cov.pars = c(1, phi),
                            inv=TRUE, sqrt.inv = TRUE,
                            det = FALSE)$sqrt.inverse
  }
  sivx <- crossprod(siv, temp.list$xmat)
  xivx <- crossprod(sivx)
  sivy <- crossprod(siv, temp.list$z)
  xivy <- crossprod(sivx, sivy)
  betahat <- solve(xivx, xivy)
  res <- as.vector(temp.list$z - temp.list$xmat %*% betahat)
  if(!fix.nugget | (round(1e+12 * nugget) == 0)){
    res <- as.vector(temp.list$z - temp.list$xmat %*% betahat)
    ssres <- as.vector(crossprod(crossprod(siv,res)))
    if(method == "ML")
      sigmasq <- ssres/n
    else
      sigmasq <- ssres/(n - beta.size)
  }
  if(fix.nugget){
    if(nugget > 0)
      tausq <- nugget
  }
  else tausq <- tausq * sigmasq
  betahat.var <- solve(xivx)
  if(sigmasq > 1e-12) betahat.var <- sigmasq * betahat.var
#  if(!fix.nugget & phi < 1e-16){
#    tausq <- sigmasq + tausq
#    sigmasq <- 0
#  }
  n.model.pars <- beta.size+7
  par.su <- data.frame(status=rep(-9,n.model.pars))
  ind.par.su <- c(rep(0, beta.size), ip$f.tausq, 0, 0, ip$f.kappa,
                  ip$f.psiR, ip$f.psiA,ip$f.lambda)
  par.su$status <- ifelse(ind.par.su,"fixed", "estimated")
  par.su$values <- round(c(betahat, tausq, sigmasq, phi, kappa, psiR, psiA, lambda), dig=4)
  if(beta.size == 1) beta.name <- "beta"
  else beta.name <- paste("beta", 0:(beta.size-1), sep="")
  row.names(par.su) <- c(beta.name, "tausq", "sigmasq", "phi", "kappa",
                             "psiR", "psiA", "lambda")
  par.su <- par.su[c((1:(n.model.pars-3)), n.model.pars-1, n.model.pars-2, n.model.pars),] 
  ##
  ## Preparing output
  ##
  lik.results <- list(cov.model = cov.model,
                      nugget = tausq,
                      cov.pars=c(sigmasq, phi),
                      kappa = kappa,
                      beta = as.vector(betahat),
                      beta.var = betahat.var,
                      lambda = lambda,
                      aniso.pars = c(psiA = psiA, psiR = psiR),
                      method = method,
                      loglik = loglik.max,
                      npars = npars,
                      AIC = (loglik.max - npars),
                      BIC = (loglik.max - 0.5 * log(n) * npars),
                      parameters.summary = par.su,
                      info.minimisation.function = lik.minim,
                      max.dist = max.dist,
                      trend.matrix= temp.list$xmat,
                      transform.info = list(fix.lambda = fix.lambda,
                        log.jacobian = log.jacobian.max))
  ##
  ## Likelihood results for the model without spatial correlation
  ##
  if(nospatial){
    if(fix.lambda){
      beta.ns <- solve(crossprod(temp.list$xmat), crossprod(temp.list$xmat, temp.list$z))
      ss.ns <- sum((as.vector(temp.list$z - temp.list$xmat %*% beta.ns))^2)
      if(method == "ML"){
        nugget.ns <- ss.ns/n
        loglik.ns <- (n/2)*((-log(2*pi)) - log(nugget.ns) - 1) + temp.list$log.jacobian
      }
      if(method == "RML"){
        nugget.ns <- ss.ns/(n-beta.size)
        loglik.ns <- ((n-beta.size)/2)*((-log(2*pi)) - log(nugget.ns) -1) + temp.list$log.jacobian
      }
      npars.ns <- beta.size + 1 + fix.lambda
      lambda.ns <- lambda
    }
    else{
      bc.list <- list(n = n, beta.size = beta.size,
                      data = data, xmat = temp.list$xmat,
                      method = method)
      if(is.R())
        lik.lambda.ns <- optim(par=1, fn = boxcox.ns, method="L-BFGS-B",
                               lower=limits$lambda["lower"], upper=limits$lambda["upper"], bc.list=bc.list)
      else
        lik.lambda.ns <- nlminb(par=1, fn = boxcox.ns,
                                lower=limits$lambda["lower"], upper=limits$lambda["upper"], data=data,
                                bc.list = bc.list)
      bc.list <- NULL
      if(is.R()) gc(verbose=FALSE)
      lambda.ns <- lik.lambda.ns$par
      tdata.ns <- ((data^lambda.ns)-1)/lambda.ns
      beta.ns <- solve(crossprod(temp.list$xmat),crossprod(temp.list$xmat,tdata.ns))
      ss.ns <- sum((as.vector(tdata.ns - temp.list$xmat %*% beta.ns))^2)
      if(is.R())
        value.min.ns <- lik.lambda.ns$value
      else
        value.min.ns <- lik.lambda.ns$objective
      if(method == "ML"){
        loglik.ns <- (- value.min.ns)+ (n/2)*((-log(2*pi)) + log(n) - 1)
        nugget.ns <- ss.ns/n
      }
      if(method == "RML"){
        nugget.ns <- ss.ns/(n-beta.size)
        loglik.ns <- (- value.min.ns)+ ((n-beta.size)/2)*((-log(2*pi)) +
                                                          log(n-beta.size) - 1)
      }      
      npars.ns <- beta.size + 1 + fix.lambda
    }
    lik.results$nospatial <- list(beta.ns = beta.ns, variance.ns = nugget.ns,
                                  loglik.ns = loglik.ns, npars.ns = npars.ns,
                                  lambda.ns = lambda.ns)
  }
  ##
  ## Computing residuals and predicted values
  ## (isolated components of the model)
  ##
  if (components) {
    if(!fix.psiR & !fix.psiA)
      if(psiR != 1 | psiA != 0)
        coords <- coords.aniso(coords, aniso.pars=c(psiA, psiR))
    trend.comp <- temp.list$z - res
    invcov <- varcov.spatial(coords = coords, cov.model = cov.model, 
                             kappa = kappa, nugget = tausq,
                             cov.pars = c(sigmasq, phi), inv=TRUE)$inverse 
    covmat.signal <- varcov.spatial(coords = coords, cov.model = cov.model, 
                                    kappa = kappa, nugget = 0,
                                    cov.pars = c(sigmasq, phi))$varcov
    spatial.comp <- as.vector(covmat.signal %*% invcov %*% res)
    predict.comp <- trend.comp + spatial.comp
    residual.comp <- as.vector(temp.list$z - predict.comp)
    residual.std <- as.vector(invcov %*% residual.comp)
    residual.trend.std <- as.vector(invcov %*% res)
    s2.random <- (crossprod(res,invcov) %*% res)/(n - beta.size)
    s2 <- (crossprod(residual.comp,invcov) %*% residual.comp)/(n - beta.size)
  }
  if (length(lik.results$beta.var) == 1)
    lik.results$beta.var <- as.vector(lik.results$beta.var)
  if (length(lik.results$beta) > 1){
    if(inherits(trend, "formula"))
      beta.names <- c("intercept", paste("covar", 1:(ncol(temp.list$xmat)-1), sep = ""))
    else
      if (trend == "1st")
        beta.names <- c("intercept", "x", "y")
      else
        if (trend == "2nd")
          beta.names <- c("intercept", "x", "y", "x2", "xy", "y2")
    names(lik.results$beta) <- beta.names
  }
  if (components) {
    lik.results$model.components <- data.frame(trend = trend.comp, spatial = spatial.comp, residuals = residual.comp)
    lik.results$s2 <- s2
    lik.results$s2.random <- s2.random
  }
  lik.results$call <- call.fc
  ##
  ## Assigning classes
  ##
  if(is.R())
    class(lik.results) <- c("likGRF", "variomodel")
  else{
    if(version$major <= 4)
      class(lik.results) <- c("likGRF", "variomodel")
    else oldClass(lik.results) <- c("likGRF", "variomodel")
  }
  ##
  ## Some warning messages about particular possible results
  ##
  if(messages.screen){
    if((lik.results$cov.pars[1] < (0.01 * (lik.results$nugget + lik.results$cov.pars[1])))& lik.results$cov.pars[2] > 0)
      cat("\nWARNING: estimated sill is less than 1 hundredth of the total variance. Consider re-examine the model excluding spatial dependence\n" )      
    if((lik.results$cov.pars[2] > (10 * max.dist)) & lik.results$cov.pars[1] > 0 )
      cat("\nWARNING: estimated range is more than 10 times bigger than the biggest distance between two points. Consider re-examine the model:\n 1) excluding spatial dependence if estimated sill is too low and/or \n 2) taking trends (covariates) into account\n" ) 
    if(((lik.results$cov.pars[2] < (0.1 * min.dist)) & (lik.results$cov.pars[1] > 0)) & lik.results$cov.pars[2] > 0)
      cat("\nWARNING: estimated range is less than 1 tenth of the minimum distance between two points. Consider re-examine the model excluding spatial dependence\n" ) 

  }
  ##
  return(lik.results)
}

"negloglik.GRF" <-
  function(pars, fp, ip, temp.list)
### pars : values for the parameters to be estimated
  ## sequence is c(phi, tausq, kappa, lambda, psiR, psiA, sigmasq)
### fixed pars: parameters considered fixed
### ind.pars : list indicating which are fixed and which are to be estimated
  ##
  ## Warning:
  ##  if fix.nugget = TRUE and nugget > 0 ,
  ## sigmasq should be passed and fp$nugget is the value of the nugget
  ## otherwise the RELATIVE nugget should be passed
{
  n <- temp.list$n
  p <- temp.list$beta.size
  log.jacobian <- temp.list$log.jacobian
  ## Obligatory parameter:
  phi <- pars[1]
  ## Others
  if(ip$f.tausq){
    if(fp$tausq > 0){
      npars.min <- length(pars)
      sigmasq <- pars[npars.min]
    }
    else sigmasq <- 1
  }
  else sigmasq <- 1
  if(ip$f.tausq & ip$f.kappa & ip$f.lambda & ip$f.psiR & ip$f.psiA){
    tausq <- fp$tausq
    kappa <- fp$kappa
    lambda <- fp$lambda
    psiR <- fp$psiR
    psiA <- fp$psiA
  }
  if(ip$f.tausq & ip$f.kappa & ip$f.lambda & ip$f.psiR & !ip$f.psiA){
    tausq <- fp$tausq
    kappa <- fp$kappa
    lambda <- fp$lambda
    psiR <- fp$psiR
    psiA <- pars[2]
  }
  if(ip$f.tausq & ip$f.kappa & ip$f.lambda & !ip$f.psiR & ip$f.psiA){
    tausq <- fp$tausq
    kappa <- fp$kappa
    lambda <- fp$lambda
    psiR <- pars[2]
    psiA <- fp$psiA
  }
  if(ip$f.tausq & ip$f.kappa & ip$f.lambda & !ip$f.psiR & !ip$f.psiA){
    tausq <- fp$tausq
    kappa <- fp$kappa
    lambda <- fp$lambda
    psiR <- pars[2]
    psiA <- pars[3]
  }
  if(ip$f.tausq & ip$f.kappa & !ip$f.lambda & ip$f.psiR & ip$f.psiA){
    tausq <- fp$tausq
    kappa <- fp$kappa
    lambda <- pars[2]
    psiR <- fp$psiR
    psiA <- fp$psiA
  }
  if(ip$f.tausq & ip$f.kappa & !ip$f.lambda & ip$f.psiR & !ip$f.psiA){
    tausq <- fp$tausq
    kappa <- fp$kappa
    lambda <- pars[2]
    psiR <- fp$psiR
    psiA <- pars[3]
  }
  if(ip$f.tausq & ip$f.kappa & !ip$f.lambda & !ip$f.psiR & ip$f.psiA){
    tausq <- fp$tausq
    kappa <- fp$kappa
    lambda <- pars[2]
    psiR <- pars[3]
    psiA <- fp$psiA
  }
  if(ip$f.tausq & ip$f.kappa & !ip$f.lambda & !ip$f.psiR & !ip$f.psiA){
    tausq <- fp$tausq
    kappa <- fp$kappa
    lambda <- pars[2]
    psiR <- pars[3]
    psiA <- pars[4]
  }
  if(ip$f.tausq & !ip$f.kappa & ip$f.lambda & ip$f.psiR & ip$f.psiA){
    tausq <- fp$tausq
    kappa <- pars[2]
    lambda <- fp$lambda
    psiR <- fp$psiR
    psiA <- fp$psiA
  }
  if(ip$f.tausq & !ip$f.kappa & ip$f.lambda & ip$f.psiR & !ip$f.psiA){
    tausq <- fp$tausq
    kappa <- pars[2]
    lambda <- fp$lambda
    psiR <- fp$psiR
    psiA <- pars[3]
  }
  if(ip$f.tausq & !ip$f.kappa & ip$f.lambda & !ip$f.psiR & ip$f.psiA){
    tausq <- fp$tausq
    kappa <- pars[2]
    lambda <- fp$lambda
    psiR <- pars[3]
    psiA <- fp$psiA
  }
  if(ip$f.tausq & !ip$f.kappa & ip$f.lambda & !ip$f.psiR & !ip$f.psiA){
    tausq <- fp$tausq
    kappa <- pars[2]
    lambda <- fp$lambda
    psiR <- pars[3]
    psiA <- pars[4]
  }
  if(ip$f.tausq & !ip$f.kappa & !ip$f.lambda & ip$f.psiR & ip$f.psiA){
    tausq <- fp$tausq
    kappa <- pars[2]
    lambda <- pars[3]
    psiR <- fp$psiR
    psiA <- fp$psiA
  }
  if(ip$f.tausq & !ip$f.kappa & !ip$f.lambda & ip$f.psiR & !ip$f.psiA){
    tausq <- fp$tausq
    kappa <- pars[2]
    lambda <- pars[3]
    psiR <- fp$psiR
    psiA <- pars[4]
  }
  if(ip$f.tausq & !ip$f.kappa & !ip$f.lambda & !ip$f.psiR & ip$f.psiA){
    tausq <- fp$tausq
    kappa <- pars[2]
    lambda <- pars[3]
    psiR <- pars[4]
    psiA <- fp$psiA
  }
  if(ip$f.tausq & !ip$f.kappa & !ip$f.lambda & !ip$f.psiR & !ip$f.psiA){
    tausq <- fp$tausq
    kappa <- pars[2]
    lambda <- pars[3]
    psiR <- pars[4]
    psiA <- pars[5]
  }
  if(!ip$f.tausq & ip$f.kappa & ip$f.lambda & ip$f.psiR & ip$f.psiA){
    tausq <- pars[2]
    kappa <- fp$kappa
    lambda <- fp$lambda
    psiR <- fp$psiR
    psiA <- fp$psiA
  }
  if(!ip$f.tausq & ip$f.kappa & ip$f.lambda & ip$f.psiR & !ip$f.psiA){
    tausq <- pars[2]
    kappa <- fp$kappa
    lambda <- fp$lambda
    psiR <- fp$psiR
    psiA <- pars[3]
  }
  if(!ip$f.tausq & ip$f.kappa & ip$f.lambda & !ip$f.psiR & ip$f.psiA){
    tausq <- pars[2]
    kappa <- fp$kappa
    lambda <- fp$lambda
    psiR <- pars[3]
    psiA <- fp$psiA
  }
  if(!ip$f.tausq & ip$f.kappa & ip$f.lambda & !ip$f.psiR & !ip$f.psiA){
    tausq <- pars[2]
    kappa <- fp$kappa
    lambda <- fp$lambda
    psiR <- pars[3]
    psiA <- pars[4]
  }
  if(!ip$f.tausq & ip$f.kappa & !ip$f.lambda & ip$f.psiR & ip$f.psiA){
    tausq <- pars[2]
    kappa <- fp$kappa
    lambda <- pars[3]
    psiR <- fp$psiR
    psiA <- fp$psiA
  }
  if(!ip$f.tausq & ip$f.kappa & !ip$f.lambda & ip$f.psiR & !ip$f.psiA){
    tausq <- pars[2]
    kappa <- fp$kappa
    lambda <- pars[3]
    psiR <- fp$psiR
    psiA <- pars[4]
  }
  if(!ip$f.tausq & ip$f.kappa & !ip$f.lambda & !ip$f.psiR & ip$f.psiA){
    tausq <- pars[2]
    kappa <- fp$kappa
    lambda <- pars[3]
    psiR <- pars[4]
    psiA <- fp$psiA
  }
  if(!ip$f.tausq & ip$f.kappa & !ip$f.lambda & !ip$f.psiR & !ip$f.psiA){
    tausq <- pars[2]
    kappa <- fp$kappa
    lambda <- pars[3]
    psiR <- pars[4]
    psiA <- pars[5]
  }
  if(!ip$f.tausq & !ip$f.kappa & ip$f.lambda & ip$f.psiR & ip$f.psiA){
    tausq <- pars[2]
    kappa <- pars[3]
    lambda <- fp$lambda
    psiR <- fp$psiR
    psiA <- fp$psiA
  }
  if(!ip$f.tausq & !ip$f.kappa & ip$f.lambda & ip$f.psiR & !ip$f.psiA){
    tausq <- pars[2]
    kappa <- pars[3]
    lambda <- fp$lambda
    psiR <- fp$psiR
    psiA <- pars[4]
  }
  if(!ip$f.tausq & !ip$f.kappa & ip$f.lambda & !ip$f.psiR & ip$f.psiA){
    tausq <- pars[2]
    kappa <- pars[3]
    lambda <- fp$lambda
    psiR <- pars[4]
    psiA <- fp$psiA
  }
  if(!ip$f.tausq & !ip$f.kappa & ip$f.lambda & !ip$f.psiR & !ip$f.psiA){
    tausq <- pars[2]
    kappa <- pars[3]
    lambda <- fp$lambda
    psiR <- pars[4]
    psiA <- pars[5]
  }
  if(!ip$f.tausq & !ip$f.kappa & !ip$f.lambda & ip$f.psiR & ip$f.psiA){
    tausq <- pars[2]
    kappa <- pars[3]
    lambda <- pars[4]
    psiR <- fp$psiR
    psiA <- fp$psiA
  }
  if(!ip$f.tausq & !ip$f.kappa & !ip$f.lambda & ip$f.psiR & !ip$f.psiA){
    tausq <- pars[2]
    kappa <- pars[3]
    lambda <- pars[4]
    psiR <- fp$psiR
    psiA <- pars[5]
  }
  if(!ip$f.tausq & !ip$f.kappa & !ip$f.lambda & !ip$f.psiR & ip$f.psiA){
    tausq <- pars[2]
    kappa <- pars[3]
    lambda <- pars[4]
    psiR <- pars[5]
    psiA <- fp$psiA
  }
  if(!ip$f.tausq & !ip$f.kappa & !ip$f.lambda & !ip$f.psiR & !ip$f.psiA){
    tausq <- pars[2]
    kappa <- pars[3]
    lambda <- pars[4]
    psiR <- pars[5]
    psiA <- pars[6]
  }
  ##
  ## Absurd values
  ##
  if(kappa < 1e-04) return(1e+32)
  if(round(1e+16*(tausq+sigmasq)) == 0) return(1e+32)
  ##
  ## Anisotropy
  ##
  if(!ip$f.psiR | !ip$f.psiA){
    coords.c <- coords.aniso(temp.list$coords, aniso.pars=c(psiA, psiR))
    if(is.R()) assign(".likGRF.dists.vec", as.vector(dist(coords.c)), pos=1)
    else assign(".likGRF.dists.vec", as.vector(dist(coords.c)), where=1)
  }
  ##
  ## Box-Cox transformation
  ##
  if(!ip$f.lambda){
    if(round(lambda, dig=4) == 1) {
      log.jacobian <- 0
    }
    else {
      if(any(temp.list$z <= 0))
        stop("Transformation not allowed for zero or negative data")
      data <- temp.list$z^(lambda - 1)
      if(any(data <= 0)) log.jacobian <- log(prod(data))
      else log.jacobian <- sum(log(data))
      data <- NULL
    }
    if(round(lambda, dig=4) == 0)
      data <- log(temp.list$z)
    else data <- ((temp.list$z^lambda) - 1)/lambda
  }
  else data <- temp.list$z
  ##
  ## Computing likelihood
  ##
  ## NOTE: Likelihood for Independent observations 
  ##       arbitrary criteria used here:
  ##       (phi < 1-e16) or (sigmasq < 1-e16)  ==> independence
  ##
  if((phi < 1e-16) | (sigmasq < 1e-16)){
    if(ip$f.tausq)
      iv <- list(sqrt.inverse = diag(x=1/sqrt((tausq+sigmasq)), n),
                 log.det.to.half = (n/2) * log(tausq+sigmasq))
    else
      iv <- list(sqrt.inverse = diag(x=1/sqrt((1+tausq)), n),
                 log.det.to.half = (n/2) * log(1+tausq))
  }
  else{
    iv <- varcov.spatial(dists.lowertri = .likGRF.dists.vec,
                         cov.model = temp.list$cov.model, kappa=kappa,
                         nugget = tausq, cov.pars=c(sigmasq, phi),
                         sqrt.inv = TRUE, det = TRUE)
  }
  sivx <- crossprod(iv$sqrt.inverse, temp.list$xmat)
  xivx <- crossprod(sivx)
  sivy <- crossprod(iv$sqrt.inverse, data)
  xivy <- crossprod(sivx, sivy)  
  betahat <- solve(xivx, xivy)
  res <- data - temp.list$xmat %*% betahat
  ssres <- as.vector(crossprod(crossprod(iv$sqrt.inverse,res)))
  if(temp.list$method == "ML"){
    if(ip$f.tausq & (tausq > 0))
      negloglik <- iv$log.det.to.half +  0.5 * ssres - log.jacobian
    else
      negloglik <- (n/2) * log(ssres) +  iv$log.det.to.half - log.jacobian
  }
  if(temp.list$method == "RML"){
    if(length(as.vector(xivx)) == 1) {
      choldet <- 0.5 * log(xivx)
    }
    else {
      chol.xivx <- chol(xivx)
      choldet <- sum(log(diag(chol.xivx)))
    }
    if(ip$f.tausq & (tausq > 0))
      negloglik <- iv$log.det.to.half +  0.5 * ssres + choldet - log.jacobian
    else
      negloglik <- ((n-p)/2) * log(ssres) +  iv$log.det.to.half +
        choldet - log.jacobian
  }
  if(negloglik > 1e+32) negloglik <- 1e32
  return(negloglik)
}

"boxcox.ns" <- function(lambda, bc.list)
{
  data <- bc.list$data
  n <- bc.list$n
  ##
  if(round(lambda, dig=4) == 1) {
    log.jacobian <- 0
    y <- data
  }
  else {
    if(any(data <= 0))
      stop("Transformation option not allowed when there are zeros or negative data")
    Jdata <- data^(lambda - 1)
    if(any(Jdata <= 0))
      log.jacobian <- log(prod(Jdata))
    else log.jacobian <- sum(log(Jdata))
    Jdata <- NULL
    if(is.R()) gc(verbose=FALSE)
    if(round(lambda, dig=4) == 0)
      y <- log(data)
    else y <- ((data^lambda) - 1)/lambda
  }
  beta.ns <- solve(crossprod(bc.list$xmat), crossprod(bc.list$xmat, y))
  ss.ns <- sum((as.vector(y) - as.vector(bc.list$xmat %*% beta.ns))^2)
  if(bc.list$method == "ML")
    neglik <- (n/2) * log(ss.ns) - log.jacobian
  if(bc.list$method == "RML")
    neglik <- ((n-bc.list$beta.size)/2) * log(ss.ns) - log.jacobian
  ##
  return(as.vector(neglik))
}

"likfit.limits" <-
  function(phi = c(lower=0, upper=+Inf),
           sigmasq = c(lower=0, upper=+Inf),
           tausq.rel = c(lower=0, upper=+Inf),
           kappa = c(lower=0, upper=+Inf),
           lambda = c(lower=-3, upper=3),
           psiR = c(lower=1, upper=+Inf),
           psiA = c(lower=0, upper=2*pi)
           )
{
  if(length(phi) != 2)
    stop("phi must be a 2 components vector with lower and upper limits for the parameter phi") 
  if(length(sigmasq) != 2)
    stop("phi must be a 2 components vector with lower and upper limits for the parameter phi") 
  if(length(tausq.rel) != 2)
    stop("phi must be a 2 components vector with lower and upper limits for the parameter phi") 
  if(length(kappa) != 2)
    stop("phi must be a 2 components vector with lower and upper limits for the parameter phi") 
  if(length(lambda) != 2)
    stop("phi must be a 2 components vector with lower and upper limits for the parameter phi")
  if(length(psiR) != 2)
    stop("phi must be a 2 components vector with lower and upper limits for the parameter phi") 
  if(length(psiA) != 2)
    stop("phi must be a 2 components vector with lower and upper limits for the parameter phi") 
  if(phi[1] >= phi[2])
    stop("parameter phi: lower limit greater or equal upper limit")
  if(sigmasq[1] >= sigmasq[2])
    stop("parameter sigmasq: lower limit greater or equal upper limit")
  if(tausq.rel[1] >= tausq.rel[2])
    stop("parameter tausq.rel: lower limit greater or equal upper limit")
  if(kappa[1] >= kappa[2])
    stop("parameter kappa: lower limit greater or equal upper limit")
  if(lambda[1] >= lambda[2])
    stop("parameter lambda: lower limit greater or equal upper limit")
  if(psiR[1] >= psiR[2])
    stop("parameter psiR: lower limit greater or equal upper limit")
  if(psiA[1] >= psiA[2])
    stop("parameter psiA: lower limit greater or equal upper limit")
  names(phi) <- c("lower", "upper")
  names(sigmasq) <- c("lower", "upper")
  names(tausq.rel) <- c("lower", "upper")
  names(kappa) <- c("lower", "upper")
  names(lambda) <- c("lower", "upper")
  names(psiR) <- c("lower", "upper")
  names(psiA) <- c("lower", "upper")
  return(list(phi = phi, sigmasq = sigmasq,
              tausq.rel = tausq.rel, kappa = kappa,
              lambda = lambda, psiR = psiR, psiA = psiA))
}

"print.likGRF" <-
  function(obj, digits = "default", ...)
{
  if(is.R() & digits == "default") digits <- max(3, getOption("digits") - 3)
  else digits <- options()$digits
  est.pars <- as.vector(obj$parameters.summary[obj$parameters.summary[,1] == "estimated",2])
  names.est.pars <- dimnames(obj$parameters.summary[obj$parameters.summary[,1] == "estimated",])[[1]]
  names(est.pars) <- names.est.pars
  cat("likfit: estimated model parameters:\n")
  print(round(est.pars, digits=digits))
  cat("\nlikfit: maximised log-likelihood = ")
  cat(round(obj$loglik, digits=digits))
  cat("\n")
  return(invisible())
}  

"summary.likGRF" <-
  function(obj, ...)
{
  names.pars <- dimnames(obj$parameters.summary)[[1]]
  summ.lik <- list()
  if(obj$method == "ML")
    summ.lik$method <- "maximum likelihood"
  if(obj$method == "RML")
    summ.lik$method <- "restricted maximum likelihood"
  summ.lik$mean.component <- obj$beta
  names(summ.lik$mean.component) <- names.pars[1:length(obj$beta)]
  summ.lik$cov.model <- obj$cov.model
  summ.lik$spatial.component <- obj$parameters.summary[c("sigmasq", "phi"),]
  summ.lik$spatial.component.extra <- obj$parameters.summary[c("kappa", "psiA", "psiR"),]
  summ.lik$nugget.component <- obj$parameters.summary[c("tausq"),, drop=FALSE]
  summ.lik$transformation  <- obj$parameters.summary[c("lambda"),, drop=FALSE]
  summ.lik$likelihood <- c(log.L = obj$loglik, n.params = as.integer(obj$npars),
                               AIC = obj$AIC, BIC = obj$BIC)
  summ.lik$estimated.pars <- dimnames(obj$parameters.summary[obj$parameters.summary[,1] == "estimated",])[[1]]
  likelihood.info <- c(log.L = obj$loglik, n.params = as.integer(obj$npars),
                       AIC = obj$AIC, BIC = obj$BIC)
  summ.lik$call <- obj$call
  class(summ.lik) <- "summary.likGRF"
  return(summ.lik)
}

"print.summary.likGRF" <-
  function(obj, digits = "default", ...)
{
  if(is.R() & digits == "default") digits <- max(3, getOption("digits") - 3)
  else digits <- options()$digits
  cat("Summary of the parameter estimation\n")
  cat("-----------------------------------\n")
  cat(paste("Estimation method:", obj$method, "\n"))
  cat("\n")
  ##
  ## Estimates of the model components
  ## Model: Y(x) = X\beta + S(x) + e 
  ##
  cat("Parameters of the mean component (trend):")
  cat("\n")
  print(round(obj$mean.component, digits=digits))
  cat("\n")
  ##
  cat("Parameters of the spatial component:")
  cat("\n")
  cat(paste("   correlation function:", obj$cov.model))
  cat(paste("\n      (estimated) variance parameter sigmasq (partial sill) = ", round(obj$spatial.component[1,2], dig=digits)))
  cat(paste("\n      (estimated) cor. fct. parameter phi (range parameter)  = ", round(obj$spatial.component[2,2], dig=digits)))
  if(obj$cov.model == "matern" | obj$cov.model == "powered.exponential" |
     obj$cov.model == "cauchy" | obj$cov.model == "gneiting.matern"){
    kappa <- obj$spatial.component.extra["kappa",2]
    if(obj$spatial.component.extra["kappa",1] == "estimated")
      cat(paste("\n      (estimated) extra parameter kappa =", round(kappa, digits=digits)))
    else{
      cat(paste("\n      (fixed) extra parameter kappa = ", kappa))
      if(obj$cov.model == "matern" & round((1e12 *kappa)  == 0.5))
      cat(" (exponential)")
    }
  }
  cat("\n")
  ##
  aniso <-  obj$spatial.component.extra[c("psiA", "psiR"),]
  psiApsiR <- obj$spatial.component.extra[c("psiA", "psiR"),2]
  cat("   anisotropy parameters:")
  if(aniso["psiA",1] == "estimated")
    cat(paste("\n      (estimated) anisotropy angle =",
              round(psiApsiR[1], digits=digits),
              " (",round((psiApsiR[1]*360)/(2*pi), dig=1), "degrees )"))
  else
    cat(paste("\n      (fixed) anisotropy angle =", psiApsiR[1],
              " (",(psiApsiR[1]*360)/(2*pi), "degrees )"))
  if(aniso["psiR",1] == "estimated")
    cat(paste("\n      (estimated) anisotropy ratio =",
              round(psiApsiR[2], digits=digits)))
  else
    cat(paste("\n      (fixed) anisotropy ratio =", psiApsiR[2]))
  cat("\n")
  cat("\n")  
  cat("Parameter of the error component:")
  if(obj$nugget.component[,1] == "estimated")
    cat(paste("\n      (estimated) nugget = ", round(obj$nugget.component[,2], dig=digits)))
  else
    cat(paste("\n      (fixed) nugget =", obj$nugget.component[,2]))
  cat("\n")
  cat("\n")
  cat("Transformation parameter:")
  cat("\n")
  lambda <- obj$transformation[,2]
  if(obj$transformation[,1] == "estimated")
    cat(paste("      (estimated) Box-Cox parameter =", round(lambda, dig=digits)))
  else{
    cat(paste("      (fixed) Box-Cox parameter =", lambda))
    if(lambda == 1) cat(" (no transformation)")
    if(lambda == 0) cat(" (log-transformation)")
  }
  cat("\n")
  cat("\n")
  cat("Maximised Likelihood:")
  cat("\n")
  print(round(obj$likelihood, digits=digits))
  cat("\n")
  cat("Call:")
  cat("\n")
  print(obj$call)
  cat("\n")
  invisible(obj)
}

"loglik.GRF" <-
  function(geodata, coords=geodata$coords, data=geodata$data, cov.model="exp", cov.pars, nugget=0, kappa=0.5, lambda=1, psiR=1, psiA=0, trend="cte", method="ML", compute.dists=T)
{
  if (method == "REML" | method == "reml" | method == "rml") 
    method <- "RML"
  if(method == "ML" | method == "ml")
    method <- "ML"
  n <- nrow(coords)
  xmat <- trend.spatial(trend=trend, coords=coords)
  beta.size <- ncol(xmat)
  z <- data
  sigmasq <- cov.pars[1]
  phi <- cov.pars[2]
  ##
  ## Absurd values
  ##
  if(kappa < 1e-04) return(1e+32)
  if(round(1e+16*(nugget+sigmasq)) == 0) return(1e+32)
  ##
  ## Anisotropy
  ##
  if(psiR != 1 | psiA != 0){
    coords.c <- coords.aniso(coords, aniso.pars=c(psiA, psiR))
    .likGRF.dists.vec <-  as.vector(dist(coords.c))
  }
  else
    if(compute.dists) .likGRF.dists.vec <-  as.vector(dist(coords))
  ##
  ## Box-Cox transformation
  ##
  if(round(lambda, dig=4) == 1) {
    log.jacobian <- 0
  }
  else {
    if(any(z <= 0))
      stop("Transformation not allowed for zero or negative data")
    data <- z^(lambda - 1)
    if(any(data <= 0)) log.jacobian <- log(prod(data))
    else log.jacobian <- sum(log(data))
    data <- NULL
    if(round(lambda, dig=4) == 0)
      data <- log(z)
    else data <- ((z^lambda) - 1)/lambda
  }
  ##
  ## Computing likelihood
  ##
  ## NOTE: Likelihood for Independent observations 
  ##       arbitrary criteria used here:
  ##       (phi < 1-e16) or (sigmasq < 1-e16)  ==> independence
  ##
  if((phi < 1e-16) | (sigmasq < 1e-16)){
    iv <- list(sqrt.inverse = diag(x=1/sqrt((nugget+sigmasq)), n),
               log.det.to.half = (n/2) * log(nugget+sigmasq))
  }
  else{
    iv <- varcov.spatial(dists.lowertri = .likGRF.dists.vec,
                         cov.model = cov.model, kappa=kappa,
                         nugget = nugget, cov.pars=c(sigmasq, phi),
                         sqrt.inv = TRUE, det = TRUE)
  }
  sivx <- crossprod(iv$sqrt.inverse, xmat)
  xivx <- crossprod(sivx)
  sivy <- crossprod(iv$sqrt.inverse, data)
  xivy <- crossprod(sivx, sivy)  
  betahat <- solve(xivx, xivy)
  res <- data - xmat %*% betahat
  ssres <- as.vector(crossprod(crossprod(iv$sqrt.inverse,res)))
  if(method == "ML"){
    negloglik <- (n/2)*(log(2*pi)) + iv$log.det.to.half +  0.5 * ssres - log.jacobian
  }
  if(method == "RML"){
    if(length(as.vector(xivx)) == 1) {
      choldet <- 0.5 * log(xivx)
    }
    else {
      chol.xivx <- chol(xivx)
      choldet <- sum(log(diag(chol.xivx)))
    }
    negloglik <- iv$log.det.to.half +  0.5 * ssres + choldet - log.jacobian
    xx.eigen <- eigen(crossprod(xmat), symmetric = TRUE, only.values = TRUE)
    negloglik <- negloglik + ((n-beta.size)/2)*(log(2*pi)) - 0.5 * sum(log(xx.eigen$values))
  }
  if(negloglik > 1e+32) negloglik <- 1e32
  return(-negloglik)
}



"maijun" <-
structure(list(coords = structure(c(402.95294, 501.70493, 556.32618, 
573.40431, 702.42284, 668.54416, 435.8477, 434.01245, 432.46225, 
162.14585, 169.47663, 717.84675, 713.34332, 726.66302, 618.46885, 
661.88082, 506.71667, 561.7874, 585.63949, 531.71056, 308.97183, 
306.10469, 150.12196, 155.3399, 620.00181, 475.05766, 366.87187, 
479.49527, 615.72353, 599.07288, 566.56758, 484.6613, 445.54377, 
498.29528, 477.94465, 353.18336, 338.52463, 267.23204, 676.87847, 
538.88998, 340.96468, 203.93695, 254.26247, 719.46243, 768.50871, 
721.65469, 687.75359, 599.06043, 541.89327, 449.6865, 357.51524, 
222.33953, 297.32299, 184.17268, 401.98944, 365.36206, 331.68065, 
294.95009, 674.31593, 518.6908, 562.79775, 399.91232, 309.69511, 
308.21592, 397.59888, 206.29791, 289.38003, 246.80304, 279.55302, 
231.82964, 245.59459, 410.80902, 400.01532, 405.11185, 422.47533, 
457.73378, 388.47179, 299.07823, 320.4695, 317.7274, 328.1532, 
341.96468, 346.05942, 371.75809, 377.09969, 395.46678, 244.79297, 
232.08612, 212.36295, 250.92841, 249.11474, 262.98869, 264.83559, 
277.18097, 267.48289, 287.42489, 291.21859, 296.594, 295.11083, 
211.66139, 189.31521, 165.97002, 164.28177, 184.98705, 711.89977, 
749.59433, 702.73299, 650.76351, 640.73071, 600.65702, 421.28296, 
451.44393, 419.33262, 377.71707, 357.49565, 372.27262, 287.35834, 
285.6835, 224.32153, 218.99256, 262.66451, 217.24028, 212.63015, 
249.35852, 245.90862, 251.23978, 197.7439, 172.85298, 187.83331, 
568.34196, 541.68923, 493.34111, 471.68756, 445.03235, 479.99208, 
333.22941, 400.32814, 343.6607, 320.24316, 331.8482, 235.32685, 
226.53333, 692.54456, 164.52841, 428.771, 445.27065, 447.04177, 
272.2959, 261.6707, 286.54044, 317.90506, 288.37001, 286.29984, 
334.53158, 181.56741, 214.88336, 207.27952, 127.43265, 128.81455, 
212.86691, 138.89773, 212.59809, 131.62153, 181.97682, 148.69468, 
162.12624, 154.86652, 112.64927, 83.6448399999998, 92.2122400000002, 
461.96806, 400.63795, 456.12037, 445.22857, 423.22788, 400.9933, 
426.92592, 356.79285, 446.50166, 381.74695, 373.39454, 254.18138, 
308.77851, 271.01967, 309.41998, 240.19632, 177.84707, 199.12352, 
209.20941, 187.56295, 210.66083, 185.11924, 196.16437, 188.1198, 
154.59123, 176.26524, 211.03701, 70.3728200000003, 77.4256500000004, 
110.26775, 113.43465, 185.89772, 378.94032, 362.2088, 354.61628, 
375.85998, 359.22457, 443.1808, 359.3556, 370.04023, 374.89923, 
345.88871, 450.35979, 445.05926, 242.10715, 339.85273, 338.04194, 
262.48003, 306.9234, 299.16275, 285.24334, 329.84407, 279.96009, 
256.09127, 271.01967, 313.5301, 278.71589, 247.38436, 267.83959, 
297.28745, 343.23287, 311.44174, 238.28897, 339.84732, 321.6157, 
312.41223, 281.21777, 255.20257, 273.991, 246.34864, 340.60144, 
325.80878, 263.38789, 281.3855, 267.90686, 267.86614, 323.79426, 
124.40728, 175.47115, 179.95796, 178.79807, 180.7536, 199.57391, 
177.57037, 174.02255, 223.70296, 179.08881, 189.96577, 223.34289, 
170.56964, 170.54282, 139.84991, 154.52067, 146.1385, 158.17924, 
217.21353, 142.19763, 147.67415, 131.14844, 205.80201, 192.283, 
198.18379, 114.86877, 120.51877, 98.4340199999995, 103.94229, 
105.70264, 118.72419, 119.52053, 70.36, 84.566, 93.49421, 97.3436600000001, 
92.0340700000003, 114.02601, 170.87504), .Dim = c(143, 2)),
    data = c(306.09, 200.88, 167.07, 162.77, 163.57, 178.61, 
    301.54, 282.07, 319.12, 244.67, 233.31, 224.46, 206.12, 248.99, 
    237.87, 222.87, 263.1, 236.91, 247.01, 240.58, 304.28, 351.73, 
    277.92, 323.08, 253.32, 315.33, 379.94, 197.09, 199.91, 167, 
    182.88, 197.1, 257.5, 205.16, 224.07, 212.5, 242.08, 247.79, 
    187.27, 222.54, 313.6, 269.92, 321.69, 208.89, 238.62, 248.76, 
    193.48, 240.48, 265.56, 302.13, 335.41, 330.87, 329.49, 262.81, 
    365.88, 359.08, 344.59, 366.1, 201.84, 218.27, 200.38, 229.4, 
    235.07, 236.25, 228.82, 258.12, 232.17, 248.17, 240.66, 184.59, 
    165.46, 320.31, 232.8, 266.27, 301.1, 244.78, 248.57, 299, 
    269.49, 300.89, 311.81, 317.34, 286.93, 306.64, 349.09, 289.11, 
    280.49, 237.11, 269.68, 307.47, 253.57, 226.71, 299.69, 266.62, 
    331.46, 270.67, 321.15, 258.2, 288.85, 255.76, 264.77, 271.76, 
    313.63, 252.94, 246.75, 215.95, 299.71, 203.4, 209.47, 185.63, 
    304.3, 320.05, 273.01, 295.62, 334.95, 350.18, 291.35, 319.57, 
    367.32, 332.04, 332.18, 323.82, 334.22, 345.33, 345.24, 383.13, 
    348.67, 282.54, 291.07, 245.42, 291.63, 288.5, 314.65, 331.85, 
    305.43, 285.05, 359.51, 356.82, 372.56, 413.7, 412.44, 400.96, 
    199.37)))
class(maijun) <- "geodata"

"borders" <-
structure(c(670.204875, 663.718688, 656.066688, 649.971438, 642.634563, 
635.271688, 628.742125, 622.755875, 617.270813, 610.457, 604.42175, 
600.262, 592.621, 585.444625, 581.92875, 574.871563, 567.99075, 
565.6725, 560.5745, 553.86975, 546.261688, 542.435063, 537.837375, 
534.376563, 529.409438, 528.1835, 520.268156, 513.789781, 509.833156, 
505.105344, 499.977219, 491.988906, 486.094406, 478.869313, 474.045719, 
470.925344, 473.2195, 476.876406, 476.248156, 468.888031, 462.077594, 
457.88275, 453.517656, 448.93925, 440.925281, 433.288844, 425.385563, 
417.613313, 409.689906, 402.012969, 395.090313, 388.905156, 381.620781, 
373.674125, 365.810969, 358.028781, 350.718188, 343.347406, 336.159188, 
329.237156, 321.329844, 313.663188, 305.80225, 298.281313, 290.861594, 
284.227281, 277.077844, 271.504219, 263.543781, 256.520563, 248.8575, 
241.656109, 235.520094, 231.700766, 226.547563, 225.8555, 219.657406, 
216.403594, 215.340656, 216.788484, 213.318719, 212.552734, 206.664047, 
201.807359, 199.6235, 192.151453, 187.422156, 188.983141, 184.399016, 
178.437156, 174.949359, 170.373547, 163.837438, 157.839297, 155.134688, 
150.051703, 144.086359, 138.815359, 138.297594, 137.987281, 144.653203, 
149.781453, 148.382547, 153.388125, 159.686078, 155.62975, 153.212984, 
154.687313, 157.737141, 163.658594, 160.675313, 164.211328, 172.130047, 
165.728922, 158.935609, 157.750344, 159.271844, 158.399938, 155.783297, 
157.487594, 158.247594, 162.589, 165.391719, 157.077922, 157.754734, 
160.412781, 162.614203, 166.648797, 163.022797, 163.530688, 166.652172, 
171.481313, 171.281031, 165.928656, 165.506375, 165.3565, 167.316813, 
170.03325, 167.977563, 163.491438, 161.473719, 163.997516, 170.930609, 
178.861531, 185.087641, 190.730766, 190.762938, 193.086672, 195.451828, 
198.341609, 200.613078, 203.828609, 204.428203, 208.680922, 214.333047, 
221.180953, 224.122406, 226.760719, 229.690422, 231.72975, 231.220078, 
235.038641, 240.402063, 247.119406, 254.061875, 261.190891, 268.550156, 
275.370188, 282.111844, 287.672, 293.985563, 301.211688, 307.937906, 
315.517, 323.409656, 331.151406, 336.672156, 343.911, 351.417813, 
356.621688, 362.441, 367.828219, 368.792531, 375.514781, 381.666813, 
383.534469, 387.002438, 394.258938, 396.523313, 396.644688, 402.613531, 
408.970906, 416.527875, 423.576406, 430.054938, 437.537875, 444.892563, 
452.482813, 459.560219, 464.211406, 465.268031, 467.336875, 465.45875, 
472.983781, 476.725063, 480.198531, 478.720594, 479.167094, 480.899469, 
482.783188, 483.727938, 483.194875, 485.376531, 488.530688, 491.747375, 
494.156156, 499.878813, 496.505469, 499.45075, 502.862656, 495.062781, 
491.5965, 498.709813, 504.852094, 500.144813, 500.993125, 505.134438, 
503.235969, 506.228938, 508.160219, 509.2235, 503.473406, 509.178344, 
516.034031, 520.147219, 521.236875, 527.753063, 531.995563, 538.783625, 
545.707125, 553.139188, 560.024625, 566.74675, 574.19725, 582.02375, 
589.887375, 597.229313, 603.811188, 605.871, 612.286938, 613.857063, 
621.287813, 628.639063, 631.817438, 637.492438, 634.20475, 629.226563, 
630.906875, 638.968875, 632.025, 625.798625, 631.953125, 632.155188, 
639.860125, 641.00775, 637.589125, 637.24325, 636.01825, 638.107625, 
641.851813, 642.569438, 647.145438, 646.820625, 642.397875, 646.79425, 
651.9535, 655.33225, 660.207125, 666.556125, 668.464313, 671.055375, 
674.749375, 682.068125, 678.8675, 676.1265, 672.481563, 670.703063, 
672.062563, 679.603063, 686.332625, 694.170375, 700.403375, 707.086875, 
715.055063, 722.802563, 730.246063, 738.165688, 745.08375, 751.146, 
748.617188, 746.678813, 746.024563, 742.908813, 745.792625, 750.472063, 
754.7605, 756.962, 764.113563, 766.416875, 773.6025, 779.524438, 
780.54775, 781.245688, 782.124438, 785.844, 786.494, 793.205188, 
798.625563, 793.336563, 788.13475, 783.140375, 779.378563, 777.683813, 
780.69875, 787.14325, 779.648813, 772.613438, 769.153125, 769.227688, 
765.929875, 763.41925, 758.943313, 758.803438, 755.02475, 758.635313, 
755.889125, 749.726063, 741.674063, 735.94675, 729.9425, 729.948, 
732.53975, 724.347438, 732.07725, 739.425688, 747.238313, 749.815563, 
757.045125, 765.36825, 761.700813, 756.485438, 752.37975, 748.731938, 
744.853563, 738.498438, 733.378375, 731.44225, 739.654313, 743.952688, 
740.586188, 734.3525, 726.756313, 718.759688, 710.763438, 702.864375, 
695.90125, 688.046, 680.367938, 670.204875, 111.761, 107.051, 
105.242, 100.7915, 97.993, 101.1545, 105.5405, 110.7495, 116.397, 
120.5795, 119.667, 121.898, 119.9675, 120.177, 115.782, 119.532, 
115.8975, 110.499, 117.333, 120.929, 121.4415, 119.9445, 117.39, 
110.527, 104.375, 97.7945, 98.896, 97.3895, 94.0785, 94.73, 98.2715, 
98.26, 93.482, 90.0065, 83.768, 76.4215, 69.7315, 62.9205, 55.255, 
52.0755, 49.4895, 46.7695, 53.727, 59.958, 58.8065, 60.247, 60.1965, 
58.7055, 58.782, 61.107, 64.9245, 70.0245, 72.88, 73.6025, 74.997, 
75.2305, 77.7445, 78.171, 80.7555, 84.688, 84.1965, 83.564, 83.2995, 
82.2695, 80.292, 84.5905, 87.9925, 93.4435, 94.249, 90.7505, 
88.96, 92.3015, 96.4505, 103.246, 109.2565, 116.906, 122.0885, 
128.354, 136.198, 143.92, 151.001, 158.5445, 163.667, 166.2495, 
167.25, 169.1685, 165.457, 173.248, 174.298, 169.808, 168.851, 
169.669, 165.8305, 161.0635, 154.883, 161.614, 164.554, 167.945, 
176.129, 183.5995, 179.9475, 181.639, 189.343, 190.8625, 188.6405, 
192.2815, 196.7395, 201.929, 202.4355, 198.6235, 201.474, 207.0945, 
206.12, 207.3115, 205.5975, 210.408, 213.2685, 216.1175, 220.7955, 
222.779, 224.8275, 225.788, 228.3695, 228.715, 233.218, 242.7005, 
248.682, 249.457, 253.9695, 258.6515, 265.281, 265.8665, 266.8735, 
272.053, 279.341, 286.609, 293.7815, 301.3105, 309.0375, 315.479, 
323.254, 330.385, 334.1935, 335.4215, 340.129, 345.8245, 353.308, 
360.851, 368.452, 375.9125, 383.465, 390.732, 398.7125, 405.3755, 
411.065, 415.045, 422.3715, 429.8395, 437.277, 444.9855, 453.0285, 
459.9115, 465.8625, 470.151, 474.1165, 477.7425, 480.8905, 485.017, 
489.3505, 495.0075, 499.934, 503.1085, 499.2055, 496.7625, 495.3895, 
496.865, 500.318, 497.055, 497.705, 494.4745, 496.3685, 491.1985, 
491.6935, 489.49, 493.9795, 501.523, 507.9295, 504.9965, 497.152, 
503.3365, 502.907, 498.0905, 497.7135, 496.94, 493.1595, 492.287, 
489.9165, 492.5375, 493.548, 490.967, 483.0055, 482.201, 490.2815, 
491.8335, 484.8545, 482.781, 474.8165, 473.0705, 480.3425, 472.1715, 
468.8685, 476.651, 481.32, 474.7985, 475.4645, 479.503, 474.766, 
470.4505, 467.3145, 465.864, 463.654, 461.1125, 459.3925, 454.5375, 
452.687, 443.8865, 444.3065, 451.4485, 457.7515, 463.096, 470.414, 
475.6315, 476.1755, 473.357, 467.075, 459.4815, 459.3755, 464.782, 
464.875, 465.694, 462.5555, 464.8, 463.1885, 462.725, 461.2715, 
462.521, 465.8655, 464.5585, 460.5085, 456.086, 449.595, 446.604, 
444.8285, 437.737, 432.025, 425.496, 422.555, 420.4055, 419.884, 
416.7505, 413.4625, 414.2685, 411.1705, 411.021, 403.435, 401.2055, 
400.3325, 397.1535, 394.517, 389.9, 382.0995, 375.437, 367.535, 
361.227, 355.1195, 349.3295, 342.219, 335.8645, 330.8, 324.334, 
317.356, 310.5265, 307.0365, 300.0215, 292.5305, 285.4165, 277.625, 
269.8635, 267.0335, 269.963, 269.5365, 273.576, 269.09, 269.743, 
269.0615, 267.9115, 267.6705, 267.2245, 261.5795, 255.2725, 247.9015, 
240.0395, 232.673, 226.0275, 222.804, 230.096, 237.04, 239.545, 
247.0055, 246.4635, 241.737, 235.974, 228.906, 221.1565, 214.5615, 
208.1565, 204.8215, 205.266, 199.1955, 193.133, 186.8895, 180.5855, 
187.108, 194.386, 199.101, 198.6915, 194.921, 197.2955, 205.304, 
200.563, 199.8925, 205.232, 196.9885, 193.748, 188.6695, 181.1865, 
182.4225, 182.355, 185.664, 190.334, 185.579, 179.157, 179.706, 
177.326, 175.5195, 176.2825, 172.8805, 171.5365, 170.572, 165.084, 
159.1015, 152.2355, 145.1195, 138.108, 141.5515, 139.683, 135.9955, 
136.1905, 133.044, 125.3875, 126.208, 124.2705, 124.561, 124.8235, 
124.122, 121.1025, 122.3335, 121.7645, 111.761), .Dim = c(369, 
2))

"loci.paper" <-
structure(c(300.336, 647.755, 361.764, 410.1, 484.453, 317.463, 
438.781, 260.373), .Dim = c(4, 2))

"s100" <- 
structure(list(coords = matrix(c(0.80712671019136906, 0.54999807151034474, 
	0.34080554451793432, 0.13709931029006839, 0.044185692444443703, 
	0.027816359885036945, 0.72438529878854752, 0.24697387451305985, 
	0.52298387931659818, 0.2496004281565547, 0.028991738334298134, 
	0.1431424212642014, 0.086238803341984749, 0.98392054392024875, 
	0.079471597447991371, 0.47806540178135037, 0.63142601726576686, 
	0.82055020006373525, 0.93456090893596411, 0.094132198952138424, 
	0.58557709446176887, 0.6424179170280695, 0.50794567726552486, 
	0.36339767649769783, 0.51376408850774169, 0.26380650792270899, 
	0.86457986291497946, 0.0056380061432719231, 0.90725778182968497, 
	0.17779795033857226, 0.33019586699083447, 0.094165341928601265, 
	0.25778363971039653, 0.22686947789043188, 0.84611224429681897, 
	0.40495845908299088, 0.70719032734632492, 0.73298889445140958, 
	0.29306220728904009, 0.19297148520126939, 0.84977900004014373, 
	0.45702632376924157, 0.69153992086648941, 0.47588160028681159, 
	0.26541369035840034, 0.67922241427004337, 0.30641887569800019, 
	0.10858904384076595, 0.46477998839691281, 0.9680333798751235, 
	0.061719681136310101, 0.30642673885449767, 0.94701418839395046, 
	0.43085947446525097, 0.93184410920366645, 0.68919968511909246, 
	0.63498426042497158, 0.48569687129929662, 0.16337611014023423, 
	0.88779823575168848, 0.0068703116849064827, 0.447166598867625, 
	0.50651556765660644, 0.9675562740303576, 0.74790614517405629, 
	0.50318212015554309, 0.077190299052745104, 0.5888819950632751, 
	0.10819409834221005, 0.88668545568361878, 0.64707031473517418, 
	0.29039342422038317, 0.6587650584988296, 0.74858199106529355, 
	0.60409955633804202, 0.8510717460885644, 0.2984471470117569, 
	0.02076778095215559, 0.94028107263147831, 0.58688202267512679, 
	0.85059535410255194, 0.38138212589547038, 0.5775338183157146, 
	0.15835696738213301, 0.49152398156002164, 0.73334673279896379, 
	0.53632984682917595, 0.90022387076169252, 0.18626313842833042, 
	0.90458632912486792, 0.66404575202614069, 0.085919891018420458, 
	0.45504520181566477, 0.89055463997647166, 0.18480749567970634, 
	0.58583401376381516, 0.54133438318967819, 0.40625399397686124, 
	0.49531776830554008, 0.38566303066909313, 0.94544601300731301, 
	0.68326491676270962, 0.45850888034328818, 0.47200832422822714, 
	0.12232017470523715, 0.80374587560072541, 0.6233249488286674, 
	0.14205769635736942, 0.76201844355091453, 0.58405975811183453, 
	0.9560615555383265, 0.95199656672775745, 0.14106986345723271, 
	0.71920010773465037, 0.58686494454741478, 0.43981989938765764, 
	0.89451548783108592, 0.48675112426280975, 0.90151411760598421, 
	0.55364480847492814, 0.6976369135081768, 0.76206591771915555, 
	0.88213742570951581, 0.62456433195620775, 0.068266099318861961, 
	0.35663249157369137, 0.26637957291677594, 0.76928547210991383, 
	0.1357616214081645, 0.053212936967611313, 0.20887774648144841, 
	0.77715519070625305, 0.55595250288024545, 0.59507908811792731, 
	0.73963968316093087, 0.053829971700906754, 0.52866792306303978, 
	0.60828701267018914, 0.94696339964866638, 0.074820824433118105, 
	0.3783628880046308, 0.35670797154307365, 0.41060066176578403, 
	0.68247604789212346, 0.032653048168867826, 0.43100733356550336, 
	0.84172254335135221, 0.72541735647246242, 0.74625289300456643, 
	0.94756137067452073, 0.57977207330986857, 0.90216260589659214, 
	0.95192688051611185, 0.96433238359168172, 0.65749259013682604, 
	0.15415125247091055, 0.64798716781660914, 0.40060388622805476, 
	0.24714981345459819, 0.38501224247738719, 0.11247721174731851, 
	0.62429935811087489, 0.41561616118997335, 0.58564516250044107, 
	0.52078143507242203, 0.45504265557974577, 0.92241132538765669, 
	0.83999337395653129, 0.7630884712561965, 0.8352356250397861, 
	0.14569711638614535, 0.60043479455634952, 0.52603476820513606, 
	0.26287564961239696, 0.79045206587761641, 0.76730799023061991, 
	0.1661572502925992, 0.98308196477591991, 0.77707860944792628, 
	0.70516562322154641, 0.010910267475992441, 0.9912497908808291, 
	0.17780000669881701, 0.20805035438388586, 0.37358486047014594, 
	0.38444906054064631, 0.76250095805153251, 0.82685943180695176, 
	0.71121034119278193, 0.95129011943936348, 0.98082405980676413, 
	0.68740117410197854, 0.21624275902286172, 0.94962813565507531, 
	0.25771586783230305, 0.44893033849075437, 0.13499378738924861, 
	0.96311050420626998, 0.17356992745772004, 0.52014265954494476),
	ncol=2),
	data = c(0.91718875217951368,
	1.1483233540906543, 1.0327562997057935, 0.12195476725214387, 
	0.61529877761349172, -0.55060654281636467, 1.7038148867550866, 
	1.7479010203012721, 2.1008690730653319, 0.44122460880913639, 
	-0.0095363565905963402, -0.80144833308901664, -0.0019308963727725586,
	1.1419438128609063, 0.53867551458861018, 0.56112476038952641, 
	1.0843426059672516, 1.4942248674711278, 1.4517737091769016, 
	-0.75662243696875009, 1.7425368412090507, 1.9720699430624418, 
	0.86903050344309996, -0.22093192187119662, 1.3117251861204724, 
	1.561038128566387, 1.8435008499911785, 0.33317766272990318, 
	1.2230304764424147, 0.5450238236175865, 1.5829329116145006, 
	-0.81495805142910149, 0.054556701779654139, 0.40569741750324506, 
	1.4226655841979632, 1.7652989931105356, 1.1854706379314226, 
	1.7544353678761651, -1.1506248460666451, 1.5533570862881627, 
	1.1649510111067605, 0.25947990947309718, 1.5774264394821504, 
	0.83798307695578866, 1.6430874651847667, 1.7172314260223402, 
	0.99391755385086633, 0.12860306201604563, 0.80566651788376364, 
	0.7657770248006176, 0.1032904988989201, -1.1676954743268042, 
	0.71914878338197463, 0.35970927077184017, 1.7784226571403636, 
	2.0795654236158581, 2.8678969027446781, 0.56002127601571894, 
	0.27749095652446709, 1.792041134066106, 0.2001096320311031, 
	0.23775632618307183, 1.3085486213567772, 1.9449011488221781, 
	0.40213187039401721, 0.97964048725903174, -0.23850943482183704, 
	2.1561205832889683, -0.61099806597416217, 1.5462031744184936, 
	1.496030375120436, 0.35389800284972772, 1.9687174190519297, 
	2.1715776639009392, 1.965981721784692, 1.6202284891942846, 
	1.6372050699149807, 0.061662134593765031, 1.8101610142077365, 
	1.7263088905969155, 0.060908553883130323, -0.72982403159029297, 
	1.2103336419402468, 0.64394793010828033, -0.066323181821390756, 
	1.1626951165139294, 2.443485524259847, 1.8451807976038588, 
	1.1248445718036915, 1.3800251669296828, 1.3995387979357528, 
	1.3111491751405291, 1.606818085489534, 1.4293800865039437, 
	0.13473358016350365, 1.2848223192212394, 0.75724609135845966, 
	0.18618489551416972, 0.8910846759306511, 0.073563986156599426), 
	cov.model = "exponential", nugget = 0, cov.pars = c(1., 
	0.29999999999999999), kappa = 0.5, lambda=1), class="geodata")


"s121" <- 
structure(list(coords = matrix(c(0., 0.10000000000000001, 0.20000000000000001, 
	0.30000000000000004, 0.40000000000000002, 0.5, 0.60000000000000009,
	0.70000000000000007, 0.80000000000000004, 0.90000000000000002, 1.,
	0., 0.10000000000000001, 0.20000000000000001, 0.30000000000000004,
	0.40000000000000002, 0.5, 0.60000000000000009, 0.70000000000000007,
	0.80000000000000004, 0.90000000000000002, 1., 0., 0.10000000000000001,
	0.20000000000000001, 0.30000000000000004, 0.40000000000000002, 0.5,
	0.60000000000000009, 0.70000000000000007, 0.80000000000000004, 
	0.90000000000000002, 1., 0., 0.10000000000000001, 0.20000000000000001,
	0.30000000000000004, 0.40000000000000002, 0.5, 0.60000000000000009,
	0.70000000000000007, 0.80000000000000004, 0.90000000000000002, 1.,
	0., 0.10000000000000001, 0.20000000000000001, 0.30000000000000004,
	0.40000000000000002, 0.5, 0.60000000000000009, 0.70000000000000007,
	0.80000000000000004, 0.90000000000000002, 1., 0., 0.10000000000000001,
	0.20000000000000001, 0.30000000000000004, 0.40000000000000002, 0.5,
	0.60000000000000009, 0.70000000000000007, 0.80000000000000004, 
	0.90000000000000002, 1., 0., 0.10000000000000001, 0.20000000000000001,
	0.30000000000000004, 0.40000000000000002, 0.5, 0.60000000000000009,
	0.70000000000000007, 0.80000000000000004, 0.90000000000000002, 1.,
	0., 0.10000000000000001, 0.20000000000000001, 0.30000000000000004,
	0.40000000000000002, 0.5, 0.60000000000000009, 0.70000000000000007,
	0.80000000000000004, 0.90000000000000002, 1., 0., 0.10000000000000001,
	0.20000000000000001, 0.30000000000000004, 0.40000000000000002, 0.5,
	0.60000000000000009, 0.70000000000000007, 0.80000000000000004, 
	0.90000000000000002, 1., 0., 0.10000000000000001, 0.20000000000000001,
	0.30000000000000004, 0.40000000000000002, 0.5, 0.60000000000000009,
	0.70000000000000007, 0.80000000000000004, 0.90000000000000002, 1.,
	0., 0.10000000000000001, 0.20000000000000001, 0.30000000000000004,
	0.40000000000000002, 0.5, 0.60000000000000009, 0.70000000000000007,
	0.80000000000000004, 0.90000000000000002, 1., 0., 0., 0., 0., 0., 0.,
	0., 0., 0., 0., 0., 0.10000000000000001, 0.10000000000000001, 
	0.10000000000000001, 0.10000000000000001, 0.10000000000000001, 
	0.10000000000000001, 0.10000000000000001, 0.10000000000000001, 
	0.10000000000000001, 0.10000000000000001, 0.10000000000000001, 
	0.20000000000000001, 0.20000000000000001, 0.20000000000000001, 
	0.20000000000000001, 0.20000000000000001, 0.20000000000000001, 
	0.20000000000000001, 0.20000000000000001, 0.20000000000000001, 
	0.20000000000000001, 0.20000000000000001, 0.30000000000000004, 
	0.30000000000000004, 0.30000000000000004, 0.30000000000000004, 
	0.30000000000000004, 0.30000000000000004, 0.30000000000000004, 
	0.30000000000000004, 0.30000000000000004, 0.30000000000000004, 
	0.30000000000000004, 0.40000000000000002, 0.40000000000000002, 
	0.40000000000000002, 0.40000000000000002, 0.40000000000000002, 
	0.40000000000000002, 0.40000000000000002, 0.40000000000000002, 
	0.40000000000000002, 0.40000000000000002, 0.40000000000000002, 0.5,
	0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.60000000000000009,
	0.60000000000000009, 0.60000000000000009, 0.60000000000000009, 
	0.60000000000000009, 0.60000000000000009, 0.60000000000000009, 
	0.60000000000000009, 0.60000000000000009, 0.60000000000000009, 
	0.60000000000000009, 0.70000000000000007, 0.70000000000000007, 
	0.70000000000000007, 0.70000000000000007, 0.70000000000000007, 
	0.70000000000000007, 0.70000000000000007, 0.70000000000000007, 
	0.70000000000000007, 0.70000000000000007, 0.70000000000000007, 
	0.80000000000000004, 0.80000000000000004, 0.80000000000000004, 
	0.80000000000000004, 0.80000000000000004, 0.80000000000000004, 
	0.80000000000000004, 0.80000000000000004, 0.80000000000000004, 
	0.80000000000000004, 0.80000000000000004, 0.90000000000000002, 
	0.90000000000000002, 0.90000000000000002, 0.90000000000000002, 
	0.90000000000000002, 0.90000000000000002, 0.90000000000000002, 
	0.90000000000000002, 0.90000000000000002, 0.90000000000000002, 
	0.90000000000000002, 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1.), 
	ncol=2),
	data = matrix(c(0.10626965882243737, -1.1981122813439611, -1.2764807484640164, 
	-1.4875946522483703
	, -1.3961969907001148, -0.39291401586961539, 0.0095075077671084307,
	-0.31501050186117158, -0.77586391336922256, -0.98091594361900958, 
	-1.9707746407272451
	, -0.37861258734505676, -0.49300945093541254, -1.9972842009356788,
	-1.6234658862227551, -1.2233528386856014, 0.031387384602555933, 
	0.983617647476004, 0.094979766597481086, -0.2488968232543381, 
	-0.81935616685608725, -0.26622706843956506, -0.37769005990169763, 
	0.45108764740049179, -1.0158283775168302, -0.72718658319460838, 
	0.044752608252268809, 0.40228714628584022, 0.46432853710314226, 
	-0.32781963972198325, 0.17140613260617177, 0.60701236583999774, 
	-0.46125553553919035, 0.92795397616237429, 0.92537007838525209, 
	-0.030068237628211574, -0.33849036506949848, 0.25800494094701337, 
	0.31165132665428474, -0.32343463504968839, -0.061958199112938664, 
	1.7013942774963824, 0.81251846943671602, 0.16999374716208859, 
	0.84285730628025424, 1.1088377918716055, 0.64286041930516602, 
	1.265209412555548, 2.0241073315670235, 1.139510883671055, 
	-0.11857217295494206, 0.50179698030565112, 0.59645962819643894, 
	1.0834570837653779, -1.6671864688164801, 1.0935602225045795, 
	1.6686552682481497, 2.0318315437147678, 1.2380506014425503, 
	1.3137274342037677, 0.53266396650529035, 1.3854947149232875, 
	1.0046878293888926, 0.56574694726969943, 0.60645304483531537, 
	0.19613348783858184, 0.55650675824957541, 1.5391142773702005, 
	1.8476483638177177, 1.0977854444072506, 1.3908025496278154, 
	1.6726017557292778, 1.1571992482319009, 1.0871850084753216, 
	1.2905346912903681, 1.3292468202019208, 1.0535073859943505, 
	0.30998329949784165, 1.232172121142491, 0.87428312882908421, 
	1.6535692722541471, 2.3224073588549459, 2.8701787305128725, 
	1.7901579534031344, 1.5363295002173278, 0.60207673686675645, 
	-0.052753854478232373, 0.65754096584042865, 1.0043420836685371, 
	0.59684794629333426, 0.78014833497589453, 1.6946989273569262, 
	2.5530762574913202, 0.71490259097348607, 1.5964226519982327, 
	1.470656693915658, 0.42534089965769534, -0.73409962339685142, 
	0.18692112878172845, 1.3781172177983694, 1.6267472565071439, 
	2.1148495277072175, 1.8102774120377296, 1.325292610839691, 
	1.0031206236463475, 0.38898419185974542, -0.017768604750426967, 
	0.13786384462019569, -0.25550818025440336, 0.38840672443192154, 
	0.96305855609043645, 1.6464221411194109, 0.93946174077233757, 
	2.3770408320629892, 1.104449408145028, -0.051417217099113843, 
	0.42135292996431917, 0.030666446200383616, 0.26418350646648281, 
	0.57425801311375235, 1.8722244693419514, 0.79396510129976106, 
	-0.15927241930701613, -0.61551650615666398, -0.83011980812599728, 
	-2.2169734243394386
	, -1.9343393843063135, -0.90360397001483628, -0.012485563607058698,
	0.78648183289118268, 2.3910234062885407, 1.8012197964567651, 
	-0.35942809209573423, 0.34324430985208398, -0.69096638409869493, 
	-1.3211713185273866
	, -0.53364023190209819, -0.54860216772483472, -1.736610568769253, 
	-0.39473991284824228
	, 0.88337042159313206, 2.3705743652669873, 1.6312959682116726, 
	0.049006619756794531, -0.48763229258860025, -0.38630035677452956, 
	-0.089838732882574671
	, 0.20657076706868407, -0.41698880071652439, -0.41221119916892662,
	-0.14736030666094269, 0.87242246642399435, 1.9786895051271172, 
	1.4281858019690501, -1.0073581157758578, -1.2045502528593848, 
	-0.82122567545945668, -0.0057508755223721116, -1.102467820828166, 
	-1.0289982758376035
	, 0.36830308155768032, 0.87660430758876373, 2.0408185870207336, 
	2.3992507871031066, 2.3424640745777863, 0.61147940260853562, 
	0.14861633472974603, -0.06122984867906759, -0.93082513859944049, 
	-1.1265315086139731
	, -1.3716792023978104, 0.38964049863568151, 0.54446336083728164, 
	0.99681500900174969, 1.3665901456095328, 2.693211099623984, 
	0.67598809205692167, 1.6906135241946656, 0.62555149022430379, 
	0.011374162891643913, -0.52272242722750106, -0.10989808272795561, 
	0.14215839533009658, 0.93399798253570576, 1.4743282512356111, 
	1.7595229694272763, 2.0556640832370663, 1.4541174086555351, 
	1.7887158801816758, 1.3280392996337891, 1.3204834321569132, 
	1.1611429916834375, 0.36978827125521241, 0.86329492970365163, 
	0.74700334256949219, 1.245845558715617, 1.268407730329514, 
	1.846007883017573, 1.4377564893910264, 1.5423235809653604, 
	1.1770748231459296, 1.820999560448866, 0.84098158815826685, 
	0.98726141222482955, 1.6586019377047827, 1.0351493378172258, 
	0.023588526267250606, 0.54117866699969841, 1.5009051197728276, 
	2.0361556480647436, 0.93062947538238061, 1.7668324280228778, 
	1.4442705759019765, 1.5430792605952386, 0.8871111921371595, 
	0.97522836539078195, 1.3721426132319274, 0.21893197828007482, 
	-0.34538708598588985, 0.2229557350458917, 1.1958428322882302, 
	1.4534199797540679, 1.347276830664335, 1.1597456218732292, 
	1.2191138012701004, 0.66487557161539479, 1.6950140327297338, 
	0.3882927443635702, 1.770533863894908, 0.44679185828275847, 
	0.90170957565377363, 0.54544995471738666, 0.89690861528397758, 
	0.33431614630042561, 0.98428585369951105, 0.51470441412105061, 
	0.41421555591478987, 0.32888743383075969, 0.17955174626728801, 
	0.8238723716531311, 1.1272987161590982, 1.133170983913232, 
	-1.2366348721536196, -0.80653345017256695, -0.45197596051979161, 
	-0.17278857536237735
	, -0.55823219959120185, 0.23581231264694316, -0.015248585615885671,
	-2.242933060686334, -2.4734974804259648, -0.92683035500299049, 
	-0.73914344853111524, -0.55289100426119464, -0.9389607268641712, 
	-1.6906860257139065
	, -0.54944608331171152, 0.070559921094661318, -1.5490237521975236,
	-1.2039007411302936, -1.8824522056480821, -0.80950494136483864, 
	-0.96049069116604091, -0.73049816616600072, -0.41207211353324052, 
	-2.0260217682796311
	, -2.1013384979867413, -0.28575012593987437, -0.4657243125268053, 
	-1.584624369275772
	, -2.108584723563371, -0.64397792682330812, 0.026179081470906165, 
	0.47729949774072505, -0.85709213050177069, 0.37017981317509119, 
	-2.2438106417285484, -1.0041057092653958, -0.8127926411968992, 
	-0.48207071505951349, -1.2696096740797309, -1.3564214397164114, 
	-1.1250090208127863, -0.27136935716032046, -0.46840879157343313, 
	-0.64445594404150175
	, -0.99364848358545443, -0.9508956309311245, -1.0460213225544217, 
	0.13584295211899067, -0.75789463416761937, -0.66784077583804546, 
	-0.025495461882543397
	, 0.1446262779604687, -0.1253897447956179, -0.45818320166090482, 
	-0.040321488248513726
	, -0.057960320698848114, 0.54864338264835177, -0.60721446815883195,
	-0.94615150387519187, -1.2092672507906912, -0.8055931200501345, 
	-0.54447224558967811, -0.22921870075621936, 0.92944539805644466, 
	0.3284934812133129, 0.97147521148841198, 1.2523810364765673, 
	1.1245850657698686, 0.44298918989808728, -0.74076727335259041, 
	-0.56097543506111913, -0.4153926671304119, 0.58679772180984568, 
	0.25106799158948145, 0.51485488004529911, 0.15374052169933042, 
	0.03944306822550335, 0.8598997316384982, 0.33000063806535973, 
	1.0498492138599804, 0.57882083154087338, 0.74007554972220357, 
	-0.080577114386659224, 0.042688269684481944, 0.54085700568475759, 
	0.054363719871075147, -0.054989657176950507, -1.0451793176035047, 
	1.9785684251808435, 1.1888325589720821, 0.54935376294912541, 
	0.58573883362230295, 2.1916266558129354, 0.50341839686739176, 
	-0.19068124369693631, 0.52471792222827707, -0.099190528218610033, 
	-0.059103190704586799
	, -0.035556985232074199, 0.6009300978090758, 0.72679743476566716, 
	0.55122832833601654, 1.4838581409225151, 1.7169198973589992, 
	1.3483758280281828, -0.48038667841712152, 0.49725600837689576, 
	-0.32106054312628252, -0.31433771875133898, -0.31382026153604953, 
	-0.014151646544940145
	, -0.96602072271488526, 1.2205883054244187, 1.2021286573123804, 
	0.84532656371386772, 0.51034132366335483, 0.091306573129239163, 
	-0.36309639943368383, -0.14541573156859242, -0.12245056619506152, 
	0.9477975234077417, -0.32581110353950238, 0.86248371257965062, 
	-0.036718701923112766, 1.2333718553988575, -0.26118018288952699, 
	-1.2287559070682654
	, -0.41920473774314249, -0.0863292667996651, 0.71612002817017284, 
	0.31842164681860796, 0.020674236405725743, 0.14177428661385377, 
	-0.60770445809698947, -0.90576608926951574, -0.20131027809479599, 
	-1.7663973475624846
	, -2.6282527565565861, -0.50506437565430229, -0.039632398428539897,
	0.34947569076164275, 0.33184387139796079, 0.068641428746218328, 
	-0.51049594404914289, -1.3689913689152318, -1.9947925834435958, 
	-1.4873392236377077, -1.0112192892498633, -1.8891546646041131, 
	-1.5108114439470526, -0.94573408473258658, -0.075032544382333019, 
	-0.4118515028002665
	, -0.052778083449352316, -1.3064513301484639, -2.1821360127436518,
	-2.2505452966775934, -2.0732560689430257, -2.6480194470500447, 
	-1.8671357281876222, -1.287627334154916, -1.8709985418920858, 
	-0.53039434972492561, 1.1657686924834112, -0.18207256830245794, 
	-0.62771764937461538, -0.91495355115225152, -1.4438795948082614, 
	-1.4172024696450749
	, -1.2527461235362511, 0.67719799087650778, -0.7311822559004445, 
	-0.79569142325493214
	, 1.4541389397234481, 0.43307353457398512, 0.64080290532040574, 
	-0.33901871107768644, -1.872162097582782, -1.892127692254179, 
	-0.54144416702983023, 0.70073528897295223, 0.78918986154512283, 
	0.31202501779741143, 1.4329522479242343, 0.31919439593649918, 
	0.38036185484277613, 0.29662942852884139, -0.12644612400225605, 
	-1.2098411461463174, -2.3865046018466556, -1.2182304673814925, 
	0.33385380673209131, 0.63932193958169714, 0.35670373542294892, 
	-0.52774822784710695, -0.41164715998621398, -1.309359456197724, 
	0.22202711564809605, -0.59224860383137234, -0.6739264621003499, 
	-1.0363953055217534, -1.3605986607295584, 0.36657419557415988, 
	0.24651976179582516, -0.77494316423043419, -0.61198850721791365, 
	-0.80618410405362562
	, -0.92761250131039863, -0.099129662369140212, -0.65650543140073081,
	-0.11583685968891508, -0.038209749920023961, -1.2321367863108472, 
	0.68779553181984232, -0.63206930608937251, -1.1677742941014975, 
	-0.77265651258550949, -1.2167238386390666, -0.44544820569048038, 
	-0.16398783396478922
	, -0.69986342423581083, 0.06634019490000842, 0.72808597629839877, 
	-0.090107904809651557
	, 0.26646732631431641, 0.4538566932744737, -0.46896064686632849, 
	-1.3713982086796683
	, -1.334027051950238, -0.46604607877516696, 0.026370299716904182, 
	-1.2685329870549278
	, -0.31764116552955479, 0.9689014083950368, 0.90379903066864686, 
	0.10811026212260394, -0.52172607523914893, -0.26986287856065289, 
	-0.76241883183003312
	, -0.58435779809558464, -1.1995961636356112, -0.16875639089651548,
	1.6698756347753088, 2.1914727507643605, 1.9594389447804781, 
	0.30072571668014936, 0.74298539439472189, 0.54083971334903835, 
	0.97656442306149438, 0.43486711542769663, -0.20584186798853246, 
	-1.850704262792827, -0.97198171846557513, 1.1254583697564469, 
	1.1723640521053929, 1.2032154628420622, 1.9234044702440338, 
	1.4732631022564919, 0.87210486138617793, 0.80199675843698504, 
	0.077792949351967525, 0.1522264402377328, -1.5181658263119893, 
	-1.721354745428689, 0.35862707533261001, 0.9995045590801801, 
	1.3658348564740945, 1.5939876115248723, 1.247842653710078, 
	1.1070110626086422, 1.3571950762581306, 0.76472352635233165, 
	-0.21548627407571005, -0.91093718547902969, -2.6586468116524267, 
	-0.40952643791020071
	, -0.34350031194655017, -0.357240332245509, -0.48473955520510276, 
	0.74115275544699299, 0.63513945269839645, -0.20797170060815542, 
	1.5198863110839165, -0.21925459362687114, -0.92456157669257766, 
	-0.83524584308611283, -0.10709956666191599, -0.36263558023091408, 
	-0.14098845007006192
	, -0.023213587102629833, -0.014646608981612302, 0.82246265221631243,
	1.1091529547217687, 0.72322620280286354, -0.36178642551057577, 
	-1.3829726993129363, -1.6503079299908296, -0.78805210976368234, 
	-0.067959836749646285, 0.67455052129647752, -0.69059748554289424, 
	-1.2634305898706848
	, 0.18209743737136377, -0.15811299128432602, 2.2762227656278191, 
	0.28562336874722, -0.68634560371980957, -1.7569016199565752, 
	-1.1798293683472434, -0.083977547402681235, -0.22225025771823498, 
	-1.0158277041092747
	, -0.43080445794144623, -0.31092749527171271, 0.29970409387654862,
	0.71145753755361463, 0.40625259752420767, -0.98409538449268297, 
	-1.0898240711750233, 0.19660396489350507, -0.0013881138155035413, 
	-0.47078838496191966
	, -1.1337076995870747, -2.0542925399327427, -1.0907833335450237, 
	-0.056741920203136009
	, 0.1073299420329278, -0.37923908062041684, -0.20995893310164396, 
	-0.21027758534798319
	, -1.4134370863261367, -0.60129817943477537, -0.14021135895180908,
	-0.3492139889356275, -2.0953190066088361, -2.3963491830417674, 
	-0.92629823353333618, -0.51661902521560932, -0.49622711547931853, 
	0.69865776590550566, 1.001967833825127, -2.1367597611766165, 
	-1.32377881808283, 0.61721490955509528, 0.15715218368595016, 
	-2.0859153179268106, -1.5907456562881868, -0.59726354055318331, 
	-1.1539615907502117, 0.17916358719641193, -0.24421757360801766, 
	0.97726761531628403, -1.215986196130769, -0.9562303734622547, 
	-0.33735889014569626, -1.1903273911200742, -1.2367607515107997, 
	-0.46997288065519077, -1.7520248464191517, -1.5140761335284421, 
	-0.67001610595208871, 0.14083525200155039, -0.30822806589350543, 
	0.50242997126492761, -0.41176825382421806, 0.30032520204166391, 
	-1.5442982004299461, -1.8281910210783803, -1.5622655952419942, 
	-0.40920252621016301, -0.028661080857780911, 0.082143204126646663,
	1.5690067232758298, 2.6002513252710435, 0.11076573600805818, 
	-0.71667988710954866, 0.36263580745538349, -1.6230301952370059, 
	-1.1672497156654809, -0.89587722318541008, -0.67490136025718361, 
	-1.2628463322335119
	, -0.77545281945052591, -0.0019110767435148975, 1.8269765889324372,
	-1.0259100235968484, -0.57062880854615616, -0.44197637871469181, 
	-1.2304734397813135
	, 0.21541037061975868, -0.16275849249028682, -0.97551749299769408,
	-0.71546642944399952, -1.1307860980072755, -0.51535797811013717, 
	0.58543348046565902, -1.2235889207167547, -0.5652689515523569, 
	-0.10162191422538719, 0.064579438584112442, 1.1404106231553239, 
	-0.077137956960760604, -0.64749098037819741, -0.75180896997008329,
	-0.79290928694517904, -0.84785997822486325, -1.3039543898346804, 
	-0.26961813184072803
	, 0.11053590394892929, -0.34831535963174304, 1.0117744713918837, 
	1.0113891719058099, -0.23492827670122837, 0.84023774046142763, 
	-1.2062412701919412, -0.6044888934092868, -0.47191543065439068, 
	-0.81782036886482601, 0.90901213972060313, 0.23274484979726226, 
	-0.28166619432443701, -0.31478424205799638, -0.3356041552682098, 
	0.34814808628199373, 0.13049596003862904, 0.12661337655830254, 
	-0.60725489555059242, -0.30527928568616836, -1.1049644221652906, 
	0.80714190724428825, 0.52742240211887226, -0.0066266189748139864, 
	-1.3160403549161575
	, -0.71600219041422386, 0.23860187280486489, -0.73963014940102922,
	-1.2757527581807258, -0.54536173133954202, -1.0416106687135667, 
	-0.66065986755682626, 0.27411818762879225, 0.023654674152486552, 
	-1.0859548458348149
	, -0.9899743135556841, -0.47405894072595711, -0.3426945540883537, 
	-1.4598682995240415
	, -0.66651646167580747, -1.6022363309839391, -0.36116611136255766,
	-0.53448981169862619, -0.26951212398371999, 0.058189624287460211, 
	-1.5151722497411655
	, -0.5160930310643066, -0.12839728558645463, 0.1764198753162981, 
	-0.53587413187480293
	, 0.26854489459221309, -0.16255227418626192, 0.095171342400642164,
	0.91283741036228572, 0.16862209233215705, -0.29681671085120698, 
	-2.3661945556609436, -0.80910405849314349, -0.073338823495950195, 
	0.39547013700363687, 0.26538914711661687, 0.079405383672902369, 
	0.34297340476183458, 0.35716263619352367, 1.6333375957572764, 
	-0.34317917524680502, -0.56886505675169408, -1.8631386191696149, 
	-1.0235020851175802
	, 0.19309516653892256, 0.8626101420770671, 0.58048679566610994, 
	0.70938944476299826, 0.29590248084962389, 0.2010599837091494, 
	0.57811900686740902, 0.99149586421752822, 0.71768026582741906, 
	0.79256273495955343, -0.016654379200003643, 1.0076170579960111, 
	0.84654514839097006, 0.80956879995679365, 0.51369439483449986, 
	-0.28842937683535103, -0.44738629056701057, 0.43459647633586845, 
	-0.29811249224394049
	, 0.4025722424032801, -0.19633705340297186, 0.8572789110882959, 
	0.68507833533302742, 1.5219878012205514, 1.5681928098186215, 
	1.1420436306548505, -0.066742940998877023, -0.82235462005421778, 
	-0.34294940816540798
	, -0.45899269096747214, -0.30977144136333612, -0.093450563690781752,
	0.55561535501457782, 0.68761026970287831, 0.040466867154147268, 
	0.32062241493825328, -0.72445771865076181, -0.71437381526378918, 
	-0.050054294578726899
	, 0.17126365938975727, -1.5936033405489383, 0.36183214091570715, 
	0.40486376095121868, 0.37849514840109111, -0.055683198314327713, 
	0.13683201933301906, 1.3103872667349932, 0.46033659658109788, 
	0.68940228932482028, 1.2022105306471127, 0.39409646092605144, 
	-1.2803068576295764, -1.2448827707274162, 0.49208679184289222, 
	1.0238361970179641, 1.0482608536292801, 0.93876807709072407, 
	1.2539539871170999, 0.79700351400360647, 1.0792271223616172, 
	1.1889267427320631, 0.898292128812613, -1.770641153506024, 
	-0.15441068041046874, -0.11038871198127748, 1.1757244860110057, 
	1.4447600660921904, 1.4785323951830631, 1.21402629703139, 
	1.4525756113075807, 0.5818198257249293, 0.16759038972826978, 
	-0.29359534548990607, -0.91542230764826493, 0.17759808718626094, 
	0.55371755491992536, 1.4543333627874893, 2.314219327043455, 
	0.93532868525101165, 0.7327294992664688, 1.0139228092200547, 
	0.15570780095630435, -0.44856114183169682, -0.89550537673116137, 
	-0.70038128346433615
	, 0.56603315076852612, 1.1002206669028538, 0.95806284400662955, 
	2.4656672884013102, 1.0529798725845405, 0.85599461833924384, 
	0.32146046526259175, -0.19696412689065138, -0.96879570150671102, 
	-0.6171149615700674
	, 0.273399655601053, 0.068296904663158825, 0.06522604246702754, 
	0.13663506786987031, 1.6839557031093011, 1.7547573617118006, 
	0.89422069426174455, 0.021759275488763809, 0.2820319905145024, 
	-1.0119279733966118, -0.48887621453642693, -0.2838412788822855, 
	-0.27680797694725739, 0.45485844675478249, 0.86169026107920921, 
	1.3429094563039603, 2.3331830176131381, 0.76867580093294052, 
	0.61123308327455028, 0.20283581066709974, -0.86336661451209951, 
	-1.200210244656736, 0.89146649656293675, 1.4983251371648469, 
	2.0583133057278307, 1.8074187438814262, 2.0687074192426231, 
	2.9654509644319376, 2.2389320838872084, 0.71317937502941475, 
	0.86805878860348529, -1.0385511279339061, 0.28161830944586769, 
	2.5650353329519429, 1.4792871317431471, 0.82019337830139138, 
	0.1589188090993974, 1.3425425174010892, 0.74085399868664881, 
	0.29393715016585503, 0.11419580750992958, 0.48395978337921314, 
	0.50337328300179718, 0.33108968826263313, 2.6075301382074194, 
	1.9686495508223838, 0.59399511869525479, 0.29188186870010491, 
	0.63475988326687593, 0.78796602580213781, 0.98451998867445512, 
	1.6293793815604158, 1.3606459388145353, 0.66581950641994914, 
	0.31664666626226023, 1.1421979263163751, 0.8146227655362096, 
	-0.38855993453363302, -0.93728935430959004, -0.52382126358977266, 
	1.1117210481166109, 0.95305820195296009, 1.5037469329388564, 
	1.7035528398108062, 1.2729433977073921, 0.24313021612830138, 
	-0.62764964523857647, -0.42237402317406331, -0.58632937329358448, 
	-0.98755883010820067
	, 0.30809275832129707, 0.40860273955348425, 0.51624371341840658, 
	2.4285977047035074, 2.5798038121203017, 1.8964208856679321, 
	0.96622567763204981, 0.15780162417159116, 0.03544733540474572, 
	0.11322679684047998, 0.0057317832446593352, 0.76484408945182425, 
	0.94483742702817641, 0.43326510231365556, 1.5620063870669321, 
	2.3277934332603785, 1.1637597285133676, 1.2492435862472056, 
	0.50671240957274943, -0.12269671709960553, 0.022914329055151512, 
	0.77996947143216078, 0.93547655577399802, 0.014868027655510502, 
	-0.072986779466438861, -0.46661247277393025, 0.27985702940261065, 
	1.7487543370205854, 1.2654445893108497, -0.085164411539716628, 
	0.14613253427279099, 0.94545180138539853, 0.38946572562137066, 
	-0.24598354746974149, -0.76890120293598097, -0.32763958770980939, 
	0.0038485885774023798, 0.39133902346494692, 0.61111085697149203, 
	1.8252913771353438, 0.054786835314653087, 0.99230767915947404, 
	1.1797958932685604, 0.36224440125766716, -0.046932887462179627, 
	-0.032316329631057218, 0.12940720180744389, 0.91667104696438373, 
	0.56468752435701097, 1.3257188851308648, 0.87073687063845917, 
	1.0425425891244191, 1.7278855276565424, 1.5885636564796766, 
	-0.2301808183668643, -1.0656973451136849, -0.032676605370133192, 
	0.053700766709079382, -0.25004073429932605, -0.86886241633216033, 
	0.061588401848512629, -0.2723265615937287, 0.47587883750380955, 
	0.5686966428063861, 0.17000468212253589, 0.13006440863439564, 
	-1.2442925068924013, -1.3667856115075963, -1.5888578076326298, 
	-1.2079112311216473, -1.1865147572093346, -0.060850604406349117, 
	0.20854963629294074, -0.42260229797547905, 0.84699475260876878, 
	0.50626415470401609, 0.32951856573287169, -0.94198946834956265, 
	0.0086758346689801286, -1.0027704493889935, -0.026934180908065708,
	-0.50666249920280026, -1.0717366515716633, -0.22214206786532215, 
	-0.63406500991343584
	, -0.58497280501298998, -0.24164876416404479, -1.5174487730945867,
	-1.7801147190787914, -1.0613340086908905, -0.18527606687354428, 
	0.18384173275611976, -1.0468107736699139, 0.16359970380979399, 
	0.3376767001092722, 0.27322653612363229, 0.61644355639187276, 
	-1.2981641239547188, -0.91010300796384525, -2.4881873348053158, 
	-1.6041999743227635, -1.1717051281556081, -1.4625589386454152, 
	-1.668980798292157, -1.0701759998415115, -0.19251773973771025, 
	0.68621602412197802, -0.2959949501661262, -1.3389515124511826, 
	-1.3219567338587217, -1.8344678641255092, -1.7205590220490459, 
	-0.61459202816510672, -1.8815388113628706, -2.5217202250437629, 
	-1.1849493758270606, -1.4981541407488863, 0.9092631204161743, 
	-0.47781365813728655, -1.2841126817490445, -1.4249402173670191, 
	-1.0432451316091091, -0.34727463133707809, -1.0643307719401818, 
	-1.3897031152606463, -0.27870472730776896, -0.57816410051198064, 
	-0.46121077507893588
	, -0.34697505436088916, 0.60841810518769512, -0.023130699640576572,
	0.22264193852296177, -0.38545528312347516, 0.65454823149002639, 
	-0.2279626929805999, 0.023348079669575469, -0.30422083011909695, 
	-0.73538633456720659
	, -0.084006966585696036, -0.33115663576231025, -0.12328856069352769,
	-0.94816846687217682, -0.6511685770015917, -0.56083326170411119, 
	-0.12465957625536804
	, 0.94220432095349238, 0.50898830503748216, 0.32997376773733972, 
	0.081853895396140505, -0.35661253941518778, -0.77541507985128044, 
	-2.0548217947921916
	, -1.0933700247518749, -0.82851390965662441, -0.39014374721071599,
	0.14286648969928023, 0.4754563149671408, 1.5763371362466956, 
	0.40679032569406337, 0.3784444854074186, 0.84766877501240634, 
	-1.1633912747484372, -0.85707258022140576, -0.71973862146934997, 
	-1.0042644823198346
	, -0.65071019341538794, 0.65524489528082397, 0.60189163880166796, 
	0.85168365860482842, 1.2426591311101209, 0.68756735790763457, 
	0.23611628053467376, -0.38468640866916637, -0.34811541844039418, 
	-0.56832381982415259
	, -0.21296437380866956, 0.2969709854632388, 1.0693082722988736, 
	0.30742126651944146, 0.72164829309143541, 0.35000845502919359, 
	-0.23975447484755225, -0.27361776542236937, 0.41120802402052614, 
	1.0281824597751836, -1.1890899904574506, -1.7029018888926337, 
	-1.0469335766837953, -1.140427419048635, -0.89800120888374335, 
	0.25117280826601346, 0.037898788308503406, 0.39560695969919474, 
	0.53975092125715618, 1.3021906587793355, 1.2135683084482876, 
	-0.41773123330967132, -1.9055798835200419, -2.1852623113531244, 
	-2.3741278770706242, -3.1416995791271889, -0.60493270507651919, 
	-0.25822755583090551, 0.6841408994046978, 1.2179448433636746, 
	0.99960101809999169, 0.44266472817742408, 0.44297010001737491, 
	-1.1264004753033812, -1.6532250667112376, -2.1424619900676394, 
	-1.9254993211899254, -1.0302226724253243, -1.3599054782622975, 
	-0.47414081720255524, -0.44963349240416312, -0.17967220903996375, 
	0.14615353677819179, -0.68089681638886301, -0.60125522236711082, 
	-0.89159216544728559
	, -1.2042687228162812, -0.95449358560320641, 0.090795640343886874,
	-1.5762191923623601, -0.78699013412750007, -0.902876406924911, 
	0.37427612388690673, -0.42196303897934268, -0.095313769955223171, 
	0.33392649009918479, 0.14562262132292916, -0.88558161319759543, 
	-0.39150406524804127, -0.079794898470753897, -0.30986843516813162,
	-0.92005078897409143, -0.52511779623789689, 0.46953781795447797, 
	0.28092948755869629, 0.61250519251771396, 0.5224460030515502, 
	-0.1520674637118368, -0.32315713350667591, 0.10413926613834854, 
	0.7608198153435084, -0.85648911465764443, -0.64488754518058677, 
	0.065064705132137315, 0.33139529818005026, -0.18953073573092563, 
	0.34141625187397528, 1.4250249866345484, 0.71001011067449937, 
	-0.50482647242941159, -0.32813868917671163, -0.056988609673853205,
	-0.083916589972417285, 0.39107626888597369, 0.12034921346418508, 
	0.2650284795322902, 0.040718815255349303, -0.78485021343171391, 
	-1.1258512761387796, -0.22020044975582601, -0.4784250563517608, 
	0.38388841606267687, 0.58594317865477596, -0.13171333588738843, 
	0.61768987513336626, 0.46435616152684767, 0.96195636358027559, 
	0.30933653601977817, -0.47042451007091862, -0.73683069628399167, 
	-0.50791900791141997
	, -0.62447814377292876, 0.10526662339267485, 0.039009852257351985,
	0.48128915789865301, 0.63721236637272072, 0.33508879163117167, 
	1.0677572655846583, 1.1664760397547784, 0.286053872625465, 
	0.73923761792732556, 0.16556251172124126, -0.062045022425682023, 
	-0.83592783976098928
	, 0.46336921021969529, 0.12342469674050094, 0.84366972143807961, 
	0.92183247962014836, 0.33371702675449877, 1.4529519748263884, 
	-0.26615548323297905, 0.56536426865225131, 0.19104376414491811, 
	0.21848149858209737, 0.0081995559365257691, 0.62700745226893595, 
	0.60716642338886084, 1.1002971952680494, 0.81420981476246135, 
	0.7962557274431612, 0.20077039266440133, 0.30672697962200907, 
	1.0125169416004471, 0.54745640814458107, 0.49366625918418661, 
	-0.33720392874818012, 1.1113504050254843, 2.2805843649066428, 
	1.3988733653328427, 1.4300746532298474, 0.73466006209787504, 
	1.0001159696605795, 1.4486918207624415, 0.80357545138289643, 
	0.92873688183447933, 0.50513366031585738, 0.13166587893853629, 
	1.0387364768807181, 2.211797340324563, 2.7801723876056466, 
	1.8678575288076389), ncol=10), cov.model = "exponential",
	nugget = 0, cov.pars = c(1., 0.29999999999999999), kappa = 0.5, lambda=1), class = "grf")

